/**
 * This tool outputs the parsed abstract syntax tree of a Java program.
 * J2CDumpAST takes the same arguments as JavaGenCCode.
 */

import AST.*;

class J2CDumpAST extends Frontend {
	public static void main(String args[]) {
		if(!compile(args))
			System.exit(1);
	}

	public static boolean compile(String args[]) {
		return new J2CDumpAST().process(
				args,
				new BytecodeParser(),
				new JavaParser() {
					public CompilationUnit parse(java.io.InputStream is, String fileName) throws java.io.IOException, beaver.Parser.Exception {
						return new parser.JavaParser().parse(is, fileName);
					}
				});
	}

	protected void processErrors(java.util.Collection errors, CompilationUnit unit) {
		super.processErrors(errors, unit);
		System.out.println(unit.dumpTree());
	}

	protected void processNoErrors(CompilationUnit unit) {
		System.out.println(unit.dumpTree());
	}

	protected void initOptions() {
		super.initOptions();
		program.options().addKeyValueOption("-o");
		program.options().addKeyOption("--deadcode-elim");
		program.options().addKeyOption("--sourcelines");
		program.options().addKeyOption("--devirtualize");
		program.options().addKeyOption("--inline"); // Not Implemented yet?
		program.options().addKeyOption("--noredundantpush");
		program.options().addKeyOption("--with-reflection");
	}

	protected  void printUsage() {
		printVersion();
		System.out.println(
				"\nJ2CDumpAST\n\n" +
				"Usage: java J2CDumpAST <options> <source files>\n" +
				"  -verbose					 Output messages about what the compiler is doing\n" +
				"  -classpath <path>		 Specify where to find user class files\n" +
				"  -sourcepath <path>		 Specify where to find input source files\n" + 
				"  -bootclasspath <path>	 Override location of bootstrap class files\n" + 
				"  -extdirs <dirs>			 Override location of installed extensions\n" +
				"  -o <file>				 Name of resulting C file\n" +
				"  --deadcode-elim			 \n" +
				"  --sourcelines			 \n" +
				"  --devirtualize			 \n" +
				"  --inline					 \n" +
				"  --noredundantpush		 \n" +
				"  --with-reflection		 \n" +
				"  -g						 \n" +
				"  -d <directory>			 Specify where to place generated files\n" +
				"  -help					 Print a synopsis of standard options\n" +
				"  -version					 Print version information\n"
				);
	}

	protected String url() { return "(http://www.robot.lth.se/java)";}
	protected String name() { return "ljrt compiler"; }
	protected String version() { return "devel"; }
}
