import AST.*;
import java.io.*;

class JavaGenCCode extends Frontend {
	public static void main(String args[]) {
		if(!compile(args))
			System.exit(1);
	}

	public static boolean compile(String args[]) {
		JavaGenCCode obj = new JavaGenCCode();
		boolean status = obj.process(
									 args,
									 new BytecodeParser(),
									 new JavaParser() {
										 public CompilationUnit parse(java.io.InputStream is, String fileName) throws java.io.IOException, beaver.Parser.Exception {
											 return new parser.JavaParser().parse(is, fileName);
										 }
									 }
									 );
		if (status) {
			obj.genCode();
		} 
		return status;
	}

	protected void genCode() {
			program.lookupType("java.lang","Thread");
			program.lookupType("java.lang","InterruptedException");

			program.dummy();			

			if (program.options().hasValueForOption("-o")) {	
				program.setOutFile(program.options().getValueForOption("-o"));
			} else {
				printUsage();
				System.exit(1);
			}

			// Generate reachability information
			try {
				PrintStream rf;
				//rf = new PrintStream(new FileOutputStream("reachability.classes"));
				//program.printClassReachability(rf);
				rf = new PrintStream(new FileOutputStream("reachability.methods"));
				program.printMethodReachability(rf);
			} catch (FileNotFoundException e) {
				e.printStackTrace();
				System.exit(1);
			}

			// Generate C structs
			program.genStructs();

			// Class indices are generated		
			program.genIndices();

			// If --with-reflection, generate data structures
			if (program.withReflection()) {
				// 	System.out.println("** generate reflection");
				program.genReflectionStructs();
			}

			// Generate prototypes
			program.genPrototypes();

			// Generate C Code	
			program.genCCode();

			// Generate initialization for attributes
			program.genInitFields();		
			// Generate C Code for init function
			program.genInit();
			// 		unit.generateClassfile();

	}


	protected void initOptions() {
		super.initOptions();
		program.options().addKeyValueOption("-o");
		program.options().addKeyOption("--deadcode-elim");
		program.options().addKeyOption("--sourcelines");
		program.options().addKeyOption("--devirtualize");
		program.options().addKeyOption("--inline"); // Not Implemented yet?
		program.options().addKeyOption("--noredundantpush");
		program.options().addKeyOption("--with-reflection");
	}

	protected void processNoErrors(CompilationUnit unit) {
		unit.transformation();
	    // Quick fix to make sure Thread is included
	    // Necessary since main() is a native thread method
	}

	protected  void printUsage() {
		printVersion();
		System.out.println(
      "\nJavaGenCCode\n\n" +
      "Usage: java JavaGenCCode <options> <source files>\n" +
      "  -verbose                  Output messages about what the compiler is doing\n" +
      "  -classpath <path>         Specify where to find user class files\n" +
      "  -sourcepath <path>        Specify where to find input source files\n" + 
      "  -bootclasspath <path>     Override location of bootstrap class files\n" + 
      "  -extdirs <dirs>           Override location of installed extensions\n" +
      "  -o <file>                 Name of resulting C file\n" +
      "  --deadcode-elim           \n" +
      "  --sourcelines             \n" +
      "  --devirtualize            \n" +
      "  --inline                  \n" +
      "  --noredundantpush         \n" +
      "  --with-reflection         \n" +
      "  -g                        \n" +
      "  -d <directory>            Specify where to place generated files\n" +
      "  -help                     Print a synopsis of standard options\n" +
      "  -version                  Print version information\n"
		);
	}

    protected String url() { return "(http://www.robot.lth.se/java)";}
	protected String name() { return "ljrt compiler"; }
	protected String version() { return "devel"; }
}
