import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;

public class ImageFrame extends JFrame implements ActionListener {

	private GUI gui;
	ImagePanel imagePanel;
	boolean firstFrame = true;
	ImageListener listener;

	public ImageFrame(GUI gui, String addr) {
		super(addr);

		this.gui = gui;

		imagePanel = new ImagePanel();
		JButton btn = new JButton("Disconnect");
		btn.addActionListener(this);

		this.getContentPane().setLayout(new BoxLayout(getContentPane(), BoxLayout.Y_AXIS));
		this.getContentPane().add(imagePanel);
		this.getContentPane().add(btn);
		this.setLocationRelativeTo(null);
		this.pack();

		onConnect(addr);
	}
	
	private void onConnect(String addr) {
		int separator = addr.indexOf(':');
		
		if (separator == -1) {
			gui.addStatusLine("malformed address string: \""+addr+'"');
			onClose();
			return;
		}

		String host = addr.substring(0, separator);
		String portStr = addr.substring(separator+1, addr.length());
		int port;

		try {
			port = Integer.parseInt(portStr);
		} catch (Exception e) {
			gui.addStatusLine("malformed address string: \""+addr+'"');
			onClose();
			return;
		}

		listener = new ImageListener(gui, this, host, port);
		listener.start();
	}

	public void onDisconnect() {
		listener.stopListening();
		try {
			listener.join();
		} catch (InterruptedException e) {
		}
		onClose();
	}

	public void onClose() {
		setVisible(false);
		dispose();
	}

	public void setImage(byte[] data) {
		imagePanel.setImage(data);
		if (firstFrame) {
			this.pack();
	    	this.setVisible(true);
			firstFrame = false;
		}
	}

	public void actionPerformed(ActionEvent a) {
		onDisconnect();
	}
}
