import java.io.*;
import java.net.*;

class ImageListener extends Thread {

	private GUI gui;
	private DatagramSocket socket;
	private ImageFrame frame;
	private String host;
	private int port;
	private boolean stop = false;

	public ImageListener(GUI gui, ImageFrame frame, String host, int port) {
		this.gui = gui;
		this.frame = frame;
		this.host = host;
		this.port = port;
	}

	public void run() {

		try {
			socket = new DatagramSocket();
			socket.connect(InetAddress.getByName(host), port);

			gui.addStatusLine("connected to "+host+":"+port);
			frame.setVisible(true);

		} catch (IOException e) {
			gui.addStatusLine("could not connect to "+host+":"+port+" ("+e+")");
			frame.onClose();
			return;
		}

		try {
			// Send hello message
			String msg = "JClient";
			DatagramPacket hello = new DatagramPacket(msg.getBytes(), msg.length());
			socket.send(hello);

			while (!shouldStop()) {
				byte[] data = new byte[32000];
				DatagramPacket packet = new DatagramPacket(data, data.length);
				socket.receive(packet);

				frame.setImage(packet.getData());
			}
		} catch (IOException e) {
			gui.addStatusLine(e.toString());
			frame.onClose();
		}
		socket.close();
	}

	public synchronized void stopListening() {
		stop = true;
	}
	
	public synchronized boolean shouldStop() {
		return stop;
	}
}
