/* vim: set noet sw=4 ts=4: */
import se.lth.cs.camera.*;
import java.io.*;
import java.net.*;

class JAxis extends Thread {
	public static void main(String[] args) {
		int port = 3800;
		if (args.length < 1) {
			System.out.println("no port specified, using default port 3800");
		} else {
 	   		port = Integer.parseInt(args[0]);
		}

		Axis211A camera = new Axis211A();
		MotionDetector md = new MotionDetector();
		camera.connect();

		ImageMonitor monitor = new ImageMonitor();

		CaptureThread cap = new CaptureThread(camera, md, monitor);
		cap.start();

		try {
			while (true) {
				System.out.println("listening on port "+port);
				DatagramSocket socket = new DatagramSocket(port);

				// Wait for new client
				DatagramPacket hello = new DatagramPacket(new byte[256], 256);
				socket.receive(hello);

				byte[] data = hello.getData();
				System.out.print("client connected: ");
				for (int i = 0; i < data.length; ++i) {
					System.out.print((char)data[i]);
				}
				System.out.println("");

				socket.connect(hello.getAddress(), hello.getPort());

				boolean motion = false;
				try {
					byte[] img = new byte[Axis211A.IMAGE_BUFFER_SIZE];
					while (socket.isConnected()) {
						int bytes = monitor.getImage(img);
						DatagramPacket packet = new DatagramPacket(img, bytes);
						socket.send(packet);
					}
				} catch (IOException e) {
					System.out.println("IOException occurred: "+e.getMessage());
				}
				socket.close();

			}
		} catch (IOException e) {
			System.out.println("IOException occurred: "+e.getMessage());
		}
	}
}

public class ImageMonitor {
	private boolean newImg = false;
	boolean motion = false;
	private byte[] data;
	private int size;

	synchronized int getImage(byte[] dest) {
		try {
			while (!newImg) wait();
			System.arraycopy(data, 0, dest, 0, size);
		} catch (InterruptedException e) {
		}
		return size;
	}

	synchronized void setImage(byte[] data, int size) {
		this.data = data;
		this.size = size;
		notifyAll();
	}

	synchronized boolean getMotion() {return motion;}
	synchronized void setMotion(boolean m) {motion = m;}
}

public class CaptureThread extends Thread {
	private ImageMonitor monitor;
	private MotionDetector motion;
	private Axis211A camera;

	public CaptureThread(Axis211A c, MotionDetector d, ImageMonitor m) {
		camera = c;
		motion = d;
		monitor = m;
	}

	public void run() {
		long t0 = System.currentTimeMillis();
		long t1;
		long delay = 500;
		byte[] data = new byte[Axis211A.IMAGE_BUFFER_SIZE];
		int size;
		try {
			while (!isInterrupted()) {
				size = Axis211A.IMAGE_BUFFER_SIZE;

				camera.getJPEG(data, 0);
				monitor.setImage(data, size);

				if (!monitor.getMotion() && motion.detect()) {
					monitor.setMotion(true);
					delay = 100;
				}

				t0 += delay;
				t1 = System.currentTimeMillis();
				if (t0 > t1) sleep(t0-t1);
			}
		} catch (InterruptedException e) {
		}
	}
}
