// Virtual robot
//import se.lth.robot.virtualirb6.*;
//import se.lth.robot.servo.*;
// Real robot
import se.lth.cs.robot.irb6.*;
import se.lth.cs.robot.servo.*;

public class Irb6SDSblock implements CascadePID_Enum {
    public static void main(String[] args) {

	float h_sample = 0.005f;
	float h_ref    = 10f;
	float h_sync   = 0.2f;

	Motors motors = new Motors();

	SDSoutput[] driveBlock = motors.getDrives();
	PosSDSblock[] posBlock = new PosSDSblock[Motors.NBR_OF_JOINTS];
	VelSDSblock[] velBlock = new VelSDSblock[Motors.NBR_OF_JOINTS];
	Synchronizer syncBlock = new Synchronizer((int)(h_sync*1000), posBlock, velBlock, motors);
	Sampler sampler = new Sampler((int)(h_sample*1000), posBlock, velBlock, driveBlock);
	float[] ref = new float[Motors.NBR_OF_JOINTS];
	ref[0] = 1000.0f;
	ref[1] = 0.0f;
	ref[2] = 0.0f;
	ref[3] = 0.0f;
	ref[4] = 0.0f;
	RefGen rg = new RefGen((int)(h_ref*1000), ref, posBlock);

	for (int i=0; i<Motors.NBR_OF_JOINTS;  i++) {

	    posBlock[i] = new PosSDSblock();
	    velBlock[i] = new VelSDSblock();

	    Parameters posp = posBlock[i].getParameters();
	    posp.val[K_] =  2.00f;
	    posp.val[Ti_] = 1.20f;
	    posp.val[Tr_] = 1.00f;
	    posp.val[Td_] = 0.00f;
	    posp.val[N_] = 10f;
	    posp.val[beta_] = 1.00f;
	    posp.val[h_] = h_sample;
	    posp.val[I_on_] = 0.0f;
	    posp.val[ndot_] = 0.0f;
	    posp.val[maxv_] = 200.0f;
	    posp.val[minv_] = -200.0f;
	    posBlock[i].setParameters(posp);

	    Parameters velp = velBlock[i].getParameters();
	    velp.val[K_] =  1.0f;
	    velp.val[Ti_] = 0.10f;
	    velp.val[Tr_] = 0.08f;
	    velp.val[Td_] = 0.0f;
	    velp.val[N_] = 10f;
	    velp.val[beta_] = 1.0f;
	    velp.val[h_] = h_sample;
	    velp.val[I_on_] = 0.0f;
	    velp.val[ndot_] = 0.0f;
	    velp.val[maxv_] = 50.0f;
	    velp.val[minv_] = -velp.val[maxv_];
	    velBlock[i].setParameters(velp);

	    velBlock[i].addListener(driveBlock[i]);
	}

	sampler.start();
	rg.start();

	syncBlock.initSync();
	syncBlock.doSync();

	System.out.println("All joints synched!");

	rg.startGenerating(syncBlock.getSyncPos());
    }
}
