// Virtual robot
//import se.lth.robot.virtualirb6.*;
//import se.lth.robot.servo.*;
// Real robot
import se.lth.cs.robot.irb6.*;
import se.lth.cs.robot.servo.*;

public class Synchronizer {

    private PosSDSblock[] posBlock;
    private VelSDSblock[] velBlock;
    private Motors motors;

    private boolean[] sync;
    private JointSyncher[] syncs;
    private float[] syncPos;

    private Sample[] theSample;

    private int interval = 200;
    private long nextTime;

    
    public Synchronizer(int period, PosSDSblock[] iposBlock, VelSDSblock[] ivelBlock, Motors imotors) {

	interval = period;
	
	posBlock = iposBlock;
	velBlock = ivelBlock;
	motors = imotors;      

	sync = new boolean[Motors.NBR_OF_JOINTS];
	syncs = new JointSyncher[Motors.NBR_OF_JOINTS];
	syncPos = new float[Motors.NBR_OF_JOINTS];

	theSample = new Sample[Motors.NBR_OF_JOINTS];
	for (int i=0; i<theSample.length; i++) {
	    theSample[i] = new Sample();
	}
    }

    
    public void initSync() {
	for (int i=0; i<sync.length; i++) {
	    sync[i] = false;
	    syncs[i] = new JointSyncher(i);
	}
    }


    public void doSync() {
	nextTime = System.currentTimeMillis() + interval;
	while (!allSync()) {

	    for (int i=0;i<sync.length;i++) {

		if (!sync[i]) {

		    // Check if we passed synch point and
		    // find reference to sync joint i
		    boolean tmp = sync[i];
		    sync[i] = syncs[i].sync();
		    System.out.println("Joint " + i + ": " + tmp + " -> " + sync[i]);

		    // Sending velocity reference to velBlock
		    theSample[i].siso = syncs[i].getReference();
		    System.out.println("sync ref joint " + i + "= "+theSample[i].siso); 
		    velBlock[i].reference(theSample[i]);

		    if(sync[i]) {

			// Joint is synched
			motors.reset(i+1);
			motors.analogOut(i+1, 0); 

			System.out.print("Joint " + i + " synchronized");

			// Set reference to keep joint still until all joints
			// are synchronized
			syncPos[i] = motors.getPosition(i+1);
			theSample[i].siso = syncPos[i];
			posBlock[i].reference(theSample[i]);
			posBlock[i].addListener(velBlock[i]);

			if(allSync()) { 
			    // This is done once when all joints are synched
			    motors.setSafeBit();
			}
		    }
		}
	    }
	    waitNextPeriod();
	}
    }

    
    private void waitNextPeriod() {
	long time = System.currentTimeMillis();
	long sleepTime = (nextTime - time);
	nextTime = nextTime + interval;
	try {
	    if(sleepTime > 0) {
		Thread.sleep(sleepTime);
	    }
	} catch(Exception e) {
	    System.out.println("Exception in Synchronizer.waitNextPeriod"); 
	}
    }


    public float[] getSyncPos() {
	return syncPos;
    }

    
    private boolean allSync() {
	boolean allSync = true;
	for (int i=0;i<sync.length;i++) {
	    if (allSync && !sync[i])
		allSync = false;
	}
	return allSync;
    }

    
    /**
     * Inner class handling synchronization of a single joint
     */
    private class JointSyncher {

	private int joint;
	private float ref;
	private boolean syncFromProperSide;

	
	public JointSyncher(int ijoint) {
	    joint = ijoint;
	    syncFromProperSide = getSyncSwitch(joint);
	}

	
	public float getReference(){
	    return ref;
	}

	
	public boolean sync() {
	    boolean sync;

	    if (!syncFromProperSide) { // joint was on the wrong side
		if(getSyncSwitch(joint)) {
		    sync = false;
		    //passed switch: now at proper side
		    syncFromProperSide = true;
		    ref = 0.0f;
		} else {
		    //still on wrong side
		    sync = false;
		    ref = -2*syncSpeed(joint);
		}
	    } else {
		if(!getSyncSwitch(joint)) {
		    ref = 0.0f;
		    sync = true;

		} else {
		    sync = false;
		    //drive towards syncPos
		    ref = syncSpeed(joint);
		}
	    }

	    return sync;
	}

	
	private float syncSpeed(int j){
	    float speed = 4.0f;
	    switch (j) {
		case 0 : return speed;
		case 1 : return -speed;
		case 2 : return -speed;
		case 3 : return speed;
		case 4 : return -speed; 
	    }
	    return 0;	
	}

	
	private boolean getSyncSwitch(int j) {

	    boolean tmp;
	    tmp = motors.checkSync(j);
	    // System.out.println("getSyncSwitch(" + j + "): " + tmp);
	    return !tmp;  // HERE BE DRAGONS! sign mismatch between
	    // the java and modula code.
	}


	/* Read the sync switch nbrSamples times, until they are
	 * all the same, in order to avoid glitches.
	 * It was in the modula code, is it necessary? */
	private boolean getSyncSwitch1(int j) {

	    boolean allsame;
	    final int nbrSamples = 3;
	    boolean tmp[] = new boolean[nbrSamples];

	    do {
		int i;
		for(i=0; i<nbrSamples; i++) {
		    tmp[i] = motors.checkSync(j);
		}
		allsame = true;
		for(i=0; i<nbrSamples-1; i++) {
		    allsame = allsame && (tmp[i] == tmp[i+1]);
		}

		//if(!allsame) {
		//	System.out.println("getSync: not all same!");
		//}
	    } while (!allsame);

	    return !tmp[0];
	}
    }
}
