package regtest;

import regtest.paket.*;

public class ArrayTest extends TestCase {

	private static int INDEX = 2;
	private static int[] sPA;
	private int[] aPA;

	private int ix;

	private static int errCount = 0;

	public static int testAll() {
		System.out.println("  ArrayTest start");
		testAlloc();
		testIndex();
		testStaticIndex();
		testSumStatic();
		testMultiDim();
		testComplexIndex();
		testCreate();
		testInitializers();
		testAsObject();
		System.out.println("  ArrayTest finished");
		return errCount;
	}

	private static void testIndex() {
		int err = 0;
		int ix = 2;
		sPA = new int[5];
		sPA[ix] = 3;
		if (sPA[ix] != 3) {
			System.out.println("    testIndex failed");
			err++;
		}
		errCount += err;
	}

	private static void testStaticIndex() {
		int err = 0;
		sPA = new int[5];
		sPA[INDEX] = 3;
		if (sPA[INDEX] != 3) {
			System.out.println("    testStaticIndex failed");
			err++;
		}
		errCount += err;
	}

	private static void testAlloc() {
		int err = 0;
		sPA = new int[5];
		if (sPA.length != 5) {
			System.out.println("    testAlloc failed for static attribute array");
			System.out.print("      sPA.length = ");
			System.out.print(sPA.length);
			System.out.println(" != 5");
			err++;
		}
		int[] loc = new int[6]; 
		if (loc.length != 6) {
			System.out.println("    testAlloc failed for local array");	    
			System.out.print("      loc.length = ");
			System.out.print(sPA.length);
			System.out.println(" != 6");
			err++;
		}
		ArrayTest at = new ArrayTest();
		at.aPA = new int[7];
		if (at.aPA.length != 7) {
			System.out.println("    testAlloc failed for attribute array");	    
			System.out.print("      at.aPA.length = ");
			System.out.print(sPA.length);
			System.out.println(" != 7");
			err++;
		}
		errCount += err;
	}

	private static void testSumStatic() {
		int err = 0;
		sPA = new int[5];
		for (int i = 0; i<5; i++) {
			sPA[i] = i;
		}
		int sum = 0;
		for (int i = 0; i<5; i++) {
			sum += sPA[i];
		}
		if (sum != 10) {
			err++;
		}

		if (err > 0) {
			System.out.println("    testSumStatic failed!");
			errCount += err;
		}
	}

	private static void testMultiDim() {
		int err = 0;
		int x = 0;
		int a[][] = new int[3][3];
		for (int i=0; i<3; i++) {
			for (int j=0; j<3; j++) {
				a[i][j] = x;
				x++;
			}
		}
		if (a[1][1] != 4) {
			err++;
		}
		if (err > 0) {
			System.out.println("    testMultiDim failed: expected 4, was "+a[1][1]);
			errCount += err;
		}

		int init_2dim[][] = { {1, 2}, null };
		errCount += check(init_2dim[0][1] == 2,
				"    testMultiDim failed(2)\n"+
				"      init_2dim[0][1] = "+init_2dim[0][1]+" (2)");
		errCount += check(init_2dim[1] == null,
				"    testMultiDim failed(3)\n"+
				"      init_2dim[1] = "+init_2dim[1]+" (null)");
	}

	private static void testComplexIndex() {
		ArrayTest at = new ArrayTest();
		int array[] = new int[3];
		at.ix = 1;
		array[at.ix++] = at.ix;
		int i = at.ix-1;
		errCount += check(array[i] == 2,"    testComplexIndex failed.\n"+
				"      array["+i+"]="+array[at.ix-1]+" (array[1]=2)");
	}

	private static void testCreate() {
		Klass[] klass = new regtest.paket.Klass[10];
		errCount += check(klass.length == 10,"    testCreate failed.\n"+
				"      klass.length = "+klass.length+" (klass.length=10)");
	}

	private static void testInitializers() {
		int[] a = {1,2,3};
		errCount += check(a[1]==2,"    testInitializers failed 1.\n"+
				"      a[1] = "+a[1]+" (2)");
		ArrayKlass aa = new ArrayKlass();

		// Make below lines compile ASAP!

		errCount += check(aa.aMethod(new int[] {4,5,6,7}) == 4,
				"      testInitializers failed 2.\n"+
				"      aa.aMethod(new int[] {4,5,6,7}) = "+
				aa.aMethod(new int[] {4,5,6,7})+" (4)");
		errCount += check(aa.ka[3]==4,"    testInitializers failed 3.\n"+
				"      aa.ka[3] = "+
				aa.ka[3]+" (4)");
		errCount += check(ArrayKlass.typeString[2].equals("pos"),
				"    testInitializers failed 4.\n"+
				"      ArrayKlass.typeString[2].equals(pos) = "+
				ArrayKlass.typeString[2].equals("pos")+" (true)");
	}

	private static void testAsObject() {
		ArrayKlass x = new ArrayKlass();
		ArrayKlass y = x;
		int[] a = {1,2,3};
		int[] b = {0,0}; // (int[]) a.clone(); Lacks implementation		
		b[1] = 4;
		errCount += check(b[1]==4,"    testAsObject failed 1.\n"+
				"      b[1] = "+b[1]+" (4)");
		// Not useful unless someone implements java.lang.Object.clone()
		// 	errCount += check(a[1]==2,"    testAsObject failed 2.\n"+
		// 			  "      a[1] = "+a[1]+" (2)");
	}

}

class ArrayKlass {
    public int aMethod(int[] ii) {
		return ii.length;
    }

    public int obs = 3;
    public int[] ka = {1,2,3,4,5};
    public static String[] typeString = { "trq", "vel", "pos", "volt", "curr",
										  "sync", "asyn", "pwm1", "pwmm", "step" };
}
