/**
 * Copyright (C) 2006-????  Anders Nilsson <anders.nilsson@cs.lth.se>
 *                    2010  Jesper Öqvist  <jesper.oqvist@cs.lth.se>
 * 
 * This file is part of ljrt.
 *
 */
package regtest;

import java.io.*;

public class ExceptionTest extends TestCase {

	private static int errCount = 0;

	public static int testAll() {
		// TODO: Add tests for abrupt termination of catch clauses!
		System.out.println("  ExceptionTest start");
		testEmptyCatch();
		testEmptyFinally();
		testSimpleException();
		testExceptionControlFlow1();
		testBreak();
		testControlFlow1();
		testExceptionObject();
		System.out.println("  ExceptionTest finished");
		return errCount;
	}

	public static void testSimpleException() {
		T t = new T();
		errCount = t.throwIt();
		if (errCount == 2) {
			System.out.println("	try not executed");
		} else if (errCount == 1) {
			System.out.println("	catch not executed");
		}
	}

	public static void testExceptionControlFlow1() {
		int i = 0;
		try {
			i += 1;
			throw new Exception();
		} catch (Exception e) {
			i += 2;
		} finally {
			i += 4;
		}
		errCount += check(i == 7,"	  testExceptionControlFlow1 failed 1.\n"+
						  "		 i = "+i+" (7)");			
		i = 0;
		try {
			i += 1;
		} catch (Exception e) {
			i += 2;
		} finally {
			i += 4;
		}
		errCount += check(i == 5,"	  testExceptionControlFlow1 failed 2.\n"+
						  "		 i = "+i+" (5)");			
		i = 0;
		try {
			i += 1;
		} finally {
			i += 4;
		}
		errCount += check(i == 5,"	  testExceptionControlFlow1 failed 3.\n"+
						  "		 i = "+i+" (5)");			
		i = 0;
		try {
			try {
				i += 1;
				throw new Exception();
			} finally {
				i += 4;
			}
		} catch (Exception e) {
			i += 8;
		}
		errCount += check(i == 13,"    testExceptionControlFlow1 failed 4.\n"+
						  "		 i = "+i+" (13)");			
	}

	/**
 	 * This is just to test if there are any segfaults / other runtime
 	 * issues generated from an empty catch clause.
 	 */
	public static void testEmptyCatch() {
		try {
			throw new Exception();
		} catch (Exception e) {
		}
	}

	/**
 	 * This is just to test if there are any segfaults / other runtime
 	 * issues generated from an empty catch finally block.
 	 */
	public static void testEmptyFinally() {
		try {
		} finally {
		}
	}

	/**
 	 * Test breaking out of catch clauses or try blocks.
 	 */
	public static void testBreak() {
		int value = 0;
		while (true) {
			value += 1;
			try {
				break;
			} finally {
			}
		}
		errCount += check(value == 1, "    testBreak failed (1)");

		value = 0;
		while (value < 100) {
			try {
				value += 10;
				break;
			} finally {
				value = 1;
			}
		}
		errCount += check(value == 1, "    testBreak failed (2)");

		value = 0;
		while (value < 100) {
			value += 1;
			try {
				throw new Exception();
			} catch (Exception e) {
				break;
			} finally {
			}
		}
		errCount += check(value == 1, "    testBreak failed (3)");

		value = 0;
		while (value < 100) {
			value += 1;
			try {
				throw new Exception();
			} catch (Exception e) {
			} finally {
				break;
			}
		}
		errCount += check(value == 1, "    testBreak failed (3)");
	}

	// FIXME
	public static int testControlFlow1() {
loop1:
		for (int count = 0; count < 10; ++count) {
			try {
				int i = 1+2;
				while (true) {
					if (i > 0) {
						//System.out.println("returning");
						return 4;
					} else if (i < 0)
						continue;
					continue loop1;
				}
			} catch (Error e) {
			} catch (Exception e) {
			} finally {
				//System.out.println("in finally");
				//return 3;
			}
			int k = 4;
		}
		return -1;
	}

	public static void testExceptionObject() {
		int i = 13;
		try {
			throw new TestException();
		} catch (TestException e) {
			i -= e.a;
		}
		errCount += check(i == 8,"	  testExceptionObject failed 1.\n"+
						  "		 i = "+i+" (8)");	
		i = 12;
		try {
			throw new SubTestException();
		} catch (TestException e) {
			i -= e.a;
		}
		errCount += check(i == 7,"	  testExceptionObject failed 2.\n"+
						  "		 i = "+i+" (7)");	
	}

	private static class TestException extends Exception {
		int a = 5;
	}

	private static class SubTestException extends TestException {
	}

	private static class T	{
		public int throwIt() {
			int result = 2;
			try {
				result -= 1;
				TestException e = new TestException();
				//		System.out.println("Throwing exception: "+result);
				throw e;
				//		System.out.println("After throw: should never come here");
			} catch (TestException e) {
				//		System.out.println("Caught exception: "+result);
				result -= 1;
			}
			//		System.out.println("Returning result: "+result);
			return result;
		}
	}

}
