/**
 * Copyright (C) 2009-2011 Jesper Öqvist <jesper.oqvist@cs.lth.se>
 *
 * This file is part of ljrt.
 */
package regtest;

public class ForTest extends TestCase {

	private static int errCount = 0;

	public static int testAll() {
		System.out.println("  ForTest start");
		testFor();
		testForEach();
		errCount += testContinue();
		errCount += testBreak();
		errCount += testDereference();
		errCount += testMutliUpdate();
		errCount += testMultiInit();
		errCount += testEmpty();
		errCount += testGCPopRoot();
		System.out.println("  ForTest finished");
		return errCount;
	}

	/**
 	 * Test case for basic for loops.
 	 */
	private static void testFor() {
		for (int i = 0; i < 7; ++i) {
			if (i != 0) {
				errCount++;
				System.out.println("    testFor failed(1)");
				break;
			}
			i = 7;
		}

		int[] intArray = new int[] { 2, 4, 16 };
		int sum = 0;
		for (int i = 0; i < intArray.length; ++i)
			sum += intArray[i];
		errCount += check(sum == 22, "    testFo failed(2)\n"+
				"      sum == "+sum+" (expected 22)");

		int i;
		for (i = 0; i < 13; ++i) {}
		errCount += check(i == 13, "    testFor failed(3)");

	}

	/**
 	 * Test case for enhanced for statement.
 	 *
 	 * Test iterating over a simple array.
 	 */
	private static void testForEach() {
		int[] intArray = new int[] { 2, 4, 16 };
		int sum = 0;
		for (int val : intArray)
			sum += val;
		errCount += check(sum == 22, "    testForEach failed(1)\n"+
				"      sum == "+sum+" (expected 22)");
	}

	private static int testContinue() {
		int k = 0;
		for (int i = 0; i < 10; ++i) {
			if (k >= 10) {
				System.out.println("    testContinue failed(1)");
				return 1;
			}
			k += 1;
			continue;
		}
		k = 0;
		for (int i = 0; i < 10; ++i) {
			if (i < 10) continue;
			k = 10;
		}
		if (k != 0) {
			System.out.println("    testContinue failed(2)");
			return 1;
		}
		return 0;
	}

	private static int testBreak() {
		int i;
		for (i = 0; i < 10; ++i) {
			if (i == 3) break;
		}
		if (i != 3) {
			System.out.println("    testBreak failed");
			return 1;
		}
		return 0;
	}

	private static int testDereference() {
		ForTestClass obj = new ForTestClass();
		for (; true; obj.a = obj.a +1) {
			++obj.a;
			break;
		}
		if (obj.a != 4) {
			System.out.println("    testDereference failed");
			return 1;
		}
		return 0;
	}

	private static int testMutliUpdate() {
		int a = 0;
		for (int i = 0; i < 10; ++i, a+=3) {
		}
		if (a != 30) {
			System.out.println("    testMutliUpdate failed");
			return 1;
		}
		return 0;
	}

	private static int testMultiInit() {
		int j = 0;
		for (int i = 0, a = 3; i > -3; i--) {
			j = i;
		}
		if (j != -2) {
			System.out.println("    testMultiInit failed");
			return 1;
		}
		return 0;
	}

	private static int testEmpty() {
		int i;
		for (i = 0; i < 10; ++i);
		if (i != 10) {
			System.out.println("    testEmpty failed");
		}
		return 0;
	}

	private static int testGCPopRoot() {
		// The translated code for this method needs to be manually inspected
		// to confirm that GC commands are inserted in their proper locations.
		for ( ; true; ) {
			ForTestClass obj = new ForTestClass();
			break;
		}
		for (int i = 0; i < 10; ++i) {
			ForTestClass obj = new ForTestClass();
			continue;
		}
		return 0;
	}

}

class ForTestClass {
	public int a = 3;
	public int b = 4;
}

