/**
 * Copyright (C) 2009-2010 Jesper Öqvist <jesper.oqvist@cs.lth.se>
 *
 * This file is part of ljrt.
 */
package regtest;

/**
 * Unit tests for generics designed to fit in the Ljrt/regressiontests test suite.
 *
 * TODO: could use a lot more tests here!
 */
public class GenericsTest extends TestCase {

	private static int errCount = 0;

	public static int testAll() {
		System.out.println("  GenericsTest start");
		testStack();
		testTypeBound();
		System.out.println("  GenericsTest finished");
		return errCount;
	}

	private static void testStack() {
		Stack<String> stack = new Stack();
		String str1 = new String("item 1");
		String str2 = new String("item 2");
		String str3 = new String("object 3");

		stack.push(str1);
		stack.push(str2);
		stack.push(str3);

		errCount += check(stack.pop() == str3, "    testStack failed (1)");
		errCount += check(stack.pop() == str2, "    testStack failed (2)");
		errCount += check(stack.pop() == str1, "    testStack failed (3)");
		errCount += check(stack.pop() == null, "    testStack failed (4)");
	}

	private static void testTypeBound() {
		GenericTypeAContainer<TypeB> container;
		
		container = new GenericTypeAContainer(new TypeB());

		errCount += check(container.myA.identify().equals("B"),
				"    testTypeBound failed");
	}

}

public class Stack <T> {
	Item<T> next;

	public Stack() {
		next = null;
	}

	public T pop() {
		T result = null;

		if (next != null) {
			result = next.data;
			next = next.next;
		}
		return result;
	}

	public void push(T d) {
		Item<T> item;
 	       
		item = new Item<T>(d);
		item.next = next;
		next = item;
	}


}

public class Item<T> {
	public T data;
	public Item<T> next;

	public Item(T d) {
		data = d;
		next = null;
	}
}

class TypeA {
	public String identify() {return new String("A");}
}

class TypeB extends TypeA {
	public String identify() {return new String("B");}
}

public class GenericTypeAContainer<T extends TypeA> {
	public T myA;

	public GenericTypeAContainer(T a) {
		myA = a;
	}
}
