/**
 * Copyright (C) 2006-????  Anders Nilsson <anders.nilsson@cs.lth.se>
 *                    2011  Jesper Öqvist  <jesper.oqvist@cs.lth.se>
 * 
 * This file is part of ljrt.
 *
 */
package regtest;

import regtest.paket.*;

public class MethodTest extends TestCase{

    private static int errCount = 0;

    public static int testAll() {
		System.out.println("  MethodTest start");
		testStaticCall();
		testDismissedRetVal();
		testSequenceCall();
		testConstructorArg();
		testQCEArg();
		testParenthesisCall();
		testStringParamConcat();
		testVarArgs();
		System.out.println("  MethodTest finished");
		return errCount;
    }

    private static void testStaticCall() {
		int err = 0;
		vsMethod();
		Klass.vsMethod();
		int i = 0;
		i = sMethod(i);
		if (i != 1) {
			System.out.println("    testStaticCall failed.");
			err++;
		}
    }

    private static void testDismissedRetVal() {
		int err = 0;
		int i = 0;
		Object o = refMethod();
		refMethod();
		if (i != 0) {
			System.out.println("    testDismissedRetVal failed.");
			err++;
		}
    }

    private static void testSequenceCall() {
		int err = 0;
		int i = 0;
		i = new Klass().intMethod();
		if (i != 3) {
			System.out.println("    testSequenceCall 1 failed.");
			err++;
		}
		i = 0;
		i = refMethod().hashCode();
		if (i == 0) {
			System.out.println("    testSequenceCall 2 failed.");
			err++;
		}
    }

    private static void testQCEArg() {
		errCount += check(qceMethod(1>0 ? "foo" : "bar") == 1,"    testQCEArg failed.");
    }

    private static void testConstructorArg() {
		int i = qceMethod("foo");
		errCount += check(i==1,"    testQCEArg failed 1. i="+i+" (1)");
		i = qceMethod(new String("bar"));
		errCount += check(i==2,"    testQCEArg failed 2. i="+i+" (2)");
    }

    private static void testParenthesisCall() {
		byte[] bytes;
// 		String ss = new String("bar");
// 		bytes = (ss).getBytes();
 		bytes = (new String("bar")).getBytes();
		errCount += check(bytes[1] == 'a',"    testParenthesisCall failed.\n"+
						  "      bytes[1] == "+(char) bytes[1]+" (a)");
    }
    
    private static void testStringParamConcat() {
    	Dummy dummy = new Dummy();
    	String str = dummy.op("Call " + dummy.get()); 
    	String ref = "Call OK";
    	errCount += check(str.equals(ref),
    	                  "    testStringParamConcat failed.\n" +
    	                  "      expected \"" + ref + "\",\n" +
    	                  "      got \"" + str + "\".");
    }

    private static int qceMethod(String s) {
		if (s.equals("foo")) {
			return 1;
		} else {
			return 2;
		}
    }

    private static int sMethod(int a) {
		return a+1;
    }

    private static void vsMethod() {
    }

    private static Object refMethod() {
		return new Object();
    }
    
	public static class Dummy {
		public String get() {
			return "OK";
		}
		
		public String op(String str) {
			return str;
		}
	}

	private static void testVarArgs() {
		int sum = addNumbers(2, 4, 16);
		errCount += check(sum == 22, "    testVarArgs failed (1)\n");
	}

	private static int addNumbers(int ... numbers) {
		int sum = 0;
		for (int i = 0; i < numbers.length; ++i) {
			sum += numbers[i];
		}
		return sum;
	}
}
