package regtest;

public class StmtTest extends TestCase {

	private static int[] a;
	private static int errCount = 0;
	private int tmpi;
	private static int TMP;


	public static int testAll() {
		System.out.println("  StmtTest start");
		testCond();
		testBoolean();
		testWhile();
		testDo();
		testFor();
		testSwitch();
		testIfNoBlock();
		testIfElse();
		testReturn();
		System.out.println("  StmtTest finished");
		return errCount;
	}

	private static void testCond() {
		int err = 0;
		int i = 0;
		StmtTest t;
		t = new StmtTest();
		if (t.getOne() != 1) {
			System.out.println("    testCond faild in simple virtual method.");
		}
		if (t.getBool()) {
			System.out.println("    testCond failed in simple virtual method.");
			err++;
		}
		if (getStaticBool()) {
			System.out.println("    testCond failed in simple static method.");
			err++;
		}
		errCount += err;
	}


	private static void testBoolean() {
		int err = 0;
		boolean flag = 3 < 4;
		if (!flag) {
			System.out.println("    testBoolean failed(1)");
			err++;
		}
		flag = 3 > 4;
		if (flag) {
			System.out.println("    testBoolean failed(2)");
			err++;
		}
		flag = (3 == 4);
		if (flag) {
			System.out.println("    testBoolean failed(3)");
			err++;
		}
		int a = 4;
		flag = (a == 4);
		if (!flag) {
			System.out.println("    testBoolean failed(4)");
			err++;
		}
		errCount += err;
	}



	private static void testWhile() {
		int err = 0;
		a = new int[5];
		a[4] = 1;
		int i = 0;
		while (a[i] == 0) {
			i++;
		}
		errCount += check(i == 4,"    testWhile failed 1.\n"+
				"      i = "+i+" (4)");

		StmtKlass sk = new StmtKlass();
		i = 5;
		while (sk.cont()) {
			i++;
		}
		errCount += check(i == 5,"    testWhile failed 2.\n"+
				"      i = "+i+" (5)");

	}

	private static void testDo() {
		int i;
		StmtKlass sk = new StmtKlass();
		i = 5;
		do {
			i++;
		} while (sk.cont());
		errCount += check(i == 6,"    testDo failed 1.\n"+
				"      i = "+i+" (6)");

	}


	private static void testFor() {
		int err = 0;
		a = new int[5];
		a[4] = 1;
		int i = 0;
		for (int j=0; a[j] == 0; j++) {
			// 	    System.out.print(a[i]);
			// 	    System.out.print(" ");
			i++;
		}
		if (i != 4) {
			System.out.println("    testFor failed.");
			err++;
		}
		errCount += err;
	}

	private static void testSwitch() {
		int err = 0;
		a = new int[5];
		a[3] = 1;
		int i = 0;
		int k = 0;
		StmtTest t = new StmtTest();
		switch (a[3]) {
			case 0: 
				i=0;
				break;
			case 2:
				k=i*t.getOne();
				i=1;
				break;
			default:
				i = 4;
				a[i] = a[i-1];
		}
		if (a[i] != 1) {
			System.out.print("    testSwitch failed 1. a[i] = ");
			System.out.print(a[i]);
			System.out.println(" (!=1");
			err++;
		}
		errCount += err;
		i = 2;
		switch (i) {
			case StaticClass.A: 
				i=5;
				break;
			case StaticClass.B: 
				i=6;
				break;
			case StaticClass.C: 
				i=7;
				break;
		}
		errCount += check(i == 6,"    testSwitch failed 2.\n"+
				"      i = "+i+" (6)");	

		i = 2;
		StmtKlass sk = new StmtKlass();
		i = sk.caseTest(i);
		errCount += check(i == 7,"    testSwitch failed 3.\n"+
				"      i = "+i+" (7)");
	}

	private static void testIfNoBlock() {
		int err = 0;
		int tmp = 0;
		StmtTest st = new StmtTest();
		// Simple if-statement
		if (err==0)
			st.tmpi++;
		if (st.tmpi != 1) {
			System.out.println("    testIfNoBlock failed in 1.");
			err++;
		}

		// if () else if
		st.tmpi = 0;
		if (st.tmpi==1)
			st.tmpi += 1; // Phony, shoudn't occur
		else if (st.tmpi==0)
			st.tmpi += 5;
		if (st.tmpi != 5) {
			System.out.println("    testIfNoBlock failed in 2.");
			err++;
		}
		errCount += err;
	}

	private static void testIfElse() {
		// Don't know if this test is motivated, AndersN
		int err = 0;
		TMP = 0;
		// Simple if-statement
		if (err == 0) {
			TMP++;
		}
		if (TMP != 1) {
			System.out.println("    testIfElse failed in 1.");
			err++;
		}

		// if () else if
		if (TMP == 255) {
		}
		else {
			int i = tmpMethod(TMP); 
			{
				TMP++;	    
			}
		}
		if (TMP != 2) {
			System.out.println("    testIfElse failed in 2.");
			err++;
		}
		errCount += err;
	}



	private static int tmpMethod(int i) {
		if (i > 0) {
			return (int) tmpMethod(i-1)+2;
			// 	    return tmpMethod(i-1)+2;
		} else {
			return 0;
		}
	}

	private static void testReturn() {
		int err = 0;
		int i = 0;
		i = tmpMethod(tmpMethod(2));
		if (i != 8) {
			System.out.println("    testReturn failed 1.");
			System.out.print("      tmpMethod(tmpMethod(2) = ");
			System.out.print(i);
			System.out.println(" != 8");
			err++;
		}
		errCount += err;
		StmtKlass s = new StmtKlass();
		i = s.aMethod();
		errCount += check(i == 13,"    testReturn failed 2.\n"+
				"      i = "+i+" (13)");
		errCount += check(s.a == 13,"    testReturn failed 3.\n"+
				"      s.a = "+s.a+" (13)");
		errCount += check(s.b == 13,"    testReturn failed 4.\n"+
				"      s.b = "+s.b+" (13)");
		StmtKlassGrandChild sk = new StmtKlassGrandChild();
		errCount += check(sk.bMethod() == 13,"    testReturn failed 5.\n"+
				"      sk.bMethod() = "+sk.bMethod()+" (13)");
	}

	private int getOne() {
		return 1;
	}

	private boolean getBool() {
		return false;
	}

	private static boolean getStaticBool() {
		return false;
	}

	class StaticClass {
		public final static int A = 1;
		public final static int B = 2;
		public final static int C = 3;	    
	}

}

class StmtKlass {
    protected volatile int a,b;
    private final int A=0,B=1,C=2;
    public int aMethod() {
		b = 13;
		return a = b;
    }
    public int caseTest(int i) {
        switch (i) {
        case A: 
            i=5;
            break;
        case B: 
            i=6;
            break;
        case C: 
            i=7;
            break;
        }
        return i;
    }

	public boolean cont() {
		return false;
	}
}

class StmtKlassChild extends StmtKlass {
    
}

class StmtKlassGrandChild extends StmtKlassChild {
    public int bMethod() {
		b = a = 13;
		return b;
    }    
}
