package regtest;

public class StringTest extends TestCase {

	private static int errCount = 0;

	public static int testAll() {
		System.out.println("  StringTest start");
		testSimpleString();
		testLength();
		testAddString();
		testAddPrimitive();
		testConcat();
		testConcatSpecial();
		testConcatNull();
		System.out.println("  StringTest finished");
		return errCount;
	}

	private static void testSimpleString() {
		int err = 0;
		String str = new String("simple");
		if (!str.equals("simple")) {
			System.out.println("    testSimpleString failed(1)");
			err += 1;
		}
		str = "simple2";
		if (!str.equals("simple2")) {
			System.out.println("    testSimpleString failed(2)");
			err += 1;
		}
		str = "simple3".toString();
		if (!str.equals("simple3")) {
			System.out.println("    testSimpleString failed(3)");
			err += 1;
		}
		errCount += err;
	}

	private static void testCharAt() {
		int err = 0;
		if ("fdgfda72".charAt(5) != 'a') {
			System.out.println("    testCharAt");
			err += 1;
		}
		errCount += err;
	}

	private static void testLength() {
		int err = 0;
		if ("".length() != 0) {
			System.out.println("    testLength failed(1)");
			err++;
		}
		if ("abcdefg".length() != 7) {
			System.out.println("    testLength failed(2)");
			err++;
		}
		errCount += err;
	}

	private static void testAddString() {
		int err = 0;
		String s = new String("Hello");
		s = s+" World";
		errCount += check(s.equals("Hello World"),"    testAddString failed.\n"+
				s+" (Hello World)");
	}

	private static void testAddPrimitive() {
		int err = 0;
		String s = new String(" Hello ");
		String ss;
		ss = s+10;
		errCount += check(ss.equals(" Hello 10"),"    testAddPrimitive failed int 1.\n"+
				"      "+ss+" ( Hello 10)");
		ss = 10+s;
		errCount += check(ss.equals("10 Hello "),"    testAddPrimitive failed int 2.\n"+
				"      "+ss+" (10 Hello)");
		long l = 112233445566L;
		ss = l+s;
		errCount += check(ss.equals("112233445566 Hello "),"    testAddPrimitive failed long.\n"+
				"      "+ss+" (112233445566 Hello )");
		boolean b = true;
		ss = s+b;
		errCount += check(ss.equals(" Hello true"),"    testAddPrimitive failed boolean.\n"+
				"      "+ss+" ( Hello true)");
		float f = 3.14F;
		ss = s+f;
		errCount += check(ss.equals(" Hello 3.14"),"    testAddPrimitive failed float.\n"+
				"      "+ss+" ( Hello 3.14)");
		double d = 3.141592653589793;
		ss = s+d;
		errCount += check(ss.equals(" Hello 3.141592653589793"),
				"    testAddPrimitive failed double.\n"+
				"      "+ss+" ( Hello 3.141592653589793)");
		errCount += err;
		char c = '!';
		ss = s+c;
		errCount += check(ss.equals(" Hello !"),"    testAddPrimitive failed char.\n"+
				"      "+ss+" ( Hello !)");
		errCount += err;
	}

	private static void testConcat() {
		String str;

		str  = "Sun "+"and Java";
		errCount += check(str.equals("Sun and Java"),
				"    testConcat failed(1)");

		str = "sitting "+"in "+"a "+"tree";
		errCount += check(str.equals("sitting in a tree"),
				"    testConcat failed(2)");

		str = "K."+"I."+"S."+"S."+"I."+"N."+"G.";
		errCount += check(str.equals("K.I.S.S.I.N.G."),
				"    testConcat failed(3)");

		str = 1 + "two";
		errCount += check(str.equals("1two"),
				"    testConcat failed(4)");

		str = 1 + (2 + "three");
		errCount += check(str.equals("12three"),
				"    testConcat failed(5)");

		str = "Just ";
		str += 1;
		str += " more!";
		errCount += check(str.equals("Just 1 more!"),
				"    testConcat failed(6)\n"+
				"      expected \"Just 1 more!\",\n"+
				"      got \""+str+"\"");
	}

	private static void testConcatSpecial() {
		String str;

		str = " \n\t\b\"" + "\"\"\r\n";
		errCount += check(str.equals(" \n\t\b\"\"\"\r\n"),
				"    testConcatSpecial failed");
	}

	private static void testConcatNull() {
		String s = null;
		s = "World is "+s;
		errCount += check(s.equals("World is null"),"    testAddString failed.\n"+
				s+" (Hello World)");
	}


}
