package regtest;

public class SwitchTest {

	public static int testAll() {
		int errCount = 0;
		System.out.println("  SwitchTest start");
		errCount += testEmpty();
		errCount += testCase();
		errCount += testDefault();
		errCount += testFallthrough();
		errCount += testFieldLabel();
		errCount += testExpression();
		errCount += testDeclaration();
		errCount += testNested();
		errCount += testGCPopRoot();
		System.out.println("  SwitchTest finished");
		return errCount;
	}

	private static int testEmpty() {
		switch (0) {
			case 1:
			case 2:
			case 3:
		}
		return 0;
	}

	private static int testCase() {
		int b = 4;
		switch (3) {
			case 1:
				b = 8;
				break;
			case 2:
				b = -2;
				break;
			case 3:
				b = 2;
				break;
			case 4:
				b = 4;
				break;
			default:
				b = 0;
		}
		if (b != 2) {
			System.out.println("    testCase failed");
			return 1;
		}
		return 0;
	}

	private static int testFieldLabel() {
		int b = 4;
		switch (3) {
			case SwitchTestClass.A:
				b = 8;
				break;
			case SwitchTestClass.B:
				b = -2;
				break;
			case SwitchTestClass.C:
				b = 2;
				break;
			case SwitchTestClass.D:
				b = 4;
				break;
			default:
				b = 0;
		}
		if (b != 2) {
			System.out.println("    testFieldLabel failed.");
			return 1;
		}
		return 0;
	}


	private static int testDefault() {
		int b = 4;
		switch (120) {
			case 1:
				b = 8;
				break;
			case 2:
				b = -2;
				break;
			case 3:
				b = 2;
				break;
			case 4:
				b = 4;
				break;
			default:
				b = 0;
		}
		if (b != 0) {
			System.out.println("    testDefault failed");
			return 1;
		}
		return 0;
	}

	private static int testFallthrough() {
		int b = 4;
		switch (3) {
			case 1:
				b = 8;
				break;
			case 2:
				b = -2;
				break;
			case 3:
				b = 2;
			case 4:
				b = 4;
				break;
			default:
				b = 0;
		}
		if (b != 4) {
			System.out.println("    testFallthrough failed(1)");
			return 1;
		}

		switch (1) {
			case 1:
				b = 8;
			case 2:
				b = -2;
			case 3:
				b = 2;
			case 4:
				b = 4;
			default:
				b = 0;
		}
		if (b != 0) {
			System.out.println("    testFallthrough failed(2)");
			return 1;
		}
		return 0;
	}

	private static int testExpression() {
		SwitchTestClass obj = new SwitchTestClass();
		obj.a = 40;
		obj.b = 10;
		switch (obj.a - obj.b) {
			case 30:
				break;
			default:
				System.out.println("    testExpression failed");
				return 1;
		}
		return 0;
	}

	private static int testDeclaration() {
		switch (1) {
			case 1:
				int a;
			case 2:
				a = 3;
				break;
			case 3:
				SwitchTestClass obj = new SwitchTestClass();
				if (obj.a != 3) {
					System.out.println("    testDeclaration failed");
					return 1;
				}
		}
		return 0;
	}

	private static int testNested() {
		int a[] = {3, 2, 1};
		switch (a[1]) {
			case 1:
				switch (1) {
					case 1:
					case 2:
					case 3:
						break;
				}
				break;
			case 2:
				switch (a[0]) {
					case 1:
					case 2:
						break;
					case 3:
						return 0;
				}
				break;
			case 3:
				switch (a[2]) {
					case 1:
					case 2:
					case 3:
				}
		}
		System.out.println("    testNested failed");
		return 1;
	}

	private static int testGCPopRoot() {
		switch (1) {
			case 1:
				SwitchTestClass obj = new SwitchTestClass();
				break;
			default:
				System.out.println("    testGCPopRoot failed");
				return 1;
		}
		return 0;
	}

}

class SwitchTestClass {
	public static final int A = 1;
	public static final int B = 2;
	public static final int C = 3;
	public static final int D = 4;

	public int a = 3;
	public int b = 4;
}

