package regtest;

public class ThreadTest extends TestCase {

    private static int errCount = 0;

    public static int testAll() {
		System.out.println("  ThreadTest start");
		testThreadSimple();
		//testSynchronizedMethods();
		//testSynchronizedBlocks();
		System.out.println("  ThreadTest finished");
		return errCount;
    }

	private static void testSynchronizedBlocks() {
	}

	private static void testSynchronizedMethods() {
		Monitor mon = new Monitor();
		IncThread inc = new IncThread(mon);
		PollThread poll = new PollThread(mon);

		try {
			inc.start();
			poll.start();

			inc.join();
			poll.interrupt();
			errCount += check(inc.iterations == 10000 && poll.iterations > 0,
					"    testSynchronizedMethods failed"+
					"      (inc="+inc.iterations+", poll="+poll.iterations+")");
		} catch (Exception e) {
			errCount += check(true, "    testSynchronizedMethods failed by interrupt");
		}
	}

    private static void testThreadSimple() {
		PhonyThread pt = new PhonyThread();
		pt.start();
	    try {
			Thread.sleep(500);
	    } catch (InterruptedException e) {
			System.out.println("Thread was interrupted");
	    }
		errCount += check(pt.a==15,"    testThreadSimple failed.\n"+
			  	"      pt.a = "+pt.a+" (15)");
    }

}

class Monitor {
	int value=0;

	public synchronized void modify() {
		value++;
		notifyAll();
	}

	public static void tweak(Monitor m, Monitor u, int v) {
		// The purpose here is to check that the calling thread
		// actually leaves the monitor n when the return stmt
		// is executed.
		synchronized (m) {
			for (int i = 0; i < m.value; ++i) {
				synchronized (u) {
					u.value += v;
					if (u.value == 10) return;
				}
			}
			m.value += v;
		}
	}

	public synchronized int waitNext(int old) throws InterruptedException {
		while (old == value) {
			wait();
		}
		return value;
	}
}

class PollThread extends Thread {
	private Monitor mon;
	public int iterations = 0;
	public PollThread(Monitor m) {mon = m;}
	public void run() {
		try {
			int v = mon.waitNext(0);
			while (true) {
				iterations++;
				v = mon.waitNext(v);
			}
		} catch (InterruptedException e) {}
	}
}

class IncThread extends Thread {
	private Monitor mon;
	public int iterations = 0;
	public IncThread(Monitor m) {mon = m;}
	public void run() {
		// do a fixed nmbr of loops
		for (int i = 0; i < 10000; ++i) {
			mon.modify();
			iterations++;
		}
	}
}

class TweakThread extends Thread {
	private Monitor mon;
	public int iterations = 0;
	public TweakThread(Monitor m) {mon = m;}
	public void run() {
		try {
			int v = mon.waitNext(0);
			while (!isInterrupted()) {
				v = mon.waitNext(v);
				iterations++;
			}
		} catch (InterruptedException e) {}
	}
}

class PhonyThread extends Thread {
    public int a;

    public void start() {
		a = 5;
		super.start();
    }

    public void run() {
		for (int i=0; i<10; i++) {
	    	a++;
	    	try {
				sleep(10);
	    	} catch (InterruptedException e) {
				System.out.println("Thread was interrupted");
	    	}
		}
    }
}
