package todo;
import done.*;
import se.lth.cs.realtime.semaphore.MutexSem;

public class AlarmClock extends Thread {

	private static ClockOutput	output;
	private static int		hh = 0, mm = 0, ss = 0;
	private static int		hh_a = 0, mm_a = 0, ss_a = 0;
	private static boolean		alarm = false;
	private static int		alarmTicks = 0;
	private static MutexSem		sem; 

	/**********************************************************/
	/******************** PUBLIC INTERFACE ********************/
	/**********************************************************/
	public AlarmClock(ClockInput i, ClockOutput o) {
		sem = new MutexSem();
		output = o;
		new AlarmControl(this, i).start();
	}

	public void resetAlarm() {
		sem.take();
		alarmTicks = 0;
		sem.give();
	}

	public static final void setAlarm(boolean alarmFlag) {
		sem.take();
		alarm = alarmFlag;
		sem.give();
	}

	public static final void setAlarmTime(int time) {
		sem.take();
		hh_a = time / 10000;
		mm_a = (time / 100) % 100;
		ss_a = (time) % 100;
		sem.give();
	}

	public static final void setTime(int time) {
		sem.take();
		hh = time / 10000;
		mm = (time / 100) % 100;
		ss = (time) % 100;
		sem.give();
	}

	public void run() {
		long t0, t1;
		t0 = System.currentTimeMillis();
		t1 = t0 + 1000;

		while (true) {
			t0 = System.currentTimeMillis();
			if (t0 < t1) {
				try {sleep(t1-t0);} catch (InterruptedException e) {}
			}
			t1 += 1000;
			//try {sleep(1000);} catch (InterruptedException e) {}
			incrementTimer();
			showTime();
			doAlarm();
		}
	}

	/***********************************************************/
	/******************** PRIVATE INTERFACE ********************/
	/***********************************************************/
	private static final void doAlarm() {
		sem.take();
		if (alarm && hh == hh_a && mm == mm_a && ss == ss_a) {
			alarmTicks = 20;
		}

		if (alarm && alarmTicks > 0) {
			alarmTicks--;
                        output.doAlarm();
		}
		sem.give();
	}

	private static final void incrementTimer() {
		sem.take();
		ss += 1;
		if (ss == 60) {
			ss = 0;
			mm += 1;
			if (mm == 60) {
				mm = 0;
				hh = (hh + 1) % 24;
			}
		}
		sem.give();
	}

	private static final void showTime() {
		sem.take();
		output.showTime(hh * 10000 + mm * 100 + ss);
		sem.give();
	}

}
