import se.lth.cs.util.byteArrayAccess.*;
import se.lth.cs.util.linux.*;

/* This class provides access methods for a TestObject  */
public class TestObjectAccess {
	private final int i_offset = 0;
	private final int f_offset = 4;
	private final int v_offset = 8;

	private ByteArrayAccess baa;

	public TestObjectAccess() {
		byte sharedMem[];
		System.out.println("Warning: This is a testing only constructor");
		System.out.println("Allocating normal Java byte[].");
		sharedMem = new byte[1024];
		sharedMem[ByteArrayAccess.ENDIANNESS_OFFSET] = ByteArrayAccess.BIG_ENDIAN;
		sharedMem[ByteArrayAccess.ENDIANNESS_OFFSET + 1] = 0;
		sharedMem[ByteArrayAccess.ENDIANNESS_OFFSET + 2] = 0;
		sharedMem[ByteArrayAccess.ENDIANNESS_OFFSET + 3] = 0;
		baa = ByteArrayAccessFactory.createByteArrayAccess(sharedMem);
	}

	public TestObjectAccess(String modName, String symName) {
		byte sharedMem[];
		sharedMem = ImmobileObjectFactory.createByteArray(modName, symName, 1024);
		baa = ByteArrayAccessFactory.createByteArrayAccess(sharedMem);
	}

	public final int i(){
		return baa.getInt(i_offset);
	}

	public final void seti(int i) {
		baa.setInt(i_offset, i);
	}

	public final float f(){
		return baa.getFloat(f_offset);
	}

	public final void setf(float f) {
		baa.setFloat(f_offset, f);
	}

	public final float v(int i){
		final int base = v_offset + 4*i;
		return baa.getFloat(base);
	}

	public final void setv(int i, float f) {
		final int base = v_offset + 4*i;
		baa.setFloat(base, f);
	}

	public final FooParamAccess p(int index){
		return new FooParamAccess(baa, index);
	}

	public void printArray(){
		baa.printArray(15);
	}

        public static void readTest() {
		System.out.println("TestObjectAccess.printMem()");

		String modName = "vme_read";
		String symName = "slave_global_memory";
		TestObjectAccess t = new TestObjectAccess(modName, symName);

		t.printArray();

		System.out.print("i=");
		System.out.println(t.i());
		System.out.print("f=");
		System.out.println(t.f());
		for(int i=0; i<3; i++) {
			System.out.print("v(");
			System.out.print(i);
			System.out.print(")=");
			System.out.println(t.v(i));
		}
		for(int i=0; i<3; i++) {
			System.out.print("p(");
			System.out.print(i);
			System.out.print(").K=");
			System.out.println(t.p(i).Ti());
		}
	}
}
