import se.lth.cs.throttlenet.*;
import se.lth.cs.throttlenet.event.*;


/**
 * The <code>Receiver</code> class is a simple test application for ThrottleNet 
 * communication.
 *
 * @author  Patricia Grudziecka
 * @author  Daniel Nyberg
 * @version 0.8, 12/11/03
 * @since   ThrottleNet0.8
 */
public class Receiver extends Thread {
	
	/**
	 * Data vector with different data sizes in bytes, allowed in each packet 
	 * sent over ThrottleNet.
	 */
	private int dataVector[];
	
	
	/**
	 * Period vector with adjusted period times to guaratee succeeded data trans-
	 * missions.
	 */
	private int periodVector[];
	
	
	/**
	 * Receive socket.
	 */
	ThrottleNetReceiveSocket sock = null;
	
	
	/**
	 * Constructor.
	 */
	public Receiver() {
		dataVector = new int[16];
		periodVector = new int[16];

		dataVector[0] = 46;  periodVector[0] = 15;
		dataVector[1] = 100;  periodVector[1] = 30;
		dataVector[2] = 150;  periodVector[2] = 45;
		dataVector[3] = 200;  periodVector[3] = 55;
		dataVector[4] = 250;  periodVector[4] = 65;
		dataVector[5] = 300;  periodVector[5] = 70;
		dataVector[6] = 350;  periodVector[6] = 80;
		dataVector[7] = 400;  periodVector[7] = 90;
		dataVector[8] = 450;  periodVector[8] = 95;
		dataVector[9] = 500;  periodVector[9] = 100;
		dataVector[10] = 600;  periodVector[10] = 110;
		dataVector[11] = 700; periodVector[11] = 120;
		dataVector[12] = 800; periodVector[12] = 130;
		dataVector[13] = 900; periodVector[13] = 140;
		dataVector[14] = 1000; periodVector[14] = 150;
		dataVector[15] = 1500; periodVector[15] = 160;

		try {
			start();
		} catch(Exception e) {}
		
	}
	
	private void tramstest() {
		byte[] data;
		byte[] msg;
		String s;
		s = "Goodbye, world!     ";
		data = s.getBytes();
		int size = 30;
		int period = 100;
		String tname;
		tname = "herbie";
		ThrottleNetTransmitSocket transmitSocket = null;
		try {
			String rname;
			rname = "c3po";
			// Open a receive socket with the specified connection name, data size 
			// and period time.
			sock = new ThrottleNetReceiveSocket(rname, size, period);
			
			System.out.println("Receiver: created sockets");
			// Synchronize the sender and the receiver
			msg = sock.receive();
			System.out.println("Receiver: got message:");
			int ll = msg.length;
			for(int i=0; i<ll; i++) {
				char tmpc = msg[i];
				System.out.print(tmpc);
			}
			System.out.println();
			transmitSocket = new ThrottleNetTransmitSocket(tname, size, period);
			System.out.println("Receiver: created TransmitSocket:");
			transmitSocket.transmit(data);
			
			System.out.println("Receiver: done");
			sock.close();
			transmitSocket.close();
		} catch(Exception e) {
			System.out.println("fel");
		}
	}

	
	public void run_new() {
		tramstest();
	}

	/**
	 * Run method.
	 */
	public void run() {
		int i;
		for(i = 0; i < 2; i++) {
			int dataSize = dataVector[i];
			int period = periodVector[i];
			testTime(dataSize, period);
		}
		for(i = 0; i < 6; i++) {
			int dataSize = dataVector[i];
			int period = periodVector[i];
			testTime(dataSize, period);
		}

	}
	
	
	/**
	 * Time test program.
	 *
	 * @param  size   the data size in each packet.
	 * @param  period the period time to send one packet.
	 */
	private void testTime(int size, int period) {
		byte[] data;
		String tname;
		tname = "herbie";
		ThrottleNetTransmitSocket transmitSocket = null;
		try {
			String rname;
			rname = "c3po";
			// Open a receive socket with the specified connection name, data size 
			// and period time.
			sock = new ThrottleNetReceiveSocket(rname, size, period);
			
			// Synchronize the sender and the receiver
			data = sock.receive();
			transmitSocket = new ThrottleNetTransmitSocket(tname, size, period);
			transmitSocket.transmit(data);
			
			System.out.println("// real measure");
			// Real measure of round trip time
			int i;
			for(i = 0; i < 1000; i++) {
				data = sock.receive();
				transmitSocket.transmit(data);
			}
			System.out.println("done");
			sock.close();
			transmitSocket.close();
		} catch(Exception e) {
			System.out.println("fel");
		}
	}
	
	/**
	 * Main method.
	 */
	public static void main(String args[]) {
		Receiver receiver = new Receiver();
	}
	
}


