import se.lth.cs.throttlenet.*;
import java.util.*;

/**
 * The <code>Receiver</code> class is a simple test application for ThrottleNet 
 * communication.
 *
 * @author  Patricia Grudziecka
 * @author  Daniel Nyberg
 * @version 0.8, 12/11/03
 * @since   ThrottleNet0.8
 */
public class Transmitter extends Thread {

	
	/**
	 * Data vector with different data sizes in bytes, allowed in each packet 
	 * sent over ThrottleNet.
	 */
	public int dataVector[];
	
	
	/**
	 * Period vector with adjusted period times to guaratee succeeded data trans-
	 * missions.
	 */
	private int periodVector[];
	
	/**
	 * Transmit socket.
	 */
	ThrottleNetTransmitSocket sock = null;
	ThrottleNetReceiveSocket receiveSocket = null;
	
	
	/**
	 * Constructor.
	 */
	public Transmitter() {
		dataVector = new int[16];
		periodVector = new int[16];
		dataVector[0] = 46;   periodVector[0] =15;
		dataVector[1] = 100;  periodVector[1] = 30;
		dataVector[2] = 150;  periodVector[2] = 45;
		dataVector[3] = 200;  periodVector[3] = 55;
		dataVector[4] = 250;  periodVector[4] =65;
		dataVector[5] = 300;  periodVector[5] = 70;
		dataVector[6] = 350;  periodVector[6] = 80;
		dataVector[7] = 400;  periodVector[7] = 90;
		dataVector[8] = 450;  periodVector[8] = 95;
		dataVector[9] = 500;  periodVector[9] = 100;
		dataVector[10] = 600; periodVector[10] = 110;
		dataVector[11] = 700; periodVector[11] = 120;
		dataVector[12] = 800; periodVector[12] = 130;
		dataVector[13] = 900; periodVector[13]= 140;
		dataVector[14] = 1000;periodVector[14] =150;
		dataVector[15] = 1500;periodVector[15] = 160;

		try {
			start();
		} catch(Exception e) {
		}
	}
	
	private void tramstest() {
		byte[] data;
		byte[] answer;
		String s;
		int size = 30;
		int period = 100;
		s = "Hello, world!     ";
		data = s.getBytes();
		//data = new byte[size];
		try {
			String tname;
			tname = "c3po";
			String rname;
			rname = "herbie";
			// Open transmit and receive socket with the specified connection name,
			// dataSize and period time.
			receiveSocket = new ThrottleNetReceiveSocket(rname, size, period);
			sock = new ThrottleNetTransmitSocket(tname, size, period);
			System.out.println("Transmitter: created sockets");
			// Synchronize the sender and the receiver
			sock.transmit(data);
			System.out.println("Transmitter: transmitted");
			answer = receiveSocket.receive();
			System.out.println("Transmitter: got message:");
			int ll = answer.length;
			for(int i=0; i < ll; i++) {
				char tmpc = answer[i];
				System.out.print(tmpc);
			}
			System.out.println();
		
			receiveSocket.close();
			sock.close();
			
			sleep(1000);
			
		} catch(Exception e) {
			System.out.println("fel");
			//e.printStackTrace();
		}	
	}

	public void run_new() {
		tramstest();	
	}
	/**
	 * Run method.
	 */
	public void run() {
		int i;
		System.out.println("\n\n+++++++++++++++ Warming up... +++++++++++++++\n\n");
		for(i = 0; i < 2; i++) {
			int dataSize = dataVector[i];
			int period = periodVector[i];
			testTime(dataSize, period, false);
		}
		System.out.println("\n\n=============== Real test!! ===============\n\n");
		for(i = 0; i < 6; i++) {
			int dataSize = dataVector[i];
			int period = periodVector[i];
			testTime(dataSize, period, true);
		}
	}
	
	
	/**
	 * Time test program.
	 *
	 * @param  size   the data size in each packet.
	 * @param  period the period time to send one packet.
	 */
	private void testTime(int size, int period, boolean printable) {
		byte[] data;
		data = new byte[size];
		try {
			String tname;
			tname = "c3po";
			String rname;
			rname = "herbie";
			// Open transmit and receive socket with the specified connection name,
			// dataSize and period time.
			receiveSocket = new ThrottleNetReceiveSocket(rname, size, period);
			sock = new ThrottleNetTransmitSocket(tname, size, period);
			if(printable) {
				System.out.print("Test size: ");
				System.out.print(size);
				System.out.print(",  period: ");
				System.out.print(period);
			}
			// Synchronize the sender and the receiver
			sock.transmit(data);
			data = receiveSocket.receive();
			
			// Real measure of round trip time
			long theFirstTime = System.currentTimeMillis();
			int i;
			for(i = 0; i < 1000; i++) {
				sock.transmit(data);
				data = receiveSocket.receive();
			}
			long theSecondTime = System.currentTimeMillis();
			long time = theSecondTime - theFirstTime;
			//double dtime = (double)time / 1000.0;
			if(printable) {
				System.out.print("   Result Time: ");
				System.out.println(time);
			}
			receiveSocket.close();
			sock.close();
			
			sleep(1000);
			
		} catch(Exception e) {
			System.out.println("fel");
			//e.printStackTrace();
		}	
	}
	
	
	/**
	 * Main method.
	 */
	public static void main(String args[]) {
		Transmitter transmitter = new Transmitter();
	}
}

