#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef HOST_RTAI_KERNEL
#include <stdlib.h>
#endif
#include "gc.h"
#include "gc-ms.h"
#include "gci_common.h"
#include "gci_sched.h"
#include <math.h>

#ifdef CPU_UTILIZATION_LOGGING
#include "cpu_utilization_logger.h"
#endif

#if !(defined HOST_AVR || defined HOST_RTAI_KERNEL)
#include <stdio.h>
#endif

#ifdef GC_TIMER
#include "gc_timer.h"
#endif

#undef PARANOID_CHECKS

#define PHASE_SWEEP_DONE 0
#define PHASE_MARK_ROOTS 1
#define PHASE_MARK 2
#define PHASE_MARK_DONE 3
#define PHASE_SWEEP 4

#ifdef HOST_AVR
#include <avr/io.h>
/* #include <iomacros.h>  */

#define BEEP	outp(inp(PORTE) & ~0x04,PORTE);       
#define SILENT	outp(inp(PORTE) | ~0x04,PORTE);
extern void print(char);
extern void printMsg(int,char);
#endif

#ifdef HOST_AVR
extern void transmitString(char*);
#endif

extern struct block theBlock;
/* GC DATA: should really be in a struct or something... */
GC___REF(ObjectHead) mark_stack;
void *gc_heap_base;
// #if 1 || ... <- needs fix!!! /J
#if 1 || GC_THREAD == GC_THREAD_STORK || GC_THREAD == GC_THREAD_RTAIKERNEL
// OBS!! OBS!! FIX!!!!!!
// DRAGONS ETC
//uint8 gc_heap[GC_HEAPSIZE];
// OBS!! OBS!!
int gc_started;
#endif

static int gc_phase;
static int gc_black_value;  	// toggle this to clear mark bits for the next
static int gc_white_value;	// sweep phase

#ifdef GC_FREE_CONTIGOUS_HACK
static void * gc_free_contigous;
static int gc_contigous_size;
#endif
static int sweepsize;

#ifdef SANITY_CHECKS
static int mark_stack_size;
#endif
struct GC___REF(ObjectHead_real) sweep;
// Snapshots of values used for GC amount calculation. Taken at the
// beginning of each GC cycle.
int gc_A;		// number of allocated blocks
int gc_F;		// amount of free memory
int gc_cur_amount; 	// number of bytes allocated this cycle

float workfctr;
int workcnt;

int gc_live_count;
int gc_max_live = 500;

//static int no_cycles = 0;


#ifdef TIME_METRIC
#define MIN_OBJSIZE 20
#define MARK_FACTOR 0.3
#define GC_CYCLE_LENGTH 1000
#define GC_INIT_HOLDOFF 0       // Wait this time to give the system a chance 
                                // to reach a stable state before starting GC.
                                // May not be necessary.
static uint32 gc_cycle_length;  // effective GC cycle length
static uint32 gc_cycle_olength; // nominal GC cycle length
static unsigned long gc_cycle_start;   // start time for current cycle
static unsigned long gc_cycle_end;     // start_time + cycle_length
static uint32 maxlive;
static uint32 livecount;

#endif
/* END GC DATA */

inline void gc_do_mark_obj(ObjectHead *obj);
void gc_mark_init(void);
void gc_mark(void);
void gc_sweep_init(void);
void gc_sweep(void);
inline void gc_shade_ptr(GC___REF(ObjectHead) *o);
inline void gc_shade(GC___REF(ObjectHead) o);
inline void gc_push(GC___REF(ObjectHead) ptr);

#define gc_is_black(o) (o->gc.next == (void *)gc_black_value)
#define gc_is_white(o) (o->gc.next == (void *)gc_white_value)
#define gc_color_black(o) o->gc.next = (void *)gc_black_value  
#define gc_color_white(o) o->gc.next = (void *)gc_white_value  

#define gc_michael_jackson do {\
	int tmp = gc_black_value; \
	gc_black_value = gc_white_value; \
	gc_white_value = tmp; \
} while(0)	

void debug_check_heap(char *arg)
{
	struct ObjectHead_real *scan=theBlock.bottom;
	
	while ((pointer)scan < (pointer) theBlock.top && scan->m.size > 0) {
		 //printf("scan = %d, top = %d\n", scan, theBlock.top);
		if(scan->m.next==(struct memdq*)MAGIC_ALLOCED && 
		   scan->m.prev==(struct memdq*)MAGIC_ALLOCED ){
#ifdef GC_PRINT_WARNINGS
#if !(defined HOST_AVR || defined HOST_RTAI_KERNEL)
				if( !gc_is_white(scan) && !gc_is_black(scan)){
					fprintf(stderr,"WARNING: debug_check_heap: from(%s) grey object found \n", arg);
					fprintf(stderr,"         scan = %x, scan->gc.next= %x\n", scan, scan->gc.next);
				}	
#endif
#endif
			}
		scan = (GC___REF(ObjectHead_real)) (((pointer)scan)+scan->m.size);
	}
}
void gc_full_cycle(void);
void debug_print_mark_stack(void) {
#ifndef HOST_AVR
	ObjectHead * o=mark_stack;

	printf("mark_stack (%x):", mark_stack);
	while(o && (pointer) o != -2){
		printf(" %x ", o);
		o = o->gc.next;
	}
	printf(".\n");
#endif
}

int gc_init(int heapsize) {
	int result = 0;
	static int called = 0;
	if (!called) {
#if 1 || GC_THREAD == GC_THREAD_STORK		
		gc_started = 0;
#endif		
		called = 1;
		result = 1;
		gc_black_value = 0;
		gc_white_value = 1;
		gc_live_count = gc_max_live;
		gc_ms_init(heapsize);
		gci_sched_init();
	}
#ifndef GC_IS_INCREMENTAL
	// HERE BE DRAGONS!!! I don't really know why this works... :-/
	gc_full_cycle();
#endif
	return result;
}

void gc_ms_init(int heapsize) {
	
	gc_create_heap(heapsize);
	gc_roots_gc_init(&gc_shade_ptr);
#ifdef SANITY_CHECKS
	mark_stack_size = 0;
#endif	
}

int gc_create_heap(unsigned int size)
{

	mark_stack = (void *)-2;
	
	gc_phase = PHASE_SWEEP_DONE;

#if GC_THREAD == GC_THREAD_STORK || GC_THREAD == GC_THREAD_RTAIKERNEL 
#if defined INIT_ON_FREE || defined INIT_ON_GC
	do_memset(gc_heap, 0, GC_HEAPSIZE);
#endif
	gc_heap_base = &gc_heap;

	if( 1 ) {
#else
#if defined INIT_ON_FREE || defined INIT_ON_GC
/* 	if(gc_heap_base = (void *) calloc(1, size)) { */
#if defined HOST_AVR
#if GC_THREAD == GC_THREAD_TOBBE
	if(gc_heap_base = (void *) myMalloc(size)) {
#elif GC_THREAD == GC_THREAD_T2
	if(gc_heap_base = ealloc(size)) {
#endif
#elif defined HOST_RTAI_KERNEL
        gc_heap_base = &gc_heap;
#else
	if(gc_heap_base = (void *) malloc(size)) {
#endif
	  do_memset(gc_heap_base, 0, size);
		// calloc initialises the memory to zero...
#else
	if(gc_heap_base = (void *) malloc(size)) {
#endif
#endif

#if GC_THREAD == GC_THREAD_STORK || GC_THREAD == GC_THREAD_RTAIKERNEL
		mempool_create(gc_heap_base, GC_HEAPSIZE);
#else
		mempool_create(gc_heap_base, size);
#endif		
		gc_A = 1;
		gc_F = theBlock.freemem;
		sweep = theBlock.bottom;
		return 1;
	} else {
		gc_A = 0;
		gc_F = 0;
		return 0;
	} 
}

inline void *gc_do_alloc(uint32 size, GC___REF(LayoutHead) tmplate)
{
	GC___REF(ObjectHead) b;

	if (b = alloc(size + sizeof(ObjectHead))){
		GC___PTR(b)->tmplate = tmplate;

#ifdef GC_IS_INCREMENTAL
		// alloc all objects black if incremental
		gc_color_black(GC___PTR(b));
#else
		gc_color_white(GC___PTR(b));
#endif

		return b;
	} else {
		return NULL;
	}
}

void conclude_gc_cycle(void)
{
#ifdef CPU_UTILIZATION_LOGGING
	cpu_utilization_logger_gc_log_conclude();
#endif
	//printf("conclude gc cycle\n");
	while(!gc());
}	

void *gc_new(uint32 size, void  *tmplate)
{
	void *b;
	gc_cur_amount += size;

#ifndef GC_ONLY_TIME
#ifdef GC_IS_INCREMENTAL
	if (!gc_is_working()) {
		gc_live_count++;
		if(gc_start_new_cycle()) {
			gc_mark_init();
		}
	}
	trigger_gc();
#endif
#endif

/* 	printMsg(1,'r'); */
	if (b = (void *) gc_do_alloc(size, tmplate)){
		return b;
	} else {
		/* If alloc fails, perform GC and try again */
#ifdef GC_IS_INCREMENTAL
#ifdef GC_PRINT_WARNINGS
#ifndef HOST_AVR
		printf("gc.c: *** completing GC cycle ***\n");
#endif
#endif
		if(gc_phase == PHASE_SWEEP_DONE) gc_mark(); 
		conclude_gc_cycle();
		if(b = gc_do_alloc(size,tmplate) )
			return b;
		else {
#ifdef GC_PRINT_WARNINGS
#ifndef HOST_AVR
			printf("gc.c: *** performing full GC cycle ***\n");
#endif
#endif
			/* last chance: do full GC cycle and try again... */
			gc_mark_init(); conclude_gc_cycle(); //run full GC cycle
		}
#else //GC_IS_INCREMENTAL

		gc_full_cycle();
#endif //GC_IS_INCREMENTAL
		if(b = gc_do_alloc(size,tmplate) )
			return b;
		else {
#ifdef GC_PRINT_WARNINGS
#ifndef HOST_AVR
			printf("gc.c: out of memory when allocating object\n");
#endif
#ifdef HOST_AVR
			transmitString("Out of memory?\n");
#endif
#endif
			return 0;
		}
	}
}

inline void* gc_free_obj(GC___REF(ObjectHead_real) ptr, unsigned int size)
{
	if(ptr) {
		return hfree_obj((struct memdq*)ptr,size);
	} else {
#ifdef GC_PRINT_WARNINGS
		printf("WARNING: gc_free_obj: ptr == 0, ignoring\n");
#endif
		return 0;
	}
}

inline int gc_free(void *ptr, unsigned int size)
{
	if(ptr) {
		return hfree(ptr, size); 
	} else {
		return 0;
	}
}

inline void gc_shade_ptr(GC___REF(ObjectHead)* p)
{
	gc_shade(*p);
}

inline void gc_shade(GC___REF(ObjectHead) ptr)
/* 	Colour grey if 	1) we are in the mark phase
			2) the pointer is valid
			3) the object is white
*/
{
	if(ptr != NULL && gc_phase <= PHASE_MARK && gc_is_white(GC___PTR(ptr))){
		gc_push(ptr);
	}	
}

#ifndef UPDATE_MACRO
inline void gc_update(void **a, void *c)
{
#ifdef SANITY_CHECKS	
	if(GC___PTR(a)) {
#endif			
		*GC___PTR(a) = GC___PTR(c);
#ifdef SANITY_CHECKS	
		if(GC___PTR(c)) {
			// don't shade NULL...
#endif			
			gc_shade(GC___PTR(c));
#ifdef SANITY_CHECKS	
		}
	} 
#ifndef HOST_AVR
	else {
		printf("WARNING! gc_update: trying to assign a value to *NULL.\n");
	}
#endif
#endif			
}
#endif

inline void gc_push(GC___REF(ObjectHead) ptr)
{
#ifdef SANITY_CHECKS
	if(gc_phase >PHASE_MARK) {
		printf("gc_push called in the sweep phase???\n");
	}
	if(ptr!=0 && GC___PTR(ptr)!=(void *)0x01 &&
           ((GC___PTR(ptr) <= (GC___REF(ObjectHead)) theBlock.top) && 
            (GC___PTR(ptr) >= (GC___REF(ObjectHead)) theBlock.bottom)) ) {
#endif
		GC___PTR(ptr)->gc.next = mark_stack;
		mark_stack = GC___PTR(ptr);
#ifdef SANITY_CHECKS
		mark_stack_size++;
	} else {
		fprintf(stderr, "WARNING! gc_push: erroneus ptr...%x\n",ptr);
		fprintf(stderr, "         gc_push: top=%x, bottom=%x\n",theBlock.top, theBlock.bottom);
		if(ptr && GC___PTR(ptr) != (void *) 0x01){
			fprintf(stderr, "         gc_push: next=%x\n",GC___PTR(ptr)->gc.next);
		}
	}
#endif
}

inline GC___REF(ObjectHead) gc_pop(void)
{
	GC___REF(ObjectHead) b=IVM_NULL;

	b = mark_stack;
	mark_stack = mark_stack->gc.next;
#ifdef SANITY_CHECKS
	mark_stack_size--;
#endif	
	return b;
}

inline int gc(void)
// returns zero if transition to PHASE_SWEEP_DONE
// i.e. a gc cycle has been completed
{
#ifdef CPU_UTILIZATION_LOGGER
#if CPU_TIME > 0
    cpu_utilization_logger_setGCRunning();
#endif
#endif
#ifdef GC_TIMER
	gc_timer_enter();
#endif
	//printf("GC: %d\n", gc_phase);
	switch(gc_phase) {
		case PHASE_SWEEP_DONE:
// HERE BE DRAGONS!
// Should we check if we should start a new cycle here?
//			if(gc_start_new_cycle()) {
				gc_mark_init();
				// NB! continue to mark...
//			} else {
//				break;
//			}	
		case PHASE_MARK_ROOTS:
		  	if(gc_roots_mark_one()) {
				//gc_mark_progress++;
				workcnt++;
			} else {
				gc_phase = PHASE_MARK;
			}	
			break;
		case PHASE_MARK: 
			if(gc_mark_inc()) {
				gc_phase = PHASE_MARK_DONE;
			}	
			break;
		case PHASE_MARK_DONE:
			gc_sweep_init();
				// NB! continue to sweep...
		case PHASE_SWEEP:
			if(gc_sweep_inc()) {
				gc_phase = PHASE_SWEEP_DONE;
				gci_sched_cycle_end();
#ifdef GC_TIMER
				gc_timer_exit();
				gc_timer_cycle_end();
				printf("GC_TIMER: (%ld, %ld)\n", gc_timer_cycle_time_ms(), gc_timer_cycle_etime_ms());
#endif				
			}
			break;
		default:
#if !(defined HOST_AVR || defined HOST_RTAI_KERNEL)
			fprintf(stderr,"WARNING: unknown gc phase?\n");	
#endif
			break;
	}

#ifdef GC_TIMER
	gc_timer_exit();
#endif
#ifdef CPU_UTILIZATION_LOGGER
#if CPU_TIME > 0
	cpu_utilization_logger_setGCnotRunning();
#endif
#endif
	return gc_phase == PHASE_SWEEP_DONE;
}

void gc_mark(void)
{
	gc_mark_init();

	while(gc_roots_mark_one());
	while(!gc_mark_inc());
}

void gc_sweep(void)
{
	gc_sweep_init();
	while(!gc_sweep_inc());
}

void gc_full_cycle(void)
{
	gc_mark();
	gc_sweep();
}

void gc_mark_init(void)
{
#ifdef GC_PRINT_DEBUG
	//printf("gc.c: gc_mark_init\n");
#endif

#ifdef CPU_UTILIZATION_LOGGING
	cpu_utilization_logger_gc_cycle_start();
#endif
#ifdef GC_TIMER
	gc_timer_cycle_start();
#endif
	gci_sched_init_cycle();

#ifdef TIME_METRIC
	maxlive = theBlock.totalmem / MIN_OBJSIZE;
#endif
	workfctr = (GC_MAX_ROOTS + gc_live_count) / (float)(GC_MARKRATIO(theBlock.freemem));
	workcnt = 0;
 
	gc_roots_initiate_gc_cycle();
	gc_phase = PHASE_MARK_ROOTS;
}

inline void checkAndPush(GC___REF(ObjectHead) *ib)
{
	GC___REF(ObjectHead) b = *ib;
	if(GC___PTR(b)) { 
#ifdef SANITY_CHECKS
		if(gc_phase >PHASE_MARK) {
			printf("checkAndPush called after mark phase has ended???\n");
		}
		if((GC___PTR(b) < (GC___REF(ObjectHead))theBlock.top) && (GC___PTR(b) >= (GC___REF(ObjectHead))theBlock.bottom) ) {
#endif
			if(gc_is_white(GC___PTR(b))){ 
				// only push white
				gc_push(b);
			}
#ifdef SANITY_CHECKS
		} else {
			fprintf(stderr,"WARNING! CheckAndPush: erroneous non-null pointer (%x) next == %x\n", GC___PTR(b), GC___PTR(b)->gc.next);
		}
#endif
	}
}

inline void gc_do_mark_obj(ObjectHead *obj) {
#ifdef SANITY_CHECKS
	if( obj && obj->tmplate==0){
		printf("gc_do_mark_obj: tmplate == NULL??? (obj = %x)\n",obj);
	}
#endif
	if( !gc_is_black(obj)) {
		gc_parse_object(obj, checkAndPush);
		gc_color_black(obj);
	} else {
		// printf(marked...\n");
	}	
}	
		
inline int gc_mark_inc(void)
{
        ObjectHead *obj;

//	printf("marking one increment...\n");
	if( mark_stack != (void *)-2) {
		//debug_print_mark_stack();
		//gc_mark_progress++;
		workcnt++;
		obj = gc_pop();
		//printf("marking %x\n", obj);
		gc_do_mark_obj(obj);
		return 0;
	} else {
		return 1;
	}
}	
				
				

void gc_sweep_init(void)
{
#ifdef GC_PRINT_DEBUG
	//printf("gc.c: gc_sweep_init\n");
#endif
	gc_phase = PHASE_SWEEP;
	workfctr = theBlock.totalmem / (float) (1+ theBlock.freemem);
	workcnt = 0;

#ifdef SANITY_CHECKS
	if(mark_stack != (void *)-2 ) {
		fprintf(stderr,"WARNING! gc_sweep_init: mark stack not empty?");
	}
	if(mark_stack_size != 0 ) {
		fprintf(stderr,"WARNING! gc_sweep_init: mark_stack_size not zero (%d)\n?", mark_stack_size);
		debug_print_mark_stack();
		mark_stack_size = 0;
	}
#ifdef PARANOID_CHECKS	
	debug_check_heap("sweep init");
#endif	
#endif

#ifdef GC_FREE_CONTIGOUS_HACK
	gc_free_contigous = 0;
	gc_contigous_size = 0;
#endif
	sweep = (ObjectHead_real *) theBlock.bottom; 
	//gc_mark_progress = gc_max_live;
	gc_live_count = 1;

}

inline int align_up(int i)
{
	if(i % theBlock.alignment) {
		return (i + theBlock.alignment - 1) & -theBlock.alignment;
	} else
		return i;
}

#ifdef GC_FREE_CONTIGOUS_HACK
void gc_contigous_found_white()
{
	gc_contigous_size += sweep->m.size;
	if(gc_free_contigous==0) {
		//printf("*** contigous obj start at %x\n", sweep);
		gc_free_contigous = sweep;
#ifdef INIT_ON_GC
		do_memset((void *)(((pointer)sweep)+sizeof(struct memdq)), 0, sweep->m.size - sizeof(struct memdq));
	} else {
		do_memset(sweep, 0, sweep->m.size);
#endif
	}
	//printf("incrementing gc_contigous_size with %d, new value = %d\n", sweep->m.size, gc_contigous_size);
}

void gc_contigous_block_end()
{
	int freeresult;

	if(gc_free_contigous) {
		//printf("*** contigous obj end at %x, size = %d (start = %x)\n", sweep, gc_contigous_size, gc_free_contigous);
		if( freeresult=gc_free_obj(gc_free_contigous, gc_contigous_size)) {	
			sweep = freeresult;
			sweepsize = sweep->m.size;
			gc_free_contigous = 0;
			gc_contigous_size = 0;
		} else {
			printf("WARNING! gc_sweep_inc: free failed!\n");
		}
	}
}
#endif

inline int gc_sweep_inc(void)
// preforms one sweep increment.
// returns zero there is more heap left to sweep
{
	void* freeresult;
	//printf("sweeping one increment...\n");
	//printf("       sweep = %x, sweep->gc.next= %x sweep->m.size = %d\n", sweep, sweep->gc.next, sweep->m.size);
	sweepsize = sweep->m.size;
	if ((pointer)sweep < (pointer) theBlock.top && sweep->m.size > 0) {
		//printf("sweep = %x, top = %x\n", sweep, theBlock.top);
		if(sweep->m.next==(struct memdq*)MAGIC_ALLOCED && 
		   sweep->m.prev==(struct memdq*)MAGIC_ALLOCED ){
			if(gc_is_white(sweep)) {
#if GC_HAS_FINALIZERS
				if(sweep->tmplate->finalizer) {
				// HERE BE DRAGONS!
					sweep->tmplate->finalizer((void*) ((pointer)sweep)+sizeof(memdq));
				}
#endif
				// Better destroy monitor after java finalizer
				// has been run (in case it is synchronized)
				ljrt_monitor_free((void*) ((pointer)sweep)+sizeof(memdq));
#ifdef GC_FREE_CONTIGOUS_HACK
				gc_contigous_found_white(sweep);
#else
				if( freeresult=gc_free_obj(sweep, 0) ) {
					sweep = (GC___REF(ObjectHead_real)) freeresult;
					sweepsize = sweep->m.size;
				} 
#if !(defined HOST_AVR || defined HOST_RTAI_KERNEL)
				else {
					fprintf(stderr,"WARNING: gc_sweep_inc: free failed\n");
				}
#endif
#endif
			} else {
#ifdef SANITY_CHECKS
				if( !gc_is_black(sweep) && !gc_is_white(sweep)){
					fprintf(stderr,"WARNING: gc_sweep_inc: grey object found -- coloring white anyway\n");
					fprintf(stderr,"         sweep = %x, sweep->gc.next= %x\n", sweep, sweep->gc.next);
				}	
#endif
#ifdef GC_FREE_CONTIGOUS_HACK
				gc_contigous_block_end();
#endif
				gc_live_count++;
			}
		}

#ifdef GC_FREE_CONTIGOUS_HACK
		else {
#if 1
			gc_contigous_block_end();
#else
			// unlink freed objects and do all coalescing in GC.
			// This is a bad strategy for incremental GC, as it will
			// "allocate" free objects (temporarily) in order to do
			// coalescing, which might cause out-of-memory situations
			// (which will always be fixed by completing the cycle, but
			// is bad for incrementality. For some workloads, it will 
			// quickly cause the scheduling to degenerate to batch-gc.
			freelist_unlink(sweep);
			gc_contigous_size += sweep->m.size;
			if(gc_free_contigous==0) {
				//printf("*** contigous obj start at free object %x\n", sweep);
				gc_free_contigous = sweep;
			}
#ifdef INIT_ON_GC
			else {
				do_memset(sweep, 0, sizeof(struct memdq));
			}
#endif
#endif
		}
#endif

		//printf("incrementing %x with %d\n", sweep, sweep->m.size);	
		workcnt+=sweepsize;
		sweep = (GC___REF(ObjectHead_real)) (((pointer)sweep)+sweepsize);
		//printf("sweep is now %x\n", sweep);	
		return 0;
	} else {	 
#ifdef GC_FREE_CONTIGOUS_HACK
		gc_contigous_block_end();
#endif
		gc_michael_jackson;
		//printf("GC: sweep done...\n");
		gc_cur_amount = 0;
		sweep = theBlock.bottom;
		gc_A = theBlock.allocCount;
		gc_F = theBlock.freemem;
		//debug_print_freelist();
#ifdef GC_PRINT_DEBUG
	printf("gc.c: completed cycle: freemem = %d\n", gc_freemem());
#endif
//	no_cycles++;
//	printf("Completed %d\n",no_cycles);
		return 1;
	}
}
int gc_marking_roots(void)
{
	return (gc_phase == PHASE_MARK_ROOTS);
}

void gc_finalize_all(void) {
#ifdef GC_PRINT_WARNINGS
	printf("****Running finalizers for ALL objects on the heap...\n");
#endif
	struct GC___REF(ObjectHead_real) fobj = (ObjectHead_real *) theBlock.bottom; 
	int fsize = fobj->m.size;
	while ((pointer)fobj < (pointer) theBlock.top && fobj->m.size > 0) {
		//printf("sweep = %x, top = %x\n", sweep, theBlock.top);
		if(fobj->m.next==(struct memdq*)MAGIC_ALLOCED && 
		   fobj->m.prev==(struct memdq*)MAGIC_ALLOCED ){
#if GC_HAS_FINALIZERS
			if(fobj->tmplate->finalizer) {
			// HERE BE DRAGONS!
				fobj->tmplate->finalizer((void*) ((pointer)fobj)+sizeof(memdq));
			}
#endif
		}
		// Better destroy monitor after java finalizer
		// has been run (in case it is synchronized)
		ljrt_monitor_free((void*) ((pointer)fobj)+sizeof(memdq));
		fobj = (GC___REF(ObjectHead_real)) (((pointer)fobj)+fsize);
	}
}


inline int gc_is_working(void) {return gc_phase != PHASE_SWEEP_DONE;}
inline int gc_workcnt(void) {return workcnt;}
inline float gc_workfctr(void) {return workfctr;}
inline unsigned int gc_freemem(void) {return theBlock.freemem;}
inline unsigned int gc_allocated(void) {return gc_cur_amount;}
inline int gc_heapsize(void) {return theBlock.totalmem;}

