#ifndef __GC_IMP_INCREMENTAL_MS_H__
#define __GC_IMP_INCREMENTAL_MS_H__
//
// This file defines the implementation layer macros for the compacting
// incremental mark sweep collector.

extern int gc_marking;
extern struct ObjectHead *markstk;

//
// Internally used macros
//

/* REF: Macro for specifying reference types.
 *   Examples: Object *x is written REF(Object) x
 *             (Object *) x is written (REF(Object)) x;
 */
#define GC___REF(ref_type) ref_type **

/* PTR: Macro for accessing a reference.
 *   Example: x->y is written PTR(x)->y
 */
#define GC___PTR(ref) (*ref)

/* GC___CHK: Macro checking reference assignments. Must be called
 *      after having assigned a reference a new value.
 *   Example: x = y; is written x = y; GC___CHK(x);
 *            x->y = z; is written GC___PTR(x)->y = z; GC___CHK(PTR(x)->y);
 */
#define GC___CHK(reference) \
if (gc_marking && (reference) && (*((ObjectHead **) (reference)))->gc.mark==0) \
{ \
    (*((ObjectHead **) (reference)))->gc.mark = markstk; \
    markstk = *((ObjectHead **) (reference)); \
}

/* address of pointer */

#define GC___ADDR_OF_PTR(ref) &GC___PTR(ref)

/* "real" pointer (i.e., not reference) assignment */
#ifdef IAR
#define GC___ASSIGN(target, source) target=source; GC___CHK(target);
#else
/* #define GC___ASSIGN(target, source) (void *)target=source; GC___CHK(target); */
/* Fix for gcc 4.0. AndersN */
#define GC___ASSIGN(target, source) target=source; GC___CHK(target);
#endif


// gc header information 
typedef struct GCFields {
  void *mark;
  void *fp;
} GCFields;


#endif
