#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define INIT_ON_FLIP

#ifdef GC_PRINT_WARNINGS
#if !(defined HOST_AVR || defined HOST_RTAI_KERNEL)
#include <stdio.h>
#endif
#endif

#include "gc.h"
#include "gc_types.h"
#include "gci_common.h"
#include "gc_roots.h"

#define GC_CYCLE_FINISHED 0
#define GC_EVACUATE_ROOTS 1
#define GC_SCAN_FROMSPACE 2

extern int gc_initialized;
void *gc_heap_base;
#if GC_THREAD == GC_THREAD_STORK || GC_THREAD == GC_THREAD_RTAIKERNEL
uint8 gc_heap[GC_HEAPSIZE];
#endif

static void *tospace, *fromspace, *allocpos, *evacpos;
ObjectHead *scanp;

void gc_evacuate(GC___REF(ObjectHead) *o);
void gc_cycle_init(void);
int gc_inc(void);

static unsigned int semi_size;
static unsigned int L_max;
static unsigned int gc_working, gc_phase;

int gc_init(int asked_heapsize){
	int heapsize; 
	int result;

	// The border between tospace and fromspace should be
	// (at least) 64 bits  ==  8 bytes aligned. The size is 
	// rounded down in order to fit in the statically allocated 
	// heap used on some targets.
	heapsize = round_down(asked_heapsize, 16);
	result = gc_create_heap(heapsize);

	if(result) {
		tospace=(void *) gc_heap_base;
		semi_size = heapsize / 2;
		fromspace=(void *) (((unsigned int) gc_heap_base)+semi_size);
		evacpos=tospace;
		allocpos=tospace+semi_size;
	
		// HERE BE DRAGONS!!!
		L_max = semi_size / 2;
#ifdef GC_PRINT_DEBUG
		printf("*** semi_size = 0x%x\n", semi_size);
		printf("*** fromspace = 0x%x -- 0x%x\n", fromspace, fromspace+semi_size);
		printf("*** tospace = 0x%x   -- 0x%x\n", tospace, tospace+semi_size);
		printf("*** evacpos = 0x%x\n", evacpos);
		printf("*** allocpos = 0x%x\n", allocpos);
#endif
#ifdef INIT_ON_FLIP
		memset(tospace, 0, semi_size);
#endif
		gc_roots_gc_init(&gc_evacuate);
#ifdef GC_IS_INCREMENTAL
		gc_cycle_init();
#endif
	}
	return result;
}

static inline int gc_enough_free_for(uint32 size)
{
	//return( !(semi_size - (allocpos - tospace) >= size ) );
	return allocpos - size > evacpos;
}

static inline int do_more_gc(void)
/* The GC work invariant is: do more GC work if
 * \frac{e}{L_{max}} >= \frac{a}{F_{min}}
 * This guarantees sufficient progress in the new-driven case.
 * In the concurrent case, a safety margin is required [Hen98]
 */
{
	unsigned int e = evacpos - tospace;
	unsigned int a = tospace + semi_size - allocpos;
	//int result = gc_working && (e / L_max) >= (a/(semi_size - L_max));
	int result =  gc_working && ( e*semi_size >= (a+e) * L_max);	
#ifdef GC_PRINT_DEBUG_DMGC
	printf("do_more_gc() returns %d (working == %d)\n", result, gc_working);
#endif
	return result;
}

void *gc_new(uint32 size,void *tmplate)
{
	void *result;
#ifdef GC_IS_INCREMENTAL
	if(!gc_working) gc_cycle_init();
	//while(cont && do_more_gc()) cont=gc_inc();
#ifdef GC_HAS_THREADS
	if( do_more_gc() ) {
		if( CURRENT_THREAD_IS_HP)
			gc_signal_allocation();
		else
			perform_gc();
	}
#else
	perform_gc();
#endif
	if( !gc_enough_free_for(size) ) {
#ifdef GC_PRINT_WARNINGS
		printf("performing full GC cycle\n");
#endif
		gc();
	}
#else
	if( !gc_enough_free_for(size) ) {
		gc();
	}
#endif
	if( !gc_enough_free_for(size) ) {
#ifdef GC_PRINT_WARNINGS
		printf("performing second GC cycle\n");
#endif
		gc();
	}

	if( gc_enough_free_for(size) ) {
		allocpos -= size;
		result = allocpos;
#ifndef INIT_ON_FLIP
		memset(result, 0, size);
#endif
		//GC___CPTR(ObjectHead,result)->tmplate = tmplate;
		//GC___CPTR(ObjectHead,result)->gc.fp = result;
		GC___ASSIGN(((ObjectHead *) result)->tmplate, tmplate);
		((ObjectHead *) result)->gc.fp = result;
	} else {
#ifdef GC_PRINT_WARNINGS
		printf("gc_incremental_copy.c: out of memory when allocating object\n");
#endif
#if HAS_EXCEPTIONS == 0 && !defined HOST_RTAI_KERNEL
		exit(1);
#else
		result = 0;
#endif
	}
#ifdef GC_PRINT_DEBUG_NEW
	printf("gc_new: returns 0x%x\n", result);
#endif
	return result;
}

int gc_create_heap(unsigned int size)
{
#if GC_THREAD == GC_THREAD_STORK || GC_THREAD == GC_THREAD_RTAIKERNEL
        gc_heap_base = &gc_heap;

	return 1;

#else
#if defined HOST_AVR
        if(gc_heap_base = (void *) myMalloc(size)) {
#else
        if(gc_heap_base = (void *) malloc(size)) {
#endif
		return 1;
        } else {
                return 0;
        }
#endif
}

void gc_cycle_init(void)
{
	void *tmp;
// This safety feature should probably be possible to turn off
// #ifdef SANITY_CHECKS
	if(gc_working) {
#ifdef GC_PRINT_DEBUG
		printf("gc_incremental_copy.c: gc_cycle_init() called but GC is working???\n");
#endif
		return;
	}
// #endif
// end
#ifdef GC_HAS_THREADS
	if(gc_initialized) {
#else
	if(1) {
#endif

		/* flip */
		tmp = fromspace;
		fromspace = tospace;
		tospace = tmp;
#ifdef GC_PRINT_DEBUG_FLIP
		printf("gc_incremental_copy.c: flip!\n");
		printf("        fromspace = 0x%x -- 0x%x\n", fromspace, fromspace+semi_size);
		printf("        tospace   = 0x%x -- 0x%x\n", tospace, tospace+semi_size);
#endif
		evacpos=tospace;
		allocpos=tospace+semi_size;
		scanp = tospace;
#ifdef INIT_ON_FLIP
		memset(tospace, 0, semi_size);
#endif

		gc_roots_initiate_gc_cycle();
		gc_working = 1;
		gc_phase = GC_EVACUATE_ROOTS;
	}
#ifdef GC_PRINT_DEBUG
	else {
		printf("gc_incremental_copy.c: gc_cycle_init() called but GC is not initialized. Ignoring.\n");
	}
#endif
}

static inline int inFromspace(void *o)
{
	return o >= fromspace && o < (fromspace + semi_size);
}

void gc_evacuate(GC___REF(ObjectHead) *o)
{
	ObjectHead *obj = *o;

	if(inFromspace(obj) ){
		if( obj->gc.fp == obj ) {
			// evacuate
			int osize = getsize(obj);
			//printf("evacuating 0x%x    evacpos = 0x%x, allocpos = 0x%x\n", obj, evacpos, allocpos);
#ifdef GC_PRINT_WARNINGS
			if(evacpos + osize > allocpos) {
				printf("gc_incremental_copy.c: memory deadlock!/\n");
				printf("   osize = %d = 0x%x\n", osize, osize);
				printf("   evacpos = 0x%x, ep + osize = 0x%x\n", evacpos, evacpos+osize);
				printf("   allocpos = 0x%x\n", allocpos);
#if !(defined HOST_RTAI_KERNEL || HOST_AVR)
				//exit(1);
#endif
			}
#endif
			memcpy(evacpos, obj, osize);
			obj->gc.fp = evacpos;
			((ObjectHead *)evacpos)->gc.fp = evacpos;
			evacpos += osize;
		}	
		*o = obj->gc.fp;
		//printf("already evacuated 0x%x\n", obj);
		//printf("updating pointer 0x%x\n", o);
	} // else already evacuated...
	//else printf("gc_evacuate ignored 0x%x\n", obj);
}

int gc_inc(void)
/* Perform one GC increment.
 * Returns 0 if GC cycle was completed and 1 if more work to be done
 */
{
	int result;
#ifdef GC_PRINT_DEBUG_INC
	printf("gc_incremental_copy.c: gc_inc: phase == %d\n", gc_phase);
#endif
	switch (gc_phase){
		case GC_EVACUATE_ROOTS:
			if(!gc_roots_mark_one()) {
				gc_phase = GC_SCAN_FROMSPACE;
			}
			result = 1;	
			break;
		case GC_SCAN_FROMSPACE:
			if(scanp < evacpos) {
				int osize;
				osize = getsize(scanp);
				gc_parse_object(scanp, gc_evacuate);	
				scanp = (ObjectHead *) (void *) (((uint8 *) scanp)+osize);
				result = 1;
			} else {
				result = 0;
				gc_working = 0;
				gc_phase = GC_CYCLE_FINISHED;
			}
			break;
		case GC_CYCLE_FINISHED:
#ifdef GC_PRINT_DEBUG
			printf("gc_incremental_copy.c: gc_inc called but phase == GC_CYCLE_FINISHED???\n");
			result = 0;
			break;	
#endif
		default:
#ifdef GC_PRINT_WARNINGS
			printf("gc_incremental_copy.c: gc_inc called but phase is corrupted???\n");
			result = 0;
#endif			
	}
	return result;
}

void gc(void)
{
	if(!gc_working) gc_cycle_init();
	while( gc_inc() );
}

#ifdef GC_IS_INCREMENTAL
#ifdef GC_ONLY_TIME
void perform_gc(void)
{
	int i;
	// make increments larger
	while(gc_working) {
		for(i=0; gc_working && i<20; i++) {
			gc_inc();
		}
	gc_unlock();
	gc_lock();
	}
}
#else
void perform_gc(void)
{
	while(do_more_gc()) {
		gc_inc();
	}
}
#endif

int gc_marking_roots(void)
{
        return (gc_phase==GC_EVACUATE_ROOTS) ;
}
#endif
