/*
  gc_incremental_ms/gc.h,
  The garbage collector interface.

  Copyright (C) 2000 Roger Henriksson <roger@cs.lth.se>

  This library is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
/* 
 * R.Henriksson 990303 
 */

#ifndef __gc_h
#define __gc_h

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef GC_DEBUG_MS
#include <assert.h>
#endif

#include "gc_types.h"
#include "gci_common.h"

/* Some compilers don't accept array size 0, and some warns when
 * hard coded array accesses exceed the size 1 as given below. If
 * so please set JVL_MIN_ARRAY_SIZE to some greater value.*/
//#define JVL_MIN_ARRAY_SIZE 1

/* Temporary switches */
#define EXIT_IVM exit(1)
#define ivmPrintf printf


/* Struct definer macro for getting padding byte in AVR */
// #ifdef ADD_PADDING
// #define STRUCT_REF(_type, _name) REF(_type) _name; unsigned char _name##padder
// #else
// #define STRUCT_REF(_type, _name) REF(_type) _name
// #endif

#ifdef HOST_KOALA
/* Due to problems with the koala target */
/* we use these from Anders Blomdell */
static unsigned int GC_MARKRATIO(unsigned int heapsize) {
  return ((heapsize >> 2) + (heapsize >> 4) + (heapsize >> 6) + (heapsize >> 8));
}

static unsigned int GC_TRESHOLD(unsigned int heapsize) {
  return ((heapsize >> 2) + (heapsize >> 4) + (heapsize >> 6) + (heapsize >> 8));
}
#else
// in gci_sched.h
//#define GC_MARKRATIO(heapsize) ((heapsize)/3.0)
//#define GC_TRESHOLD(heapsize) ((heapsize)/3.0)
#endif

void *gc_new(uint32 size,void *tmplate);

/* Try to deallocate object on the heap.
 * It works if the object is the latest allocated.
 * Theoretically, the heap is used as a stack.
 * Useful for activation records allocated on the heap.
 * However, activations that creates objects cannot
 * be deallocated in this way. The heap should maybe be 
 * organised in an object part and one activation part.
 */
void fastDealloc(GC___REF(void) objectStart, int objectSize);

/* Where should these go? */

int gc_is_ref(void *object, void *field);

/*void gc_mark_object(ObjectHead **o);
*/
int gc_marking_roots(void);

void gc_signal_allocation(void);
#endif
