/* gc_timer.c
 * 
 * S. G. Robertz 20010124
 */

#include "gc_timer.h"
#include <stdio.h>

timeval	etime_last_cycle, /* time on CPU */
	time_last_cycle, /* wall-clock time */ 
	elapsed_this_cycle,  /* accumulator for current cycle etime */
	time_istart, time_iend, /* current increment start/stop times */
	time_cstart, time_cend; /* current cycle start/stop times */

/* Routines to call when entering/exiting GC code */

void gc_timer_enter(void)
{
	gettimeofday(&time_istart, NULL);
}
void gc_timer_exit(void)
{
	long secs, usec;
	gettimeofday(&time_iend, NULL);
	secs = (time_iend.tv_sec  - time_istart.tv_sec);
	usec = (time_iend.tv_usec - time_istart.tv_usec);
	if(usec < 0) {
		fprintf(stderr, "time before %d.%d\n", secs, usec);
		secs -= 1;
		usec += 1000000;
		fprintf(stderr, "time after %d.%d\n", secs, usec);
	}
	elapsed_this_cycle.tv_sec  += secs;
	elapsed_this_cycle.tv_usec += usec;
/*	
	elapsed_this_cycle.tv_sec  += (time_iend.tv_sec  - time_istart.tv_sec);
	elapsed_this_cycle.tv_usec += (time_iend.tv_usec - time_istart.tv_usec);
*/
}

/* Routines to call when starting/ending a GC cycle */

void gc_timer_cycle_start(void)
{
	gettimeofday(&time_cstart, NULL);
/*	timerclear(&elapsed_this_cycle);
*/
	elapsed_this_cycle.tv_sec = 0;
	elapsed_this_cycle.tv_usec = 0;
}
void gc_timer_cycle_end(void)
{
	gettimeofday(&time_cend, NULL);
	etime_last_cycle.tv_sec  = elapsed_this_cycle.tv_sec;
	etime_last_cycle.tv_usec = elapsed_this_cycle.tv_usec;
	time_last_cycle.tv_sec  = (time_cend.tv_sec  - time_cstart.tv_sec);
	time_last_cycle.tv_usec = (time_cend.tv_usec - time_cstart.tv_usec);
	if(time_last_cycle.tv_usec < 0) {
		time_last_cycle.tv_sec -= 1;
		time_last_cycle.tv_usec += 1000000;
	}
}

/* Get time estimates */

void gc_timer_cycle_time(timeval *tv)
{
	tv->tv_sec = time_last_cycle.tv_sec;
	tv->tv_usec = time_last_cycle.tv_usec;
}

void gc_timer_cycle_etime(timeval *tv)
{
	tv->tv_sec = etime_last_cycle.tv_sec;
	tv->tv_usec = etime_last_cycle.tv_usec;
}

/* Get times in *micro*seconds */

long gc_timer_cycle_time_ms(void)
{
	/*
	long result = time_last_cycle.tv_sec * 1000;
	result += (time_last_cycle.tv_usec / 1000);
	*/
	long result = time_last_cycle.tv_sec * 1000000;
	result += (time_last_cycle.tv_usec);
	return result;
}

long gc_timer_cycle_etime_ms(void)
{
	/*
	long result = etime_last_cycle.tv_sec * 1000;
	result += (etime_last_cycle.tv_usec / 1000);
	*/
	long result = etime_last_cycle.tv_sec * 1000000;
	result += (etime_last_cycle.tv_usec);
	return result;
}
