/*
  gci_common.c,
  The common (algorithm independent) implementation of the GCI 

  Copyright (C) 2000 Roger Henriksson <roger@cs.lth.se>, 
                     Sven Gesteg�d Robertz <sven@cs.lth.se>

  This library is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// THESE SHOULD BE MOVED TO config.h
#undef GC_DEBUG_FULL_CYCLE 
#undef GC_DEBUG_COUNT_ALLOC

#ifdef CPU_UTILIZATION_LOGGING
#include "cpu_utilization_logger.h"
#endif

#include "gc.h"
#include "gc_roots.h"
#include "gci_common.h"

#if defined GC_PRINT_WARNINGS && !(defined HOST_AVR || defined HOST_RTAIKERNEL)
#include <stdio.h>
#endif

#include <stdarg.h>

#if GC_THREAD != GC_THREAD_RTAIKERNEL  	// get size_t
#include <stdlib.h>
#else
#include <linux/types.h>
#endif

//#include <assert.h>
#include <string.h>

#ifdef GC_HAS_THREADS
int gc_exited;
#endif

#ifdef LPMEM
extern int gc_non_critical;
#endif
uint32 getsize(ObjectHead *ptr) 
{
  GC___REF(LayoutHead) tmplate;

  tmplate = ptr->tmplate;
#ifdef GC_PRINT_DEBUG
  if(tmplate == 0) {
	printf("gci_common.c: getsize: tmplate == NULL (ptr==%x)\n", ptr);
  }
#endif
  return (GC___PTR(tmplate)->objectSize==1) 
    ? ((Array *) ptr)->objectSize
    : GC___PTR(tmplate)->objectSize;
}

/* memory methods */

//HERE BE DRAGONS
static GC__IMPL_REF(LayoutHead, META_LAYOUT);
GC__IMPL_REF(LayoutHead, VAL_ARRAY_LAYOUT); // AN, tmp fix for arrays. Should be removed.
GC__IMPL_REF(LayoutHead, REF_ARRAY_LAYOUT); // AN, tmp fix for arrays. Should be removed.
/* static GC__IMPL_REF(LayoutHead, VAL_ARRAY_LAYOUT); */
static GC__IMPL_REF(Uint8Array, LAYOUT_GCINFO);
static GC__IMPL_REF(Uint8Array, VAL_ARRAY_GCINFO);
static GC__IMPL_REF(Uint8Array, REF_ARRAY_GCINFO);

/**
 *
 */
GC___REF(Uint8Array) newGCInfo(int iNbrElements) 
{
  GC___REF(Uint8Array) a = IVM_NULL;
  uint16 arrSize = iNbrElements * sizeof(uint8);
  uint16 objSize = sizeof(Uint8Array) + arrSize;// -4 /* considering JVL_MIN_ARRAY_SIZE*/;
  objSize = ALIGN4(objSize);
  GC___ASSIGN(a, HEAPALLOC(Uint8Array, objSize, VAL_ARRAY_LAYOUT->ref));
  GC___PTR(a)->objectSize = objSize;
  GC___PTR(a)->nbrElements = iNbrElements;
  GC___PTR(a)->arraySize = arrSize;
  return a;
}

/**
 *
 */
void gci_common_init(int heapsize)
{ 
#ifdef GC_PRINT_DEBUG
  printf("gci_common_init: heapsize = %d\n", heapsize);
#else
#if GC_THREAD == GC_THREAD_XENOMAI
//HERE BE DRAGONS!
//If this output is removed, gc_lock/unlock stops working???
//To be investigated...
printf("This line intentionally left blank.\n");
#endif
#endif
#ifdef CPU_UTILIZATION_LOGGER
#if CPU_TIME > 0
  cpu_utilization_logger_gc_log_init();
#endif
#endif
#ifdef GC_HAS_THREADS
  gc_exited = 0;
/*   outp(~0x40,PORTB);  */
  gci_thread_once_routine();  
/*   outp(~0x43,PORTB);  */
#endif
/*   outp(~0x50,PORTB);  */
  if (gc_init(heapsize)) {
//HERE BE DRAGONS
/*   outp(~0x60,PORTB);  */
    GC__IMPL_PUSH_STATIC_ROOT(META_LAYOUT);
    GC__IMPL_PUSH_STATIC_ROOT(VAL_ARRAY_LAYOUT);
    GC__IMPL_PUSH_STATIC_ROOT(LAYOUT_GCINFO);
    GC__IMPL_PUSH_STATIC_ROOT(VAL_ARRAY_GCINFO);
    GC__IMPL_PUSH_STATIC_ROOT(REF_ARRAY_LAYOUT);
    GC__IMPL_PUSH_STATIC_ROOT(REF_ARRAY_GCINFO);
    
    {
#ifdef HOST_AVR
      uint16 size = sizeof(LayoutHead);
#else
      uint16 size = ALIGN4(sizeof(LayoutHead));
#endif
      META_LAYOUT->ref = HEAPALLOC(LayoutHead, size, IVM_NULL);
      GC___ASSIGN(GC___PTR(META_LAYOUT->ref)->tmplate, META_LAYOUT->ref) /* The tmplate recursion ends.     */
      /* The gcInfo is set after it has been created */
      /*PTR(META_LAYOUT)->gcInfo = LAYOUT_GCINFO; CHK(PTR(META_LAYOUT)->gcInfo); */
      GC___PTR(META_LAYOUT->ref)->objectSize = size;
    }
    
    {
      uint16 size = sizeof(Uint8Array) /*+ 4*/; /* JVL_MIN_ARRAY_SIZE == 1 -> 4 extra bytes */
      
      /* The template pointer is set to a correct value below after the VAL_ARRAY_LAYOUT has been created. */
      LAYOUT_GCINFO->ref = (GC___REF(Uint8Array)) HEAPALLOC(Uint8Array, size, VAL_ARRAY_LAYOUT->ref); 
      GC___PTR(LAYOUT_GCINFO->ref)->objectSize = size;
      GC___PTR(LAYOUT_GCINFO->ref)->nbrElements = 1;
      GC___PTR(LAYOUT_GCINFO->ref)->arraySize = 4; /* Shall it be 4 bytes or one value? */
    }
    SETARRAY(LAYOUT_GCINFO->ref, 0, 2);
    
    /* It is necessary to identify arrays from ordinary objects. */
    /* An array layout is placed between the meta meta layout and the arrays. */
    /* Size of 1 indicates that the size should be read from the object itself, */
    /* and not from the layout. */
    /* NOTE: This '1' differs from the Ivm-gc algorithms.  */
    /*       Create a consistent 'getsize'-method. */
    VAL_ARRAY_LAYOUT->ref = HEAPALLOC(LayoutHead, sizeof(LayoutHead), META_LAYOUT->ref);
    /* Value is set as soon as it has been created. */
    /*PTR(VAL_ARRAY_LAYOUT)->gcInfo = VAL_ARRAY_GCINFO; */
    GC___PTR(VAL_ARRAY_LAYOUT->ref)->objectSize = 1; /* Mark array object size */
    
    /* Fix the GC_INFO template */
    GC___ASSIGN(GC___PTR(LAYOUT_GCINFO->ref)->tmplate, VAL_ARRAY_LAYOUT->ref)
    
    {
      uint16 size = sizeof(Uint8Array) /* + 4 */; /* JVL_MI_ARRAY_SIZE == 1 -> 4 extra bytes */
      
      /* The template pointer is set to a correct value below */
      VAL_ARRAY_GCINFO->ref = (GC___REF(Uint8Array)) HEAPALLOC(Uint8Array, size, VAL_ARRAY_LAYOUT->ref);
      GC___PTR(VAL_ARRAY_GCINFO->ref)->objectSize = size;
      GC___PTR(VAL_ARRAY_GCINFO->ref)->nbrElements = 1;
      GC___PTR(VAL_ARRAY_GCINFO->ref)->arraySize = 4;
    }
    SETARRAY(VAL_ARRAY_GCINFO->ref, 0, 1);
    
/**************** REF_ARRAY_LAYOUT *********/

    /* It is necessary to identify arrays from ordinary objects. */
    /* An array layout is placed between the meta meta layout and the arrays. */
    /* Size of 1 indicates that the size should be read from the object itself, */
    /* and not from the layout. */
    /* NOTE: This '1' differs from the Ivm-gc algorithms.  */
    /*       Create a consistent 'getsize'-method. */
    REF_ARRAY_LAYOUT->ref = HEAPALLOC(LayoutHead, sizeof(LayoutHead), META_LAYOUT->ref);
    /* Value is set as soon as it has been created. */
    /*PTR(REF_ARRAY_LAYOUT)->gcInfo = REF_ARRAY_GCINFO; */
    GC___PTR(REF_ARRAY_LAYOUT->ref)->objectSize = 1; /* Mark array object size */
    
    {
      int nbrSkip = 0;
      int sizeSizes = 0;
      uint16 size;  

      /* HERE BE DRAGONS! The following sizes are hard-coded, as I didn't
       * bother to figure out how the commented-out versions should be
       * expressed. If possible, this should be changed.
       */
      sizeSizes += sizeof(uint16); //sizeof(ArrayHead.nbrElements);
      sizeSizes += sizeof(uint16); //sizeof(ArrayHead.objectSize);
      sizeSizes += sizeof(uint32); //sizeof(ArrayHead.arraySize);

#if 0 // This is part of the object
      sizeSizes += sizeof(int);    //sizeof(RefArray._RefA_length);
#endif
	
      // Size of array head (including the skipped _refA_length)
      size = sizeof(Uint8Array) + sizeSizes + sizeof(int);

      /* The template pointer is set to a correct value below */
      REF_ARRAY_GCINFO->ref = (GC___REF(Uint8Array)) HEAPALLOC(Uint8Array, size, VAL_ARRAY_LAYOUT->ref);
      GC___PTR(REF_ARRAY_GCINFO->ref)->objectSize = size;
      GC___PTR(REF_ARRAY_GCINFO->ref)->nbrElements = 4;
      GC___PTR(REF_ARRAY_GCINFO->ref)->arraySize = 16;
  
    SETARRAY(REF_ARRAY_GCINFO->ref, 0, 1);
    
      /* This is an attempt to make a generic calculation of the number of 
       * non-pointers in the ObjectHead struct (locks, queues et cetera).
       */
      nbrSkip += (sizeof(ObjectHead) - sizeof(ObjectHead_Head) );
#ifdef GC_PRINT_DEBUG
      printf("REF_ARRAY_GCINFO: skipcount=%d\n", sizeSizes + nbrSkip);
#endif

      SETARRAY(REF_ARRAY_GCINFO->ref, 1, sizeSizes + nbrSkip);   // Skip lock ptrs etc
//      SETARRAY(REF_ARRAY_GCINFO->ref, 2, 0);         // no pointers
//      SETARRAY(REF_ARRAY_GCINFO->ref, 3, sizeSizes);  // Skip array sizes etc
    }
    SETARRAY(REF_ARRAY_GCINFO->ref, 2, 254); //Escape, read noptrs from object
    SETARRAY(REF_ARRAY_GCINFO->ref, 3, 0);
    
/*******************************************/ 	
 	


    /* Fixes: */
    GC___ASSIGN(GC___PTR(META_LAYOUT->ref)->gcInfo, LAYOUT_GCINFO->ref)
    GC___ASSIGN(GC___PTR(VAL_ARRAY_LAYOUT->ref)->gcInfo, VAL_ARRAY_GCINFO->ref)
    GC___ASSIGN(GC___PTR(REF_ARRAY_LAYOUT->ref)->gcInfo, REF_ARRAY_GCINFO->ref)
  }

#ifdef TIME_METRIC
	//HERE BE DRAGONS!!! Where is the extraneous gc_lock()???
	// **findme!!!**
	gc_unlock();
#endif

	gci_thread_init();
}

/**
 * Destroy and clean up the GC.
 * The static roots are popped.
 */
void GC_DESTROY(void) 
{
//HERE BE DRAGONS
  GC__IMPL_POP_STATIC_ROOT(REF_ARRAY_GCINFO);
  GC__IMPL_POP_STATIC_ROOT(REF_ARRAY_LAYOUT);
  GC__IMPL_POP_STATIC_ROOT(VAL_ARRAY_GCINFO);
  GC__IMPL_POP_STATIC_ROOT(LAYOUT_GCINFO);
  GC__IMPL_POP_STATIC_ROOT(VAL_ARRAY_LAYOUT);
  GC__IMPL_POP_STATIC_ROOT(META_LAYOUT);
}

/**
 * Create a layout from the object description v, that contains n numbers.
 * 's' is the size of the object created from this layout.
 */
// REF(LayoutHead) CreateLayout(unsigned int n, unsigned int v[], unsigned int s)
// {
//   REF(LayoutHead) th = IVM_NULL;
//   REF(Uint8Array) gi = IVM_NULL;

//   GC_PUSH_ROOT(th);

//   th = HEAPALLOC(LayoutHead, sizeof(MetaLayout), META_LAYOUT); CHK(th);
//   gi = newGCInfo(n); CHK(gi);

//   {
//     int i;
//     for (i = 0; i<n; i++)
//       SETARRAY(gi, i, v[i]);
//   }

//   PTR(th)->gcInfo = gi; CHK(PTR(th)->gcInfo);
//   PTR(th)->objectSize = s;

//   GC__IMPL_POP_ROOT(th);
//   return th;
// }

GC___REF(LayoutHead) gc_new_layout(int size, int descriptors, ...)
{
  va_list ap;
  int i, nextpos, pointers, gcinfopos;
  GC__IMPL_REF(LayoutHead, th);
  GC__IMPL_REF(Uint8Array, gi);
  GC__GET_THREAD_ROOT;

  GC__IMPL_PUSH_ROOT(th);
  GC__IMPL_PUSH_ROOT(gi);

  // Find out size of GCInfo array
  va_start(ap, descriptors);
  nextpos = sizeof(void*); // Kludge: tmplate pointer
  pointers = 1; // Kludge: tmplate pointer
  gcinfopos = 0;
  for (i = 0 ; i < descriptors ; i++) {
    int offset = va_arg(ap, int);
    if (offset < 0) {
#ifdef GC_PRINT_WARNINGS
      printf("Unknown offset kind %i %d\n", i, offset);
#endif
    } else {
      offset -= sizeof(GCFields); // Kludge!!
//      printf("gc_new_layout[%d] = %d\n", i, offset);
      if (offset != nextpos ) {
//	printf("%d, ", pointers);
//	printf("%d, ", offset - nextpos);
//	printf("\n");
	pointers = 1;
	nextpos = offset + sizeof(void*);
	gcinfopos += 2;
      } else {
	nextpos += sizeof(void*);
	pointers++;
      }
    }
  }
  va_end(ap);
  if (pointers > 0) { 
    gcinfopos++; 
//    if ((size - nextpos - sizeof(GCFields)) > 0) { 
      gcinfopos++; 
//    }
  }

//  printf("%d, %d\n", pointers, size - nextpos - sizeof(ObjectHead));
  
  // Create and fill in the GCInfo
  GC___ASSIGN(gi->ref, newGCInfo(gcinfopos))
  va_start(ap, descriptors);
  nextpos = sizeof(void*); // Kludge: tmplate pointer
  pointers = 1; // Kludge: tmplate pointer
  gcinfopos = 0;
  for (i = 0 ; i < descriptors ; i++) {
    int offset = va_arg(ap, int);
    if (offset < 0) {
#ifdef GC_PRINT_WARNINGS
      printf("Unknown offset kind %i %d\n", i, offset);
#endif
    } else {
      offset -= sizeof(GCFields); // Kludge!!
//      printf("gc_new_layout[%d] = %d\n", i, offset);
      if (offset != nextpos ) {
	SETARRAY(gi->ref, gcinfopos, pointers);
//	printf("gc[%d] = %d\n", gcinfopos, GETARRAY(gi, gcinfopos));
	gcinfopos++;
	SETARRAY(gi->ref, gcinfopos, offset - nextpos);
//	printf("gc[%d] = %d\n", gcinfopos, GETARRAY(gi, gcinfopos));
	gcinfopos++;
	pointers = 1;
	nextpos = offset + sizeof(void*);
      } else {
	nextpos += sizeof(void*);
	pointers++;
      }
    }
  }
  va_end(ap);
  if (pointers > 0) { 
    SETARRAY(gi->ref, gcinfopos, pointers);
//    printf("gc[%d] = %d\n", gcinfopos, GETARRAY(gi, gcinfopos));
    gcinfopos++; 
//    if ((size - nextpos - sizeof(ObjectHead)) > 0) { 
      SETARRAY(gi->ref, gcinfopos, size - nextpos - sizeof(GCFields));
//      printf("gc[%d] = %d\n", gcinfopos, GETARRAY(gi, gcinfopos));
      gcinfopos++; 
//    }
  }
  GC___ASSIGN(th->ref, HEAPALLOC(LayoutHead, sizeof(MetaLayout), META_LAYOUT->ref))
  GC___ASSIGN(GC___PTR(th->ref)->gcInfo, gi->ref)
  GC___PTR(th->ref)->objectSize = size;

#ifdef GC_HAS_FINALIZERS 
  // Per default, objects don't have finalizers
  GC___PTR(th->ref)->finalizer = (void*) 0;
#endif

  {
    // Ugly kludge...
    GC___REF(LayoutHead) result;
    result = th->ref;
    GC__IMPL_POP_ROOT(gi);
    GC__IMPL_POP_ROOT(th);
    return result;
  }
}

void *gc_new_array(GC___REF(LayoutHead) layout, int size){
  void *result;
  /*   int size; */ // Why is this declaration doubled? AndersN
  result = 0;
  size = GC___PTR(layout)->objectSize;
  if (size == 1) {
    result = (void *) gc_new(size, layout);
  } else {
#ifndef HOST_AVR 
    printf("gc_new_array: layout has size != 1???\n");
#endif
    return result;
  }
}

#if defined JAVA_2_C && HAS_EXCEPTIONS > 0 && defined JAVA_LANG_OUTOFMEMORYERROR_INDEX
#include <exception.h>
#include <gci_common.h>
#endif

#ifdef GC_DEBUG_COUNT_ALLOC
static foo_debug_alloccount=0;
#endif
void *gc_new_2(GC___REF(LayoutHead) layout)
{
  void *result;
  int size;

  result = 0;
#ifdef GC_DEBUG_COUNT_ALLOC
  printf("Alloccount: %d\n", ++foo_debug_alloccount);
#endif
  //gc_roots_debug_print_stack();
#ifdef GC_DEBUG_FULL_CYCLE
  gc_full_cycle();
#endif
  size = GC___PTR(layout)->objectSize;
  if (size == 1) {
#ifdef GC_PRINT_WARNINGS
    printf("Arrays not implemented yet");
#endif
  } else {
#ifdef LPMEM
    if(gc_non_critical && !allow_lp_mem(size)) {
#if HAS_EXCEPTIONS > 0 && defined NOLOWPRIORITYMEMORYERROR_INDEX
	//HERE BE DRAGONS...
	gc_unlock();
	EXCEPTION_THROW(NOLOWPRIORITYMEMORYERROR_INDEX);
#else
	printf("gc_new_2: LP mem allocation denied\n");
	result = 0;
#endif

    } else {
	    result = (void *) gc_new(size, layout);
    }
#else //LPMEM
    result = (void *) gc_new(size, layout);
#endif//LPMEM
  }
  if(!result) {
#if defined JAVA_2_C && HAS_EXCEPTIONS > 0 && defined JAVA_LANG_OUTOFMEMORYERROR_INDEX
//HERE BE DRAGONS...
gc_unlock();
EXCEPTION_THROW(JAVA_LANG_OUTOFMEMORYERROR_INDEX);
#elif defined GC_PRINT_WARNINGS
    printf("gc_new_2: alloc failed... not good.\n");
#endif
  } else {
#if GC_THREAD == GC_THREAD_RTAIKERNEL
  (GC___PTR(((GC___REF(ObjectHead))result)))->waitQ_first = 0;
  (GC___PTR(((GC___REF(ObjectHead))result)))->waitQ_last = 0;
  (GC___PTR(((GC___REF(ObjectHead))result)))->lock = 0;
#elif GC_THREAD == GC_THREAD_PTHREAD
  (GC___PTR(((GC___REF(ObjectHead))result)))->lock = 0;
#endif
  }
  return result;
}

void gc_parse_object(ObjectHead *obj, void (*mark_callback)(GC___REF(ObjectHead)*))
{
	uint32 noptrs;
	uint8 *pos,*parse_pos,*end_pos;
	uint32 i;
	Uint8Array *gc_info;

	//pos = (uint8 *) &obj->tmplate;
	pos = ((uint8 *) obj)+sizeof(GCFields);
	//gc_info = obj->tmplate->gcInfo;
	gc_info = GC___PTR(GC___CPTR(LayoutHead,obj->tmplate)->gcInfo);
	parse_pos = (uint8 *) gc_info->data;
	end_pos = parse_pos+gc_info->nbrElements;
	while (parse_pos<end_pos) {
		if (*parse_pos<254) {
			 noptrs = *parse_pos;
#ifdef GC_PRINT_DEBUG_PARSE
				printf("parse_obj [0x%x : 0x%x]: normal: noptrs=%d\n", obj, pos, noptrs);
#endif
		} else {
			if (*parse_pos==254) {
				noptrs = *((uint32 *) (void *) pos);
#ifdef GC_PRINT_DEBUG_PARSE
				printf("parse_obj [0x%x : 0x%x]: escape 254: noptrs=%d\n", obj, pos, noptrs);
#endif			
				parse_pos++;
				pos += sizeof(uint32);
			}
			else { /* *parse_pos==255  */
				noptrs = ((uint32) (*(++parse_pos)))<<8;
				noptrs += *(++parse_pos);
				/* This seems strange... If the result should be an uint32, the two parts should
 				 * be uint16, and the shift 16 positions... Also, as parse_pos is an uint8*, 
  				 * isn't the ++an increment by sizeof(uint8)?
  				 */
#if ! (defined HOST_AVR || defined HOST_RTAI_KERNEL)
				printf("***************************************************\n");
				printf("HERE BE DRAGONS! escape 255 does probably not work!\n");
				printf("***************************************************\n");
#endif				
#ifdef GC_PRINT_DEBUG_PARSE
				printf("parse_obj: escape 255: noptrs=%d\n", noptrs);
#endif
			}
		}
		parse_pos++;
		for(i=0;i<noptrs;i++) {
			//checkAndPush(*(ObjectHead**) (void *)pos);
			
#ifdef GC_PRINT_DEBUG_PARSE
			printf("parse_obj [0x%x : 0x%x -> 0x%x]: calling mark\n", obj, pos, *pos);
#endif
			(*mark_callback)((void **)pos);
			pos += sizeof(void *);
		}
		if (parse_pos<end_pos) {
			if (*parse_pos<254) {
#ifdef GC_PRINT_DEBUG_PARSE
				printf("parse_obj [0x%x : 0x%x]: normal skip: %d bytes\n", obj, pos, *parse_pos);
#endif
				pos += *parse_pos;
			} else {
				if (*parse_pos==254) {
					noptrs = *((uint16 *) (void *) pos);
#ifdef GC_PRINT_DEBUG_PARSE
					printf("parse_obj [0x%x : 0x%x]: escape 254 skip: %d bytes\n", obj, pos, noptrs);
#endif
					pos += sizeof(uint16)+noptrs;
				}
				else{ /* *parse_pos==255  */
					noptrs = ((uint32) (*(++parse_pos)))<<8;
					noptrs += *(++parse_pos);
#ifdef GC_PRINT_DEBUG_PARSE
					printf("parse_obj [0x%x : 0x%x]: escape 255 skip: %d bytes\n", obj, pos, noptrs);
#endif
					pos += noptrs;
				}
			}
			parse_pos++;
		}
	}	
}

int gc_field_kind(void *object, void *field, int length)
{
  GC___REF(LayoutHead) layout;
  GC___REF(Uint8Array) gi = IVM_NULL;
  int offset;
  int result;

  result = -1;
  layout = ((struct ObjectHead*)object)->tmplate;
  gi = GC___PTR(layout)->gcInfo;
  offset = field - object;
  if (offset > GC___PTR(layout)->objectSize) {
#ifdef GC_PRINT_WARNINGS
    printf("gc_field_kind: offset outside object\n");
    printf("               object=%p, field=%p\n",object, field);
#endif
  } else {
    int first, last, i, ptr;

    first = sizeof(GCFields);
    last = sizeof(GCFields);
    i = 0;
    ptr = 0;
    while (last <= offset) {
      ptr = !ptr;
      first = last;
      if (ptr) {
	last += GETARRAY(gi, i) * sizeof(void*);
      } else {
	last += GETARRAY(gi, i);
      }
      i++;
    }
    if (offset + length > last) {
      result = -1;
    } else {
      result = ptr;
    }
  }
  return result;
}

#if 0 && defined  LPMEM // moved to gci_sched
int allow_lp_mem(void)
/* This function decides whether to grant or deny requests for LP (or
 * non-critical) memory.
 *
 * The least amount of allocatable memory required to grant requests is
 * L_{NC} = (n-i)m_{HP} + f(H-A_{start}, C) - A_{LP}
 * where  m_{HP} = the amount of memory the HP process allocates in one period
 *        H = HEAPSIZE
 *        A_{start} the amount of allocated memory at the start of the cycle
 *        C = the maximum amount of live objects and floating garbage
 *        A_{LP}=the total amount of memory allocated by lp *processes* during
 *               this cycle (critical + non-critical)
 *
 * Here, it is implemented with the decreasing reserved amount
 * At the start of each cycle, the amount of memory needed by HP allocations
 * is calculated;
 * R_{HP} = M_{HP} + f(H-A_{start}, C)
 *
 * For each critical HP allocation of size osize, R_{HP} -= osize
 * 
 * If the amount of allocatable memory <= R_{HP}, fail NC alloc request
 */
{
  if( HP_reserved > LP_alloced) {
	  // HERE BE DRAGONS! allocatable memory is not necessarily == free memory
	  // There should perhaps be a function called gc_allocatable_amount or something?
          return (gc_freemem() >= (HP_reserved-LP_alloced));
   } else {
          return 0;
   }	
}
void setGCCycleTime(gctime t){
  gc_cycle_olength=t;
}
void setHPPeriod(int t){
  T_HP = t;
}
void setHPMemoryPerPeriod(int m){
  m_HP = m;
}

uint32 getHPreserved(void) {
  return HP_reserved;
}

uint32 getNClimit(void) {
  return (HP_reserved-LP_alloced);
}

uint32 getLPalloced(void){
  return LP_alloced;
}

#endif //LPMEM

unsigned int round_up(unsigned int size, unsigned int alignment )
{
    //printf("roundup: size = %d == 0x%x\n", size, size);
    size = (size + alignment - 1) & -alignment;
    //printf("roundup: rounded size = %d == 0x%x\n", size, size);
    return size;
}

unsigned int round_down(unsigned int size, unsigned int alignment )
{
    //printf("rounddown: size = %d == 0x%x\n", size, size);
    size = size & -alignment;
    //printf("rounddown: rounded size = %d == 0x%x\n", size, size);
    return size;
}

#ifdef GC_PROFILING
/* dummy functions that are called in order to generate hit counts
 * in the profile... 
 */
void push_root_dummy(void){};
void save_root_dummy(void){};
#endif
