#ifndef GCI_COMMON_H
#define GCI_COMMON_H
/*
  gc_incremental_ms/gc.c,
  The implementation of an inremental garbage collector for the IVM.

  Copyright (C) 2000 Roger Henriksson <roger@cs.lth.se>
  Copyright (C) 2002 Sven Gestegrd Robertz <sven@cs.lth.se>

  This library is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "gc.h"
#include "gc_roots.h"
#include "gc_types.h"
#include "gci_thread.h"

/* ---------------- GC INTERFACE SECTION ----------------- */

/* CPTR: Macro for accessing a reference after a type cast.
 *   Example: ((Object *) x)->y is written CPTR(Object,x)->y
 */
//#define CPTR(ref_type,ref) (*((ref_type **) ref))
#define GC___CPTR(ref_type,ref) (GC___PTR( (GC___REF(ref_type)) ref) )


/* EQUAL: Macro for comparing two references, returns true if equal.
 *   Example: x==y is written EQUAL(x,y)
 *   Changes might be required for certain types of GC algorithms.
 */
#define EQUAL(ref1,ref2) ((void *)(ref1)==(void *)(ref2))

/* NEQUAL: Macro for comparing two references, returns true if not equal.
 *   Example: x!=y is written NEQUAL(x,y)
 *   Changes might be required for certain types of GC algorithms.
 */
#define NEQUAL(ref1,ref2) ((void *)(ref1)!=(void *)(ref2))


/* HEAPALLOC: Macro for allocating a new object on the heap.
 *   Example: x = (Object *) malloc(sizeof(Object)); x->tmplate = a_template;
 *                is written x = HEAPALLOC(Object,sizeof(Object),a_template);
 *   The macro calls an allocation function, gc_new, implemented in gc.c.
 */
#define HEAPALLOC(ref_type,size,tmplate) ((GC___REF(ref_type) ) gc_new(size,tmplate))

#define ALIGN4(_value) (((_value + 3) >> 2) << 2)

uint32 getsize(ObjectHead *ptr); 
int gc_init(int heapsize);
GC___REF(Uint8Array) newGCInfo(int iNbrElements);
void gci_common_init(int heapsize);
/**
 * Destroy and clean up the GC.
 * The static roots are popped.
 */
void GC_DESTROY(void) ;

/**
 * Create a layout from the object description v, that contains n numbers.
 * 's' is the size of the object created from this layout.
 */
GC___REF(LayoutHead) gc_new_layout(int size, int descriptors, ...);

void *gc_new_2(GC___REF(LayoutHead) layout) ;
void *gc_new_array(GC___REF(LayoutHead) layout, int size) ;
void *gc_new(uint32 size,void *tmplate);
void perform_gc(void);

void gc_parse_object(ObjectHead *obj, void (*mark_callback)(GC___REF(ObjectHead)*));

int gc_field_kind(void *object, void *field, int length) ;

unsigned int round_up(unsigned int size, unsigned int alignment );
unsigned int round_down(unsigned int size, unsigned int alignment );
#endif

void gc_full_cycle(void);
