/*
  gci_sched.c,

  GC scheduling

  Copyright (C) 2004 Sven Gestegård Robertz <sven@cs.lth.se>

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "gci_sched.h"
#include "gci_thread.h"

//#define  GC_PRINT_DEBUG_ALLOCRATE 1
/****** These should be moved to the logger file *****/

#if defined CPU_UTILIZATION_LOGGER && CPU_TIME > 0 
#define LOG_GC_INCREMENT_START cpu_utilization_logger_setGCRunning();
#define LOG_GC_INCREMENT_END cpu_utilization_logger_setGCnotRunning();
#elif defined MEASURE_CGC
#define LOG_GC_INCREMENT_START gc_cgc_start();
#define LOG_GC_INCREMENT_END   gc_cgc_stop();
#else
#define LOG_GC_INCREMENT_START
#define LOG_GC_INCREMENT_END
#endif

/***** Research hacks
 *******************/
#if defined TIME_METRIC || defined MEASURE_CGC
#include <rtai/timer.h>
typedef RTIME gctime;
#endif

#ifdef TIME_METRIC
#if GC_THREAD == GC_THREAD_RTAIKERNEL || GC_THREAD == GC_THREAD_XENOMAI  
//gctime gc_cycle_start;
//gctime gc_cycle_end;
#else
//#error untested
//long gc_cycle_start;   	// start time for current cycle
//long gc_cycle_end;     	// start_time + cycle_length
typedef long gctime;
#endif
gctime gc_cycle_start;   	// start time for current cycle
gctime gc_cycle_end;     	// start_time + cycle_length
gctime gc_cycle_length;  	// effective GC cycle length
gctime gc_cycle_olength; 	// nominal GC cycle length
unsigned int gc_cycle_olength_ms; 	// nominal GC cycle length

#if defined GC_CYCLE_TIME_ADAPTIVE
extern RT_TASK adapt_thread;
static unsigned long allocrate;
static unsigned int lastAlloced;
//static gctime lastSample;
static long lastSample;
static gctime smp_threshold;
#endif // ADAPTIVE
#endif // TIME_METRIC

#ifdef MEASURE_CGC
//#include <rtai/task.h>
gctime Cgc;
gctime cgc_start;

void gc_cgc_start(void)
{
	cgc_start = rt_timer_read();
}

void gc_cgc_stop(void)
{
	gctime now = rt_timer_read();
	if(now <= cgc_start) 
		Cgc++;
	else
		Cgc += now - cgc_start;
}
void measure_cgc_hook(void * p)
{
//	RT_TASK *t = T_DESC(p);
	RT_TASK *t = rt_task_self();
	if(t==&gc_thread){
		gc_cgc_start();
	} else {
		gc_cgc_stop();
	}
}

#endif

#if defined LPMEM
unsigned int HP_reserved;
static unsigned int LP_alloced;
static unsigned int T_HP;
static unsigned int m_HP;
static unsigned int M_HP;
static unsigned int l;          // the reduction in GC cycle length, in periods
static unsigned int live_HWM;   // "C", the maximum amount of live and
                                // floating memory
int gc_non_critical;
void gc_lpmem_init_cycle()
{
	//int eff_heapsize = ((uint8*)otabletop)-gc_heap;
	int astart = gc_freemem();
	M_HP = (gc_cycle_length / (1E6*T_HP)) * m_HP;
	live_HWM = 45000;
//    HP_reserved = M_HP;
//    HP_reserved = M_HP + ((eff_heapsize-gc_freemem() < live_HWM)?live_HWM-(eff_heapsize-gc_freemem()):0);
	HP_reserved = M_HP + ((astart < live_HWM)?live_HWM-astart:0);
	LP_alloced = 0;
	l = 0;
	printf("lpmem_init_cycle: HP_reserved = %u\n", HP_reserved);
}

int allow_lp_mem(void)
/* This function decides whether to grant or deny requests for LP (or
 * non-critical) memory.
 *
 * The least amount of allocatable memory required to grant requests is
 * L_{NC} = (n-i)m_{HP} + f(H-A_{start}, C) - A_{LP}
 * where  m_{HP} = the amount of memory the HP process allocates in one period
 *        H = HEAPSIZE
 *        A_{start} the amount of allocated memory at the start of the cycle
 *        C = the maximum amount of live objects and floating garbage
 *        A_{LP}=the total amount of memory allocated by lp *processes* during
 *               this cycle (critical + non-critical)
 *
 * Here, it is implemented with the decreasing reserved amount
 * At the start of each cycle, the amount of memory needed by HP allocations
 * is calculated;
 * R_{HP} = M_{HP} + f(H-A_{start}, C)
 *
 * For each critical HP allocation of size osize, R_{HP} -= osize
 * 
 * If the amount of allocatable memory <= R_{HP}, fail NC alloc request
 */
{
  if( HP_reserved > LP_alloced) {
	  // HERE BE DRAGONS! allocatable memory is not necessarily == free memory
	  // There should perhaps be a function called gc_allocatable_amount or something?
          return (gc_freemem() >= (HP_reserved-LP_alloced));
   } else {
	   printf("LP_Alloced > HP_Reserved\n");
          return 0;
   }	
}
void setGCCycleTime(gctime t){
  gc_cycle_olength=t;
}
void setHPPeriod(int t){
  T_HP = t;
}
void setHPMemoryPerPeriod(int m){
  m_HP = m;
}

unsigned int getHPreserved(void) {
  return HP_reserved;
}

unsigned int getNClimit(void) {
  return (HP_reserved-LP_alloced);
}

unsigned int getLPalloced(void){
  return LP_alloced;
}
#endif // LPMEM

/*********** Aux macros **************/

#if 0 && defined GC_CYCLE_TIME_ADAPTIVE
#define GC_CYCLE_TIME_ADAPT 	if(gc_is_working()) gc_cycle_time_adapt();
#else
#define GC_CYCLE_TIME_ADAPT 	
#endif

/***** Functions of the central scheduler
 ***************************************/

#if defined TIME_METRIC || defined CONTEXT_SWITCH_LOGGING
unsigned long sched_currentTimeMillis(void)
{
#if GC_THREAD == GC_THREAD_PTHREAD
    struct timeval tp;
    int time;
    gettimeofday(&tp,0);
    time = ((long) tp.tv_sec*1000)+((long) tp.tv_usec/1000);
    return time;
#elif GC_THREAD == GC_THREAD_RTAIKERNEL
    return rt_get_time_ns() >>20; // Almost 1E6
    //return rt_get_time_ns() / 1000000;
#elif GC_THREAD == GC_THREAD_XENOMAI
    return rt_timer_read() / 1E6; 
#else
    printf("WARNING! sched_currentTimeMillis not implemented!\n");
    return 0;
#endif
}
#endif

#ifdef TIME_METRIC
void gc_wait_period()
{
#if GC_THREAD == GC_THREAD_RTAIKERNEL || GC_THREAD == GC_THREAD_XENOMAI
  RTIME now = rt_timer_read();
	if(gc_cycle_end > now ){
#ifdef GC_CYCLE_TIME_ADAPTIVE
		while(gc_cycle_end > now) {
		//	int msg;
		//	rt_receive_until(&adapt_thread, &msg, gc_cycle_end);
			rt_task_sleep_until(gc_cycle_end);
			now = rt_timer_read();
			//in new RTAI; woken up by unblock()
		}
#else
		rt_task_sleep_until(gc_cycle_end);
#endif
	} else {
		printf("end < now %llu < %llu \n",gc_cycle_end,now);
	}
#else
//#error gc_wait_period not implemented
	if(gc_cycle_end>sched_currentTimeMillis()) {
		long diff = gc_cycle_end - sched_currentTimeMillis();
		//printf("gc_wait_period: sleeping\n");
		usleep(1000*diff);	
		//printf(".............done\n");
	} else {
		printf("diff < 0: %lu %lu!\n", gc_cycle_end, sched_currentTimeMillis());
	}
#endif	
}
#endif

void gci_sched_init(void)
{
#ifdef MEASURE_CGC
//	rt_task_add_hook(T_HOOK_SWITCH, &measure_cgc_hook);
//	rt_task_catch(&measure_cgc_hook);
	Cgc = 0;
	printf("gci_sched_init: Cgc = %d\n", Cgc);
#endif
#ifdef TIME_METRIC
	printf("gci_sched_init\n");
#if GC_THREAD == GC_THREAD_RTAIKERNEL || GC_THREAD == GC_THREAD_XENOMAI

#ifdef GC_CYCLE_TIME_ADAPTIVE
	//lastSample = rt_get_time();
	lastSample = sched_currentTimeMillis();
	lastAlloced = 0;
	allocrate = 0;
	//smp_threshold = nano2count(100000000);
	smp_threshold = 200;
#endif
/* 	gc_cycle_start = rt_get_time()+nano2count(GC_INIT_HOLDOFF*1000000L); */
	gc_cycle_start = rt_timer_read()+(GC_INIT_HOLDOFF*1000000L);
	gc_cycle_olength_ms = GC_CYCLE_LENGTH;
/* 	gc_cycle_olength = nano2count(gc_cycle_olength_ms*1000000L); */
	gc_cycle_olength = ((gctime)gc_cycle_olength_ms)*1000000L;
#ifdef GC_PRINT_DEBUG
	printf("gc_cycle_olength=%llu\n", gc_cycle_olength);
#endif
#elif GC_THREAD == GC_THREAD_PTHREAD
#ifdef GC_CYCLE_TIME_ADAPTIVE
#error not implemented
#endif
	gc_cycle_start = sched_currentTimeMillis+GC_INIT_HOLDOFF;
	gc_cycle_olength_ms = GC_CYCLE_LENGTH;
	gc_cycle_olength = gc_cycle_olength_ms;
#else
#error gci_sched_init/time_metric not tested for this thread model
	lastSample = Times_Now();
	lastAlloced = 0;
	allocrate = 0;
	gc_cycle_start = sched_currentTimeMillis()+GC_INIT_HOLDOFF;
	gc_cycle_olength = GC_CYCLE_LENGTH;
#endif // HOST_RTAI_KERNEL
	gc_cycle_length = gc_cycle_olength;
	gc_cycle_end = gc_cycle_start + gc_cycle_length;
#endif
#ifdef LPMEM
	m_HP = 100000;
	T_HP = 10;
	gc_non_critical = IVM_FALSE;
#endif
}

void gci_sched_init_cycle(void)
{
#ifdef MEASURE_CGC
	Cgc = 0;
#endif
#ifdef TIME_METRIC
#if GC_THREAD == GC_THREAD_RTAIKERNEL || GC_THREAD == GC_THREAD_XENOMAI
	RTIME now = rt_timer_read();
	//We only need microsecond resoluition from the timer...
	if((now+1000) < gc_cycle_end) 
#else
//#error untested
	unsigned long now = sched_currentTimeMillis();
	if(now < gc_cycle_end)
#endif //HOST_RTAI_KERNEL
	{
		printf("_____________ premature cycle start %llu, %llu\n",gc_cycle_end, now);
		gc_cycle_start = now;
	}else{
		gc_cycle_start = gc_cycle_end;
	}
	gc_cycle_length = gc_cycle_olength;
	gc_cycle_end = gc_cycle_start + gc_cycle_length;
	gc_signal_allocation();
#ifdef GC_CYCLE_TIME_ADAPTIVE
	lastSample = sched_currentTimeMillis();
	lastAlloced = 0;
#endif
#ifdef GC_PRINT_DEBUG
	printf(" GC-cycle START: freemem = %d\n", gc_freemem());
	printf("initiate_gc_cycle[%llu]: gc_start = %llu, gc_end = %llu\n", now, gc_cycle_start, gc_cycle_end);
	printf("                         length = %llu\n", (gctime)gc_cycle_length/1000);
#endif

#endif
}

extern int deltaDc, deltaDb;
extern int deltaDcp, deltaDbp;
void gci_sched_cycle_end(void)
/* Inform the scheduler that the work of the current GC cycle 
 *  has been completed.
 */
{
#ifdef MEASURE_CGC
	long a1 = 141;
	long a2 = 52;
	long a3 = -2372;
	long a4 = 116;
	//gctime Cpred = Cgc + (a3-a1)*deltaDc + (a4-a2)*deltaDb;
	gctime Cpred = Cgc + (a3-a1)*deltaDc + (a4-a2)*deltaDb;

	printf("Cgc = %llu\n", Cgc/1000);
	printf("Prediction: %llu\n", Cpred/1000);
#endif
#ifdef TIME_METRIC
#ifdef HOST_RTAI_KERNEL  || GC_THREAD == GC_THREAD_XENOMAI
	RTIME now = rt_timer_read();
#else
	//#error untested
	unsigned long now = sched_currentTimeMillis();
#endif
	if(now > gc_cycle_end) {
		gc_cycle_end = now;
		printf("_____________ late cycle end\n");
	}
#ifdef GC_PRINT_DEBUG
	printf("        cycle_end[%llu]: gc_start = %llu, gc_end = %llu\n", now, gc_cycle_start, gc_cycle_end);
	printf(" GC-cycle END: freemem = %d\n", gc_freemem());
#endif

#endif
}

//#define GC_SCHED_DEBUG
int gc_start_new_cycle(void)
{
#ifdef TIME_METRIC
	return sched_currentTimeMillis() >= gc_cycle_end;
#else
	int tmp;
	tmp = ( gc_freemem() < ( (1-GC_THRESHOLD) * gc_heapsize()) );
#ifdef GC_SCHED_DEBUG
	printf("%d, %d, %f\n", gc_freemem(), gc_heapsize(), GC_THRESHOLD * gc_heapsize());
	if(tmp) {
		printf("gc_start_new_cycle returns TRUE\n");
	} else {
	}
#endif
	return tmp;
#endif	
}

int gc_continue(void) 	/* do_more_gc() */
{
#ifdef TIME_METRIC
#ifndef GC_ONLY_TIME
	//todo... return (time_now - time_start * gc_workfctr > gc_workcnt);
	return 1;
#else
	printf("WARNING: gc_continue() should not be called in GC_ONLY_TIME\n");
	return 0;
#endif
#else
	int tmp;
	tmp =  (gc_is_working() && ( (gc_allocated()*(GC_FACTOR*gc_workfctr())+GC_OFFSET) > gc_workcnt()  ));
#ifdef GC_SCHED_DEBUG
	if(tmp) {
		printf("gc_continue returns TRUE\n");
	 } else {
		printf("%d, %d, %f, %d\n", gc_allocated(), gc_workcnt(), GC_FACTOR * (float) gc_allocated()*gc_workfctr(), gc_is_working());
		printf("gc_continue returns FALSE\n");
	}
#endif
	return tmp;
#endif // TIME_METRIC
}

#ifdef GC_IS_INCREMENTAL
void trigger_gc(void) 
{
#ifndef GC_ONLY_TIME
	if(gc_continue()) {
#ifdef GC_HAS_THREADS
		if(CURRENT_THREAD_IS_HP) {
			gc_signal_allocation();
		} else {
			perform_gc();
		}
#else
		perform_gc();
#endif
	}
#else
#ifdef GC_PRINT_WARNINGS
		printf("gci_sched.c: WARNING! trigger_gc should not be called in ONLY_TIME mode\n");
#endif
#endif // GC_ONLY_TIME
}
#endif // GC_IS_INCREMENTAL

void perform_gc(void) 
{  
#if defined TIME_METRIC && defined GC_ONLY_TIME
	while(1) {
		LOG_GC_INCREMENT_START;
		gc();
		while(gc_is_working()) {
			gc();
			LOG_GC_INCREMENT_END;
			gc_unlock();
			GC_CYCLE_TIME_ADAPT;
			gc_lock();
			LOG_GC_INCREMENT_START;
		}
		LOG_GC_INCREMENT_END;
		gc_unlock();
		/* sleep until next cycle start (i.e. cycle end)*/
		gc_wait_period();
		gc_lock();
	}
#else
	// HERE BE DRAGONS!
	// The increment logging should be tighter due ti the unlock/lock
	// which might cause a context switch. But this may be too 
	// intrusive...
	LOG_GC_INCREMENT_START;
	GC_CYCLE_TIME_ADAPT;
	while(gc_continue()) {
		gc();
		LOG_GC_INCREMENT_END;
		gc_unlock();
		gc_lock();
		LOG_GC_INCREMENT_START;
	}
	//printf("perform_gc finished increment\n");
	GC_CYCLE_TIME_ADAPT;
	LOG_GC_INCREMENT_END;
#endif
}

// HERE BE DRAGONS!
// Experimental adaptive GC length

#ifdef TIME_METRIC
#ifdef GC_CYCLE_TIME_ADAPTIVE
void gc_cycle_time_adapt()
{
gctime timeRemaining;
gctime timeElapsed;
gctime tmpNow;
#if defined STORK_PPC	
#error not tested for current thread model
	Times_Time tmpnow;
	double timeRemaining;	// until out of memory at current allocrate
	int timeDiff;
	double elapsed, smp_int;
	int momCycleTime;	// calculated cycle time at this moment
	int changed = 0;	// should we reschedule?
	
	tmpnow = Times_Now();
	elapsed = Times_DurationToSeconds(Times_Diff(tmpnow, gc_cycle_start));
	smp_int = Times_DurationToSeconds(Times_Diff(tmpnow, lastSample));
#elif GC_THREAD == GC_THREAD_RTAIKERNEL || GC_THREAD == GC_THREAD_XENOMAI
	unsigned long tmpnow = sched_currentTimeMillis();
/* 	unsigned int elapsed = tmpnow - (count2nano(gc_cycle_start)>>20); */
	//unsigned int elapsed = tmpnow - ((gc_cycle_start)>>20);
	unsigned int elapsed = tmpnow - (gc_cycle_start/1E6);
	unsigned int smp_int = tmpnow - lastSample; // in milliseconds
	unsigned int momCycleTime;  // in milliseconds
	int timeDiff;
	int changed = 0;
//	it should be like below...
//	gctime tmpnow = rt_get_time();
//	gctime elapsed = tmpnow - gc_cycle_start;
//	gctime smp_int = tmpnow - lastSample;
#else
#error not implemented for current thread model
#endif
	// sample every 0.1 seconds  and apply a moving average filter
//	if(smp_int > smp_threshold) {
		// rate in bytes/second
		unsigned int tmprate;
		unsigned int allocated = gc_allocated();
		// HACK!
		if(lastAlloced > allocated) return;
		tmprate = ((allocated - lastAlloced)) / smp_int;
#if defined  GC_PRINT_DEBUG_ALLOCRATE_MORE
		printf("allocated: %lu, last: %lu, diff: %lu, int = %ui\n", allocated, lastAlloced, allocated - lastAlloced, smp_int);
#endif
		// Sanity check: filter spurious values.
		// TODO: Find out what goes wrong and fix it.
		if( tmprate < 10000000) {
			allocrate = 0.7*allocrate + 0.3*tmprate;
			lastAlloced = allocated;
			lastSample = tmpnow;
		} 
#if defined GC_PRINT_DEBUG_ALLOCRATE_MORE
		printf("allocrate = %lu\n", allocrate);
#endif
//	} 
#ifndef ONLY_MEASURE_RATE
//	if(elapsed > 50 ) 
	if(allocrate > 5)	// HACK
	{ 
		// Reserve T_cycle * allocrate for the next cycle
		// (to make sure that we won't run out of memory during
		// the next cycle even if all the garbage floats...
		//momCycleTime = 1000.0 * ( gc_freemem() + (elapsed * allocrate)) / (2.0*allocrate);
		// f�enkla!
//		momCycleTime = 500.0 * (gc_freemem() / allocrate + elapsed);
		unsigned int timeLeft = gc_freemem() / allocrate; // in milliseconds
		momCycleTime = (timeLeft + elapsed)/2; // in milliseconds
//		timeDiff = momCycleTime - gc_cycle_olength; 
		timeDiff = momCycleTime - gc_cycle_olength_ms;
#if 1
		if(timeDiff > 100) {  // 200 ms hysteresis...
//			gc_cycle_olength_ms += timeDiff; // Increase...
			gc_cycle_olength_ms = momCycleTime;
			changed = 1;
		} else if (timeDiff < 100){ // hysteresis
//			gc_cycle_olength_ms += timeDiff; // Decrease
			gc_cycle_olength_ms = momCycleTime;
			changed = 1;
		} 
#if defined GC_PRINT_DEBUG_ALLOCRATE
		printf("elapsed: %d, momCycleTime: %d, length: %d \n", elapsed, momCycleTime, gc_cycle_olength_ms);
#endif
#ifdef LPMEM
		printf("Warning: adapt() breaks LPMEM: length calculation");
#endif
	
		if(changed) {
			//failsafe
			if(gc_cycle_olength_ms < 200) {
				gc_cycle_olength_ms = 200;
				printf("Warning: adapt() length<200. Reset to 200");
			}	
			if(gc_cycle_olength_ms > 500000) {
				gc_cycle_olength_ms = 500000;
				printf("Warning: length > 500000. Reset to 50000");
			}	

			//gc_cycle_olength = nano2count(gc_cycle_olength_ms*1000000L);
			gc_cycle_olength = ((gctime)gc_cycle_olength_ms)*1000000L;
			gc_cycle_length = gc_cycle_olength;
//			printf("=== new cycle time: %d\n", gc_cycle_olength_ms);
			gc_cycle_end = gc_cycle_start + gc_cycle_length;

			//rt_send_if(&gc_thread, 1);
			rt_task_unblock(&gc_thread);
		}	
#endif
	} 
#endif
} 
#endif //GC_CYCLE_TIME_ADAPT
#endif //TIME_METRIC
// end

#ifdef LPMEM
void gc_lpmem_alloc(unsigned int size)
{
	int currentThreadPrio, gcPrio;
#if GC_THREAD == GC_THREAD_XENOMAI
	RT_TASK_INFO info;
	rt_task_inquire(&gc_thread, &info);
	gcPrio = info.cprio;
	rt_task_inquire(rt_task_self(), &info);
	currentThreadPrio = info.cprio;
#else
#error gc_lpmem_alloc not implemented for this thread model
#endif
	if(currentThreadPrio>=gcPrio){
		if(gc_non_critical) {
			// do nothing
		} else {
			HP_reserved -= size;
		}
	} else {
		int tmp;
		LP_alloced += size;
		tmp = ceil((float)LP_alloced/m_HP);
		if(l < tmp) {
			int lp_decrement;
			l = tmp;
			lp_decrement = l*T_HP;

			if(gc_cycle_olength > lp_decrement) {
				gc_cycle_length = gc_cycle_olength - lp_decrement;
			}
			gc_cycle_end = gc_cycle_start + gc_cycle_length;
#ifdef GC_PRINT_DEBUG
			printf("l=%d, tmp = %d, gc_cycle_length=%llu, (%d/%d)\n", l, tmp, gc_cycle_length, LP_alloced, m_HP);
#endif
		}
	}

}

void gc_lpmem_panic(void)
{
#if GC_THREAD == GC_THREAD_XENOMAI
    gc_cycle_end = rt_timer_read();
#else
#error gc_lpmem_panic not implemented for this thread model
#endif
}
#endif
