/*
  gci_sched.h,

  GC scheduling

  Copyright (C) 2004 Sven Gestegård Robertz <sven@cs.lth.se>

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/***** Configuration
 ******************/
#define GC_WORK_BASED_SCHEDULING
#define GC_THRESHOLD 0.6
#define GC_MARKRATIO(a) ((a)/8.0)
#define GC_FACTOR 1.5
#define GC_OFFSET 1
#define GC_MAX_ROOTS 100

#ifdef TIME_METRIC
#define GC_CYCLE_LENGTH (2000L)
#define GC_INIT_HOLDOFF (100L)    // Wait this time to give the system a chance
                                // to reach a stable state before starting GC.
                                // May not be necessary.
#endif


/***** Functions of the central scheduler
 ***************************************/

void gci_sched_init(void);

int gc_start_new_cycle(void);
int gc_continue(void);	/* do_more_gc() */
#ifdef GC_IS_INCREMENTAL
void trigger_gc(void);
#endif


/***** Functions of the GC implementation
 ***************************************/

int gc_is_working(void);
int gc_is_marking(void);
/* Returns the amount of free memory */
unsigned int gc_freemem(void);	
/* Returns the amount of allocated memory */
unsigned int gc_allocated(void);
#ifdef GC_WORK_BASED_SCHEDULING
int gc_workcnt(void);
float gc_workfctr(void);
#endif

/***** Research hacks 
 *******************/

#ifdef GC_CYCLE_TIME_ADAPTIVE
void gc_cycle_time_adapt(void);
#endif

#ifdef GC_LPMEM
void gc_lpmem_alloc(void);	// This is probably not a good name
void gc_lpmem_init_cycle(void);
void gc_lpmem_panic(void);
#endif
