/*
  gci_thread.c,
  The thread model dependent (algorithm independent) implementation of the GCI 

  Copyright (C) 2003 Sven Gesteg�d Robertz <sven@cs.lth.se>

*/
#include <gci_thread.h>
#include <gci_sched.h>

#ifdef CPU_UTILIZATION_LOGGING
#include "cpu_utilization_logger.h"
#endif


#ifdef KERNELSPACE
#define strerror(a) "strerror not available"
#endif

#if defined GC_PRINT_WARNINGS && !(defined HOST_AVR || defined HOST_RTAIKERNEL)
#include <stdio.h>
#endif

#include <stdarg.h>

#if GC_THREAD != GC_THREAD_RTAIKERNEL  	// get size_t
#include <stdlib.h>
#else
#include <linux/types.h>
#endif

//#include <assert.h>
#include <string.h>

int gc_initialized = 0;
#ifdef GC_HAS_THREADS
// HERE BE DRAGONS! This is left in gci_common for now...
extern int gc_exited;

static int msg;

#if GC_THREAD == GC_THREAD_PTHREAD

// ******** pthread ******************************************
#include <pthread.h>
#include <java2c_lxrt.h>

pthread_mutex_t gc_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t gc_trigger_cond;
pthread_once_t key_once = PTHREAD_ONCE_INIT;
pthread_key_t gc_thread_local;
pthread_t gc_thread;
#elif GC_THREAD == GC_THREAD_STORK
// ******** STORK ********************************************
#include "Kernel.h"
#include "Monitors.h"
#include "Times.h"
#include "LCS.h"
#define printf cprintf
Monitors_Monitor gc_monitor;
Monitors_Event gc_trigger_cond;
// for RT-gc with CBS
int gc_CBSh;   // CBS handles
int hp_CBSh;
int lp_CBSh;
Nucleus_Process gc_proc; //GC process handle
// end
#elif GC_THREAD == GC_THREAD_TOBBE
// ******** tobbe ********************************************
#include <AVR_Types.h>
#include <avr/interrupt.h>
#elif GC_THREAD == GC_THREAD_T2
// ******** tobbe ********************************************
#include <AVR_Types.h>
#include <avr/interrupt.h>
#elif GC_THREAD == GC_THREAD_RTAIKERNEL
// ******** RTAI kernel **************************************
#include <rtai_sched.h>
SEM gc_mutex;
SEM gc_trigger_cond;
RT_TASK gc_thread;
#ifdef GC_ADAPTIVE_THREAD_TEST
RT_TASK adapt_thread;
//void adapt_thread_function(void *a);
#endif
#elif GC_THREAD == GC_THREAD_XENOMAI
// ******** Xenomai **************************************
#include <xenomai/native/mutex.h>
#include <xenomai/native/cond.h>
#include <xenomai/native/task.h>
//RT_MUTEX gc_mutex;
//RT_COND gc_trigger_cond;
RT_MUTEX gc_mutex;
RT_COND gc_trigger_cond;
RT_TASK gc_thread;
#ifdef GC_ADAPTIVE_THREAD_TEST
RT_TASK adapt_thread;
void adapt_thread_function(void *a);
#endif
#endif // GC_THREAD

#endif // HAS_THREADS

void gci_thread_init(void)
/** Create and initialize GC thread
 */
{
#ifdef GC_HAS_THREADS
// HERE BE DRAGONS! 
// These two lines came early in gci_common_init
// does it have to?
// Or, conversely, is it safe to create the GC thread earlier?

//  gc_exited = 0;
//  gci_thread_once_routine();  

//  end

#if GC_THREAD == GC_THREAD_PTHREAD			
//******** pthread *****************************************
	// create garbage collection thread etc.
#ifdef GC_IS_INCREMENTAL
	struct sched_param param;
	pthread_cond_init(&gc_trigger_cond, NULL);
	pthread_create(&gc_thread, 0, gc_thread_function, NULL);
 	param.sched_priority = javaPr2schedPr(_RT_NORMAL_PR_JAVA, _RT_SCHEDULING_POSIX);
	pthread_setschedparam_rt(gc_thread, _RT_SCHEDULING_POSIX, &param);
	pthread_detach(gc_thread);
#endif
	// HERE BE DRAGONS!
	// This is required to get timeslicing on Solaris (a value >=2)
	// Maybe this should be another value or another place...

	// Also: is there some way to ensure that the GC thread is
	// started right away?
#ifdef __SOLARIS__
	thr_setconcurrency(4);
#endif

#elif GC_THREAD == GC_THREAD_RTAIKERNEL			
//******** RTAIkernel **************************************

#ifdef GC_ADAPTIVE_THREAD_TEST
	rt_task_init(&adapt_thread, adapt_thread_function, 0, 10000, 3000, 0, 0);
//	if(rt_task_make_periodic_relative_ns(&adapt_thread, 0, nano2count(1000000000L)))
	if(rt_task_make_periodic(&adapt_thread, rt_get_time(), nano2count(100000000L)))
		printf("ERROR?\n");

#endif
#ifdef GC_IS_INCREMENTAL
	rt_task_init(&gc_thread, gc_thread_function, 0, 10000, 2000, 0, SIGNAL_HANDLER);
	printf("created GC task: %x\n", &gc_thread);
#ifdef CPU_UTILIZATION_LOGGER
	cpu_utilization_logger_addNamedTask(&gc_thread, "GCI");
#endif // END CPU_UTILIZATION_LOGGER

//#ifdef GC_ONLY_TIME
#if 0
	printf("making periodic...\n");
	{
		RTIME gc_period = (RTIME) GC_CYCLE_LENGTH * 1000000L;
	        //rt_task_make_periodic(&gc_thread, nano2count(GC_INIT_HOLDOFF*1000000), nano2count(10*gc_period));
	        if(rt_task_make_periodic_relative_ns(&gc_thread, GC_INIT_HOLDOFF*1000000, gc_period)
		) printf("ERROR?\n");
	}
#else
	printf("resuming gc_thread...\n");
	rt_task_resume(&gc_thread);
#endif // GC_ONLY_TIME
#endif // INCREMENTAL
#elif GC_THREAD == GC_THREAD_XENOMAI
//******** Xenomai **************************************

#ifdef GC_IS_INCREMENTAL
/* 	rt_task_init(&gc_thread, gc_thread_function, 0, 10000, 2000, 0, SIGNAL_HANDLER); */
	if (msg=rt_task_create(&gc_thread, "gc_thread", 1, 60, T_FPU)) {
	  printf("gcthread failed create: %d %s\n",msg,strerror(-msg));
	}
	if (msg=rt_task_start(&gc_thread,gc_thread_function,0)) {
	  printf("gcthread failed start: %d %s\n",msg,strerror(-msg));
	}
	printf("created GC task: %x\n", &gc_thread);
#ifdef CPU_UTILIZATION_LOGGER
	cpu_utilization_logger_addNamedTask(&gc_thread, "GCthread");
#endif // END CPU_UTILIZATION_LOGGER

#ifdef GC_ADAPTIVE_THREAD_TEST
	if (msg=rt_task_create(&adapt_thread, "adapt_thread", 1, 61,  T_FPU)) {
	  printf("adapt failed create: %d %s\n",msg,strerror(-msg));
	}
	printf("created adapt task: %x\n", &adapt_thread);
//	rt_task_create(&adapt_thread, "adapt_thread", 0, 4, T_SUSP);
//	rt_task_start(&adapt_thread, adapt_thread_function, 0);
//	if(rt_task_make_periodic_relative_ns(&adapt_thread, 0, nano2count(1000000000L)))
	if(rt_task_set_periodic(&adapt_thread, TM_NOW, 1000000000L))
		printf("ERROR?\n");

	if (msg=rt_task_start(&adapt_thread,adapt_thread_function,0)) {
	  printf("adapt failed start: %d %s\n",msg,strerror(-msg));
	}
	//rt_task_resume(&adapt_thread);
#ifdef CPU_UTILIZATION_LOGGER
	cpu_utilization_logger_addNamedTask(&adapt_thread, "GCtuner");
#endif // END CPU_UTILIZATION_LOGGER
#endif
//#ifdef GC_ONLY_TIME
#if 0
	printf("making periodic...\n");
	{
		RTIME gc_period = (RTIME) GC_CYCLE_LENGTH * 1000000L;
	        //rt_task_make_periodic(&gc_thread, nano2count(GC_INIT_HOLDOFF*1000000), nano2count(10*gc_period));
	        if(rt_task_make_periodic_relative_ns(&gc_thread, GC_INIT_HOLDOFF*1000000, gc_period)
		) printf("ERROR?\n");
	}
#else
	printf("resuming gc_thread...\n");
	rt_task_resume(&gc_thread);
	printf("gci_thread_init done...\n");
#endif // GC_ONLY_TIME
#endif // INCREMENTAL
#elif GC_THREAD == GC_THREAD_STORK			
//******** STORK *******************************************
	Console_PutString("GC", 80L);
	Monitors_New(&gc_monitor, "gc_monitor", 11);
	Monitors_NewEvent(&gc_trigger_cond, gc_monitor, "gc_trigger_cond", 16);
#if 1 // USE_CBS
	hp_CBSh = LCS_CreateCBS(0.6, Times_SecondsToDuration(0.05));
	// create a CBS for the GC thread
	gc_CBSh = LCS_CreateCBS(0.005, Times_SecondsToDuration(1.0));
	lp_CBSh = LCS_CreateCBS(0.1, Times_SecondsToDuration(0.2));
	gc_proc = LCS_CreateTaskInCBS((PROC)gc_thread_function, 30000L, 
//		(STRING)"gc_thread", 11L, LCS_RelTime(0.0), 
		(STRING)"GC", 3L, Times_Add(Times_Now(), Times_SecondsToDuration(1.0)), 
		Times_SecondsToDuration(0.0), gc_CBSh);
#else
	// STORK classic
	Kernel_CreateProcess(gc_thread_function, 30000, "gc_thread", 11);
#endif	
	Console_PutString(" GC2 ", 80L);
#endif // GC_THREAD
	//gc_initialized = 1;
#endif // GC_HAS_THREADS
}

void gci_thread_start(void)
{
#ifdef GC_HAS_THREADS
	gc_initialized = 1;
#endif

}

#ifdef GC_HAS_THREADS

void gc_exit()
{
	int s;
	gc_exited = 1;
#if defined GC_HAS_THREADS
#if GC_THREAD == GC_THREAD_RTAIKERNEL			
	if(rt_task_suspend(&gc_thread))
		printf("gc_exit: rt_task_suspend failed?\n");
	if(rt_task_delete(&gc_thread))
		printf("gc_exit: rt_task_delete failed?\n");
#elif GC_THREAD == GC_THREAD_XENOMAI
	if(s=rt_task_suspend(&gc_thread))
		printf("gc_exit: rt_task_suspend failed: %d %s\n",s, strerror(-s));
	if(s=rt_task_delete(&gc_thread))
		printf("gc_exit: rt_task_delete failed: %d %s\n",s, strerror(-s));
	if(s=rt_cond_delete(&gc_trigger_cond))
		printf("gc_exit: rt_sem_delete failed: %d %s\n",s, strerror(-s));
	if(s=rt_mutex_delete(&gc_mutex))
		printf("gc_exit: rt_sem_delete failed: %d %s\n",s, strerror(-s));
#endif
#endif
#ifdef GC_HAS_FINALIZERS
	/* cleanup, to be sure to run all finalizers, etc */
#ifndef HOST_AVR 
	printf("doing final GC cycle...\n");
#endif
	gc_full_cycle();
	gc_full_cycle();
	gc_finalize_all();
#endif
#ifndef HOST_AVR 
	printf("************GC EXIT***********\n");
#endif
}

/**********************************************************/
/* Thread model specific implementations of aux functions */
/**********************************************************/

#ifdef GC_ADAPTIVE_THREAD_TEST
void adapt_thread_function(void *a)
{
	for(;;) {
		gc_cycle_time_adapt();
		rt_task_wait_period();
	}
}
#endif

#if GC_THREAD == GC_THREAD_PTHREAD			
//******** pthread *****************************************
void gci_thread_once_routine(void) {
	pthread_key_create(&gc_thread_local, NULL);
}

void gc_thread_init(gc_thread_t *thread)
{
	// moved this call to gc_init_incremental_ms since I don't
	// understand / trust the magic...
	//pthread_once(&key_once, gci_thread_once_routine);
	pthread_setspecific(gc_thread_local, thread);
	gc_roots_insert_thread_root(&thread->root);
#if HAS_EXCEPTIONS > 0	
	thread->env = 0;
#endif	
}

void gc_thread_exit(gc_thread_t *thread)
{
	// HERE BE DRAGONS
	// here we should destroy any (heap allocated) thread local data
	// to avoid leaking memory
	gc_roots_remove_thread_root(&thread->root);
}

gc_root *gc_thread_get_root(void)
{
	gc_thread_t *tmp;
	tmp = pthread_getspecific(gc_thread_local);
	return &tmp->root;

}

gc_thread_t *gc_thread_get_current(void)
{
	gc_thread_t *tmp;
	tmp = pthread_getspecific(gc_thread_local);
	return tmp;

}

#undef EMPTY_LOCK_TEST

#ifndef GC_NO_GC_LOCKING
void gc_lock() {
  int status;
#ifdef GC_PRINT_DEBUG
  printf("gc_lock()\n");
#endif
#ifndef EMPTY_LOCK_TEST
  status = pthread_mutex_lock(&gc_mutex);
#else
  status = 0;
#endif
  if(status != 0) {
#ifdef GC_PRINT_WARNINGS
	fprintf(stderr, "error locking gc_mutex\n");
  	// HERE BE DRAGONS! we probably should handle this error
#endif
  }
}

void gc_unlock() {
  int status;
#ifdef GC_PRINT_DEBUG
  printf("gc_unlock()\n");
#endif
#ifndef EMPTY_LOCK_TEST
  status = pthread_mutex_unlock(&gc_mutex);
#else
  status = 0;
#endif
  if(status != 0) {
#ifdef GC_PRINT_WARNINGS
	fprintf(stderr, "error unlocking gc_mutex\n");
  	// HERE BE DRAGONS! we probably should handle this error
#endif
  }
}
#endif

#ifdef GC_PRINT_DEBUG
#define GC_PRINT_DEBUG_THREAD
#endif

#ifdef GC_IS_INCREMENTAL
void gc_signal_allocation()
// should be called from within gc_new
{
	pthread_cond_signal(&gc_trigger_cond);
}

void *gc_thread_function(void *arg)
{
#ifdef GC_PRINT_DEBUG_THREAD
	printf("gc_thread_function: enter\n");
#endif
	while(!gc_exited) {
#ifdef GC_ONLY_TIME
	gc_wait_period();
#endif
#ifdef GC_PRINT_DEBUG_THREAD
	printf("gc_thread_function: waiting for lock\n");
#endif
		gc_lock();
#ifdef GC_PRINT_DEBUG_THREAD
	printf("gc_thread_function: waiting to be trigged\n");
#endif
		pthread_cond_wait(&gc_trigger_cond, &gc_mutex);
#ifdef GC_PRINT_DEBUG_THREAD
	printf("gc_thread_function: increment\n");
#endif
		perform_gc();
		gc_unlock();
	}
}
#endif
#elif GC_THREAD == GC_THREAD_STORK			
//******** STORK *******************************************
static gc_thread_t *my_thread_local;

void gci_thread_once_routine(void) {
	// create thread_local data "key"/mechanism/whatever
	my_thread_local = 0;
}

void gc_thread_init(gc_thread_t *thread)
{
	// assign thread_local data for this thread
	thread->next = my_thread_local;
	thread->pid = (long) Coroutines_GetCurrent();  // HERE BE DRAGONS...
	my_thread_local = thread;
	gc_roots_insert_thread_root(&thread->root);
	//Kernel_SetPriority(50L);
}

void gc_thread_local_remove(gc_thread_t *t)
{
	gc_thread_t *tmp = my_thread_local;
	gc_thread_t *prev = NULL;
	while(NEQUAL(tmp,NULL) && tmp->pid != t->pid) {
		prev = tmp;
		tmp = tmp->next;
	}
	if(EQUAL(prev, NULL) && NEQUAL(tmp,NULL)) {
		// removing first element...
		my_thread_local = tmp->next;
		tmp->next = NULL;
	else if(NEQUAL(tmp,NULL)) {
		prev->next = tmp->next;
		tmp->next = NULL;
	}
}

void gc_thread_exit(gc_thread_t *thread)
{
	gc_roots_remove_thread_root(&thread->root);
	gc_thread_local_remove(thread);
	Kernel_Terminate();
}

gc_thread_t *gc_thread_get_current(void)
{
	gc_thread_t *tmp = my_thread_local;
	int my_pid = (long) Coroutines_GetCurrent();  // HERE BE DRAGONS...
	//FIXME: this will segfault if the search fails...
	while(tmp->pid != my_pid) {
		tmp = tmp->next;
	}
	return tmp;
}

gc_root *gc_thread_get_root(void)
{
	//gc_thread_t *tmp;
	//tmp = // get thread_local data
	gc_thread_t *tmp = my_thread_local;
	int my_pid = (long) Coroutines_GetCurrent();  // HERE BE DRAGONS...
	while(tmp != 0 && tmp->pid != my_pid) {
		tmp = tmp->next;
	}
	//printf("leaving gc_thread_get_root : result = %x\n", &tmp->root);
	if (tmp != 0)
		return &tmp->root;
	else
		return 0;

}

void gc_lock() {
	Monitors_Enter(gc_monitor);
}

void gc_unlock() {
	Monitors_Leave(gc_monitor);
}

void gc_signal_allocation()
// should be called from within gc_new
{
	Monitors_Cause(gc_trigger_cond);
}

// HERE BE DRAGONS! QnD hack - fixme!
#if GC_ALGORITHM == GC_ALGORITHM_INCREMENTAL_MS_NONMOVING
extern int gc_started;
#endif

#ifdef TIME_METRIC
#ifdef STORK_PPC
extern Times_Time gc_cycle_start;
extern Times_Time gc_cycle_end;
extern uint32 gc_cycle_olength;
#endif
#endif
extern int gc_working;
extern int gc_sleeping;
extern Nucleus_Process gc_proc;

#ifdef GC_ONLY_TIME
#ifdef STORK_PPC
void GC_SetDeadline (Times_Time deadline)
{
  LCS_Attributes att;
  Coroutines_InterruptMask oldDisable;

  oldDisable = Coroutines_Disable();
  att = (LCS_Attributes)gc_proc->procv->userReference;
  att->deadline = deadline;
//  printf("Moving deadline: ");
//  LCS_PutTime(deadline);
//  printf("\n");
  Nucleus_MoveQueue(Nucleus_ReadyQueue, gc_proc);
  iNucleus_Schedule();
  Coroutines_Reenable(oldDisable);
}

void GC_SetDeadlineAndWaitUntil (Times_Time deadline, Times_Time time)
{
  LCS_Attributes att;
  Coroutines_InterruptMask oldDisable;

  oldDisable = Coroutines_Disable();
  gc_proc->head.nextTime = time;
  att = (LCS_Attributes)gc_proc->procv->userReference;
  att->deadline = deadline;
//  printf("Moving deadline: ");
//  LCS_PutTime(deadline);
//  printf("  wake up at ");
//  LCS_PutTime(time);
//  printf("\n");
  Nucleus_MoveQueue(Nucleus_TimeQueue, gc_proc);
  iNucleus_Schedule();
  Coroutines_Reenable(oldDisable);
}

#else
extern unsigned long gc_cycle_start;   // start time for current cycle
extern unsigned long gc_cycle_end;     // start_time + cycle_length
extern uint32 gc_cycle_olength;
#endif
#endif

#ifdef GC_IS_INCREMENTAL
void gc_thread_function(void)
{
	Times_Time t;
#if GC_ALGORITHM == GC_ALGORITHM_INCREMENTAL_MS_NONMOVING
	gc_started = 1;
#endif
	cprintf("gc_thread_function: enter\n");
#ifdef TIME_METRIC
		// Hack to get timers in sync...
		t = Times_Sub(Times_Now(),Times_SecondsToDuration( ((double) gc_cycle_olength) / 1000));
		gc_cycle_start = t;
		gc_cycle_end = t;
#endif
	//Kernel_SetPriority(100L);
	while(gc_exited == 0) {
#ifdef GC_ONLY_TIME
#ifdef STORK_PPC
		gc_lock();
		if(!gc_working) initiate_gc_cycle();
		gc_sleeping = 0;
		while(gc_working) {
			int i;
			//for(i=0; gc_working && i < 2; i++) {
			//for(i=0; gc_working && i < 5; i++) {
				gc();
			//}
			//LCS_BusyWait(1);
			gc_unlock();
			gc_lock();
		}
		gc_sleeping = 1;
		gc_unlock();
		//t=Times_Add(gc_cycle_end,Times_SecondsToDuration( ((double) gc_cycle_olength) / 1000));
		//GC_SetDeadlineAndWaitUntil(t,Times_Sub(t,Times_SecondsToDuration(0.2)));
		GC_SetDeadlineAndWaitUntil(Times_Add(gc_cycle_end,Times_SecondsToDuration( ((double) gc_cycle_olength) / 1000)), gc_cycle_end);
#endif
#else
		gc_lock();
		Monitors_Await(gc_trigger_cond);
		perform_gc();
		//printf("gc: increment\n");
		gc_unlock();
#endif
	}
	//printf("=================gc: exited================\n");
	Kernel_Terminate();
}
#endif
#elif GC_THREAD == GC_THREAD_TOBBE			
//******** tobbe *******************************************
#ifndef GC_NO_GC_LOCKING
void gc_lock(void) {
  cli();
}
void gc_unlock(void) {
  sei();
}
#endif
void gc_signal_allocation(void) {}
static gc_thread_t *my_thread_local;
extern KernelThread *currentThread;

void gci_thread_once_routine(void) {
  // create thread_local data "key"/mechanism/whatever
/*   outp(~0x41,PORTB);  */
  my_thread_local = 0;
/*   outp(~0x42,PORTB);  */
}

void gc_thread_init(gc_thread_t *thread)
{
	// assign thread_local data for this thread
	thread->next = my_thread_local;
	thread->pid = currentThread;
	my_thread_local = thread;
	gc_roots_insert_thread_root(&thread->root);
}

void gc_thread_local_remove(gc_thread_t *t)
{
	gc_thread_t *tmp = my_thread_local;
	gc_thread_t *prev = NULL;
	while(NEQUAL(tmp,NULL) && tmp->pid != t->pid) {
		prev = tmp;
		tmp = tmp->next;
	}	
	if(NEQUAL(tmp,NULL)) {
		prev->next = tmp->next;
		tmp->next = NULL;
	} 
}

void gc_thread_exit(gc_thread_t *thread)
{
	gc_roots_remove_thread_root(&thread->root);
	gc_thread_local_remove(thread);
}

gc_thread_t *gc_thread_get_current(void)
{
	gc_thread_t *tmp = my_thread_local;
	int my_pid = currentThread;
	//FIXME: this will segfault if the search fails...
	while(tmp->pid != my_pid) {
		tmp = tmp->next;
	}	
	return tmp;
}


gc_root *gc_thread_get_root(void)
{
	//gc_thread_t *tmp;
	//tmp = // get thread_local data 
	gc_thread_t *tmp = my_thread_local;
	int my_pid = currentThread;
	while(tmp != 0 && tmp->pid != my_pid) {
		tmp = tmp->next;
	}
	//printf("leaving gc_thread_get_root : result = %x\n", &tmp->root);
	if (tmp != 0)
		return &tmp->root;
	else
		return 0;

}

#elif GC_THREAD == GC_THREAD_T2
//******** tobbe *******************************************
#ifndef GC_NO_GC_LOCKING
void gc_lock(void) {
  cli();
}
void gc_unlock(void) {
  sei();
}
#endif
void gc_signal_allocation(void) {}
static gc_thread_t* my_thread_local = NULL;
extern kthread_t* current_thread;

void gci_thread_once_routine()
{
	my_thread_local = NULL;
}

void gc_thread_init(gc_thread_t* thread)
{
	/* assign thread_local data for this thread */
	thread->next = my_thread_local;
	thread->pid = (long)current_thread;
	my_thread_local = thread;
	gc_roots_insert_thread_root(&thread->root);
}

void gc_thread_local_remove(gc_thread_t* t)
{
	gc_thread_t*	tmp;
	gc_thread_t*	prev;

	tmp = my_thread_local;
	prev = NULL;
	while (NEQUAL(tmp,NULL) && tmp->pid != t->pid) {
		prev = tmp;
		tmp = tmp->next;
	}	
	if (NEQUAL(tmp,NULL)) {
		prev->next = tmp->next;
		tmp->next = NULL;
	} 
}

void gc_thread_exit(gc_thread_t *thread)
{
	/* locking required? */
	gc_roots_remove_thread_root(&thread->root);
	gc_thread_local_remove(thread);
}

gc_thread_t* gc_thread_get_current()
{
	gc_thread_t*	tmp;

	tmp = my_thread_local;
	long my_pid = (long) current_thread;
	while (tmp && tmp->pid != my_pid) {
		tmp = tmp->next;
	}	
	return tmp;
}

gc_root* gc_thread_get_root()
{
	gc_thread_t*	tmp;
	gc_root*	root;
	
	/* locking required? */
	tmp = gc_thread_get_current();
	long my_pid = (long) current_thread;
	while (tmp && tmp->pid != my_pid) {
		tmp = tmp->next;
	}
	if (tmp)
		root = &tmp->root;
	else
		root = NULL;

	return root;
}

#elif GC_THREAD == GC_THREAD_RTAIKERNEL			
//******** RTAI kernel *************************************
#ifdef PREEMPTION_POINT_HACK
volatile RT_TASK *gc_heap_is_locked = 0;
volatile int gc_reschedule_flag = 0;

void (*gc_lock_p)(void) = gc_lock_real;
void (*gc_unlock_p)(void) = gc_unlock_real;

//void (*gc_lock_gc)(void) = gc_lock_real;
//void (*gc_unlock_gc)(void) = gc_unlock_real;

void hack_ensure_unlock(void)
{
	if(gc_heap_is_locked) {
		if(gc_heap_is_locked == rt_whoami()) {
			//printf("hack_ensure_unlock %x\n", rt_whoami());
			gc_unlock_gc();
		} else {
			//printf("hack_ensure_unlock locked by %x (%x)\n", gc_heap_is_locked, rt_whoami());
			gc_lock_gc();
			gc_unlock_gc();	
		}
	}
}

void gc_lock_gc(void)
{
#ifdef CPU_UTILIZATION_LOGGER
#if LOCK_COUNTING == 1
    cpu_utilization_logger_loggLock();
#endif
#endif
    rt_sem_wait(&gc_mutex);
    gc_heap_is_locked = rt_whoami();
}

void gc_unlock_gc(void)
{
    gc_heap_is_locked = 0;
    rt_sem_signal(&gc_mutex);
    //rt_global_sti();
#ifdef CPU_UTILIZATION_LOGGER
#if LOCK_COUNTING == 1
    cpu_utilization_logger_loggLock();
#endif
#endif
}

void gc_lock(void)
{
	gc_lock_p();
}
void gc_unlock(void)
{
	gc_unlock_p();
}
void gc_lock_real(void)
{
	if(gc_heap_is_locked != rt_whoami()) {
		//printf("gc_lock_real locked by %x (I'm %x)\n", gc_heap_is_locked, rt_whoami());
		rt_sem_wait(&gc_mutex);
#if LOCK_COUNTING == 1
		cpu_utilization_logger_loggLock();
#endif
		gc_heap_is_locked = rt_whoami();
		gc_lock_p = gc_f_nop;
		gc_unlock_p = gc_f_nop;
	} else {
		//printf("I already have locked the heap...\n");
#if LOCK_COUNTING == 1
		cpu_utilization_logger_loggLazyLock();
#endif
	}
}
void gc_unlock_real(void)
{
	//printf("gc_unlock_real %x\n", rt_whoami());
	gc_heap_is_locked = 0;
#if LOCK_COUNTING == 1
	cpu_utilization_logger_loggLock();
#endif
	if(gc_reschedule_flag) gc_reschedule_flag = 0;
	rt_sem_signal(&gc_mutex);
	//if(gc_reschedule_flag) rt_task_yield();
}
void gc_f_nop(void) {
	//if(gc_reschedule_flag) printf("f_nop when reschedule flag\n");
#if LOCK_COUNTING == 1
	cpu_utilization_logger_loggLazyLock();
#endif
}

void gc_reschedule_handler(void)
{
#ifdef CONTEXT_SWITCH_LOGGING
	log_context_switch();
#endif
	if(gc_heap_is_locked) {
		gc_lock_p = gc_lock_real;
		gc_unlock_p = gc_unlock_real;
		gc_reschedule_flag = 1;
		//printf("reschedule: heap is locked (by %x, I'm %x)\n", gc_heap_is_locked, rt_whoami());
		//printf("reschedule: heap is locked (by %x)\n", gc_heap_is_locked);
		//rt_task_yield();
	} else {
		gc_lock_p = gc_lock_real;
		gc_unlock_p = gc_unlock_real;
		//printf("reschedule: heap is not locked\n");
		gc_reschedule_flag = 0;
	}
}
#else
#define gc_lock_gc gc_lock
#define gc_unlock_gc gc_unlock

#ifndef GC_NO_GC_LOCKING
void gc_lock(void) {
    //printf("gc_lock %x\n",rt_whoami());
#ifdef CPU_UTILIZATION_LOGGER
#if DELAY_MEASURE == 1
    delay = rt_get_time_ns() - first;   
    cpu_utilization_logger_logg(delay);
#elif DELAY_MEASURE == 2
    cpu_utilization_logger_startLockTimer();
#endif
#if LOCK_COUNTING == 1
    cpu_utilization_logger_loggLock();
#endif
#endif
    //rt_sem_wait(&gc_mutex);
    rt_global_cli();
}
void gc_unlock(void) {
    //printf("gc_unlock %x\n",rt_whoami());
    //rt_sem_signal(&gc_mutex);
    rt_global_sti();
#ifdef CPU_UTILIZATION_LOGGER
#if DELAY_MEASURE == 2
    RTIME stop = rt_get_time_ns();
    cpu_utilization_logger_stopLockTimer(stop);
#endif
#if LOCK_COUNTING == 1
    cpu_utilization_logger_loggLock();
#endif
#endif
}
#endif // NO_GC_LOCKING
#endif // PREEMPTION POINT HACK

#if defined PREEMPTION_POINT_HACK || defined GC_GET_THREAD_ROOT_IN_SCHEDULER
void gc_reschedule_handler(void)
{
#ifdef CONTEXT_SWITCH_LOGGING
	log_context_switch();
#endif
#ifdef GC_GET_THREAD_ROOT_IN_SCHEDULER
	thread_root = gc_thread_get_root();
	//printk("rh: gtr, %x\n", thread_root);
	//if(thread_root == 0) {
	//	printk("thread_root == 0?\n");
	//	thread_root = gc_thread_get_root();
	//	printk("retry: gtr, %x\n", thread_root);
	//}
#endif

#ifdef PREEMPTION_POINT_HACK
	if(gc_heap_is_locked) {
		gc_lock_p = gc_lock_real;
		gc_unlock_p = gc_unlock_real;
		gc_reschedule_flag = 1;
		//printk("reschedule: heap is locked (by %x, I'm %x)\n", gc_heap_is_locked, rt_whoami());
		//printk("reschedule: heap is locked (by %x)\n", gc_heap_is_locked);
		//rt_task_yield();
	} else {
		gc_lock_p = gc_lock_real;
		gc_unlock_p = gc_unlock_real;
		//printk("reschedule: heap is not locked\n");
		gc_reschedule_flag = 0;
	}
#endif
}
#endif 

#ifdef GC_IS_INCREMENTAL
void gc_signal_allocation(void) 
{
	rt_cond_signal(&gc_trigger_cond);
}

void gc_thread_function(void *a)
{
#ifdef GC_PRINT_DEBUG
	printf("gc_thread_function: enter\n");
#endif
#ifdef GC_ONLY_TIME
	gc_wait_period();
#endif
	while(!gc_exited) {
#ifdef GC_PRINT_DEBUG
	printf("gc_thread_function: waiting to be trigged\n");
#endif
		rt_cond_wait(&gc_trigger_cond);
		gc_lock_gc();
#ifdef GC_PRINT_DEBUG
		printf("gc_thread_function: increment\n");
#endif
		perform_gc();
		gc_unlock_gc();
	}
#ifdef GC_PRINT_DEBUG
	printf("gc_thread_function: exiting...\n");
#endif
}
#endif

static gc_thread_t *my_thread_local;

void gci_thread_once_routine(void) {
  rt_typed_sem_init(&gc_mutex, 1, BIN_SEM);
  rt_typed_sem_init(&gc_trigger_cond, 1, BIN_SEM);
  // create thread_local data "key"/mechanism/whatever
/*   outp(~0x41,PORTB);  */
  my_thread_local = 0;
/*   outp(~0x42,PORTB);  */
}

void gc_thread_init(gc_thread_t *thread)
{
#ifdef GC_PRINT_DEBUG
	printf("gc_thread_init 1\n");
#endif
	// assign thread_local data for this thread
	thread->next = my_thread_local;
	thread->pid = (void *) rt_whoami();
	my_thread_local = thread;
#ifdef GC_PRINT_DEBUG
	printf("gc_thread_init 2\n");
#endif
	gc_roots_insert_thread_root(&thread->root);
#ifdef GC_PRINT_DEBUG
	printf("gc_thread_init 3\n");
#endif
}

void gc_thread_local_remove(gc_thread_t *t)
{
	gc_thread_t *tmp = my_thread_local;
	gc_thread_t *prev = NULL;
	while(NEQUAL(tmp,NULL) && tmp->pid != t->pid) {
		prev = tmp;
		tmp = tmp->next;
	}	
	if(EQUAL(prev, NULL) && NEQUAL(tmp,NULL)) {
		// removing first element...
		my_thread_local = tmp->next;
		tmp->next = NULL;
	}else if(NEQUAL(tmp,NULL)) {
		prev->next = tmp->next;
		tmp->next = NULL;
	} 
}

void gc_thread_exit(gc_thread_t *thread)
{
	gc_roots_remove_thread_root(&thread->root);
	gc_thread_local_remove(thread);
}

gc_thread_t *gc_thread_get_current(void)
{
	gc_thread_t *tmp = my_thread_local;
	long my_pid = (void *) rt_whoami();
	//FIXME: this will segfault if the search fails...
	while(tmp->pid != my_pid) {
		tmp = tmp->next;
	}	
	return tmp;
}


gc_root *gc_thread_get_root(void)
{
	//gc_thread_t *tmp;
	//tmp = // get thread_local data 
	gc_thread_t *tmp = my_thread_local;
	int my_pid = rt_whoami();
	while(tmp != 0 && tmp->pid != my_pid) {
		tmp = tmp->next;
	}
	//printf("leaving gc_thread_get_root : result = %x\n", &tmp->root);
	if (tmp != 0)
		return &tmp->root;
	else
		return 0;

}
#elif GC_THREAD == GC_THREAD_XENOMAI
//******** Xenomai *************************************
#ifdef PREEMPTION_POINT_HACK
volatile RT_TASK *gc_heap_is_locked = 0;
volatile int gc_reschedule_flag = 0;

void (*gc_lock_p)(void) = gc_lock_real;
void (*gc_unlock_p)(void) = gc_unlock_real;

//void (*gc_lock_gc)(void) = gc_lock_real;
//void (*gc_unlock_gc)(void) = gc_unlock_real;

void hack_ensure_unlock(void)
{
	if(gc_heap_is_locked) {
		if(gc_heap_is_locked == rt_whoami()) {
			//printf("hack_ensure_unlock %x\n", rt_whoami());
			gc_unlock_gc();
		} else {
			//printf("hack_ensure_unlock locked by %x (%x)\n", gc_heap_is_locked, rt_whoami());
			gc_lock_gc();
			gc_unlock_gc();	
		}
	}
}

void gc_lock_gc(void)
{
#ifdef CPU_UTILIZATION_LOGGER
#if LOCK_COUNTING == 1
    cpu_utilization_logger_loggLock();
#endif
#endif
/*     rt_sem_wait(&gc_mutex); */
//    rt_mutex_lock(&gc_mutex);
    rt_mutex_lock(&gc_mutex);
    gc_heap_is_locked = rt_whoami();
}

void gc_unlock_gc(void)
{
    gc_heap_is_locked = 0;
/*     rt_sem_signal(&gc_mutex); */
//      rt_sem_v(&gc_mutex);
    rt_mutex_unlock(&gc_mutex);
    //rt_global_sti();
#ifdef CPU_UTILIZATION_LOGGER
#if LOCK_COUNTING == 1
    cpu_utilization_logger_loggLock();
#endif
#endif
}

void gc_lock(void)
{
	gc_lock_p();
}
void gc_unlock(void)
{
	gc_unlock_p();
}
void gc_lock_real(void)
{
	if(gc_heap_is_locked != rt_whoami()) {
		//printf("gc_lock_real locked by %x (I'm %x)\n", gc_heap_is_locked, rt_whoami());
		rt_mutex_lock(&gc_mutex);
#if LOCK_COUNTING == 1
		cpu_utilization_logger_loggLock();
#endif
		gc_heap_is_locked = rt_whoami();
		gc_lock_p = gc_f_nop;
		gc_unlock_p = gc_f_nop;
	} else {
		//printf("I already have locked the heap...\n");
#if LOCK_COUNTING == 1
		cpu_utilization_logger_loggLazyLock();
#endif
	}
}
void gc_unlock_real(void)
{
	//printf("gc_unlock_real %x\n", rt_whoami());
	gc_heap_is_locked = 0;
#if LOCK_COUNTING == 1
	cpu_utilization_logger_loggLock();
#endif
	if(gc_reschedule_flag) gc_reschedule_flag = 0;
	rt_mutex_unlock(&gc_mutex);
	//if(gc_reschedule_flag) rt_task_yield();
}
void gc_f_nop(void) {
	//if(gc_reschedule_flag) printf("f_nop when reschedule flag\n");
#if LOCK_COUNTING == 1
	cpu_utilization_logger_loggLazyLock();
#endif
}

void gc_reschedule_handler(void)
{
#ifdef CONTEXT_SWITCH_LOGGING
	log_context_switch();
#endif
	if(gc_heap_is_locked) {
		gc_lock_p = gc_lock_real;
		gc_unlock_p = gc_unlock_real;
		gc_reschedule_flag = 1;
		//printf("reschedule: heap is locked (by %x, I'm %x)\n", gc_heap_is_locked, rt_whoami());
		//printf("reschedule: heap is locked (by %x)\n", gc_heap_is_locked);
		//rt_task_yield();
	} else {
		gc_lock_p = gc_lock_real;
		gc_unlock_p = gc_unlock_real;
		//printf("reschedule: heap is not locked\n");
		gc_reschedule_flag = 0;
	}
}
#else
#define gc_lock_gc gc_lock
#define gc_unlock_gc gc_unlock

#ifndef GC_NO_GC_LOCKING
void gc_lock(void) {
    int s;
    //printf("gc_lock %x\n",rt_whoami());
#ifdef CPU_UTILIZATION_LOGGER
#if DELAY_MEASURE == 1
    delay = rt_get_time_ns() - first;   
    cpu_utilization_logger_logg(delay);
#elif DELAY_MEASURE == 2
    cpu_utilization_logger_startLockTimer();
#endif
#if LOCK_COUNTING == 1
    cpu_utilization_logger_loggLock();
#endif
#endif
   //if(s=rt_mutex_lock(&gc_mutex)){
   if(s=rt_mutex_lock(&gc_mutex, TM_INFINITE)){
	  printf("gc_lock: mutex_lock: %d %s\n", s, strerror(-s));
  }
/*     rthal_cli(); */
/*     rthal_disable_irq(0); // DRAGONS!!!! Correct?? Platform specific?? AndersN */
}
void gc_unlock(void) {
    //printf("gc_unlock %x\n",rt_whoami());
  int s;
  //if(s=rt_mutex_unlock(&gc_mutex)){
  if(s=rt_mutex_unlock(&gc_mutex)){
	  printf("gc_unlock: mutex_unlock: %d %s\n", s, strerror(-s));
  }
/*     rthal_sti(); */
/*     rthal_enable_irq(0); // DRAGONS!!!! Correct?? Platform specific?? AndersN */
#ifdef CPU_UTILIZATION_LOGGER
#if DELAY_MEASURE == 2
    RTIME stop = rt_get_time_ns();
    cpu_utilization_logger_stopLockTimer(stop);
#endif
#if LOCK_COUNTING == 1
    cpu_utilization_logger_loggLock();
#endif
#endif
}
#endif // NO_GC_LOCKING
#endif // PREEMPTION POINT HACK
#ifdef GC_IS_INCREMENTAL
void gc_signal_allocation(void) 
{
/* 	rt_sem_signal(&gc_trigger_cond); */
	int s;
	//if(s=rt_cond_signal(&gc_trigger_cond)){
	if(s=rt_cond_signal(&gc_trigger_cond)){
		printf("gc_thread_function: rt_cond_signal: %d %s\n",s, strerror(-s));
	}
}

void gc_thread_function(void *a)
{
#ifdef GC_PRINT_DEBUG
	printf("gc_thread_function: enter\n");
#endif
	while(!gc_exited) {
#ifdef GC_ONLY_TIME
		gc_wait_period();
#ifdef GC_PRINT_DEBUG
		printf("gc_thread_function: increment\n");
#endif
		perform_gc();
#else
	int s;
#ifdef GC_PRINT_DEBUG
	printf("gc_thread_function: waiting to be trigged\n");
#endif
/* 	{ */
/* 	  RT_COND_INFO cond_info; */
/* 	  if (s=rt_cond_inquire(&gc_trigger_cond,&cond_info)) { */
/* 		printf("gc_thread_function: rt_cond_inquire: %d %s (%x,%x)\n",s, strerror(-s), &gc_trigger_cond, &cond_info); */
/* 	    } */
/* 	} */
/* 	{ */
/* 	  RT_MUTEX_INFO mutex_info; */
/* 	  if (s=rt_mutex_inquire(&gc_mutex,&mutex_info)) { */
/* 		printf("gc_thread_function: rt_mutex_inquire: %d %s (%x,%x)\n",s, strerror(-s), &gc_trigger_cond, &mutex_info); */
/* 	  } else { */
/* 	    printf("gc_thread_function: rt_mutex_inquire: %d %s\n",mutex_info.nwaiters, mutex_info.name); */
/* 	  } */
/* 	} */
	gc_lock_gc();
	//if( s=rt_cond_wait(&gc_trigger_cond,&gc_mutex,TM_INFINITE)){ //DRAGONS, Correct?? AndersN
	if( s=rt_cond_wait(&gc_trigger_cond,&gc_mutex,TM_INFINITE)){ 
		printf("gc_thread_function: rt_cond_wait: %d %s (%x,%x)\n",s, strerror(-s), &gc_trigger_cond, &gc_mutex);
	}
#ifdef GC_PRINT_DEBUG
		printf("gc_thread_function: increment\n");
#endif
		perform_gc();
		gc_unlock_gc();
#endif
	}
#ifdef GC_PRINT_DEBUG
	printf("gc_thread_function: exiting...\n");
#endif
}
#endif

static gc_thread_t *my_thread_local;

void gci_thread_once_routine(void) {
/*   rt_typed_sem_init(&gc_mutex, 1, BIN_SEM); */
/*   rt_typed_sem_init(&gc_trigger_cond, 1, BIN_SEM); */
  int s;
/*   if(s=rt_sem_create(&gc_mutex,"gc_mutex", 1, S_PRIO)){ */
  if(s=rt_mutex_create(&gc_mutex,"gc_mutex")){
#ifdef GC_PRINT_WARNINGS
	printf("gc_thread_once...: rt_mutex_create: %d %s\n",s, strerror(-s));
#endif
  }
  if(s=rt_cond_create(&gc_trigger_cond,"gc_cond")){ //, 1, S_PRIO);
/*   if(s=rt_sem_create(&gc_trigger_cond,"gc_cond", 0, S_PULSE)){ */
#ifdef GC_PRINT_WARNINGS
	printf("gc_thread_once...: rt_cond_create: %d %s\n",s, strerror(-s));
#endif
  }
#ifdef GC_PRINT_DEBUG
#endif
  printf("gci_thread_once_routine %x %x\n", &gc_mutex, &gc_trigger_cond);
  // create thread_local data "key"/mechanism/whatever
/*   outp(~0x41,PORTB);  */
  my_thread_local = 0;
/*   outp(~0x42,PORTB);  */
}

void gc_thread_init(gc_thread_t *thread)
{
	// assign thread_local data for this thread
	thread->next = my_thread_local;
/* 	thread->pid = (void *) rt_whoami(); */
	thread->pid = (void *) rt_task_self();
	my_thread_local = thread;
	gc_roots_insert_thread_root(&thread->root);
}

void gc_thread_local_remove(gc_thread_t *t)
{
	gc_thread_t *tmp = my_thread_local;
	gc_thread_t *prev = NULL;
	while(NEQUAL(tmp,NULL) && tmp->pid != t->pid) {
		prev = tmp;
		tmp = tmp->next;
	}	
	if(EQUAL(prev, NULL) && NEQUAL(tmp,NULL)) {
		// removing first element...
		my_thread_local = tmp->next;
		tmp->next = NULL;
	}else if(NEQUAL(tmp,NULL)) {
		prev->next = tmp->next;
		tmp->next = NULL;
	} 
}

void gc_thread_exit(gc_thread_t *thread)
{
	gc_roots_remove_thread_root(&thread->root);
	gc_thread_local_remove(thread);
}

gc_thread_t *gc_thread_get_current(void)
{
	gc_thread_t *tmp = my_thread_local;
/* 	long my_pid = (void *) rt_whoami(); */
	long my_pid = (void *) rt_task_self();
	//FIXME: this will segfault if the search fails...
	while(tmp->pid != my_pid) {
		tmp = tmp->next;
	}	
	return tmp;
}


gc_root *gc_thread_get_root(void)
{
	//gc_thread_t *tmp;
	//tmp = // get thread_local data 
	gc_thread_t *tmp = my_thread_local;
/* 	int my_pid = rt_whoami(); */
	int my_pid = rt_task_self();
	while(tmp != 0 && tmp->pid != my_pid) {
		tmp = tmp->next;
	}
	//printf("leaving gc_thread_get_root : result = %x\n", &tmp->root);
	if (tmp != 0)
		return &tmp->root;
	else
		return 0;

}

#else			
//******** unknown (not implemented) thread model **********
void gc_lock() {
#ifdef GC_PRINT_WARNINGS
  fprintf(stderr, "gc_lock() not implemented");
#endif
}
void gc_unlock() {
#ifdef GC_PRINT_WARNINGS
  fprintf(stderr, "gc_unlock() not implemented");
#endif
}
#endif

#else // GC_HAS_THREADS

extern gc_root *gc_thread_roots;

void gc_thread_init(gc_thread_t *thread)
{
	gc_roots_insert_thread_root(&thread->root);
}

void gc_thread_exit(gc_thread_t *thread)
{
	// HERE BE DRAGONS
	// here we should destroy any (heap allocated) thread local data
	// to avoid leaking memory
	gc_roots_remove_thread_root(&thread->root);
}

gc_root *gc_thread_get_root(void)
{
	return gc_thread_roots;
}
#endif  


