/*
  gci_thread.h,
  The thread model dependent (algorithm independent) implementation of the GCI 

  Copyright (C) 2003 Sven Gestegrd Robertz <sven@cs.lth.se>

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "gc.h"
#include "gc_roots.h"
#include "gci_common.h"

void gci_thread_init(void);
void gci_thread_start(void);
void gc_thread_init(gc_thread_t *thread);
void gc_thread_exit(gc_thread_t *thread);
gc_root *gc_thread_get_root(void);

#ifdef GC_HAS_THREADS
#ifdef GC_CYCLE_TIME_ADAPTIVE
#define GC_ADAPTIVE_THREAD_TEST 1
#endif
// HERE BE DRAGONS! Implement the macro below
#define CURRENT_THREAD_IS_HP (1)

gc_thread_t *gc_thread_get_current(void);

#if defined PREEMPTION_POINT_HACK || defined GC_GET_THREAD_ROOT_IN_SCHEDULER
#define SIGNAL_HANDLER gc_reschedule_handler
#elif defined CONTEXT_SWITCH_LOGGING
#define SIGNAL_HANDLER log_context_switch
void log_context_switch(void);
#else
#define SIGNAL_HANDLER 0
#endif

#if defined PREEMPTION_POINT_HACK || defined GC_GET_THREAD_ROOT_IN_SCHEDULER
void gc_reschedule_handler(void);
#endif

#ifdef PREEMPTION_POINT_HACK
#define HACK_ENSURE_UNLOCK hack_ensure_unlock();
void gc_lock(void);
void gc_unlock(void);
void gc_lock_real(void);
void gc_unlock_real(void);
void gc_f_nop(void);
extern void (*gc_lock_p)(void);
extern void (*gc_unlock_p)(void);
//extern int gc_heap_is_locked;
//extern int gc_reschedule_flag;
void hack_ensure_unlock(void);
#else
#define HACK_ENSURE_UNLOCK 
#ifndef GC_NO_GC_LOCKING
void gc_lock();
void gc_unlock();
#endif
#endif //PREEMPTION_POINT_HACK
void gci_thread_once_routine(void);
void gc_exit();
#ifdef GC_IS_INCREMENTAL
void gc_signal_allocation(void); // should be called from within gc_new
#endif

#if GC_THREAD == GC_THREAD_PTHREAD			
//************************************************ pthread
#ifdef GC_IS_INCREMENTAL
void *gc_thread_function(void *arg);
#endif
#elif GC_THREAD == GC_THREAD_STORK			
//************************************************ STORK

void gc_thread_local_remove(gc_thread_t *t);
#ifdef GC_IS_INCREMENTAL
void gc_thread_function(void);
#if defined GC_ONLY_TIME && defined STORK_PPC
void GC_SetDeadline (Times_Time deadline);
void GC_SetDeadlineAndWaitUntil (Times_Time deadline, Times_Time time);
#endif
#endif

#elif GC_THREAD == GC_THREAD_TOBBE			
//************************************************ tobbe

void gc_thread_local_remove(gc_thread_t *t);

#elif GC_THREAD == GC_THREAD_T2
//************************************************ T2

void gc_thread_local_remove(gc_thread_t *t);

#elif GC_THREAD == GC_THREAD_RTAIKERNEL			
//************************************************ RTAI kernel
#ifdef GC_IS_INCREMENTAL
void gc_thread_function(void *a);
extern RT_TASK gc_thread;
#endif
void gc_thread_local_remove(gc_thread_t *t);

#elif GC_THREAD == GC_THREAD_XENOMAI
//************************************************ Xenomai
#ifdef GC_IS_INCREMENTAL
void gc_thread_function(void *a);
extern RT_TASK gc_thread;
#endif
void gc_thread_local_remove(gc_thread_t *t);
#endif

#else // GC_HAS_THREADS

#endif  
