#ifndef ONCE_MEM_H
#define ONCE_MEM_H


/* Memory pool configuration ***************************************/
//
/**** Moved to config.h ****/
//#undef POWER_OF_TWO_ALIGN
// define this to make the block size a power of two
//
//#undef SANITY_CHECKS   
// define this to turn on some misc. sanity checks, mostly checking
// if pointers point into the heap, and that objects about to be freed
// are actually allocated.
//
// define to turn on coalescing of dead objects before calling free
//#undef GC_FREE_CONTIGOUS_HACK
//
//#undef INIT_ON_GC
// define this if the GC takes care of object initialisation
//
//#define INIT_ON_FREE
// define this to turn on initialising (with zeroes) when freeing blocks
//
//#undef INIT_ON_ALLOC
// define this to turn on initialising (with zeroes) when allocating blocks

#ifndef POWER_OF_TWO_ALIGN
#define HEAP_BLOCK_ALIGN_STEP 4
// Define the minimum block size alignment
#endif

// Fix for pointer arithmetic to work on both 32- and 64 bit machines
typedef long pointer;


/* configuration end ***********************************************/

/*#define MAGIC_ALLOCED 0xd530d53 */
#define MAGIC_ALLOCED 0xdeadbeef 

struct memdq {
	struct memdq *prev, *next;
	unsigned int size;
};

struct memtail {
	unsigned int size;
};

struct block {
	struct memdq head;
	void *obase;
	unsigned int  osize;
	unsigned int  oalign;
	void  *bottom;
	void  *top;
	unsigned int  alignment;
	unsigned int  freemem;
	unsigned int  totalmem;
	unsigned int allocCount;
};

void debug_print_freelist(void);

// round up size passed to alloc/free to a size that will be used
// for allocation
pointer roundup(pointer size);

// Constructor: gives the base and size of the arena in which memory is
// to be carved out.
void mempool_create( void *base  ,unsigned int size );
  
// get size bytes of memory
void * alloc( unsigned int size);
void * alloc_obj( unsigned int size);
    
// free size bytes of memory back to the pool
int hfree(void * ptr, unsigned int size);
void * hfree_obj(struct memdq * ptr, unsigned int size);

unsigned int get_freemem(void);

#include <string.h>
void do_memset(void *s, int c, size_t n);
#endif 
