/*
  testgc.c,
  A demonstration of how to use the garbage collector interface.

  Copyright (C) 2002 Anders Ive <ive@cs.lth.se>

  This library is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <string.h>
#include "gc.h"

typedef GC_STRUCT_BEGIN(Empty)
GC_STRUCT_END(Empty) Empty;

typedef GC_STRUCT_BEGIN(TestObject)
  GC_STRUCT_VAR(int, x);
  GC_STRUCT_VAR(int, y);
  GC_STRUCT_VAR(int, z);
  GC_STRUCT_REF(Object, o1);
  GC_STRUCT_REF(Object, o2);
  GC_STRUCT_VAR(int, a);
  GC_STRUCT_REF(Object, o3);
  GC_STRUCT_REF(Object, o4);
  GC_STRUCT_VAR(int, b);
  GC_STRUCT_VAR(int, c);
  GC_STRUCT_REF(Object, o5);
GC_STRUCT_END(TestObject) TestObject;

typedef GC_STRUCT_BEGIN(Refs)
  GC_STRUCT_REF(TestObject, r1);
  GC_STRUCT_REF(Refs, r2);
  GC_STRUCT_REF(Mixed, r3);
  GC_STRUCT_REF(TestObject, r4);
  GC_STRUCT_REF(Refs, r5);
  GC_STRUCT_REF(Mixed, r6);
GC_STRUCT_END(Refs) Refs;

typedef GC_STRUCT_BEGIN(Mixed)
  GC_STRUCT_REF(TestObject, t1);
  GC_STRUCT_VAR(int, x);
  GC_STRUCT_REF(TestObject, t2);
  GC_STRUCT_VAR(int, y);
  GC_STRUCT_REF(Refs, r1);
  GC_STRUCT_VAR(int, z);
  GC_STRUCT_REF(Refs, r2);
  GC_STRUCT_VAR(int, a);
  GC_STRUCT_REF(Mixed, m1);
  GC_STRUCT_VAR(int, b);
  GC_STRUCT_REF(Mixed, m2);
  GC_STRUCT_VAR(int, c);
  GC_STRUCT_REF(Mixed, m3);
GC_STRUCT_END(Mixed) Mixed;

GC_GLOBALS

GC_REF_FUNC_BEGIN(TestObject, retref, int i, GC_PARAM(TestObject, ref))
  GC_PARAM_REF(TestObject, ref);
  GC_PUSH_PARAM(ref);
  GC_FUNC_ENTER;

  GC_RETURN_REF(ref);

  printf("should not come here\n");

  GC_FUNC_LEAVE;
  GC_POP_PARAM(ref);
GC_REF_FUNC_END(TestObject, ref)

GC_VAR_FUNC_BEGIN(int, retvar, int i, GC_PARAM(TestObject, ref))
  GC_PARAM_REF(TestObject, ref);
  GC_PUSH_PARAM(ref);
  GC_FUNC_ENTER;

  GC_RETURN_VAR(i);

  printf("should not come here\n");

  GC_FUNC_LEAVE;
  GC_POP_PARAM(ref);
GC_VAR_FUNC_END(int, retvar)

GC_PROC_BEGIN(nothing, int i, GC_PARAM(TestObject, ref))
  GC_PARAM_REF(TestObject, ref);
  GC_PUSH_PARAM(ref);
  GC_FUNC_ENTER;

  GC_RETURN;

  printf("should not come here\n");

  GC_FUNC_LEAVE;
  GC_POP_PARAM(ref);
GC_PROC_END(nothing)

int main(int nbrStrs, char** strs)
{
  printf("\n*** initialisation  started ***\n");

  GC_INIT(10000);

  /* Set static roots */

  printf("\n*** allocation  started ***\n");
  {
    GC_REF(TestObject, ref);
    GC_REF(LayoutHead, TestObject_layout);
    GC_REF(LayoutHead, Refs_layout);
    GC_REF(LayoutHead, Mixed_layout);

    GC_PUSH_ROOT(ref);
    GC_PUSH_ROOT(TestObject_layout);
    GC_PUSH_ROOT(Refs_layout);
    GC_PUSH_ROOT(Mixed_layout);

    GC_NEW_LAYOUT(TestObject_layout, Empty, 0);
    GC_NEW_LAYOUT(TestObject_layout, 
		  TestObject,
		  5,
		  GC_LAYOUT_FIELD(TestObject, o1),
		  GC_LAYOUT_FIELD(TestObject, o2),
		  GC_LAYOUT_FIELD(TestObject, o3),
		  GC_LAYOUT_FIELD(TestObject, o4),
		  GC_LAYOUT_FIELD(TestObject, o5));
    GC_NEW_LAYOUT(Refs_layout, 
		  Refs,
		  5,
		  GC_LAYOUT_FIELD(Refs, r1),
		  GC_LAYOUT_FIELD(Refs, r2),
		  GC_LAYOUT_FIELD(Refs, r3),
		  GC_LAYOUT_FIELD(Refs, r4),
		  GC_LAYOUT_FIELD(Refs, r5),
		  GC_LAYOUT_FIELD(Refs, r6));
    GC_NEW_LAYOUT(Mixed_layout, 
		  Mixed,
		  5,
		  GC_LAYOUT_FIELD(Mixed, t1),
		  GC_LAYOUT_FIELD(Mixed, t2),
		  GC_LAYOUT_FIELD(Mixed, r1),
		  GC_LAYOUT_FIELD(Mixed, r2),
		  GC_LAYOUT_FIELD(Mixed, m1),
		  GC_LAYOUT_FIELD(Mixed, m2),
		  GC_LAYOUT_FIELD(Mixed, m3));


    {
      int i = 11;

      GC_NEW(ref, TestObject_layout);

      GC_REF_FUNC_CALL(ref, retref, i, GC_PASS(ref));
      GC_VAR_FUNC_CALL(i, retvar, i, GC_PASS(ref));
      GC_PROC_CALL(nothing, i, GC_PASS(ref));
    }

/*     { */
/*       int i; */
/*       for (i=0; i<1000000; i++) { */
/* 	GC_NEW(ref1, layout); */
/* 	GC_SET(ref1, x, i); */
/* 	GC_SET(ref1, y, 0xAAAAAAA1); */
/* 	GC_SET(ref1, z, 0xAAAAAAA2); */
/* 	GC_SET(ref1, c, 0xFFFFFFF1); */
/* 	if (i % 100000 == 0) { */
/* 	  GC_ASSIGN(ref3,ref2); */
/* 	  GC_FUNC_CALL(ref2, saveref, i, GC_PASS(ref1)); */
/* //	  printf("%x\n", ref1.ref); */
/* //	  printf("%x\n", ref2.ref); */
/* //	  printf("%x\n", ref3.ref); */
/* 	} */
/*       } */
/*     } */
/*     { */
/*       int x1, x2, x3; */
/*       GC_GET(x1, ref1, x); */
/*       GC_GET(x2, ref2, x); */
/*       GC_GET(x3, ref3, x); */
/*       printf("%d %d %d\n", x1, x2, x3); */
/*     } */

    GC_POP_ROOT(Mixed_layout);
    GC_POP_ROOT(Refs_layout);
    GC_POP_ROOT(TestObject_layout);
    GC_POP_ROOT(ref);
  }
  printf("\n*** execution finished ***\n");

  /* GC_POP_ROOTS static roots */

  GC_EXIT
  GC_DESTROY();
  return 0;
}
