/*
  estgc.c,
  A demonstration of how to use the garbage collector interface.

  Copyright (C) 2002 Anders Ive <ive@cs.lth.se>

  This library is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <string.h>
#include "gc.h"
#include "gci_common.h"

#ifdef EXCEPTION_TEST
#include "exception.h"
#endif

#if GC_THREAD == GC_THREAD_PTHREAD
#include <pthread.h>
static pthread_mutex_t sem = PTHREAD_MUTEX_INITIALIZER;
#elif GC_THREAD == GC_THREAD_STORK
#include "Kernel.h"
#include "Semaphores.h"
static Semaphores_Semaphore pass_sem;
static char *foo_pass_param;
#define printf cprintf
#endif

typedef GC_STRUCT_BEGIN(Empty)
GC_STRUCT_END(Empty) Empty;

typedef GC_STRUCT_BEGIN(Object)
  GC_STRUCT_VAR(int, x);
  GC_STRUCT_VAR(int, y);
  GC_STRUCT_VAR(int, z);
  GC_STRUCT_REF(Object, o1);
  GC_STRUCT_REF(Object, o2);
  GC_STRUCT_VAR(int, a);
  GC_STRUCT_REF(Object, o3);
  GC_STRUCT_REF(Object, o4);
  GC_STRUCT_VAR(int, b);
  GC_STRUCT_VAR(int, c);
  GC_STRUCT_REF(Object, o5);
GC_STRUCT_END(Object) Object;

GC_GLOBALS

GC_REF(Object, saved);

#ifdef FOO_EXCEPTION_TEST
GC_REF_FUNC_BEGIN(Object, saveref, int i, jmp_buf *exception_env, GC_PARAM(Object, saveThis))
#else
GC_REF_FUNC_BEGIN(Object, saveref, int i, GC_PARAM(Object, saveThis))
#endif
#ifdef EXCEPTION_TEST
  EXCEPTION_THROWS;
#endif
  GC_PARAM_REF(Object, saveThis);
  GC_PUSH_PARAM(saveThis);
  GC_FUNC_ENTER;
#ifdef EXCEPTION_TEST
  if( i == 500000) {
    printf("saveref: throwing exception!!!   i == %d\n",i);
    EXCEPTION_THROW(2); 
  }	  
  if( i == 700000) {
    printf("saveref: throwing exception!!!   i == %d\n",i);
    EXCEPTION_THROW(1); 
  }	  
#endif
  printf("saveref: %8d @%x\n", i, 
	 // Highly dubious code...
#ifdef GC_NODEBUG
	 saveThis.ref
#else
	 saveThis.ref.ref
#endif
    );
  if (!GC_IS_NULL(saved)) {
    GC_SET_REF(saveThis, o1, saved);
    GC_SET_REF(saveThis, o5, saved);
  }
  GC_ASSIGN(saved, saveThis);
    {
      GC_REF(Object, tmp);
      GC_PUSH_ROOT(tmp);
      
      GC_ASSIGN(tmp, saved);
      printf("old: ");
      while (!GC_IS_NULL(tmp)) {
	int val;

	GC_GET(val, tmp, x);
	printf("%8d, ", val);
	GC_GET_REF(tmp, tmp, o1);
      }
      printf("\n");
      GC_POP_ROOT(tmp);
    }

  GC_RETURN_REF(saveThis);
  printf("should not come here\n");

  GC_FUNC_LEAVE;
  GC_POP_PARAM(saveThis);
GC_REF_FUNC_END(Object, saveref)

#ifdef GC_HAS_THREADS
#if GC_THREAD == GC_THREAD_PTHREAD
void *concurrent(void *name) {
#ifdef EXCEPTION_TEST
jmp_buf exception_env;
#endif
#elif GC_THREAD == GC_THREAD_STORK
void concurrent(void) {
  char *name = foo_pass_param;
  Semaphores_Signal(pass_sem);
#endif
  GC_THREAD_INIT;
#if GC_THREAD == GC_THREAD_STORK
  Kernel_SetPriority(50L);
#endif  
  printf("concurrent: %s\n", name);
  GC_BEGIN_ROOT(LayoutHead, layout);
  GC_BEGIN_ROOT(Object, ref1);
  GC_BEGIN_ROOT(Object, ref2);

  GC_NEW_LAYOUT(layout, 
		Object,
		5,
		GC_LAYOUT_FIELD(Object, o1),
		GC_LAYOUT_FIELD(Object, o2),
		GC_LAYOUT_FIELD(Object, o3),
		GC_LAYOUT_FIELD(Object, o4),
		GC_LAYOUT_FIELD(Object, o5));
  
  {
#ifdef EXCEPTION_TEST
    EXCEPTION_TRY {
#endif
    int i;
    for (i=0; i<1000000; i++) {
      GC_NEW(ref1, layout);
      GC_SET(ref1, x, i);
      GC_SET(ref1, y, 0xAAAAAAA1);
      GC_SET(ref1, z, 0xAAAAAAA2);
      GC_SET(ref1, c, 0xFFFFFFF1);
      if (i % 100000 == 0) {
#if GC_THREAD == GC_THREAD_PTHREAD
pthread_mutex_lock(&sem);
#endif
#ifdef EXCEPTION_TEST
	EXCEPTION_TRY {
#endif
	GC_REF_FUNC_CALL(ref2, saveref, i, GC_PASS(ref1));
#ifdef EXCEPTION_TEST
	} EXCEPTION_CATCH(2) {
		printf("%s: exception in saveref. i == %d\n",name, i);	
	} 
	EXCEPTION_AFTER_CATCH
#endif
#if GC_THREAD == GC_THREAD_PTHREAD
pthread_mutex_unlock(&sem);
#endif
	//sleep(1);
#if GC_THREAD == GC_THREAD_STORK
	Kernel_WaitTime(2);
#endif
	
      }
    }
#ifdef EXCEPTION_TEST
  } EXCEPTION_CATCH(1) {
    printf("%s: exception type 1\n",name);	
#ifdef GC_HAS_THREADS
#if GC_THREAD == GC_THREAD_PTHREAD
	  pthread_mutex_unlock(&sem);
#endif
#endif
  }
  EXCEPTION_FINALLY {
    printf("%s: **FINALLY**\n",name);	
  }
  EXCEPTION_AFTER_CATCH
//  }
#endif
  }
  GC_END_ROOT(ref2);
  GC_END_ROOT(ref1);
  GC_END_ROOT(layout);
  printf("concurrent: %s\n", name);
#if GC_THREAD == GC_THREAD_STORK  
  // HERE BE DRAGONS! fix me!
  Kernel_SetPriority(150L);
  Kernel_Terminate();
#endif  

  GC_THREAD_EXIT;
}
#endif

#if GC_THREAD == GC_THREAD_STORK
int foo_java_main()
#else
int main(int nbrStrs, char** strs)
#endif
{
#ifdef GC_HAS_THREADS
#if GC_THREAD == GC_THREAD_PTHREAD
  pthread_t t1, t2, t3;
#endif
#endif

  printf("\n*** initialisation  started ***\n");

  GC_INIT(100000);
  {

  /* Set static roots */
  GC_PUSH_ROOT(saved);

#ifdef GC_HAS_THREADS
#if GC_THREAD == GC_THREAD_PTHREAD
  pthread_create(&t1, 0, concurrent, "t1");
  pthread_create(&t2, 0, concurrent, "t2");
  pthread_create(&t3, 0, concurrent, "t3");
#elif GC_THREAD == GC_THREAD_STORK
  Semaphores_New(&pass_sem, 1, "ps", 3);
  Semaphores_Wait(pass_sem);
  foo_pass_param = "t1";
  Kernel_CreateProcess(concurrent, 30000, "t1", 3);
  Semaphores_Wait(pass_sem);
  foo_pass_param = "t2";
  Kernel_CreateProcess(concurrent, 30000, "t2", 3);
  Semaphores_Wait(pass_sem);
  foo_pass_param = "t3";
  Kernel_CreateProcess(concurrent, 30000, "t3", 3);
#endif
#endif

  printf("\n*** allocation  started ***\n");

#if GC_THREAD == GC_THREAD_STORK
  Kernel_SetPriority(50L);
#endif  
  {
    GC_REF(Object, ref1);
    GC_REF(Object, ref2);
    GC_REF(Object, ref3);
    GC_REF(LayoutHead, layout);

    GC_PUSH_ROOT(ref1);
    GC_PUSH_ROOT(ref2);
    GC_PUSH_ROOT(ref3);
    GC_PUSH_ROOT(layout);

    {
      GC_NEW_LAYOUT(layout, Empty, 0);
      GC_NEW_LAYOUT(layout, 
		    Object,
		    5,
		    GC_LAYOUT_FIELD(Object, o1),
		    GC_LAYOUT_FIELD(Object, o2),
		    GC_LAYOUT_FIELD(Object, o3),
		    GC_LAYOUT_FIELD(Object, o4),
		    GC_LAYOUT_FIELD(Object, o5));
    }

    GC_NEW(ref1, layout);
    GC_NEW(ref2, layout);
    GC_NEW(ref3, layout);

    {
      int i;
      for (i=0; i<1000000; i++) {
	GC_NEW(ref1, layout);
	GC_SET(ref1, x, i);
	GC_SET(ref1, y, 0xAAAAAAA1);
	GC_SET(ref1, z, 0xAAAAAAA2);
	GC_SET(ref1, c, 0xFFFFFFF1);
	if (i % 100000 == 0) {
	  GC_ASSIGN(ref3,ref2);
#ifdef GC_HAS_THREADS
#if GC_THREAD == GC_THREAD_PTHREAD
	  pthread_mutex_lock(&sem);
#endif
#endif
#ifdef EXCEPTION_TEST
	EXCEPTION_TRY {
#endif
	  GC_REF_FUNC_CALL(ref2, saveref, i, GC_PASS(ref1));
#ifdef EXCEPTION_TEST
	} EXCEPTION_CATCH(2) {	
		printf("main: exception in saveref. i == %d\n",i);	
	} EXCEPTION_CATCH_MORE(1) {	
		printf("main: caught exception 1 in saveref. i == %d. I can take this...\n",i);	
	} EXCEPTION_FINALLY {
		printf("main: finally ************* %d\n", i);
	}
	EXCEPTION_AFTER_CATCH
//	}
#endif
#ifdef GC_HAS_THREADS
#if GC_THREAD == GC_THREAD_PTHREAD
	  pthread_mutex_unlock(&sem);
#endif
#endif
//	  printf("%x\n", ref1.ref);
//	  printf("%x\n", ref2.ref);
//	  printf("%x\n", ref3.ref);
#if GC_THREAD == GC_THREAD_STORK
	Kernel_WaitTime(2);
#endif
	}
      }
    }
    {
      int x1, x2, x3;
      GC_GET(x1, ref1, x);
      GC_GET(x2, ref2, x);
      GC_GET(x3, ref3, x);
      printf("%d %d %d\n", x1, x2, x3);
    }
foo_label:
    GC_POP_ROOT(layout);
    GC_POP_ROOT(ref3);
    GC_POP_ROOT(ref2);
    GC_POP_ROOT(ref1);
  }
  printf("\n*** execution finished ***\n");

#ifdef GC_HAS_THREADS
#if GC_THREAD == GC_THREAD_PTHREAD
  pthread_join(t1, 0);
  pthread_join(t2, 0);
  pthread_join(t3, 0);
#elif GC_THREAD == GC_THREAD_STORK  
  // HERE BE DRAGONS! fix me!
  Kernel_SetPriority(150L);
  Kernel_Terminate();
#endif  
#endif
  printf("\n*** leaving main ***\n");

  /* GC_POP_ROOTS static roots */
  GC_POP_ROOT(saved);

  }
  GC_EXIT
  GC_DESTROY();
  return 0;
}
