/**
 * An absolute minimalist implementation of java.lang.String
 *
 * 2010 Jesper Öqvist <jesper.oqvist@cs.lth.se>
*/

package java.lang;

public final class String
{
	/* package access for StringBuffer */
  final char[] value;
  final int count;
  final int offset;

  public String()
  {
    value = null;
    offset = 0;
    count = 0;
  }

  public String(String str)
  {
    value = str.value;
    offset = str.offset;
    count = str.count;
    //cachedHashCode = str.cachedHashCode;
  }

  /**
   * Returns the number of characters contained in this String.
   *
   * @return the length of this String
   */
  public int length()
  {
    return count;
  }

  /**
   * Returns the character located at the specified index within this String.
   *
   * @param index position of character to return (base 0)
   * @return character located at position index
   * @throws IndexOutOfBoundsException if index &lt; 0 || index &gt;= length()
   *         (while unspecified, this is a StringIndexOutOfBoundsException)
   */
  public char charAt(int index)
  {
//     if (index < 0 || index >= count)
//       throw new StringIndexOutOfBoundsException(index);
    return value[offset + index];
  }

  /**
   * Predicate which compares anObject to this. This is true only for Strings
   * with the same character sequence.
   *
   * @param anObject the object to compare
   * @return true if anObject is semantically equal to this
   * @see #compareTo(String)
   * @see #equalsIgnoreCase(String)
   */
  public boolean equals(Object anObject)
  {
//     if (! (anObject instanceof String))
//       return false;
    String str2 = (String) anObject;
    if (count != str2.count)
      return false;
    if (value == str2.value && offset == str2.offset)
      return true;
    int i = count;
    int x = offset;
    int y = str2.offset;
    while (--i >= 0)
      if (value[x++] != str2.value[y++])
        return false;
    return true;
  }

  /**
   * Returns this, as it is already a String!
   *
   * @return this
   */
  public String toString()
  {
    return this;
  }
}
