/* Modified for use with Java2C translator
 *
 * Anders Nilsson <anders.nilsson@cs.lth.se>
*/

package java.lang;

public class Thread implements Runnable {


	/* Identifier of this Thread Object*/
	private long target;

	/* Whether or not to interrupt this thread */
	// Moved interruptedFlag to java.lang.Object instead so that we
	// may have the same monitor implementation for both Thread and
	// RTThread. AndersN
// 	private boolean interruptedFlag;
	/* Thread priority */
	private int priority;
   /* Runnable object */
	private Runnable targetRun;
	/* Pointer to the TSD - Oooops */
	private long TSD;
	/* Wether of not this thread is alive */
	private boolean isAlive;

	/* Priorities */
	public static final int MAX_PRIORITY = 10;
	public static final int MIN_PRIORITY = 1;
	public static final int NORM_PRIORITY = 5;
	
    /** The thread name, non-null. */
    volatile String name;


	/**
	 * Method to construct a java.lang.Thread Object.
 	 * Invokes a native method in java_lang_Thread.c file.
	 * @returns java.lang.Thread Object
	 **/ 
    public Thread() {
		if (false) {
			InterruptedException e = new InterruptedException();
		}
		long auxLong;
		int  auxInt;
 		this.interruptedFlag = false;
 		auxLong = this.createThread();
		this.target = auxLong; 
		this.targetRun = null;
 		this.inheritPriority();		
    }
	

    public Thread(String name) {
	this();
	this.name = name;
    }

	/**
	 * Method to construct a java.lang.Thread Object.
 	 * Invokes a native method in java_lang_Thread.c file.
	 * @param   target Runnable object to invoke in start().
	 * @returns java.lang.Thread Object
	 **/ 
    public Thread(Runnable target) {
		long auxLong;
		int  auxInt;
		this.interruptedFlag = false;
		auxLong = this.createThread();
		this.target = auxLong; 
		this.inheritPriority();		
		this.targetRun = target;
    }

    /**
     * Get this Thread's name.
     *
     * @return this Thread's name
     */
    public final String getName()
    {
	if (name == null) {
	    return new String();
	} else {
	    return name;
	}
    }


    private native void inheritPriority();
		
	/**
	 * Native method coupled to the constructor of the class java.lang.Thread
	 * @returns A long representing the ID of the thread instance
	 **/ 
    
    private final native long createThread();

	/**
	 * Creates a new thread and executes run() method defined in this Thread 
	 * class. Invokes a native method in java_lang_Thread.c file.	 
	 **/ 
    public native void start();

	/**
	 * It should be overriden with the portion of code that the developer
	 * wants to be executed by the newly created Thread object.
	 **/ 	
    public void run() {}
	

	/**
	 * Specify a number of milliseconds during the Thread object will be slept.
	 * @param millis Number of millisecondsthe thread will be slept
	 **/ 
    public static native void sleep(long millis) throws InterruptedException;

    public boolean isAlive() {
	    return isAlive;
    }

  public final boolean isDaemon()
  {
//    VMThread t = vmThread;
//    return t == null ? daemon : t.isDaemon();
	return false;
  }
	/**
	 * Specify a number of milliseconds and a number of nanoseconds during the Thread 
	 * object will be slept.
	 * @param millis Number of millisecondsthe thread will be slept
	 **/ 
    public static native void sleep(long millis,int nanos) throws InterruptedException;

	/** 
	 * Tests if this thread has been interrupted. The interrupted status is unaffected by this method
	 * @return Status of the interrupted flag.
	 **/
	 public boolean isInterrupted() {
		boolean tmp;
		tmp = this.interruptedFlag;
	 	return tmp;
	 }

	/** 
	 * Tests if this thread has been interrupted. The interrupted status is cleared by this method
	 * @return Status of the interrupted flag.
	 **/
	 public static boolean interrupted() {
		 Thread t = currentThread();
		 boolean flag = t.interruptedFlag;
		 t.interruptedFlag = false;
		 return flag;
	 }

	/** 
	 * Sets interruptedFlag to true
	 **/
	 public native void interrupt() ;
	 
	/** 
	 * @return A reference to the object which made the call to currentThread(). 
	 **/
	 public static native Thread currentThread();

	/** 
	 * Returns this thread's priority  
	 * @return this thread's priority 
	 **/
	 public final int getPriority() {
		int tmp;
		tmp = this.priority;
	 	return tmp;
	 }

	/** 
	 * Changes the priority of this thread. 
	 * @return this thread's priority 
	 * @throw  IllegalArgumentException if the priority is not in the range MIN_PRIORITY
	 *  	   MAX_PRIORITY.
	 **/
	 public final native void setPriority(int priority) throws IllegalArgumentException ;
	 

	/** 
	 * Waits for this thread to die.
	 * @throw  InterruptedException if the thread that is waiting is interrupted.
	 **/
	 public final native void join() throws InterruptedException ;

	/** 
	 * Waits at most millis milliseconds for this thread to die.
	 * @throw  InterruptedException if the thread that is waiting is interrupted.
	 **/
	 public final native void join(long millis) throws InterruptedException ;
	 
	/** 
	 * Waits at most millis milliseconds plus nanos nanoseconds for this thread to die.
	 * @throw  InterruptedException if the thread that is waiting is interrupted.
	 **/
	 public native void join(long millis, int nanos) throws InterruptedException ;

	/** 
	 * Causes the currently executing thread object to temporarily pause and allow other 
	 * threads to execute.
	 **/
	public static native void yield();
}
