/* $Id: Monitor.java,v 1.7 2005/08/19 09:37:16 d02en Exp $       
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime;

/**
 * Support for the concept of monitors ([Hoare 1974]) by imposing further rules, such 
 * as limitations on external syncronization. An object implementing
 * a <code>Monitor</code> must be thread safe as expressed by the 
 * base-interface <code>Synchronized</code>, but also the following applies for code
 * that uses/inherits/implements a <code>Monitor</code> object:
 * <ul>
 * <li> Using only the keyword <code>synchronized</code>, or even implementing the 
 *      interface <code>Synchronized</code>, it can be the case that other objects 
 *      calls wait and notify on an object that needs to be thread safe. That imposes
 *      restriction on method implementations, typically making the less readable or 
 *      less efficient (taking care of excessive unknown notifications, and the like).
 *      By implementing this interface, you state that such external synchronization is 
 *      forbidden.
 * <li> Methods that are non-blocking (not calling <code>wait</code>) in a non-abstract 
 *      base class may not contain calls to wait in a subclass. During <code>wait</code>
 *      the object is temporarily unlocked, which other monitor methods (as defined in 
 *      the base class) are designed to handle. By not permitting subclass methods to 
 *      introduce waiting, some cases of deadlock is avoided and (more important) 
 *      concurrency faults due to unexpected unlocking when base class methods call 
 *      virtual (non-final) methods that unexpectedly breaks the critical section.
 * <li> Special support for multiple conditions is enabled, that is, having multiple 
 *      <code>Condition</code> objects for calling wait/notify on more specific monitor
 *      conditions This is made possible by the first item above. For compiled Java this
 *      depends on the compiler. For JVM execution the application/applet should
 *      be started via <code>Main.main</code>, for best efficiency.
 * </ul>
 * These items all contribute to making a <code>Monitor</code> object being more 
 * conform with definitions in the literature, but also better corresponding to
 * typical implementations in available RTOS:s, thereby facilitating compilation to 
 * efficient executables for such platforms.
 * 
 * @see License
 */

public interface Monitor extends Synchronized {}
