/* $Id: RTError.java,v 1.5 2005/08/19 09:37:16 d02en Exp $        
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime;

/**
 * Unchecked exception to be thrown when handling time, concurrency, or 
 * resources are interrupted or violated in such a way that there is normally
 * nothing better to do than closing down the application. In rare cases, 
 * an <code>RTError</code> may be caught, but typically then only for handling
 * shutdown or emergency situations. Refer to <code>java.lang.Error</code>
 * for constructors and their arguments.
 * 
 * @see License
 */

public class RTError extends Error {

  /**
   * The time stamp of this error, in number of (system dependent)
   * ticks, referring to the time when the error occurred (was created).
   */
  protected volatile long timestamp;

  /**
   * Sequence number, ordering errors even within the same tick.
   */
  protected volatile int number;

  /**
   * Sequencing number to be assigned to the next error. 
   */
  private static volatile int nextErrorNum;

  public RTError() {
    super();
    number = nextErrorNum++;
    timestamp = System.currentTimeMillis();
  }

  public RTError(String error) {
    super(error);
    number = nextErrorNum++;
    timestamp = System.currentTimeMillis();
  }

  public RTError(String error, Throwable cause) {
    super(error, cause);
    number = nextErrorNum++;
    timestamp = System.currentTimeMillis();
  }

  public RTError(Throwable cause) {
    super(cause);
    number = nextErrorNum++;
    timestamp = System.currentTimeMillis();
  }

  public long getMillis() {
    return timestamp;
  }

  public int getSequenceNumber() {
    return number;
  }
}
