/* $Id: RTSystem.java,v 1.8 2005/08/15 13:24:44 d02en Exp $        
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime;

/**
 * @see License
 */

// Under construction; this class will contain system dependent properties
// that may need to be available from within the software. One example is
// the type of priority inheritance used.

public class RTSystem implements RTSystemProperties {

// HERE BE DRAGONS !!! Remove me
//  public static boolean NATIVE_SEM = RTSystemProperties.NATIVE_SEM;
// !!
	
  static int tickMillis = (OS_TICK_MICRO>0) ? (DEPLOYED ? OS_TICK_MICRO : 100) : 0;
  static int tickNanos = 0;
  static int defaultEventBufferSize = 64;
  static int maxEventBufferSize = 1024;
  
  /**
   * @return default generic buffer size.
   */
  public static int getDefaultEventBufferSize() {
    return defaultEventBufferSize;
  }

  /**
   * @return upper limit of buffer sizes.
   */
  public static int getMaxEventBufferSize() {
    return maxEventBufferSize;
  }

  /**
   * @return clock tick period in number of milli-seconds.
   */
  public static int getTickMillis() {
    return tickMillis;
  }

  /**
   * @return clock tick period in number of nano-seconds.
   */
  public static int getTickNanos() {
    return tickNanos;
  }

  //----------------------------------- Features for virtual/application time:

  /** Watchdog thread for application time handling. */
//   private static Thread ticker;
//   static {
//     if (!DEPLOYED)
// //       (ticker = new HiddenThread()).start();
//        ticker = new HiddenThread();
//   }

  /** Tick counter for hidden thread below. */
  static transient volatile int seq;

  /** Application time lag compared to system real time. */
  static transient volatile int lag;

  /** Permitted delay for ticker thread before lag increases. Must be > tickMillis+1 below. */
  protected static transient volatile int tol = 1000;

  /**
   * The system time most recently obtained by the hidden thread below.
   */
  private static transient volatile long systime;

  /** 
   * Last time returned by <code>RTSystem.currentTimeMillis</code>. 
   * No other method should use this attribute, which is needed to ensure
   * that obtained time never goes backwards, even if <code>lag</code>
   * increases suddenly <code>PRIORITY_STRICT</code> is <code>false</code>.
   */
  private static transient volatile long last = System.currentTimeMillis();

  /**
   * Obtain current 'real' time. If the JVM is not suspended (due to
   * breakpoints and the like), if priorities are strictly
   * considered for the scheduling, if deadlines are met, and if the
   * time-base is based on unscaled real time, the returned time
   * will be equal to <code>System.currentTimeMillis()</code>. For
   * any two subsequent calls (from any threads), the second call is
   * guaranteed to return a value greater than or equal to the value
   * returned from the first call. That is, the time never decreases.
   *
   * @return Real time adjusted with respect to platform or
   * debugging delays, permitting real-time programs to be tested
   * from a concurrency point of view using a JVM or OS not
   * providing real-time.
   */
    public static native long currentTimeMillis();
//  {
// //     if (DEPLOYED)
//       return System.currentTimeMillis();
// //     else {
// //       synchronized (RTSystem.class) { // For the usage of last.
// //         long now = System.currentTimeMillis() - lag;
// //         if (now > last)
// //           last = now;
// //         return last;
// //       }
// //     }
//   }

    public static native long currentTimeMicros();


  /**
   * Causes the currently executing thread to sleep (temporarily cease
   * execution) until the specified moment in time.
   * Time is measured in milliseconds, starting from 00am, January 1st
   * 1970, UTC. The thread does not lose ownership of any monitors.
   *
   * @param  wakeUpTime   the time to wake up.
   * @throws RTInterrupted if another thread has interrupted this thread.
   * @throws RTDelayed if execution continues more than
   *         <code>maxTimeLag</code> too late.
   */
  public static native void sleepUntil(long wakeUpTime) throws RTInterrupted, RTDelayed; 
  public static native void sleepUntilMicros(long wakeUpTime) throws RTInterrupted, RTDelayed; 
//  public static void sleepUntil(long wakeUpTime) throws RTInterrupted, RTDelayed {
//    /*
//     * Note: All sleeping here must refer to java.lang.Thread.sleep.
 //    */
 //   int diff = (int) (wakeUpTime - RTSystem.currentTimeMillis());
  //  int slack = Integer.MAX_VALUE;
 //   RTThread rtt = RTThread.currentRTThread();
 //   if (rtt != null)
 //     slack = rtt.getMaxTimeLag();
 //   if (DEPLOYED) {
 //     if (diff > 0) {
 //       try {
 //         Thread.sleep(diff);
 //       }
 //       catch (InterruptedException exc) {
 //         throw new RTInterrupted("RTThread.sleepUntil", exc);
 //       };
 //     }
 //     else if (diff < -slack) {
 //       throw new RTDelayed(wakeUpTime, -diff);
 //     }
 //   }
 //   else {
 //     // Get a local snapshot of current delay state:
 //     int lag0 = lag;
 //     long now = System.currentTimeMillis();
 //     long t0 = now - lag; // Initial emulated real time.
 //     long dt = wakeUpTime - t0; // Remaining time before timeout.
 //     int s = seq; // Lag updated w.r.t. this seq number.
 //     while (dt > 0) {
 //       try {
 //         Thread.sleep(dt);
 //         s = seq;
 //         Thread.sleep(tickMillis);
 //       }
 //       catch (InterruptedException exc) {
 //         throw new RTInterrupted("RTThread.sleepUntil", exc);
 //       };
 //       while (seq == s)
 //         RTThread.yield();
 //       // If there was a suspend during await, get new time lag:
 //       if (lag0 < lag)
 //         lag0 = lag;
 //       // Update timeout, and continue loop if virtual time is not up:
 //       now = currentTimeMillis();
 //       dt = wakeUpTime - (now - lag); // timeout_time - current_time
 //     }
 //   }
 // }

  /**
   * Causes the currently executing thread to sleep (temporarily cease
   * execution) for the specified number of milliseconds. The thread
   * does not lose ownership of any monitors.
   *
   * @param  millis   the length of time to sleep in milliseconds.
   * @throws IllegalArgumentException  if the value of millis is negative.
   * @throws RTInterrupted if another thread has interrupted this thread.
   * @throws RTDelayed if execution continues more than
   *         <code>maxTimeLag</code> too late.
   * @see    java.lang.Thread#sleep(long)
   */
  public static void sleep(long millis) throws RTInterrupted, RTDelayed {
    if (millis < 0) {
      throw new IllegalArgumentException("timeout value is negative");
    }
    long wakeUp = currentTimeMillis() + millis;
    sleepUntil(wakeUp);
  }
  
  /**
   * Causes the currently executing thread to sleep (cease execution)
   * for the specified number of milliseconds plus the specified number
   * of nanoseconds. The thread does not lose ownership of any monitors.
   *
   * @param  millis   the length of time to sleep in milliseconds.
   * @param  nanos    0-999999 additional nanoseconds to sleep.
   * @throws IllegalArgumentException  if the value of millis is negative.
   * @throws RTInterrupted if another thread has interrupted this thread.
   * @throws RTDelayed if execution continues more than
   *         <code>maxTimeLag</code> too late.
   * @see    java.lang.Thread#sleep(long, int)
   */
  public static void sleep(long millis, int nanos) throws RTInterrupted, RTDelayed {
    if (millis < 0) {
      throw new IllegalArgumentException("timeout value is negative");
    }

    if (nanos < 0 || nanos > 999999) {
      throw new IllegalArgumentException("nanosecond timeout value out of range");
    }

    if (tickMillis > 0) {
      if (nanos != 0)
        millis++;
    }
    else {
      nanos = nanos * (nanos / tickNanos + 1);
      if (nanos > 999999) {
        nanos -= tickNanos;
        millis++;
      }
    }
    if (nanos == 0) {
      long wakeUp = currentTimeMillis() + millis;
      sleepUntil(wakeUp);
    }
    else {
      throw new UnsupportedOperationException("nano sleep not available");
    }
  }

  /**
   * Thread used to detect suspended JVM execution.
   */
//   private static class HiddenThread extends java.lang.Thread {

//     /**
//      * Increment <code>time</code> (used by
//      * <code>RTSystem.currentTimeMillis()</code>) each tick, keeping
//      * it within the tolerance <code>tol</code> from system time by
//      * adjusting the internal <code>lag</code>.
//      */
//     public void run() {
//       setPriority(MAX_PRIORITY);
//       systime = System.currentTimeMillis();
//       long next = systime + tickMillis; // Next real time for time update.
//       long dt;
//       next = next / tickMillis * tickMillis; // Truncate for nicer values.
//       try {
//         while (!interrupted()) {
// 		System.out.println("HiddenThread is running");
//           dt = next - systime;
//           if (dt > 0)
//             Thread.sleep(dt);
//           systime = System.currentTimeMillis();
//           dt = systime - next; // Detected lag increase.
//           if (dt > tol) { // if real time too far ahead.
//             dt = (dt / tickMillis + 1) * tickMillis;
//             next += dt;
//             lag += dt;
//           }
//           else
//             next += tickMillis;
//           seq++;
//         }
//       }
//       catch (InterruptedException exc) {
//         System.out.println("Ticker terminated!");
//         System.exit(1);
//       }
//     }
//   }

}
