package LabComm.Serial;

// import java.io.IOException;
// import java.io.ObjectInputStream;
// import java.io.ObjectOutputStream;
// import java.io.PipedInputStream;
// import java.io.PipedOutputStream;
// import java.io.Serializable;
import java.io.*;

import DataStructure.LabObject;
import DataStructure.LabStruct;



/**
 * The <code>UtilsSerial</code> class is a util class that contains methods
 * for converting serialized objects to arrays of bytes 
 * and vice versa.
 * 
 * It works using the interface Serializable and let every serializable object to convert its content
 * into array of bytes. While reading, the object is reconstructed from this 
 * content.
 * 
 * 
* @author Juan Periset, Ferran Carlas
*/

public final class UtilsSerial {
	/**
	 * Private buffer to store data.
	 */
	private byte[] b=new byte[BUFFERSIZE];
	/**
	 *Boolean used to syncronize methodes write and read of this class.
	 */
	private boolean availableRead;
	
	
	/**
	 * Piped output stream used for conversion.
	 */
 	 private PipedOutputStream pout;
	
	
	/**
	 * Piped input stream used for conversion.
	 */
	 private PipedInputStream pin;
	
	
	/**
	 * Object output stream used for conversion of objects to byte arrays.
	 */
	private ObjectOutputStream out;
	
	
	/**
	 * Object input stream used for conversion of byte arrays to objects.
	 */
	private ObjectInputStream in;
	

	 static final int  PIPESIZE=1024;
	 public static final int  BUFFERSIZE=8192;
	
	/**
	 * Constructs an utility object and initializes all streams. 
	 *
	 * @exception IOException if an I/O error occurs.
	 */
	public UtilsSerial() throws IOException {
		initialize();
	}
	
	
	/**
	 * Initializes the streams.
	 *
	 * @exception IOException if an I/O error occurs.
	 */
	private void initialize() throws IOException {
		pout = new PipedOutputStream();
		pin = new PipedInputStream(pout);
		out = new ObjectOutputStream(pout);
		out.flush();
		in = new ObjectInputStream(pin);
	}
	
	public synchronized void write( byte[] buff){
		b=buff;
		availableRead=true;
		notifyAll();
	}
	
	/**
	 * Writes length number of bytes of the buffer starting from offset in PipedOutputStream.
	 * @param buff It is the buffer from whom we take the information.
	 * @param offset
	 * @param length
	 * @exception IOException if an I/O error occurs.
	 */
	public synchronized int  writePout( byte[] buff,int offset,int length){
		try{
			pout.write(buff,offset,length);
			System.out.println("Acaba de escribir");
			
		}catch(IOException e) {
// 		    System.out.println("Inside utils writePout"+e);
		    System.out.println("Inside utils writePout");
		}
		return length;
	
	}
	/**
	 * Returns private buffer b.
	 * @return a byte[].
	 */
	public synchronized byte[] read(){
		try {
			while(!availableRead) wait();
		}catch(InterruptedException e){
// 		    System.out.println(e);}
		    System.out.println("UtilsSerial.read: CAughtException");
		}
		return b;
	}
	/**
	 * Reads from PipedInputStream and stores the bytes in the buffer b.
	 * @return the number of bytes read.
	 * @exception IOException if an I/O error occurs.
	 */
	public int readPin(byte[] b,int offset,int length){
		int bytes=-1;
		try {
			bytes=pin.read(b,offset,length);
				
		}catch(IOException e){
// 		    System.out.println("Inside utils readPin"+e);}
		    System.out.println("Inside utils readPin");}
		return bytes;
	}
	/**
	 * Converts an object to an array of bytes using piped streams.
	 * 
	 * @param s the serialized object to convert.
	 *
	 * @return the array of bytes that represents the serialized object.
	 * 
	 * @exception IOException if an I/O error occurs.
	 */
	public  void convert(Serializable s) throws IOException {
		out.reset();
		out.writeObject(s);
		out.flush();
	}
	
	/**
	 * Returns a buffer of bytes where the object already serialized is stored. 
	 * This method uses ReadThreadS to avoid the size limitation of ObjectInput and ObjectOutput. 
	 * The object must implement serializable.
	 * @param s
	 * @return byte[]
	 * @throws IOException
	 */
	
	public  byte[] convertE(Serializable s) throws IOException {
		byte[] b;
		ReadThreadS rt;
		
			rt=new ReadThreadS(this);
			rt.start();
			convert(s); 
			
			// When there is not more data to be read, the readT thread is interrupted	
			boolean data=true;
				while(data){
					if(this.available()!=0)
						data=true;
					else{
						data=false;
						rt.interrupt();
					}
				}
				// Writes in the byte[] b the array which has been created in readThread
				b=read();
				
				return b;
	}
	
	/**
	 * Converts an array of bytes to an object using piped streams.
	 * 
	 * @param s the array of bytes to convert.
	 *
	 * @return the serialized object created from the bytes.
	 * 
	 * @exception IOException if an I/O error occurs.
	 */
	public  Serializable convertS() throws IOException {
		
		Serializable obj=null;
		try {
			pout.flush();
	
		obj=(Serializable) in.readObject();
		
			
		} catch(Exception e) {
			throw new IOException("Could not read a whole Object\n" + e);
		}
		return obj;
	}
	
	/** It has been created to try Serialize a LabObject when it contains a LAbStruct but it is impossible
	 * because LabStruct has a not empty constructor, this constructor is absolutely neccesary because
	 * if it does not exist the array of Attributes can't be initialized.
	 */
	public  Serializable convertSL() throws IOException {
		
		LabObject obj;
		LabStruct ls=new LabStruct(20);
		try {
			pout.flush();
	for(int i=0;i<20;i++){
		obj=(LabObject) in.readObject();
		ls.setAttribute(LabStruct.type_double,obj,i);
	}
	obj=ls;
			
		} catch(Exception e) {
			throw new IOException("Could not read a whole Object\n" + e);
		}
		return obj;
	}

	
	/**
	 * Calculates the size of an serialized object, in bytes.
	 * 
	 * @param o the serialized object.
	 *
	 * @return the size of the serialized object.
	 * 
	 * @exception IOException if an I/O error occurs.
	 */
	public int getObjectSize(Serializable o) throws IOException {
		return convertE(o).length;
	}
	/**
	 * Returns the number of bytes that can be read from PipedInputStream.
	 * @return the number of bytes that can be read from PipedInputStream.
	 * @exception IOException if an I/O error occurs.
	 */
	public int available(){
		
		int bytes=-1;
		try{
			bytes=pin.available();
		}catch(IOException e) {
// 		    System.out.println(e);
		    System.out.println("UtilsSerial.available: Caught Exception");
		}
		return bytes;
	}
	
	
}
