package RTComm;

import se.lth.cs.throttlenet.*;
import java.io.*;
import DataStructure.*;
import DataStructure.LabMatrix.*;
import LabComm.Serial.*;
import LabComm.External.UtilsExternal;
import LabComm.External.WriteThread;




/**
 * The <code>RTReceive</code> class describes an intermediate layer, between the application 
 * and ThrottleNet, for receiving objects. The class implements methods to register the 
 * ThrottleNet socket and receive the information.
 * 
 * It uses an external package, LabComm, to avoid the size limitation of ObjectInput and ObjectOutput.  
 * 
 * @author  Juan Periset & Ferran Carlas
 */
public class RTReceive extends RTSocket{
	
	private int dead_line;
	private ThrottleNetReceiveSocket trs;
	private byte[] b=new byte[UtilsExternal.BUFFERSIZE];
	
/**
* Constructs and initializes a RTReceive .
* 
* @param   name			the name of the connection.
* @param   size			the larget object (LabObject) that can be sent fulfilling temporal restrictions.
* @param   period_us	the time between two packets in microseconds.
* @exception IOException     if an I/O error occurs.
*/
public RTReceive(String name, LabObject size ,int period_us)throws IOException {

this(name, size,period_us,Integer.MAX_VALUE);
}

/**
 * Constructs and initializes a  RTReceive .
 * 
 * @param   name			the name of the connection.
 * @param   size			the larget object (LabObject) that can be sent fulfilling temporal restrictions.
 * @param   period_us		the time between two packets in microseconds.
 * @param   dead_line		time to deliver each packet in microseconds.
 * @exception IOException     if an I/O error occurs.
 */
public RTReceive(String name, LabObject size,int period_us,int dead_line) throws IOException{
		super(name,size,period_us);
		this.dead_line=dead_line;
		this.trs=new ThrottleNetReceiveSocket();
		
	}
	
/**
 *  Registers the ThrottleNet receive socket.
 * @exception IOException     if an I/O error occurs.
 */
public void register()throws IOException {
	//Registers the connection
	trs.register(name,size,(int)period_us,dead_line);
	System.out.println("Register");
}	
/**
 * Receives an externalizable object reconstructed from a byte array.
 * 
 * @exception IOException     if an I/O error occurs.
 * @exception PacketLossException if a packet is lost. 
 * @exception ClassNotFoundException if no definition for the class could be
 *            found.
 * @return LabStruct received
 */
public Externalizable  receive(Externalizable object)throws PacketLossException, IOException,ClassNotFoundException{
	// Receives data in a byte array
	b=trs.receive();
	System.out.println("Receive");
	//Writes the buffer using utils and externalizes at the same time this buffer
	// storeing the buffer in a externalizable object.
	private WriteThread wt;
	wt=new WriteThread(b,utilsE);
	wt.start();
	utilsE.externalizeRead(object);
	System.out.println(object.getClass());
	//Interrupts WriteThread when there is no more data available
	boolean data=true;
	while(data){
		if(utilsE.available()!=0)
			data=true;
		else{
			data=false;
			wt.interrupt();
		}
	}
	
	return object;
	}
/*
/**
 * Don't works, it has been created to try receive a LabObject which contains a LabStruct
 
public LabObject receive(int j)throws PacketLossException, IOException,ClassNotFoundException{
	// Receives data in a byte array
	
	LabObject obj;
	b=trs.receive();
	System.out.println("Receive");
	//Closes the ThrottleNetReceiveSocket
	trs.close();
	//Writes the buffer using utils and externallizes at the same time this buffer
	// storeing the buffer in a LabStruct.
	wts=new WriteThreadS(b,utilsC);
	wts.start();
	obj=(LabObject)utilsC.convertSL();
	
	boolean data=true;
	while(data){
		if(utilsC.available()!=0)
			data=true;
		else{
			data=false;
			wts.interrupt();
		}
	}
	return obj;
	}**/

/**
 * Receives a Serializable object reconstructed from a byte array.
 * 
 * @exception IOException     if an I/O error occurs.
 * @exception PacketLossException if a packet is lost. 
 * @exception ClassNotFoundException if no definition for the class could be
 *            found.
 * @return Serializable received
 */
public Serializable receive()throws PacketLossException, IOException,ClassNotFoundException{
	// Receives data in a byte array
	Serializable obj=null;
	b=trs.receive();
	System.out.println("Receive");
	//Writes the buffer using utils and externallizes at the same time this buffer
	// storeing the buffer in a serializable object.
	private WriteThreadS wts;
	wts=new WriteThreadS(b,utilsC);
	wts.start();
	obj=utilsC.convertS();
//	Interrupts WriteThreadS when there is no more data available
	boolean data=true;
	while(data){
		if(utilsC.available()!=0)
			data=true;
		else{
			data=false;
			wts.interrupt();
		}
	}
	return obj;
}	
}
