package RTComm;


import se.lth.cs.throttlenet.ThrottleNetTransmitSocket;
import java.io.*;
import DataStructure.*;
import LabComm.External.ReadThread;
import LabComm.Serial.ReadThreadS;
import LabComm.External.UtilsExternal;
import LabComm.Serial.UtilsSerial;


/**
* The <code>RTTransmit</code> class describes an intermediate layer, between the application and ThrottleNet, for
* transmiting LabObject. The class implements methods to connect to GlobeThrottle and 
* transmit the externalized or serialized data structure.
* 
* It uses an external package, LabComm, to avoid the size limitation of ObjectInput and ObjectOutput.
* 
* @author  Juan Periset & Ferran Carlass
*/

public class RTTransmit extends RTSocket{
	private int dead_line;
	private byte[] b;
	private LabStruct labstr;
	private ThrottleNetTransmitSocket tts;
	
/**
* Constructs and initializes a  RTTransmit .
* 
* @param   name			the name of the connection.
* @param   size			the largest object (LabObject) that can be sent to fulfill the real time restrictions.
* @param   period_us	the time between two packets in microseconds.
* @exception IOException     if an I/O error occurs.
*/
	
public RTTransmit(String name, LabObject size,int period_us)throws IOException {

	this(name,size, period_us,Integer.MAX_VALUE);
}
/**
* Constructs and initializes a  RTTransmit .
* 
* @param   name			the name of the connection.
* @param   size			the largest object (LabObject) that can be sent to fulfill the real time restrictions.
* @param   period_us	the time between two packets in microseconds.
* @param   dead_line		time to deliver each packet in microseconds.
* @exception IOException     if an I/O error occurs.
*/
public RTTransmit(String name, LabObject size,int period_us,int dead_line) throws IOException{
		super(name,size,period_us);
		this.dead_line=dead_line;
		this.tts= new ThrottleNetTransmitSocket();
	}

/**
* Opens a connection with the specified properites.
* 
* @exception IOException     if an I/O error occurs.
*/	
public void connect ()throws IOException{

		
// connects to GlobeThrottle, if the connection is not available, waits and tries again.
		tts.connect(name,size,(int)period_us);
		System.out.println("Connected");	
} 
/**
* Transmits an externalized object over ThrottleNet.
* @param   exter	The externalizable object which is transmited
*/

 public void transmit(Externalizable exter){
 		
 		//Reads from the PipedInputStream and save all the information in a byte array,
 		//At the same time  PipedOutputStream is filled up using externalize method.
		
     try{	
	 ReadThread rt;
	 UtilsExternal utilsE=new UtilsExternal();
	 rt=new ReadThread(utilsE);
	 rt.start();
	 utilsE.externalize(exter); 
				
		// When there is not more data to be read, the ReadThread thread is interrupted	
			boolean data=true;
				while(data){
					if(utilsE.available()!=0)
						data=true;
					else{
						data=false;
						rt.interrupt();
					}
				}
				// Writes in the byte[] b the array which has been created in ReadThread
				byte[] b=utilsE.read();
				
				//Transmits the byte array with all the data
				tts.transmit(b);
				
 		System.out.println("END TRANSMITERTHREAD");
 		}catch(IOException e){System.out.println(e);}
 		catch(InterruptedException ei){System.out.println(ei);}
 		}
 /**
 * Transmits a Serializable object over ThrottleNet.
 * @param   serial the Serializable object which is transmited
 */
 public void transmit(Serializable serial){
 	//Reads from the PipedInputStream and save all the information in a byte array,
	//At the same time  PipedOutputStream is filled up using convert method.
	
	try{
	    ReadThreadS rts;
	    UtilsSerial utilsC=new UtilsSerial();
	    rts=new ReadThreadS(utilsC);
	    rts.start();
	    utilsC.convert(serial); 
				
		// When there is not more data to be read, the ReadThreadS thread is interrupted	
			boolean data=true;
				while(data){
					if(utilsC.available()!=0)
						data=true;
					else{
						data=false;
						rts.interrupt();
					}
				}
				// Writes in the byte[] b the array which has been created in ReadThreadS
				byte[] b=utilsC.read();
				
				//Transmits the byte array with all the data
				tts.transmit(b);
				
 		System.out.println("END TRANSMITERTHREAD");
 		}catch(IOException e){System.out.println(e);}
 		catch(InterruptedException ei){System.out.println(ei);}
 		}
}
