package RTComm;

import se.lth.cs.throttlenet.PacketLossException;
import se.lth.cs.throttlenet.ThrottleNetReceiveSocket;
import java.io.*;


import LabComm.Serial.UtilsSerial;
import LabComm.Serial.WriteThreadS;




/**
 * The <code>Receive</code> class describes an intermediate layer, between the application 
 * and ThrottleNet, for receiving serializable objects. The class implements methods to register the 
 * ThrottleNet socket and receive the information.
 * 
 * It uses an external package, LabComm, to avoid the size limitation of ObjectInput and ObjectOutput.   
 * 
 * @author  Juan Periset & Ferran Carlas
 */
public class Receive extends Socket{
	
	private int dead_line;
	private ThrottleNetReceiveSocket trs;
	private byte[] b=new byte[UtilsSerial.BUFFERSIZE];
	
/**
* Constructs and initializes a Receive .
* 
* @exception IOException     if an I/O error occurs.
*/
public Receive(String name,Object besteffort,int period_us)throws IOException {

this(name, besteffort, period_us, Integer.MAX_VALUE);
}

/**
 * Constructs and initializes a  Receive .
 * 
 * @param   name			the name of the connection.
 * @param   size			the amount of data to be received in the packet.
 * @param   period_us		the time between two packets in microseconds.
 * @param   dead_line		time to deliver each packet in microseconds.
 * @exception IOException     if an I/O error occurs.
 */
public Receive(String name,Object besteffort,int period_us,int dead_line) throws IOException{
		super(name,besteffort, period_us);
		this.dead_line=dead_line;
		this.trs=new ThrottleNetReceiveSocket();
		
	}
	
/**
 *  Registers the ThrottleNet receive socket.
 * @exception IOException     if an I/O error occurs.
 */
public void register()throws IOException {
	//Registers the connection
	trs.register(name,size,(int)period_us,dead_line);
	System.out.println("Register");
}
/**
 * Close the connection
 *
 */
public void close(){
	trs.close();
}
/**
 * Receives an serializable object reconstructed from a byte array.
 * 
 * @exception IOException     if an I/O error occurs.
 * @exception PacketLossException if a packet is lost. 
 * @exception ClassNotFoundException if no definition for the class could be
 *            found.
 * @return serializable object received
 */
public Serializable receive()throws PacketLossException, IOException,ClassNotFoundException{
	
	
	b=trs.receive();
	//System.out.println("Receive");
	//Closes the ThrottleNetReceiveSocket
	//Writes the buffer using utils and externallizes at the same time this buffer
	// storing the buffer in a serializable object.
	WriteThreadS wts;
	wts=new WriteThreadS(b,utilsC);
	wts.start();
	Serializable object=utilsC.convertS();
//	Interrupts WriteThreadS when there is no more data available
	boolean data=true;
	while(data){
		if(utilsC.available()!=0)
			data=true;
		else{
			data=false;
			wts.interrupt();
		}
	}
	return object;
	}
}
