package RTComm;


import java.io.*;
import LabComm.Serial.UtilsSerial;


/**
 * The <code>Socket</code> class implements a socket used 
 * to send objects through ThrottleNet. Socket is used in both, real and non-real time transmissions.
 * 
 * @author  Juan Periset & Ferran Carlas
 * @version 
 * @see ThrottleNetSocket
 */
abstract class Socket {	
	
	/**
	 * The utility needed for converting serialized objects to
	 * arrays of bytes and vice versa. 
	 * utilsS is used to calculate the size if the largest object that can be sent
	 * fulfilling real time restrictions.
	 */
	protected UtilsSerial utilsS;
	
	/**
	 * The utility needed for converting serialized objects to
	 * arrays of bytes and vice versa.
	 * utilsC is used to send and receive serializable objects.
	 */
	
	protected UtilsSerial utilsC;
	

	/** 
	 * The name of the connection between two nodes.
	 *
	 */
	protected String name;
	
	
	/**
	 * The amount of data to be sent in bytes.
	 * 
	 */
	protected int size;
	
	
	protected Object besteffort;
	
	
	/**
	 * The time between two packets in microseconds.
	 * 
	 */
	protected long period_us;
	
	
	/**
	 * Constructs a Socket.
	 * 
	 * @param   name      the name of the connection.
	 * @param   period_us the time between two packets in microseconds.
	 * 
	 * 
	 */
	public Socket(String name,int period_us)
											throws IOException {
		this.name = name;
		utilsS=new UtilsSerial();
		utilsC=new UtilsSerial();
		this.period_us = period_us;
		
	}
	
	public Socket(String name, Object besteffort, int period_us)throws IOException{
	
		this.name = name;
		utilsS=new UtilsSerial();
		this.size=utilsS.getObjectSize((Serializable) besteffort);
		this.period_us = period_us;
		
	}	
}
