package RTComm;


import se.lth.cs.throttlenet.ThrottleNetTransmitSocket;
import java.io.*;

import DataStructure.*;
import LabComm.Serial.ReadThreadS;
import LabComm.Serial.UtilsSerial;



/**
* The <code>Transmit</code> class describes an intermediate layer, between the application and ThrottleNet, for
* transmiting data in not real time. The class implements methods to connect to GlobeThrottle and 
* transmit the date.
* 
* It uses an Labcomm serial package, LabComm.Serial, to avoid the size limitation of ObjectInput and ObjectOutput.
* 
* @author  Juan Periset & Ferran Carlass
*/

public class Transmit extends Socket{
	private int dead_line;
	private byte[] b;
	private LabStruct labstr;
	private ThrottleNetTransmitSocket tts;
	
/**
* Constructs and initializes a  Transmit .
* 

* @exception IOException     if an I/O error occurs.
*/
	
public Transmit(String name, Object besteffort,int period_us)throws IOException {

	this(name, besteffort, period_us, Integer.MAX_VALUE);
}
/**
* Constructs and initializes a  Transmit .
* 
* @param   name			the name of the connection.
* @param   size			the amount of data in the ThrottleNet packet.
* @param   period_us	the time between two packets in microseconds.
* @param   dead_line		time to deliver each packet in microseconds.
* @exception IOException     if an I/O error occurs.
*/
public Transmit(String name, Object besteffort,int period_us,int dead_line) throws IOException{
		super(name,besteffort, period_us);
		this.dead_line=dead_line;
		this.tts= new ThrottleNetTransmitSocket();
	}
/**
* Opens a connection with the specified properites.
* 
* @exception IOException     if an I/O error occurs.
*/	
public void connect ()throws IOException{

		
// connects to GlobeThrottle, if the connection is not available, waits and tries again.
		tts.connect(name,size,(int)period_us);
		System.out.println("Connected");	
} 
/**
 * Closes the connection
 *
 */
public void close(){
	tts.close();
}
/**
* Transmits a serializable object over ThrottleNet.
*/

 public void transmit(Serializable serial){
 	try{
	    ReadThreadS trs:
	    UtilsSerial utilsC=new UtilsSerial();
	    rts=new ReadThreadS(utilsC);
	    rts.start();
	    utilsC.convert(serial); 
		
		
		// When there is not more data to be read, the ReadThreadS is interrupted	
		boolean data=true;
			while(data){
				if(utilsC.available()!=0)
					data=true;
				else{
					data=false;
					rts.interrupt();
				}
			}
			// Writes in the byte[] b the array which has been created in ReadThreadS
			b=utilsC.read();
			tts.transmit(b);
 	}catch(IOException e){System.out.println(e);}
 	catch(InterruptedException ei){ System.out.println(ei);}
 	}
 	
 		
}
