/* Socket.java -- Client socket i
 * Minimal implementation for LJRT lib
 */

package java.net;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;


public class Socket
{

	private static final native int createSocket();
	private static final native void connect(int sockfd, int adr, int port, int timeout) throws IOException;
	private static final native void close(int sockfd);

	/**
	 * True if the socket is bound.
	 */
	private boolean bound;
	private int myfd;


	void setFd(int fd) {
		myfd = fd;
	}


	public Socket() {
		myfd = createSocket();
	}

	public Socket(String host, int port)
		throws UnknownHostException, IOException {
		this();
		connect(new InetSocketAddress(InetAddress.getByName(host), port));
	}

	public void connect(SocketAddress endpoint) throws IOException {
		connect(endpoint, 0);
	}

	public void connect(SocketAddress endpoint, int timeout)
		throws IOException {
		try {
			//	getImpl().connect(endpoint, timeout);
			byte a[] = ((InetSocketAddress)endpoint).getAddress().getAddress();
			int port = ((InetSocketAddress)endpoint).getPort();
			System.out.println("connect to:"+a[0]+"."+a[1]+"."+a[2]+"."+a[3]);

			// Network Byte Order
			int adr = ((int)a[0] & 0xFF) |
				(((int)a[1] & 0xFF) << 8) |
				(((int)a[2] & 0xFF) << 16) |
				(((int)a[3] & 0xFF) << 24);

			System.out.println("Socket.connect():"+myfd+" adr = "+adr+" port = "+port);
			connect(myfd, adr, port, timeout);
		} catch (IOException exception) {
			close();
			throw exception;
		} catch (RuntimeException exception) {
			close();
			throw exception;
		} catch (Error error) {
			close();
			throw error;
		}
	}

	/**
	 * Returns an InputStream for reading from this socket.
	 *
	 * @return The InputStream object
	 *
	 * @exception IOException If an error occurs or Socket is not connected
	 */
	public InputStream getInputStream() throws IOException {
		//    if (isClosed())
		//	throw new SocketException("socket is closed");

		//    if (! isConnected())
		//	throw new IOException("not connected");
		if(myfd != 0) {
			return new java.io.SimpleTCPInputStream(myfd);
		} else {
			System.out.println("not connected");
			return null;
		}
		//    return getImpl().getInputStream();
	}

	/**
	 * Returns an OutputStream for writing to this socket.
	 *
	 * @return The OutputStream object
	 *
	 * @exception IOException If an error occurs or Socket is not connected
	 */
	public OutputStream getOutputStream() throws IOException {
		//    if (isClosed())
		//	throw new SocketException("socket is closed");

		//    if (! isConnected())
		//	throw new IOException("not connected");

		return new java.io.SimpleTCPOutputStream(myfd);
		//    return getImpl().getOutputStream();
	}

	public void close() {
		close(myfd);
		myfd = 0;
	}

	public String toString() {
		//    try
		//	{
		//	if (isConnected())
		//	  return ("Socket[addr=" + getImpl().getInetAddress() + ",port="
		//		 + getImpl().getPort() + ",localport="
		//		 + getImpl().getLocalPort() + "]");
		//	}
		//    catch (SocketException e)
		//	{
		// This cannot happen as we are connected.
		//	}

		//    return "Socket[unconnected]";
		return "Socket[toString not completely implemented]";
	}

	/* TODO: fix these methods! */
	public boolean isConnected() {
		return myfd!=0;
	}

	public boolean isClosed() {
		return myfd==0;
	}

	public boolean isInputShutdown() {
		return myfd==0;
	}

	public boolean isOutputShutdown() {
		return myfd==0;
	}


}
