/**
 * Created: 1998-12-17
 * LastEditDate: 1999-09-16
 */

package se.lth.control;
/**
 * The <code>DoublePoint</code> class represents a location in a
 * two-dimensional (<i>x</i>,&nbsp;<i>y</i>) double-valued coordinate space.
 *
 * @author Leif Andersson
 * @version 1998-12-17
 */
public class DoublePoint implements java.io.Serializable {
  
  /**
   * The <i>x</i> coordinate. 
   * @serial 
   */
  public double x;

  /**
   * The <i>y</i> coordinate. 
   * @serial  
   */
  public double y;

  /**
   * Constructs and initializes a point at the origin 
   * (0.0D,&nbsp;0.0D) of the coordinate space. 
   */
  public DoublePoint() {
    this(0D, 0D);
  }

  /**
   * Constructs and initializes a point with the same location as
   * the specified <code>Point</code> object.
   * @param       p a point.
   * @since       JDK1.1
   */
  public DoublePoint(DoublePoint p) {
    this(p.x, p.y);
  }

  /**
   * Constructs and initializes a point at the specified 
   * (<i>x</i>,&nbsp;<i>y</i>) location in the coordinate space. 
   * @param       x   the <i>x</i> coordinate.
   * @param       y   the <i>y</i> coordinate.
   */
  public DoublePoint(double x, double y) {
    this.x = x;
    this.y = y;
  }

  /**
   * Returns a hashcode for this Double.
   *
   * @return  a <code>hash code</code> value for this object. 
   */
  public int hashCode() {
    long xbits = Double.doubleToLongBits(x);
    long ybits = Double.doubleToLongBits(y);
    return (int)((xbits ^ (xbits >> 27)) ^ (ybits ^ (ybits >> 29))) ;
  }
 
  /**
   * Compares this object against the specified object.
   * The result is <code>true</code> if and only if the argument is 
   * not <code>null</code> and is a <code>DoublePoint</code> object
   * such that the respective components are equal according to the rules
   * of <code>Double.equal()</code>.
   *
   * @param   obj   the object to compare with.
   * @return  <code>true</code> if the objects are the same;
   *          <code>false</code> otherwise.
   */
  public boolean equals(Object obj) {
    long l, lo;
    if ((obj == null) || (!(obj instanceof DoublePoint))) return false;
    l = Double.doubleToLongBits(x);
    lo = Double.doubleToLongBits(((DoublePoint)obj).x);
    if (l != lo) return false;
    l = Double.doubleToLongBits(y);
    lo = Double.doubleToLongBits(((DoublePoint)obj).y);
    return (l == lo);
  }
    
  /**
   * Returns a String representation of this DoublePoint object.
   *
   */
  public String toString() {
    return ("[" + x + ", " + y + "]");
  }


  


  
} // DoublePoint
