/*
 * MainFrame.java
 *
 * Created: 2000-09-02
 * Edited:  2000-09-02
 *
 */

package se.lth.control;

import java.awt.*;
import javax.swing.*;
import java.awt.event.*;


/**
 * <code>MainFrame</code> contains a static convenience method to
 * handle the initialization of a Swing based application. Assuming
 * the main panel of the application is a <code>JPanel</code>
 * subclass, <code>MyPanel</code>, the main class for the application
 * would reduce to
 *
 * <pre>
 * public class MyApplication
 * public static main(String[] args) {
 *   MainFrame.setFrame(new MyPanel(), "MyApplication");
 * }  
 * </pre> <p>
 *
 * There is also a static method to display a frame containing the
 * text "Loading, please wait...". It may be used when the loading and
 * initialization of the main panel takes some time, so that the user
 * is kept reasonably happy. The main program in that case would be
 *
 * <pre>
 * public class MyApplication
 * public static main(String[] args) {
 *   MainFrame.showLoading();
 *   MainFrame.setFrame(new MyPanel(), "MyApplication");
 * }  
 * </pre> <p>
 *
 * Note that the instantiation of <code>MyPanel</code> should not take
 * place until after the return of
 * <code>MainFrame.showLoading()</code> so that the "Loading ... "
 * frame is displayed before the instantiation starts.
 *
 *
 * @author <a href="mailto:leif@control.lth.se">Leif Andersson</a>
 * @version 1.0
 * @since 1.0 */
public class MainFrame {

  /**
   * Displays a frame with the text "Loading, please wait...".
   *
   */
  public static void showLoading() {
    System.out.println("Starting, please wait...");
    loading = new JFrame("   ");
    label.setFont(new Font("SansSerif",Font.BOLD,20));
    loading.getContentPane().add(label);
    loading.pack();
    Dimension fd = loading.getSize();
    loading.setLocation((sd.width - fd.width)/2,
			(sd.height - fd.height)/2);
    loading.setVisible(true);
    try { Thread.sleep(1000); } catch(Exception e) {}
  }

  private static Dimension sd = Toolkit.getDefaultToolkit().getScreenSize();
  private static JLabel label = new JLabel("   Loading, please wait...   ");
  private static JFrame loading;

  /**
   * Handles the initialization of the main <code>JFrame</code> of an
   * application. If a "Loading, please wait..." frame has been
   * displayed, then it is removed immediately before the main
   * <code>JFrame</code> is shown.
   *
   * @param jp The <code>JPanel</code> that should be the
   * <code>contentPane</code> of the <code>JFrame</code>.
   *
   * @param name The name to be displayed for the frame.
   */
  public static void setPanel(JPanel jp, String name) {
    JFrame jf = new JFrame(name);
    jf.setContentPane(jp);
    jf.pack();
    jf.addWindowListener(new WindowAdapter() {
	public void windowClosing(WindowEvent e) {System.exit(0);}
      });
    Dimension fd = jf.getSize();
    jf.setLocation((sd.width - fd.width)/2,
		   (sd.height -fd.height)/2);
    if (loading != null) { loading.setVisible(false); }
    loading = null;
    jf.setVisible(true);
  }

} // MainFrame
