/**
 * StringPosition.java
 *
 *
 * Created: Sat Apr 25 21:11:47 1998
 * LastEditDate: 1999-09-16
 *
 * @author Leif Andersson
 * @version 1999-06-15
 */
package se.lth.control;
import java.awt.*;

/**
 * StringPosition is a class for placing a string in one of eight
 * possible positions relative to the specification point, as shown by 
 * the picture:
 */ 
public class StringPosition {
  
  private Constant position;
  private Font f;
  private FontMetrics fm;
  private int height, depth;

  public static final Constant LEFT_BOTTOM = new Constant() {
    public int getx(int x, int width) {return x;}
    public int gety(int y, int height, int depth) {return y - depth;}
  };

  public static final Constant LEFT_MIDDLE = new Constant() {
    public int getx(int x, int width) { return x; }
    public int gety(int y, int height, int depth) { return y + height/2; }
  };

  public static final Constant LEFT_TOP = new Constant() {
    public int getx(int x, int width) { return x;}
    public int gety(int y, int height, int depth) {return y + height;}
  };

  public static final Constant CENTER_BOTTOM = new Constant() {
    public int getx(int x, int width){return x - width/2;}
    public int gety(int y, int height, int depth) {return y - depth;}
  };

  public static final Constant CENTER_MIDDLE = new Constant() {
    public int getx(int x, int width){return x - width/2;}
    public int gety(int y, int height, int depth) {return y + height/2;}
  };

  public static final Constant CENTER_TOP = new Constant() {
    public int getx(int x, int width){return x - width/2;}
    public int gety(int y, int height, int depth) {return y + height;}
  };

  public static final Constant RIGHT_BOTTOM = new Constant() {
    public int getx(int x, int width){return x - width;}
    public int gety(int y, int height, int depth) {return y - depth;}
  };

  public static final Constant RIGHT_MIDDLE = new Constant() {
    public int getx(int x, int width){return x - width;}
    public int gety(int y, int height, int depth) {return y + height/2;}
  };

  public static final Constant RIGHT_TOP = new Constant() {
      public int getx(int x, int width){return x - width;}
      public int gety(int y, int height, int depth) {return y + height;}
    };


  public StringPosition(Constant position) {
    this.position = position; 
  }

  private static abstract class Constant {
    abstract int getx(int x, int width);
    abstract int gety(int y, int height, int depth);
  }

  public void setFont(Font f) {
    this.f = f;
    fm = null;
  }

  public void put(String s, Graphics g,	int x, int y) {
    Font savedFont = null;
    if (f != null) {
      savedFont = g.getFont();
      g.setFont(f);
    }
    if (fm == null) {
      fm = g.getFontMetrics();
      height = fm.getAscent();
      depth = fm.getDescent();
    }
    int width = fm.stringWidth(s);
    int startx = position.getx(x,width);
    int starty = position.gety(y,height,depth);
    g.drawString(s,startx,starty);
    if (savedFont != null) { g.setFont(savedFont); };
  }
    
} // StringPosition
