/**
 * UnixServerSocket.java 
 * Copyright (C) 2008 Anders Blomdell
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package se.lth.control.io;

import java.io.File;
import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketAddress;
import java.net.SocketException;
import java.nio.channels.ServerSocketChannel;

public class UnixServerSocket extends ServerSocket {

	private UnixSocketImpl impl = new UnixSocketImpl();
	private boolean closed = false;

	public UnixServerSocket(String path, int backlog) throws IOException {
		super();
		impl.bindAndListen(new UnixSocketAddress(path), backlog);
	}
	
	public UnixServerSocket(File path, int backlog) throws IOException {
		super();
		impl.bindAndListen(new UnixSocketAddress(path), backlog);
	}

	@Override
	public Socket accept() throws IOException {
		return(new UnixSocket(impl.accept()));
	}

	@Override
	public void bind(SocketAddress arg0, int arg1) throws IOException {
		throw new IOException("Call not implemented");
	}

	@Override
	public void bind(SocketAddress arg0) throws IOException {
		throw new IOException("Call not implemented");
	}

	@Override
	public void close() throws IOException {
		impl.close();
		closed = true;
	}

	@Override
	public ServerSocketChannel getChannel() {
		return null;
	}

	@Override
	public InetAddress getInetAddress() {
		return null;
	}

	@Override
	public int getLocalPort() {
		return 0;
	}

	@Override
	public SocketAddress getLocalSocketAddress() {
		return null;
	}

	@Override
	public synchronized int getReceiveBufferSize() throws SocketException {
		throw new SocketException("Call not implemented");
	}

	@Override
	public boolean getReuseAddress() throws SocketException {
		throw new SocketException("Call not implemented");
	}

	@Override
	public synchronized int getSoTimeout() throws IOException {
		throw new SocketException("Call not implemented");
	}

	@Override
	public boolean isBound() {
		return true;
	}

	@Override
	public boolean isClosed() {
		return closed ;
	}

	@Override
	public void setPerformancePreferences(int arg0, int arg1, int arg2) {
	}

	@Override
	public synchronized void setReceiveBufferSize(int arg0) throws SocketException {
		throw new SocketException("Call not implemented");
	}

	@Override
	public void setReuseAddress(boolean arg0) throws SocketException {
		throw new SocketException("Call not implemented");
	}

	@Override
	public synchronized void setSoTimeout(int arg0) throws SocketException {
		throw new SocketException("Call not implemented");
	}

	@Override
	public String toString() {
		return "UnixServerSocket...";
	}

}
