
/**
 * AxisPoint.java
 *
 *
 * Created: 1998-12-20
 * LastEditDate: 1999-09-20
 *
 * @author Leif Andersson
 * @version
 */

package se.lth.control.plot;

/**
 * A class representing tickmarks and labels for a plot axis.
 */
public class AxisPoint  {
  /**
   * The original value to be mapped to a particular point on the axis.
   * This is not used by the actual drawing classes, but is included
   * for completeness.
   */
  public double value = 0D;

  /**
   * The positon of the point in a [0.0, 1.0] space.
   */
  public double position = 0D;

  /**
   * The label to be put on the axis point.
   */
  public String label = null;

  /**
   * The tickmark level. The highest level has value 0, increasing
   * values  mean lower level. Negative values mean no tickmark.
   */
  public int tickLevel = 0;

  /**
   * @return  a <code>hashCode</code> value for this object.
   */
  public int hashCode() {
    long l;
    int hc;
    l = Double.doubleToLongBits(value);
    hc = (int)(l ^ (l >> 32));
    l = Double.doubleToLongBits(position);
    hc ^= (int)(l ^ (l >> 32));
    hc ^= tickLevel;
    if (label != null) {
      hc ^= label.hashCode();
    }
    return hc;
  }
   
  /**
   * Compares this object against the specified object.
   * The result is <code>true</code> if and only if the argument is
   * not <code>null</code> and is an <code>AxisPoint</code> object
   * such that the respective components are equal according to the rules
   * of <code>Double.equal()</code>, <code>String.equal()<code> 
   *
   * @param   obj   the object to compare with.
   * @return  <code>true</code> if the objects are the same;
   *          <code>false</code> otherwise.
   */
  public boolean equals(Object obj) {
    long l, lo;
    if ((obj == null) || (!(obj instanceof AxisPoint))) return false;
    l = Double.doubleToLongBits(value);
    lo = Double.doubleToLongBits(((AxisPoint)obj).value);
    if (l != lo) return false;
    l = Double.doubleToLongBits(position);
    lo = Double.doubleToLongBits(((AxisPoint)obj).position);
    if (l != lo) return false;
    if (tickLevel != ((AxisPoint)obj).tickLevel) return false;
    if (label != null) { return label.equals(((AxisPoint)obj).label); }
    else { return ((AxisPoint)obj).label.equals(label); }
  }

  /**
   * @return A String representation of this AxisPoint object.  
   */
  public String toString() {
    String s = label;
    if (s == null) { s = ""; }
    String blanks = "  ";
    StringBuffer sb = new StringBuffer(100);
    sb.append('[');
    sb.append(value);
    sb.append(blanks);
    sb.append(position);
    sb.append(blanks);
    sb.append('"');
    sb.append(s);
    sb.append('"');
    sb.append(blanks);
    sb.append(tickLevel);
    sb.append(']');
    return sb.toString();
  }
  
} // AxisPoint
