/*
 * LinearAxis.java
 *
 * Created: 2000-08-29
 * Edited:  2000-09-11
 *
 */
package se.lth.control.plot;
import se.lth.control.DoubleFormat;

/**
 * An <code>Axis</code> class with a linear scale.
 *
 * @author <a href="mailto:leif@control.lth.se">Leif Andersson</a>
 * @version 1.0
 * @since 1.0
 * @see AbstractAxis
 */
public class LinearAxis extends AbstractAxis {
  
  private double min, max;
  private DoubleFormat f = new DoubleFormat(6,2);

  /**
   * Creates a new <code>LinearAxis</code> instance.
   *
   * @param range The range of the axis. <code>range</code> must be
   * &gt;= 0.0 and should probably be a
   * power of ten times 1, 2, or 5, i. e. one of ... 0.1, 0.2, 0.5, 1
   * ... 10, 20 ...
   * @param bottom The lowest value of the axis.
   * @param divTicks The number of divisions for the tick marks.
   * <code>divTicks</code> must be  &gt;= 0. The value 0 means no tick marks.
   *
   * @param divGrid The number of divisions for the grid.
   * <code>divGrid</code> must be &gt;= 0. The value 0 means no grid.
   */
  public LinearAxis(double range, double bottom, int divTicks, int divGrid) {
    AxisPoint ap;
    double val,divT, divG;
    if (range <= 0D) {
      throw(new IllegalArgumentException("range <= 0"));
    }
    min = bottom; max = min + range;
    if (divTicks < 0) {
      throw(new IllegalArgumentException("divTicks < 0"));
    }
    if (divTicks > 0) {
      divT = range/divTicks;
      val = Math.round(bottom/divT) * divT;
      if (val < min - eps*range) { val += divT; }
      while (val <= max) {
	ap = new AxisPoint();
	ap.value = val;
	if (range < 10D) { ap.label = f.format(val); }
	else {ap.label = Integer.toString((int)val); }
	ap.position = (val - min)/range;
	points.add(ap);
	val = val + divT;
      }
    }
    if (divGrid < 0) {
      throw(new IllegalArgumentException("divGrid < 0"));
    }
    if (divGrid > 0) {
      divG = range/divGrid;
      val = Math.round(bottom/divG) * divG;
      if (val < min - eps*range) { val += divG; }
      while (val <= max) {
	grid.add(new Double((val-min)/range));
	val = val + divG;
      }
    }
  }

  /**
   * Returns the minimum value on the axis
   *
   * @return the minimum value on the axis
   */
  public double getMin() { return(min); }
  /**
   * Returns the maximum value on the axis
   *
   * @return the maximum value on the axis
   */
  public double getMax() { return(max); }
  
  private static final double eps = 0.001D;
} // LinearAxis
