package se.lth.control.realtime;

/**
 * This class handles time intervals in a system independent way.
 *
 * @author <a HREF="mailto:anders.blomdell@control.lth.se">Anders Blomdell</a>
 * @version 1.0
 */
public class Duration {

  final long ticks; /* Time needs to access this */
  private static long secondsToTick(double seconds) {
    return (long)(seconds * 1000000.0);
  }

  /**
   * Creates a ticks long duration.
   */
  Duration(long ticks) {
    this.ticks = ticks;
  }

  /**
   * Creates a seconds long duration.
   */
  public Duration(double seconds) {
    ticks = secondsToTick(seconds);
  }

  /**
   * Creates a seconds long duration.
   */
  public Duration(float seconds) {
    this((double)seconds);
  }

  /**
   * Creates a 0.0 second duration.
   */
  public Duration() {
    this(0.0);
  }

  /**
   * Return a new Duration which is increment larger than this Duration.
   */
  public Duration add(Duration increment) { 
    return new Duration(ticks + increment.ticks); 
  }

  /**
   * Return the length of this duration.
   */
  public double toSeconds() {
    return (double)ticks / 1000000.0;
  }

  public String toString() {
    return Double.toString(toSeconds());
  }

}
