package se.lth.control.realtime;

/**
 * This class handles time in a system independent way.
 *
 * @author <a HREF="mailto:anders.blomdell@control.lth.se">Anders Blomdell</a>
 * @version 1.0
 */
public class Time {

  private final long ticks;
  static long epoch;

  static {
    System.loadLibrary("se_lth_control_realtime_Time");    
    epoch = currentTicks();
  }

  private static native long currentTicks();

  private Time(long ticks) {
    this.ticks = ticks;
  }

  /**
   * Creates a new time representing the same time as t.
   */
  public Time(Time t) {
    this(t.ticks);
  }

  /**
   * Returns the current system time.
   */
  public static Time now() { 
    return new Time(currentTicks()); 
  }

  /**
   * Returns a new Time representing <code>this + increment</code>. 
   */
  public Time add(Duration increment) { 
    return new Time(ticks + increment.ticks); 
  }

  /**
   * Returns a new Duration representing <code>this - t</code>. 
   */
  public Duration sub(Time t) { 
    return new Duration(ticks - t.ticks); 
  }

  public int compareTo(Time t) {
    int result;
    if (ticks < t.ticks) {
      result = -1;
    } else if (ticks == t.ticks) {
      result = 0;
    } else {
      result = 1;
    }
/*    System.out.print(toString());
    System.out.print(" <=> ");
    System.out.print(t.toString());
    System.out.print(" => ");
    System.out.println(result);
*/
    return result;
  }

  final static char[] digits = { '0', '1', '2', '3', '4', 
				 '5', '6', '7', '8', '9' };
  public String toString() {
    long i = ticks - epoch;
    char[] buf = new char[65];
    int charPos = 64;
    boolean negative = (i < 0);
    
    if (!negative) {
      i = -i;
    }
    
    while (i <= -10) {
      buf[charPos--] = digits[(int)(-(i % 10))];
      i = i / 10;
    }
    buf[charPos] = digits[(int)(-i)];
    
    if (negative) {
      buf[--charPos] = '-';
    }
    
    return new String(buf, charPos, (65 - charPos));
  }

}


