package se.lth.cs;


/**
 * The <code>Table</code> class describes a table which maps keys to values.
 * 
 * @author  Patricia Grudziecka
 * @author  Daniel Nyberg
 * @version 0.8, 12/11/03
 * @since   ThrottleNet0.5
 */
public class Table {
	
	
	/**
	 * The default size of the table.
	 */
	public static final int DEFAULT_INIT_SIZE = 250;
	
	
	/**
	 * The array which contains values.
	 */
	private Object[] elements;
	
	
	/**
	 * The array which contains keys.
	 */
	private Object[] keys;
	
	
	/**
	 * The current number of keys in the table.
	 */
	private int nbr;
	
	
	/**
	 * Constructs a table with the default size.
	 * 
	 */
	public Table(){
		this(DEFAULT_INIT_SIZE);
	}
	
	
	/**
	 * Constructs a table with the specified number of keys.
	 * 
	 * @param   initSize      	the size of the table.
	 * 
	 */
	public Table(int initSize){
		elements = new Object[initSize];
		keys = new Object[initSize];
		nbr = 0;
	}
	
	
	/**
	 * Maps the specified key to the specified value in the table.
    *
    * @param key		the table key.
	 * @param value	the value for this key.
	 * 
	 * @return 			the previous value of the specified key in this hashtable,
	 *						or null if it did not have one.
    */
	public Object put(Object key, Object value) {
		Object tmp;
		for(int i = 0; i < nbr; i++) {
			if(key.equals(keys[i])) {
				tmp = elements[i];
				elements[i] = value;
				return tmp;
			}
		}
		keys[nbr] = key;
		elements[nbr++] = value;
		return null;
	}
	
	
	/**
	 * Returns the value to which the specified key is mapped in the table.
    *
    * @param key the key.
	 * @return the value for this key.
    */
	public Object get(Object key) {
		for(int i = 0; i < nbr; i++) {
			if(key.equals(keys[i])) {
				return elements[i];
			}
		}
		return null;
	}
	
	
	/**
	 * Removes the key (and its corresponding value) from this table.
    *
    * @param key the to remove key. 
	 * @return the removed value.
    */
	public Object remove(Object key) {
		Object tmp;
		for(int i = 0; i < nbr; i++) {
			if(key.equals(keys[i])) {
				tmp = elements[i];
				elements[i] = elements[--nbr];
				keys[i] = keys[nbr];
				return tmp;
			}
		}
		return null;
	}
	
	
	/**
	 * Returns the number of keys in this table.
    *
    * @return the number of keys in this table.
    */
	public int size() {
		return nbr;
	}
	
	
	/**
	 * Returns the string representation of this table.
    *
    * @return the string representation of this table.
    */
	public String toString() {
		String s = "trottlenet.Table[";
		for(int i = 0; i < nbr-1; i++)
			s += "{" + keys[i] + "," + elements[i] + "},";
		if(nbr > 0)
			s += "{" + keys[nbr-1] + "," + elements[nbr-1] + "}";
		return (s +  "]");
	}
}


