package se.lth.cs.avr;

import se.lth.cs.realtime.semaphore.*;

public class LCDDisplay {
  
    private static int pos;
    private static int errpos;
    private static int outpos;
    MutexSem sem;

  
  public LCDDisplay() {
    setupPorts();
    pos = 0;
    errpos = 40;
    outpos = 0;
    s = new MutexSem();
  }


  private static final void moveCursor(int newpos) {
    if(newpos != pos) {
      int delta;
      char command;
      if(newpos > pos) {
        delta = newpos - pos;
        command = 0x14;
      }
      else {
        delta = pos - newpos;
        command = 0x10;
      }
      for(int i = 0; i < delta; i++) {
        writeCtrlByte(command);
      }
      pos = newpos;
    }
  }
  
  public static final void print(char c) {
//       sem.take();
    switch (c) {
      case '\n' :
        errpos = 0;
      break;
      default:
        moveCursor(outpos);
        writeDataByte(c);
        outpos++;
        pos = outpos;
      break;
    }
//     sem.give();
  }

    public static final void print(int newPos, char c) {
// 	sem.take();
	moveCursor(newPos);
	writeDataByte(c);
	moveCursor(pos);
// 	sem.give();
    }

  public static final void clear() {
//       sem.take();
    writeCtrlByte((char)0x01);
    extraDelay();
    pos = 0;
//     sem.give();
  }

  public static final void setCursor(boolean on) {
//       sem.take();
    if(on) {
      writeCtrlByte((char)0x0e);
    }
    else {
      writeCtrlByte((char)0x0c);
    }
//     sem.give();
  }

  public static final void setCursorPos(int position) {
//       sem.take();
    outpos = position;
    moveCursor(position);
//     sem.give();
  }

  public static final void setErrCursorPos(int position) {
//       sem.take();
    errpos = position;
    moveCursor(position);
//     sem.give();
  }

  public static final void setTimeCursorPos(int position) {
//       sem.take();
    if(position >= 4) {
      position += 2;
    }
    else if(position >= 2) {
      position += 1;
    }
    outpos = position;
    moveCursor(position);
//     sem.give();
  }

  public static final void errprint(char c) {
//       sem.take();
    switch (c) {
      case '\n' :
        errpos = 40;
      break;
      default:
        moveCursor(errpos);
        writeDataByte(c);
        errpos++;
        pos = errpos;
      break;
    }
//     sem.give();
  }

  public static final void intprint(int i) {
      moveCursor(outpos);
      char c3 = (char) (i % 10 + '0');
      i = i/10;
      char c2 = (char) (i % 10 + '0');
      i = i/10;
      char c1 = (char) (i % 10 + '0');
      errprint(c1);
      errprint(c2);
      errprint(c3);
      errprint(':');
  }

  public static final void timeprint(int time) {
    char s1 = (char)(time % 10 + '0');
    time /= 10;
    char s2 = (char)(time % 10 + '0');
    time /= 10;
    char m1 = (char)(time % 10 + '0');
    time /= 10;
    char m2 = (char)(time % 10 + '0');
    time /= 10;
    char h1 = (char)(time % 10 + '0');
    time /= 10;
    char h2 = (char)(time % 10 + '0');
    int status = disableInts();
    print(h2);
    print(h1);
    print(':');
    print(m2);
    print(m1);
    print(':');
    print(s2);
    print(s1);
    enableInts(status);
  }
    
  public static final void errprint(int data) {
    char c1 = (char)(data % 10 + '0');
    data /= 10;
    char c2 = (char)(data % 10 + '0');
    data /= 10;
    char c3 = (char)(data % 10 + '0');
    data /= 10;
    char c4 = (char)(data % 10 + '0');
    data /= 10;
    char c5 = (char)(data % 10 + '0');
    int status = disableInts();
    errprint(c5);
    errprint(c4);
    errprint(c3);
    errprint(c2);
    errprint(c1);
  }

  private static final native void writeDataByte(char c);
  private static final native void writeCtrlByte(char c);
  private static final native void extraDelay();
  private static final native void setupPorts();

    private static final native int disableInts();
    private static final native void enableInts(int status);
}
