package se.lth.cs.camera;

/**
 *
 * Camera API for the Axis211A model.
 *
 * @author Emma Nilsson-Nyman (emma.nyman@cs.lth.se)
 *
 */ 
public class Axis211A {

    public static final int IMAGE_BUFFER_SIZE = 100000;

    private int fd; // File descriptor

    private static native int nativeConnect();
    private static native int nativeGetJPEG(int fd, byte[] target, int offset);
    private static native void nativeClose(int fd);


    /**
     * Connects to the camera.
     *
     * @return true if connected otherwise false.
     */
    public boolean connect() {
      fd = nativeConnect();
      return (fd > 0); 
    }

    /**
     * Reads an image from the camera.
     *
     * @param target Byte array to put data in.
     * @param offset Offset from the start of the byte array.
     *
     * @return The number of read bytes.
     */
    public int getJPEG(byte[] target, int offset) {
      return nativeGetJPEG(fd, target, offset);
    }

    /**
     * Closes the camera connection.
     */
    public void close() {
      nativeClose(fd);
    }
}
