package se.lth.cs.labcomm;

import java.io.IOException;
import java.io.DataInputStream;

/**
 * The LabComm Java Library
 * 
 * Decodes primitive datatypes and String objects.
 * 
 * @author Emma Nilsson-Nyman (emma.nyman@cs.lth.se) 
 */
public class Decoder {

    private DataInputStream in;

    public Decoder(DataInputStream i) { 
	in = i;
    }
    
    /**
     * Decodes a boolean
     * @return A boolean value
     */
    public boolean decodeBoolean() throws IOException {
	return in.readBoolean();
    }
    
    /**
     * Decodes a byte
     * @return A byte value
     */
    public byte decodeByte() throws IOException {
    	return in.readByte();
    }

    /**
     * Decodes a bytea array
     * @param data Byte array to put data in
     */
    public void decodeByteArray(byte[] data) throws IOException {
        int readBytes = in.read(data, 0, data.length);
	while (readBytes < data.length) {
	  readBytes += in.read(data, readBytes, data.length-readBytes);
	}
    }
    
    /**
     * Decodes a short
     * @return A short value
     */
    public short decodeShort() throws IOException {
        return in.readShort();
    }
    
    /**
     * Decodes an integer
     * @return An integer value
     */
    public int decodeInt() throws IOException {
        return in.readInt();
    }
    
    /**
     * Decodes a long
     * @return A long value
     */
    public long decodeLong() throws IOException {
	return in.readLong();
    }
    
    /**
     * Decodes a float
     * @return A float value
     */
    public float decodeFloat() throws IOException {
	return in.readFloat();
    }
    
    /**
     * Decodes a double
     * @return A double value
     */
    public double decodeDouble() throws IOException {
	return in.readDouble();
    }
   
    /**
     * Decodes a String
     * @return A String object
     */
    public String decodeString() throws IOException {
        int length = decodeInt();
	byte[] content = new byte[length];
	decodeByteArray(content);
	return new String(content);
    }
}
