package se.lth.cs.labcomm;

import java.io.IOException;
import java.io.DataOutputStream;

/**
 * The LabComm Java Library
 * 
 * Encodes primitive datatypes and String objects. 
 * 
 * @author Emma Nilsson-Nyman (emma.nyman@cs.lth.se) 
 */
public class Encoder {

    private DataOutputStream out;

    public Encoder(DataOutputStream o) { 
	out = o;
    }
   
    /**
     * Encodes a boolean.
     * @param value Boolean value
     */
    public void encodeBoolean(boolean value) throws IOException {
      out.writeBoolean(value);
    }
    
    /**
     * Encodes a byte.
     * @param value Byte value
     */
    public void encodeByte(byte value) throws IOException {
        out.writeByte(value);
    }

    /**
     * Encodes a byte array.
     * @param value Byte array
     */
    public void encodeByteArray(byte[] value) throws IOException {
	out.write(value, 0, value.length);
    }
    
    /**
     * Encodes a short.
     * @param value Short value
     */
    public void encodeShort(short value) throws IOException {
	out.writeShort(value);
    }
    
    /**
     * Encodes an integer.
     * @param value Integer value
     */
    public void encodeInt(int value) throws IOException {
	out.writeInt(value);
    }
    
    /**
     * Encodes a long.
     * @param value Long value
     */
    public void encodeLong(long value) throws IOException {
        out.writeLong(value);
    }
    
    /**
     * Encodes a float.
     * @param value Float value
     */
    public void encodeFloat(float value) throws IOException {
        out.writeFloat(value);
    }
    
    /**
     * Encodes a double.
     * @param value Double value
     */
    public void encodeDouble(double value) throws IOException {
	out.writeDouble(value);
    }
    
    /**
     * Encodes a String.
     * @param value String object
     */
    public void encodeString(String value) throws IOException {
    	encodeInt(value.length());
	encodeByteArray(value.getBytes());
    }
}

