/* $Id: OngoingThread.java,v 1.5 2005/08/15 13:24:44 d02en Exp $        
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime;

/**
 * Real-time thread without deadlines and without explicit cyclic behavior.
 * 
 * @see License
 */

public class OngoingThread extends RTThread {

  private Runnable target = null;

  /**
   * Allocates a new <code>OngoingThread</code>.
   * <p>
   * Equivalent to <code>OngoingThread(null, null, null)</code>.
   *
   * @see OngoingThread#OngoingThread(RTThreadGroup, java.lang.Runnable, java.lang.String)
   */
  public OngoingThread() {
    this(null, null, null);
  }

  /**
   * Allocates a new <code>OngoingThread</code>.
   * <p>
   * Equivalent to <code>OngoingThread(null, target, null)</code>.
   *
   * @param target the object whose <code>run</code> method is called.
   *
   * @see OngoingThread#OngoingThread(RTThreadGroup, java.lang.Runnable, java.lang.String)
   */
  public OngoingThread(Runnable target) {
    this(null, target, null);
  }

  /**
   * Allocates a new <code>OngoingThread</code>.
   * <p>
   * Equivalent to <code>OngoingThread(group, target, null)</code>.
   *
   * @param group this thread's <code>RTThreadGroup</code>.
   * @param target the object whose <code>run</code> method is called.
   *
   * @see OngoingThread#OngoingThread(RTThreadGroup, java.lang.Runnable, java.lang.String)
   */
  public OngoingThread(RTThreadGroup group, Runnable target) {
    this(group, target, null);
  }

  /**
   * Allocates a new <code>OngoingThread</code>.
   * <p>
   * Equivalent to <code>OngoingThread(null, null, name)</code>.
   *
   * @param name this thread's name.
   *
   * @see OngoingThread#OngoingThread(RTThreadGroup, java.lang.Runnable, java.lang.String)
   */
  public OngoingThread(String name) {
    this(null, null, name);
  }

  /**
   * Allocates a new <code>OngoingThread</code>.
   * <p>
   * Equivalent to <code>OngoingThread(group, null, name)</code>.
   *
   * @param group this thread's <code>RTThreadGroup</code>.
   * @param name this thread's name.
   *
   * @see OngoingThread#OngoingThread(RTThreadGroup, java.lang.Runnable, java.lang.String)
   */
  public OngoingThread(RTThreadGroup group, String name) {
    this(group, null, name);
  }

  /**
   * Allocates a new <code>OngoingThread</code>.
   * <p>
   * The parameter target specifies an object whose method <code>run</code> will
   * be called when this thread is started.
   *
   * @param group this thread's <code>RTThreadGroup</code>.
   * @param target the object whose <code>run</code> method is called.
   * @param name this thread's name.
   *
   * @see java.lang.Runnable
   * @see OngoingThread#OngoingThread(RTThreadGroup, java.lang.Runnable, java.lang.String)
   */
  public OngoingThread(RTThreadGroup group, Runnable target, String name) {
    super(group, name);
    this.target = target;
  }

  protected void run() {
    if (target != null) {
      target.run();
    }
  }

}
