/* $Id: AckEvent.java,v 1.6 2005/08/19 09:37:16 d02en Exp $         
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime.event;

import se.lth.cs.realtime.*;
 
/**
 * Event to be used for acknowledgement of actions done, such as 
 * having received an event and/or having carried out an order.
 * 
 * @see License
 */

public class AckEvent extends RTEvent{

  /**
   * The event being acknowledged.
   */
  protected RTEvent reason;
  
  /**
   * @see RTEvent#RTEvent() Called superclass constructor with same arguments.
   */
  public AckEvent() {
    super();
    owner = null;
  }

  /**
   * @see RTEvent#RTEvent(Object) Called superclass constructor with same arguments.
   */
  public AckEvent(Object source) {
    super(source);
    owner = null;
  }

  /**
   * @see RTEvent#RTEvent(long) Called superclass constructor with same arguments.
   */
  public AckEvent(long ts) {
    super(ts);
    owner = null;
  }

  /**
   * @see RTEvent#RTEvent(Object,long) Called superclass constructor with same 
   * arguments.
   */
  public AckEvent(Object source, long ts) {
    super(source, ts);
    owner = null;
  }

  /**
   * @see RTEvent#RTEvent(Object,long,Timebase) Called superclass constructor with 
   * same arguments.
   */
  public AckEvent(Object source, long ts, Timebase timebase) {
    super(source, ts, timebase);
    owner = null;
  }

  /**
   * Obtain the reason for this event being sent.
   * @return the event that is acknowledged.
   */
  public RTEvent getReason() {
    return reason;
  }

  /**
   * Update the reason for acknowledging by storing a referens to the supplied
   * event object. The time stamp of this ack-event object (not the reason) is 
   * updated to the current real time according to the current time base (that
   * is, by default to <code>RTSystem.currentTimeMillis</code>). Other fields 
   * are kept to simplify repost of the same (anonymous) ack-event object.
   * <p>
   * However, if a reason is already set (not being null), the owner is set to 
   * the currently executing (Java or real-time) thread. That is to prevent 
   * re-sending an un-owned acknowledge referring to a reason that is actually 
   * being updated by the acknowledging thread. If the application can ensure
   * there is no such risk, this method can be called with argument null prior 
   * to setting the new reason.
   * 
   * @param event The event that is being acknowledged.
   */
  public void setReason(RTEvent event) {
    if (reason == null) {
      reason = event;
    } else {
      owner = RTThread.currentThread();
    }
    timestamp = RTSystem.currentTimeMillis(); // No time base support yet.
  }

}
