/* $Id: ActivityEvent.java,v 1.5 2005/08/19 09:37:16 d02en Exp $         
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime.event;

import se.lth.cs.realtime.Activity;
import se.lth.cs.realtime.Timebase;

/**
 * Generic type of event which informs about normal changes or advancement
 * of thread execution, such as changed period, changed priority, suspension,
 * new sample, etc. An <code>ActivityEvent</code> is seldom expected to
 * be used directly; subclasses are expected to provide desired dedicated
 * functionality.
 * 
 * @see License
 */

public class ActivityEvent extends RTEvent {

  /**
   * Like the default <code>RTEvent</code> constructor, which uses the current 
   * (RT)Thread as the source.
   */
  public ActivityEvent() {
    super();
  }

  /**
   * Constructs an <code>ActivityEvent</code> object with the specified
   * source activity, set as the source of this event.
   *
   * @param source the thread or process where the event originated.
   */
  public ActivityEvent(Activity source) {
    super(source);
  }

  /**
   * The new event gets the time stamp according to <code>ts</code>.
   * 
   * @param ts the time stamp in milliseconds of current system time.
   */
  public ActivityEvent(long ts) {
    super(ts);
  }

  /**
   * Create a new event with both source and time set.
   * 
   * @param source the thread or process where the event originated.
   * @param ts the time stamp in milliseconds of current system time.
   */
  public ActivityEvent(Activity source, long ts) {
    super(source, ts);
  }

  /**
   * 
   * Constructs an activity-event object with the specified
   * source object and time-stamp with respect to the specified time-base.
   *
   * @param source the object where the event originated.
   * @param ts the time from which the event originates.
   * @param timebase the base of time in which the time stamp is expressed.
   */
  public ActivityEvent(Activity source, long ts, Timebase timebase) {
    super(source, ts, timebase);
  }

}
