/* $Id: RTEventListener.java,v 1.4 2005/08/15 13:24:44 d02en Exp $         
 * 
 * LJRTlib - The Lund Java-based Real-Time Library
 * Copyright (C) 2005 Dept. of Computer Science, Lund University, Sweden
 * 
 * This is free software under the terms of the GNU LGPL; see License.java
 */

package se.lth.cs.realtime.event;

import java.util.*;

/**
 * Interface implemented by thread objects or other objects that are
 * able to accept <code>RTEvent</code>s via its <code>putEvent</code>
 * method.  The typical implementor of this interface appends the
 * event to some kind of queue. This normally means that event
 * processing is asynchronous to the generation of the event, opposed
 * to AWT and Swing where events are processed during dispatch (unless
 * <code>invokeLater</code> or such features is used). Also
 * <code>RTEventListener</code>s are, however, permitted to do
 * synchronous event processing (in <code>putEvent</code> in the
 * context of the caller). Hence, an <code>RTEventListener</code> does
 * not need to be a thread object, if not required by specific event
 * dispatchers, but event processing is normally assumed to be rapid.
 * 
 * @see License
 */

public interface RTEventListener extends EventListener {

  /**
   * Supply event for processing by implementing activity or object.
   *
   * @param   e   Event to process or buffer for processing
   */
  RTEvent putEvent(RTEvent e);
}
