package se.lth.cs.robot.irb6;

import se.lth.cs.robot.servo.*;

public class Motors {

    public static final int NBR_OF_JOINTS = 5;	
    public static final int J1 = 1;	
    public static final int J2 = 2;	
    public static final int J3 = 3;	
    public static final int J4 = 4;	
    public static final int J5 = 5;	
    private Drive[] drives;
    private IO io;

    /* compareSWHW is used for test purposes 
     * The method printPositionMismatches() can then be used  
     * to print the values of all arrays. The values are 
     * stored when there is a mismatch between HW and SW. */
    private boolean compareSWHW;
    private int offsetValue, value, halfrevolution;
    private int[] offset, lastvalue;
    public static final float pi = 3.14159f;
    private final float angleScale = pi*2/2048;
    private float[] resolverCorrection;
    private int temp;
    private int positionValue, realPositionValue, revValue;

    /*Variables used when compareSWHW*/
    private int[] logArray, hwArray, swArray, positionArray, realPositionArray;
    int counter, checkVal;

    public Motors() {
	    this((IO)null);
    }
    
    public Motors(IO iio) {
	 // Work-around for stupid Java rules...
	 IO tmp;
	 if(iio == null) {
            tmp = new IO();
	    tmp.nativeInit();
	    io = tmp;
	} else {
	    io = iio;
	}

	drives = new Drive[NBR_OF_JOINTS];
	for (int i=0; i<drives.length; i++) {
	    drives[i] = new Irb6Drive(this, i);
	}

	for(int i=0; i<NBR_OF_JOINTS; i++) {
	    analogOut(J1+i,0);
	}

	// ??

	offsetValue = 2048;
	halfrevolution = 1024;
	value = 0;
	offset = new int[NBR_OF_JOINTS]; 
	lastvalue = new int[NBR_OF_JOINTS];

	resolverCorrection=new float[NBR_OF_JOINTS];
	resolverCorrection[0] = pi; 
	resolverCorrection[1] = 0.0f;
	resolverCorrection[2] = 0.0f;
	resolverCorrection[3] = -2*pi; 
	resolverCorrection[4] = 1.5*pi;

	compareSWHW = false;
	if(compareSWHW) {
	    logArray = new int[500];
	    hwArray = new int[500];
	    swArray = new int[500];
	    positionArray = new int[500];
	    realPositionArray = new int[500];
	    counter = 0; //Used to avoid illegal access to memory (Java out of bounds...)
	}

	/* Set offset value to present value of position signal */
	for(int k = 0; k<offset.length; k++) {
	    int temp = read(J1+k); 
	    lastvalue[k] = (temp & 0x07ff);
	    temp = temp >> 11;
	    offset[k] = (temp - 16)*offsetValue;
	}
    }    

    
    public Drive[] getDrives(){
	return drives;
    }

    // Same as setTorque() in the vitual irb6 robot motor
    // No, not the same, AnalogOut sets the output to the specified
    // voltage, setTorque, sets a torque.
    public synchronized void analogOut(int motor, float gain) {
	int gain_in_hex = (int)((gain+(float)10.0)*((float)(4095.0/20.0)));
//	System.out.println("analogOut " + gain_in_hex);
	io.analogOut(motor, gain_in_hex);
    }
    
    public synchronized float analogIn(int motor) {
	int result = io.analogIn(motor);
//	System.out.println("analogIn " + result);
	return ((float)20.0/4095.0)*result-(float)10.0;
    }

    public synchronized void setSafeBit(){
	io.loBitOut(0x1f);
    }

    public synchronized float getPosition(int motor) {
	temp = read(motor);
	positionValue = (temp & 0x07ff); // Map out positionsignal
	realPositionValue = positionValue; //Used if compareHWSW
	revValue = temp >> 11; // Value of HW revolution counter
	revValue = revValue - 16; // Value of reseted HW counter is 16, now it is 0

	/* A change of more than a half revolution indicates a change */
	if((lastvalue[motor-J1] - positionValue) > halfrevolution) {
	    offset[motor-J1] = offset[motor-J1] + offsetValue;
	}
	else if((lastvalue[motor-J1] - positionValue) < -halfrevolution) {
	    offset[motor-J1] = offset[motor-J1] - offsetValue;
	}
	lastvalue[motor-J1] = positionValue; 
	positionValue = positionValue + offset[motor-J1];

	if(compareSWHW) {
	    /* If the HW counter is still valid */
	    if(positionValue > -32000 && positionValue < 32000) {
		if(offset[motor-J1] < 0) {
		    if((positionValue%2048) != 0.00000) {
			checkVal = (int)(positionValue/2048)-1;
		    }
		    else {
			checkVal = (int)(positionValue/2048);
		    }
		}
		else {
		    checkVal = (int)(positionValue/2048);
		}
		if(checkVal != revValue){
		    /* Change SW counter to follow HW */
		    //offset[motor-1] = revValue*2048;
		    if(counter < 500) {
			logArray[counter-J1] = motor;
			hwArray[counter-J1] = revValue;
			swArray[counter-J1] = checkVal;
			positionArray[counter-J1] = positionValue;
			realPositionArray[counter-J1] = realPositionValue;
			counter++;
		    }
		}		
	    }
	}

	float tmp = positionValue*angleScale+resolverCorrection[motor-J1];
	return (float) tmp;
    }
    
    /* Result: A printout with: Nbr of joint, value of SW revolution counter, 
     * value of HW revolution counter, value of SW position (whole working area), 
     * value of position given from HW (within one revolution*/
    public void printPositionMismatches() {
	System.out.println("joint, SW, HW, position, realposition");
	for(int s = 0; s < 500; s++) {
	    System.out.print(logArray[s]);
	    System.out.print(" ");
	    System.out.print(swArray[s]);
	    System.out.print(" ");
	    System.out.print(hwArray[s]);
	    System.out.print(" ");
	    System.out.print(positionArray[s]);
	    System.out.print(" ");
	    System.out.println(realPositionArray[s]);
	}
    }
    
    /* Read value of HW revolution counter
     * Only to be used for test purposes */
    public synchronized int readRev(int motor) {
	if(motor == 1) {
	    motor = 0x16;
	}
	else if(motor == 2) {
	    motor = 0x15;
	}
	else if(motor == 3) {
	    motor = 0x14;
	}
	else if(motor == 4) {
	    motor = 0x13;
	}
	else if(motor == 5) {
	    motor = 0x12;
	}
	else {
	    System.out.println("Wrong motor nbr");
	}
	
	io.loBitOut(motor);
	int value = io.digitalIn();
	value = (value & 0xf800); /* Use 5 revolution bits only */
	io.loBitOut(0x1f); /* No adress, no READ and RESET signals */
	return value;
    }

    /* Check zero-sense signal from motor 
     * Bit 0-2 adress, bit 3 read and bit 3 reset 
     */
    public synchronized boolean checkSync(int motor) {

	switch (motor) {
	    case 1 : motor = 0x1e; break;
	    case 2 : motor = 0x1d; break;
	    case 3 : motor = 0x1c; break;
	    case 4 : motor = 0x1b; break;
	    case 5 : motor = 0x1a; break;
	    default : System.out.println("Motors.checkSync(): " + 
			      "Motor number out of range (" + 
			      motor + ")");
	}

	io.loBitOut(motor);

	int result = io.bitIn();
	if((result & 0x02) == 0x02) {
	    return true;
	}
	return false;
    }

    /* Reset motor 
     * Bit 0-2 adress, bit 4 reset
     * Only used by one object, does not have to be synchronized */
    public void reset(int motor) {
	int bits=0x1f;
	switch (motor) {
	    case 1 : bits = 0x0e; break;
	    case 2 : bits = 0x0d; break;
	    case 3 : bits = 0x0c; break;
	    case 4 : bits = 0x0b; break;
	    case 5 : bits = 0x0a; break;
	    default : System.out.println("Motors.reset(): " + 
			      "Motor number out of range (" + 
			      motor + ")");
	}
	io.loBitOut(bits);
//////////////////
	int temp = read(motor); 
	lastvalue[motor-J1] = (temp & 0x07ff);
	temp = temp >> 11;
	offset[motor-J1] = (temp - 16)*offsetValue;
///////////////////
    }

    /* Read digital position of motor */
    private synchronized int read(int motor) {
	int bits=0x1f;
	switch (motor) {
	    case 1 : bits = 0x16; break;
	    case 2 : bits = 0x15; break;
	    case 3 : bits = 0x14; break;
	    case 4 : bits = 0x13; break;
	    case 5 : bits = 0x12; break;
	    default : System.out.println("Motors.read(): " + 
			      "Motor number out of range (" + 
			      motor + ")");
	}
	io.loBitOut(bits);
	int value = io.digitalIn();
	io.loBitOut(0x1f); /* No adress, no READ and RESET signals */
	return value;
    }

    public class Irb6Drive extends Drive {

	private int joint;
	private Motors motors;

	public Irb6Drive(Motors imotors, int ijoint) {
	    super();
	    motors = imotors;
	    joint = ijoint;
	}

	public Sample compute(Sample input) {

	    // Insert Sven code in virtual robot ??
	    motors.analogOut(J1+joint, input.siso);	    

	    input.backward = true;
	    if (input.getSource() instanceof SDSblock) {
		((SDSblock)input.getSource()).update(input);
	    }

	    return input;
	}

	public Sample sampleOutput(Sample s) {

	    s.siso = motors.getPosition(J1+joint);
	    return s;
	}

	public RTEvent putEvent(RTEvent evt) {
	    if (evt instanceof Sample) {
		compute((Sample)evt);
	    }
	    return null;
	}
    }


    /* Result: A printout with: Nbr of joint, value of SW revolution counter, 
     * value of HW revolution counter, value of SW position (whole working area), 
     * value of position given from HW (within one revolution
     */
//     public void printPositionMismatches() {
// 	System.out.println("joint, SW, HW, position, realposition");
// 	for(int s = 0; s < 500; s++) {
// 	    System.out.print(logArray[s]);
// 	    System.out.print(" ");
// 	    System.out.print(swArray[s]);
// 	    System.out.print(" ");
// 	    System.out.print(hwArray[s]);
// 	    System.out.print(" ");
// 	    System.out.print(positionArray[s]);
// 	    System.out.print(" ");
// 	    System.out.println(realPositionArray[s]);
// 	}
//     }

    /* Read value of HW revolution counter
     * Only to be used for test purposes 
     */
//     public synchronized int readRev(int motor) {
// 	int bits=0x1f;
// 	switch (motor) {
// 	    case 1 : motor = 0x16; break;
// 	    case 2 : motor = 0x15; break;
// 	    case 3 : motor = 0x14; break;
// 	    case 4 : motor = 0x13; break;
// 	    case 5 : motor = 0x12; break;
// 	    default : System.out.println("Motors.readRev(): " + 
// 			      "Motor number out of range (" + 
// 			      motor + ")");
// 	}
// 	io.loBitOut(bits);
// 	int value = io.digitalIn();
// 	value = (value & 0xf800); // Use 5 revolution bits only
// 	io.loBitOut(0x1f); // No adress, no READ and RESET signals 
// 	return value;
//     }

//     private synchronized float analogIn(int motor) {
// 	int result = io.analogIn(motor);
// 	//	System.out.println("analogIn " + result);
// 	return ((float)20.0/4095.0)*result-(float)10.0;
//     }
}
