package se.lth.cs.robot.servo;

/**
 * Simplest possible form of sampled dynamic system algorithm, simply
 * specifiying output computation based on input data and state update (as a
 * separate method to support state-consistency techniques suchas tracking
 * control and anti-windup protection).
 * 
 */
public interface Algorithm {

    /**
     * Calculate output value of the algorithm.
     * 
     * @param y
     *            the measured value
     * @param r
     *            the setpoint
     * @return the desired control signal
     */
    float compute(float y, float r);

    /**
     * Update state.
     * 
     * @param u
     *            the effective algorithm output, after succeeding saturation
     *            and the like.
     */
    void update(float u);

    /**
     * Set the parameters of the algorithms.
     * 
     * @param p
     *            the new set of parameters.
     */
    void setParameters(Parameters p);

    /**
     * Forget history and initiate data to startup values.
     */
    void reset();
    
    /**
     * Assign provided state vector to the state vector of the dynamic system defined by this algorithm.
     * @param x the new states to be used after synchronized assignment.
     */
    void getState(StateVector x);
    
    /**
     * Obtain the current state vector of the algorithm
     * @param x the object to be used for returning the state
     */
    void setState(StateVector x);
}
