/*
 * $Id: Drive.java,v 1.2 2005/09/04 19:36:21 klas Exp $
 */
package se.lth.cs.robot.servo;

/**
 * Interface to one motor; describing the type of motor, its connection via
 * actual IO ports, and the type of any built-in control. Such built-in control
 * can provide (for instance) velocity control that then is external to the
 * software (or to the part of the software that refers to this motor connection
 * object), which implies that the dimension of the output is velocity [rad/s].
 * Port numbers apply to specific hardware interfaces as expressed in specific
 * objects encapsulating the actual IO configuration; port numbers or included
 * here to facilitate parameterization of motor classes rather than being part
 * of some generic IO framework.
 */

public class Drive extends SDSblock {

	public static final short TRQ = 0;

	public static final short VEL = 1;

	public static final short POS = 2;

	public static final short VOLT = 3;

	public static final short CURR = 4;

	public static final short SYNC = 5;

	public static final short ASYN = 6;

	public static final short PWM1 = 7;

	public static final short PWMM = 8;

	public static final short STEP = 9;

	public static short SIZE = STEP + 1;

	public static String[] typeString = { "trq", "vel", "pos", "volt", "curr",
			"sync", "asyn", "pwm1", "pwmm", "step" };

//	static {
//		if (typeString.length != SIZE)
//			System.exit(1);	
//	}

	protected short type;

	protected int sensorPort;

	protected int sensorOffset;

	protected float sensorScale;

	protected int actuatorPort;

	protected int actuatorOffset;

	protected float actutorScale;

	protected int internalOffset;

	protected boolean integrating;

	protected boolean rotational;

	/**
	 * Convert a motor type number to the corresponding string.
	 * 
	 * @param type
	 *            the type of the motor connection as enumerated by the
	 *            <code>short</code> constants (starting with <code>TRQ</code>).
	 * @return the type name as a string that is three or four characters long.
	 *         The string "null" is returned the <code>type</code>does not
	 *         index a string in <code>typeString</code>.
	 */
	public static String typeName(short type) {
		return type < SIZE && type >= 0 ? typeString[type] : "null";
	}

	private static final long serialVersionUID = 1L;

}
