/*
 * $Id: Parameters.java,v 1.2 2005/09/06 12:23:07 sven Exp $
 */
package se.lth.cs.robot.servo;

import se.lth.cs.realtime.RTError;

/**
 * Parameters for feedback control, as array of float with names in separate string array.
 * 
 */
public class Parameters implements Cloneable {

	public float[] val;
	public String[] label;
	
	public static final float TRUE = 1f;
	public static final float ZERO = 0f;
	
	/**
	 * Create an empty set of parameters
	 */
	public Parameters() {
		val = new float[0];
	}

	/**
	 * Create parameter object with given sizes and names. 
	 * @param dimension the size of the parameter (float) array
	 * @param valName the names of the tunable parameters
	 * @param flagName the names of the logical (boolean) states
	 */
	public Parameters(int dimension, String[] label) {
		val = new float[dimension];
		this.label = label;
	}
	
	/**
	 * Convert a flag in terms of a floating point number into a boolean. 
	 * @param value the argument to check
	 * @return the value 1.0 (exactly) representing true and all other values
	 * resulting in a false.
	 */
	public static boolean bool(float value) {
		return value==TRUE;
	}
	
	
	/**
	 * As method <code>bool(float)</code> but using a specific element of the
	 * parameter array.
	 * @param index the index of the element to use
	 * @return the boolean value
	 */
	public boolean bool(int index) {
		return bool(val[index]);
	}
	
	/** 
	 * Duplicate.
	 */
	public Object clone() {
		try {
			return super.clone();
		} catch (CloneNotSupportedException err)  {
			throw new RTError(err);
		}
	}
}
