/*
 * $Id: StateVector.java,v 1.2 2005/09/06 12:23:07 sven Exp $
 */
package se.lth.cs.robot.servo;

import java.io.*;
import se.lth.cs.realtime.RTError;

/**
 * State vector for dynamic systems. All different types of states stored in one
 * state vector, but according to a fixed ordering and with a separate dimension 
 * attribute permitting access of specific part of the vector.
 */

public class StateVector implements Cloneable, Serializable {
    
    private static final long serialVersionUID = 1L; //TBD

    //public static class Dimension implements Cloneable, Serializable {
    public static class Dimension implements Serializable {

	    private static final long serialVersionUID = 2L; //TBD
	    short dimension;
	    short physicalStates;
	    short windupStates;
	    short trackingStates;
	    /*
	       public Object clone() throws CloneNotSupportedException {
	    //return clone();
	    return super.clone();
	    }
	    */
    }
	
	float[] state;
	Dimension size;
	String[] physicalStateName;

	/**
	 * Create empty state vector.
	 */
	public StateVector() {
		state = new float[0];
	}
	
	/**
	 * Create zero-filled state vector with specified dimension.
	 */
	public StateVector(Dimension dim) {
		state = new float[dim.dimension];
		try {
		size = (Dimension) dim.clone();
		} catch (CloneNotSupportedException exc) {
			throw new RTError(exc);
		}
	}
	
	/**
	 * Create zero-filled state vector with specified dimension.
	 * @param dimension the created length of the state array, then stored 
	 * in the internal <code>Dimension</code> attribute defining the used part
	 * of the vector (that can be less than <code>state.length</code>).
	 * @param physicalStates number of states that have physical meaning.
	 * @param windupStates
	 * @param trackingStates
	 */
	public StateVector(short dimension, short physicalStates,
			short windupStates, short trackingStates) {
		Dimension dim = new Dimension();
		dim.dimension = dimension;
		dim.physicalStates = physicalStates;
		dim.windupStates = windupStates;
		dim.trackingStates = trackingStates;
		setSize(dim);
		state = new float[dim.dimension];
	}
	
	/**
	 * @return Returns the physicalStateName.
	 */
	public String[] getPhysicalStateName() {
		return physicalStateName;
	}
	/**
	 * @param physicalStateName The physicalStateName to set.
	 */
	public void setPhysicalStateName(String[] physicalStateName) {
		this.physicalStateName = physicalStateName;
	}
	/**
	 * @return Returns the size.
	 */
	public Dimension getSize() {
		return size;
	}
	/**
	 * @param size The size to set.
	 */
	public void setSize(Dimension size) {
		this.size = size;
	}
	/**
	 * @return Returns the state.
	 */
	public float[] getState() {
		return state;
	}
    /**
     * @return Returns the specified state element.
     */
    public float getState(int ix) {
        return state[ix];
    }
	/**
	 * @param state The state to set.
	 */
	public void setState(float[] state) {
		this.state = state;
	}
    /**
     * @param state The state to set.
     * @param ix the index of the state element
     */
    public void setState(float state, int ix) {
        this.state[ix] = state;
    }
}
