package se.lth.cs.throttlenet;
import se.lth.cs.throttlenet.event.*;
import java.net.*;
import java.io.*;

/**
 * The <code>ThrottleNetSocket</code> class implements a socket for throttlenet.
 * Because this throttlenet simulation works on UDP, this socket simply
 * extends DatagramSocket.
 * 
 * @author  Patricia Grudziecka
 * @author  Daniel Nyberg
 * @version 0.4, 24/10/03
 * @see DatagramSocket
 * @since   ThrottleNet0.1
 */
abstract class ThrottleNetSocket {
	
	
	/**
	 * The datagram socket used for communication.
	 */
	protected int socket;
		
	
	/** 
	 * The name of the connection between two nodes.
	 *
	 * @serial
	 */
	protected String name;
		
	
	/**
	 * The amount of data in a throttlenet packet in bytes.
	 * 
	 */
	protected int size;
		
	
	/**
	 * The time between two packets in microseconds.
	 * 
	 */
	protected long period_us;
		
	
	/**
	 * Dummy constructor needed by Java2C compiler.
	 * 
	 */
	public ThrottleNetSocket() {
		socket = 0;
		size = 0;
		name = null;
		period_us = 0;
	}	
	
	
	/**
	 * Constructs a throttlenet socket.
	 * 
	 * @param   name      the name of the connection.
	 * @param   size      the amount of data in the throttlenet packet.
	 * @param   period_us the time between two packets in microseconds.
	 * 
	 * @exception SocketException  if the socket could not be opened, or the
	 *  						socket could not bind to the specified local port.
	 */
	public ThrottleNetSocket(String name, int size, int period_us)
											throws SocketException, IOException {
		this.name = name;
		this.size = size;
		this.period_us = period_us;
		socket = 0;
	}
	
	
	/**
	 * Attaches a throttlenet listener.
	 * 
	 * @param l throttlenet listener.
	 * 
	 */
	public void addThrottleNetListener(ThrottleNetListener l) {
		// Lacks implementations
	}
	
	
	/**
	 * Removes a throttlenet listener.
	 * 
	 * @param l throttlenet listener.
	 * 
	 */
	public void removeThrottleNetListener(ThrottleNetListener l) {
		// Lacks implementations
	}
	
	
	
}
