package se.lth.cs.throttlenet.globethrottle;


/**
 * The <code>NewConnectionSettings</code> contains the information about a 
 * specific connection and its new settings.  
 * 
 * @author  Patricia Grudziecka
 * @author  Daniel Nyberg
 * @version 0.7, 4/11/03
 * @since   ThrottleNet0.2
 */
class NewConnectionSettings {
	
	
	/**
	 * The updated fragment size.
	 */
	private int fragmentSize;
	
	
	/**
	 *  The updated fragment time.
	 */
	private int fragmentTime;
	
	
	/**
	 * The connection name.
	 */
	private String name;
	
	
	/**
	 * Constructs new connection settings.
	 * 
	 * @param name				the connection name.
	 * @param fragmentSize  the updated fragment size.
	 * @param fragmentTime	the updated fragment time.
	 */
	public NewConnectionSettings(String name, int fragmentSize, int fragmentTime) {
		this.name = name;
		this.fragmentSize = fragmentSize;
		this.fragmentTime = fragmentTime;
	}
	
	
	/**
	 * Returns the name of the connection.
	 * 
	 * @return the name of the connection.
	 *
	 */
	public String getName() { 
		return name; 
	}
	
	
	/**
	 * Returns the updated fragment size.
	 * 
	 * @return the updated fragment size.
	 *
	 */
	public int getFragmentSize() { 
		return fragmentSize; 
	}
	
	
	/**
	 * Returns the updated fragment time.
	 * 
	 * @return the updated fragment time.
	 *
	 */
	public int getFragmentTime() { 
		return fragmentTime;
	}
	
	
	/**
	 * Returns the string representation of this object.
    *
    * @return the string representation of this object.
    */
	public String toString() {
		String tmp = "NewConnectionSettings{";
		tmp += "name=" + name + ",fragmentSize=" + fragmentSize;
		return tmp + ",fragmentTime=" + fragmentTime + "}";
	}
}
