package se.lth.cs.throttlenet.globethrottle;


/**
 * The <code>SocketInfo</code> class contains all the information needed to 
 * reach two ThrottleNet nodes involved in a specific connection.
 * 
 * @author  Patricia Grudziecka
 * @author  Daniel Nyberg
 * @version 0.4, 24/10/03
 * @since   ThrottleNet0.3
 */
class SocketInfo{

	
	/**
	 * The address of receiving node.
	 * 
	 * @serial
	 */
	private String dstAddress;
	
	
	/**
	 * The address of sending node.
	 * 
	 * @serial
	 */
   private String srcAddress;
	
	
	/**
	 * The TCP port of receiving node.
	 * 
	 */
   private int dstPort;
	
	
	/**
	 * The address of sending node.
	 * 
	 */
   private int srcPort;
	
	
	/**
	 * Constructs socket information of a connection.
	 * 
	 * @param   dstAddress      the address of the receiving node.
	 * @param   dstPort      	 the TCP port of receiving node.
	 * 
	 */
   public SocketInfo(String dstAddress,int dstPort){
		this.dstAddress = dstAddress;
		this.dstPort = dstPort;
		srcAddress = null;
		srcPort = -1;
   }
	
	
	/**
	 * Sets the address of the sending node.
    *
    * @param address the address of the sending node.
    */
   public void setAddress(String address){
		srcAddress = address;
   }

	
	/**
	 * Sets the port of the sending node.
    *
    * @param port the TCP port of sending node.
    */
   public void setPort(int port){
		srcPort = port;
   }
	 
	
	/**
	 * Returns the address of the receiving node.
    *
    * @return the address of the receiving node.
    */
	public String getDstAddress(){
		return dstAddress;
	}
	 
	 
	/**
	 * Returns the address of the sending node.
    *
    * @return the address of the sending node.
    */
	public String getSrcAddress(){
		return srcAddress;
	}
	 
	
	/**
	 * Returns the port of the receiving node.
    *
    * @return the port of the receiving node.
    */
	public int getDstPort(){
		return dstPort;
	}
	
	
	/**
	 * Returns the port of the sending node.
    *
    * @return the port of the sending node.
    */
	public int getSrcPort(){
		return srcPort;
	}
	
	
	/**
	 * Returns the string representation of this object.
    *
    * @return the string representation of this object.
    */
	public String toString() {
		String tmp = "SocketInfo{";
		tmp += "srcAddress=" + srcAddress;
		tmp += ",srcPort=" + srcPort;
		tmp += ",dstAddress=" + dstAddress;
		tmp += ",dstPort=" + dstPort + "}";
		return tmp;
	}
}

