package se.lth.cs.util.byteArrayAccess;

public abstract class ByteArrayAccess {


    /* Constants */
    public static final int BIG_ENDIAN = 1;
    public static final int LITTLE_ENDIAN = 2;

    /* Shared memory header layout */
    public static final int ENDIANNESS_OFFSET = 0; // endianness 
    public static final int SIZE_OFFSET = 4;	   // size of memory block	
    public static final int ADDRESS_OFFSET = 8;    // address on defining CPU
    public static final int DATA_OFFSET = 12;	   // start of user data


    public final void setShort(short offset, short val){
        setShortInArray(mem, offset, val);
    }

    public final short getShort(int offset){
        return getShortInArray(mem, offset);
    }

    public final int getInt(int offset){
        return getIntInArray(mem, offset);
    }
    public final void setInt(int offset, int val){
        setIntInArray(mem, offset, val);
    }

    public final long getLong(int offset) {
        return getLongInArray(mem, offset);
    }

    public final void setLong(int offset, long val){
        setLongInArray(mem, offset, val);
    }

    public final float getFloat(int offset){
        return getFloatInArray(mem, offset);
    }

    public final void setFloat(int offset, float val){
        setFloatInArray(mem, offset, val);
    }

    /* These methods depend on the memory layout (endianness, etc) and
       must be defined in the corresponding subclass */

    public abstract short getShortInArray(byte[] mem, int offset);
    public abstract void setShortInArray(byte mem[], int offset, short val);
    public abstract int getIntInArray(byte[] mem, int offset);
    public abstract void setIntInArray(byte mem[], int offset, int val);
    public abstract long getLongInArray(byte[] mem, int offset);
    public abstract void setLongInArray(byte mem[], int offset, long val);
    public abstract float getFloatInArray(byte[] mem, int offset);
    public abstract void setFloatInArray(byte mem[], int offset, float val);

/* NB! The get/set methods are virtual, but their implementation does not
       have to depend on the state of the object, as the byte[] mem can be
       sent as a parameter. Therefore, if the actual implementations, in the 
       subclasses, are final the compiler could detect this and inline the 
       methods, just as if they were static.
*/

    protected byte mem[];

    public ByteArrayAccess() {
    	System.exit(1);
    }	

    public ByteArrayAccess(byte a[]) {
    	mem = a;
    }	

    public byte[] getByteArray(){
    	return mem;
    }	

   public void printArray(int len){
       System.out.println("_________SharedMem___________");
       for(int i=0; i<len; i++) {
           for(int j=0; j<4;j++) {
                System.out.print(mem[4*i+j]); 
		System.out.print(" ");
            }
            System.out.println();
        }
        System.out.println("_________   end   ___________");
   }

/* I'm not sure that these methods should be here at all...
*/ 
    public final int[] getIntVector(byte[] mem, int offset, int size) {
	int[] ret = new int[size];
	int i;
	for (i = 0; i < size; i++) {
	    ret[i] = getIntInArray(mem, offset + i * 4);
	}
	return ret;
    }

    public final void setIntVector(byte[] mem, int offset, int size, int[] val) {
	int i;
	for (i = 0; i < size; i++) {
	    setIntInArray(mem, offset + i * 4, val[i]);
	}
    }
}

