/* Ported to LJRT
 *
 * Alfred Theorin <alfred.theorin@control.lth.se>
*/

#include <javatypes.h>
#include <se_lth_control_realtime_comedi_Comedi.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <comedi.h>
#include <comedilib.h>

// TODO: Exceptions
//extern se_lth_control_realtime_comedi_ComediException_Class;
//extern GC_REF(LayoutHead, se_lth_control_realtime_comedi_ComediException_Layout);

typedef struct ComediFile {
  struct ComediFile *next;
  int useCount;
  char *name;
  comedi_t *cf;
} ComediFile;

typedef struct {
  int chan;
  int useCount;
  ComediFile *file;
  int subdev;
  int index;
} ComediPort;

typedef struct {
  ComediPort port;
  int maxdata;
  double min;
  double max;
  double delta;
} AnalogPort;
typedef struct {
  ComediPort port;
} DigitalPort;
typedef struct {
  ComediPort port;
  int min;
} EncoderPort;
static ComediFile *gFile = 0;
static struct {
  int count;
  AnalogPort *port;
} gAin = { -1, 0 };
static struct {
  int count;
  AnalogPort *port;
} gAout = { -1, 0 };
static struct {
  int count;
  DigitalPort *port;
} gDin = { -1, 0 };
static struct {
  int count;
  DigitalPort *port;
} gDout = { -1, 0 };
static struct {
  int count;
  EncoderPort *port;
} gEin = { -1, 0 };

// LJRT Hack
typedef int JNIEnv;
static JNIEnv env = 0;

static void throwComedi(JNIEnv *env, int chan, char *exceptionName)
{
/*
  GC_REF(,_tmp_1);
  GC_PUSH_ROOT(_tmp_1);
  GC_REF_FUNC_CALL(_tmp_1,
				   se_lth_control_realtime_comedi_ComediException_ComediException,
				   &se_lth_control_realtime_comedi_ComediException_Class,
				   GC_PASS(se_lth_control_realtime_comedi_ComediException_Layout));// CHECK(0);
  EXCEPTION_STORE_OBJ(_tmp_1);
  EXCEPTION_THROW(0, &se_lth_control_realtime_comedi_ComediException_Class)
  GC_POP_ROOT(_tmp_1);
  */
  // TODO
  printf("\nWARNING! ComediException thrown but ComediException not implemented. Ignored Exception of kind \"%s\"\n\n", exceptionName);
}

static void throwComediNotOpenException(JNIEnv *env, int chan)
{
  throwComedi(env, chan, "se/lth/control/realtime/comedi/ComediNotOpenException");		      
}

static void throwComediDeviceReadException(JNIEnv *env, int chan)
{
  throwComedi(env, chan, "se/lth/control/realtime/comedi/ComediDeviceReadException");		      
}

static void throwComediDeviceWriteException(JNIEnv *env, int chan)
{
  throwComedi(env, chan, "se/lth/control/realtime/comedi/ComediDeviceWriteException");		      
}

static void throwComediDeviceDoesNotExistException(JNIEnv *env, int chan)
{
  throwComedi(env, chan, "se/lth/control/realtime/comedi/ComediDeviceDoesNotExistException");		      
}

static void throwComediRangeNotFoundException(JNIEnv *env, int chan)
{
  throwComedi(env, chan, "se/lth/control/realtime/comedi/ComediRangeNotFoundException");		      
}

static void throwComediConfigurationFileNotFoundError(JNIEnv *env)
{
  throwComedi(env, 0, "se/lth/control/realtime/comedi/ComediConfigurationFileNotFoundError");		      
}

static int OpenFile(JNIEnv *env, ComediFile *f, int chan) 
{
  int result = 0;

  if (!f || !f->name) {
    throwComediDeviceDoesNotExistException(env, chan);
  } else {
    if (!f->useCount) {
      f->cf = comedi_open(f->name); 
    }
    if (f->cf) {
      f->useCount++;
      result = f->useCount;
    } else {
      throwComediNotOpenException(env, chan);
    }
  }
  return result;
}

static void CloseFile(JNIEnv *env, ComediFile *f, int chan) 
{
  if (!f || !f->name) {
    throwComediDeviceDoesNotExistException(env, chan);
  } else {
    if (f->useCount <= 0 || !f->cf) {
      throwComediNotOpenException(env, chan);
    } else {
      f->useCount--;
      if (f->useCount == 0) {
	comedi_close(f->cf);
	f->cf = 0;
      }
    }
  }
}

static int Open(JNIEnv *env, ComediPort *p, int chan) 
{
  int result = 0;

  if (!p) {
    throwComediDeviceDoesNotExistException(env, chan);
  } else if (OpenFile(env, p->file, chan)) {
    p->useCount++;
    result = p->useCount;
  }
  return result;
}

static void Close(JNIEnv *env, ComediPort *p, int chan) 
{
  if (!p) {
    throwComediDeviceDoesNotExistException(env, chan);
  } else if (p->useCount <= 0) {
    throwComediNotOpenException(env, chan);
  } else {
    p->useCount--;
    CloseFile(env, p->file, chan);
  }
}

static lsampl_t Maxdata(ComediPort *p) {
  return comedi_get_maxdata(p->file->cf, p->subdev, p->index);
}

static comedi_range *GetRange(ComediPort *p, unsigned int range) {
  return comedi_get_range(p->file->cf, p->subdev, p->index, range);
}


static lsampl_t DataRead(JNIEnv *env,
			 ComediPort *p, unsigned int range, unsigned int ref) {
  lsampl_t result = 0;

  if (!p->useCount) {
    throwComediNotOpenException(env, p->chan);
  } else if (comedi_data_read(p->file->cf, p->subdev, p->index, 
		       range, ref, &result) < 0) {
    throwComediDeviceReadException(env, p->chan);
  }
  return result;
}

static void DataWrite(JNIEnv *env, 
		      ComediPort *p, unsigned int range, unsigned int ref,
		      lsampl_t data) {
  if (!p->useCount) {
    throwComediNotOpenException(env, p->chan);
  } else if (comedi_data_write(p->file->cf, p->subdev, p->index, 
			range, ref, data) < 0) {
    throwComediDeviceWriteException(env, p->chan);
  }
}

static ComediFile *AddFile(char *name) 
{
  ComediFile *result = gFile;
  while (result != 0) {
    if (strcmp(result->name, name) == 0) { break; }
    result = result->next;
  }
  if (result == 0) {
    result = (ComediFile*) malloc(sizeof(ComediFile));
    result->next = gFile;
    result->useCount = 0;
    result->name = strdup(name);
    result->cf = 0;
    gFile = result;
  }
  return result;
}

GC_VAR_FUNC_BEGIN(JDouble, se_lth_control_realtime_comedi_Comedi_analogIn_int, JInt channel)
GC_FUNC_ENTER;
{
  double result = 0.0;

  if (channel < 0 || channel > gAin.count ||
      gAin.port[channel].port.file == 0) {
    throwComediDeviceDoesNotExistException(env, channel);
  } else {
    AnalogPort *p = &gAin.port[channel];
    result = p->min + p->delta * DataRead(env, &p->port, 0, 0);
  }
  GC_RETURN_VAR(result);
}
GC_FUNC_LEAVE;
GC_VAR_FUNC_END(JDouble, se_lth_control_realtime_comedi_Comedi_analogIn_int)

GC_PROC_BEGIN(se_lth_control_realtime_comedi_Comedi_analogOut_int_double, JInt channel, JDouble value)
GC_FUNC_ENTER;
{
  if (channel < 0 || channel > gAout.count ||
      gAout.port[channel].port.file == 0) {
    throwComediDeviceDoesNotExistException(env, channel);
  } else {
    AnalogPort *p = &gAout.port[channel];
    int intvalue;
    if (value < p->min) {
      intvalue = 0;
    } else if (value > p->max) {
      intvalue = p->maxdata;
    } else {
      intvalue = (value - p->min) / p->delta;
    }
    if (intvalue < 0) { 
      intvalue = 0; 
    } else if (intvalue > p->maxdata) {
      intvalue = p->maxdata;
    }
    DataWrite(env, &p->port, 0, 0, intvalue);
  }
}
GC_FUNC_LEAVE;
GC_PROC_END()

GC_VAR_FUNC_BEGIN(JBoolean, se_lth_control_realtime_comedi_Comedi_digitalIn_int, JInt channel)
GC_FUNC_ENTER;
{
  int result = 0;

  if (channel < 0 || channel > gDin.count ||
      gDin.port[channel].port.file == 0) {
    throwComediDeviceDoesNotExistException(env, channel);
  } else {
    DigitalPort *p = &gDin.port[channel];

    result = DataRead(env, &p->port, 0, 0) != 0;
  }
  GC_RETURN_VAR(result);
}
GC_FUNC_LEAVE;
GC_VAR_FUNC_END(JBoolean, se_lth_control_realtime_comedi_Comedi_digitalIn_int)

GC_PROC_BEGIN(se_lth_control_realtime_comedi_Comedi_digitalOut_int_boolean, JInt channel, JBoolean value)
GC_FUNC_ENTER;
{
  if (channel < 0 || channel > gDout.count ||
      gDout.port[channel].port.file == 0) {
    throwComediDeviceDoesNotExistException(env, channel);
  } else {
    DigitalPort *p = &gDout.port[channel];
    int intvalue = value != 0;
    DataWrite(env, &p->port, 0, 0, intvalue);
  }
}
GC_FUNC_LEAVE;
GC_PROC_END()

GC_VAR_FUNC_BEGIN(JBoolean, se_lth_control_realtime_comedi_Comedi_encoderIn_int, JInt channel)
GC_FUNC_ENTER;
{
  long result = 0;

  if (channel < 0 || channel > gEin.count || 
      gEin.port[channel].port.file == 0) {
    throwComediDeviceDoesNotExistException(env, channel);
  } else {
    EncoderPort *p = &gEin.port[channel];

    result = p->min + DataRead(env, &p->port, 0, 0);
  }
  GC_RETURN_VAR(result);
}
GC_FUNC_LEAVE;
GC_VAR_FUNC_END(JBoolean, se_lth_control_realtime_comedi_Comedi_encoderIn_int)

GC_PROC_BEGIN(se_lth_control_realtime_comedi_Comedi_analogInOpen_int, JInt channel)
GC_FUNC_ENTER;
{
  if (channel >= 0 && channel <= gAin.count) {
    AnalogPort *p;

    p = &gAin.port[channel];
    if (Open(env, &p->port, channel) == 1) {
      // Get range on first open of channel
      comedi_range *range;
      int flags;

      range = GetRange(&p->port, 0);
      p->maxdata = Maxdata(&p->port);
      if (range) {
	p->min = range->min;
	p->max = range->max;
	p->delta = (p->max - p->min) / p->maxdata;
      } else {
	Close(env, &p->port, channel);
	throwComediRangeNotFoundException(env, channel);
      }
    }
  } else {
    throwComediDeviceDoesNotExistException(env, channel);    
  }
}
GC_FUNC_LEAVE;
GC_PROC_END()

GC_PROC_BEGIN(se_lth_control_realtime_comedi_Comedi_analogOutOpen_int, JInt channel)
GC_FUNC_ENTER;
{
  if (channel >= 0 && channel <= gAout.count) {
    AnalogPort *p;

    p = &gAout.port[channel];
    if (Open(env, &p->port, channel)  == 1) {
      // Get range on first open of channel
      comedi_range *range;
      int flags;
      
      range = GetRange(&p->port, 0);
      p->maxdata = Maxdata(&p->port);
      if (range) {
	p->min = range->min;
	p->max = range->max;
	p->delta = (p->max - p->min) / p->maxdata;
      } else {
	Close(env, &p->port, channel);
	throwComediRangeNotFoundException(env, channel);
      }
    }
  } else {
    throwComediDeviceDoesNotExistException(env, channel);    
  }
}
GC_FUNC_LEAVE;
GC_PROC_END()

GC_PROC_BEGIN(se_lth_control_realtime_comedi_Comedi_digitalInOpen_int, JInt channel)
GC_FUNC_ENTER;
{
  if (channel >= 0 &&channel <= gDin.count) {
    DigitalPort *p;

    p = &gDin.port[channel];
    Open(env, &p->port, channel);
  } else {
    throwComediDeviceDoesNotExistException(env, channel);    
  }
}
GC_FUNC_LEAVE;
GC_PROC_END()

GC_PROC_BEGIN(se_lth_control_realtime_comedi_Comedi_digitalOutOpen_int, JInt channel)
GC_FUNC_ENTER;
{
  if (channel >= 0 && channel <= gDout.count) {
    DigitalPort *p;

    p = &gDout.port[channel];
    Open(env, &p->port, channel);
  } else {
    throwComediDeviceDoesNotExistException(env, channel);    
  }
}
GC_FUNC_LEAVE;
GC_PROC_END()

GC_PROC_BEGIN(se_lth_control_realtime_comedi_Comedi_encoderInOpen_int, JInt channel)
GC_FUNC_ENTER;
{
  if (channel >= 0 && channel <= gEin.count) {
    EncoderPort *p;
    
    p = &gEin.port[channel];
    if (Open(env, &p->port, channel) == 1) {
      p->min = -Maxdata(&p->port) / 2;
    }
  } else {
    throwComediDeviceDoesNotExistException(env, channel);    
  }
}
GC_FUNC_LEAVE;
GC_PROC_END()

GC_PROC_BEGIN(se_lth_control_realtime_comedi_Comedi_analogInClose_int, JInt channel)
GC_FUNC_ENTER;
{
  if (channel >= 0 && channel <= gAin.count) {
    Close(env, &gAin.port[channel].port, channel);
  } else {
    throwComediDeviceDoesNotExistException(env, channel);    
  }
}
GC_FUNC_LEAVE;
GC_PROC_END()

GC_PROC_BEGIN(se_lth_control_realtime_comedi_Comedi_analogOutClose_int, JInt channel)
GC_FUNC_ENTER;
{
  if (channel >= 0 && channel <= gAout.count) {
    Close(env, &gAout.port[channel].port, channel);
  } else {
    throwComediDeviceDoesNotExistException(env, channel);    
  }
}
GC_FUNC_LEAVE;
GC_PROC_END()

GC_PROC_BEGIN(se_lth_control_realtime_comedi_Comedi_digitalInClose_int, JInt channel)
GC_FUNC_ENTER;
{
  if (channel >= 0 && channel <= gDin.count) {
    Close(env, &gDin.port[channel].port, channel);
  } else {
    throwComediDeviceDoesNotExistException(env, channel);    
  }
}
GC_FUNC_LEAVE;
GC_PROC_END()

GC_PROC_BEGIN(se_lth_control_realtime_comedi_Comedi_digitalOutClose_int, JInt channel)
GC_FUNC_ENTER;
{
  if (channel >= 0 && channel <= gDout.count) {
    Close(env, &gDout.port[channel].port, channel);
  } else {
    throwComediDeviceDoesNotExistException(env, channel);    
  }
}
GC_FUNC_LEAVE;
GC_PROC_END()

GC_PROC_BEGIN(se_lth_control_realtime_comedi_Comedi_encoderInClose_int, JInt channel)
GC_FUNC_ENTER;
{
  if (channel >= 0 && channel <= gEin.count) {
    Close(env, &gEin.port[channel].port, channel);
  } else {
    throwComediDeviceDoesNotExistException(env, channel);    
  }
}
GC_FUNC_LEAVE;
GC_PROC_END()

GC_PROC_BEGIN(se_lth_control_realtime_comedi_Comedi_Init)
GC_FUNC_ENTER;
{
  FILE *conf;
  char buffer[80];
  char *data;
  int din_max, dout_max;

  conf = fopen("/etc/comedi.conf", "r");
  if (conf == 0) {
    throwComediConfigurationFileNotFoundError(env);
  } else {
    while (data = fgets(buffer, 80, conf)) {
      char file[80];
      int chan, subdev, index;
      if (sscanf(data, " ain %d %s %d:%d",
		 &chan, &file, &subdev, &index) == 4) {
	if (chan > gAin.count) { gAin.count = chan; }
      } else if (sscanf(data, " aout %d %s %d:%d", 
			&chan, &file, &subdev, &index) == 4) {
	if (chan > gAout.count) { gAout.count = chan; }
      } else if (sscanf(data, " din %d %s %d:%d", 
			&chan, &file, &subdev, &index) == 4) {
	if (chan > gDin.count) { gDin.count = chan; }
      } else if (sscanf(data, " dout %d %s %d:%d", 
			&chan, &file, &subdev, &index) == 4) {
	if (chan > gDout.count) { gDout.count = chan; }
      } else if (sscanf(data, " ein %d %s %d:%d", 
			&chan, &file, &subdev, &index) == 4) {
	if (chan > gEin.count) { gEin.count = chan; }
      }
    }
    {
      int i;
      gAin.port = (AnalogPort*) calloc(gAin.count + 1, sizeof(AnalogPort));
      for (i = 0 ; i < gAin.count ; i++) { gAin.port[i].port.chan = i; }
      gAout.port = (AnalogPort*) calloc(gAout.count + 1, sizeof(AnalogPort));
      for (i = 0 ; i < gAout.count ; i++) { gAout.port[i].port.chan = i; }
      gDin.port = (DigitalPort*) calloc(gDin.count + 1, sizeof(DigitalPort));
      for (i = 0 ; i < gDin.count ; i++) { gDin.port[i].port.chan = i; }
      gDout.port = (DigitalPort*) calloc(gDout.count + 1, sizeof(DigitalPort));
      for (i = 0 ; i < gDout.count ; i++) { gDout.port[i].port.chan = i; }
      gEin.port = (EncoderPort*) calloc(gEin.count + 1, sizeof(EncoderPort));
      for (i = 0 ; i < gEin.count ; i++) { gEin.port[i].port.chan = i; }
    }
    fseek(conf, 0, SEEK_SET);
    while (data = fgets(buffer, 80, conf)) {
      char file[80];
      int chan, subdev, index;
      ComediPort *p = 0;
      
      if (sscanf(data, " ain %d %s %d:%d", 
		 &chan, &file, &subdev, &index) == 4) {
	p = &gAin.port[chan].port;
      } else if (sscanf(data, " aout %d %s %d:%d", 
			&chan, &file, &subdev, &index) == 4) {
	p = &gAout.port[chan].port;
      } else if (sscanf(data, " din %d %s %d:%d", 
			&chan, &file, &subdev, &index) == 4) {
	p = &gDin.port[chan].port;
      } else if (sscanf(data, " dout %d %s %d:%d", 
			&chan, &file, &subdev, &index) == 4) {
	p = &gDout.port[chan].port;
      } else if (sscanf(data, " ein %d %s %d:%d", 
			&chan, &file, &subdev, &index) == 4) {
	p = &gEin.port[chan].port;
      }
      if (p) {
	p->chan = chan;
	p->file = AddFile(file);
	p->subdev = subdev;
	p->index = index;
      }
    }
    fclose(conf);
  }
}
GC_FUNC_LEAVE;
GC_PROC_END()

