/**
 * This file is part of Ljrt.
 *
 * Copyright (c) 2010-2011	Jesper Öqvist <jesper.oqvist@cs.lth.se>
 */
/* NB: strerror_r is not available on Cris */

#include <gc.h>
#include <gci_common.h>
#include <exception.h>
#include <javatypes.h>

#if defined JAVA_NET_DATAGRAMSOCKET_INDEX
#include <unistd.h>
#include <sys/socket.h>

#include <sys/types.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <stdio.h>

#include <errno.h>

#include <java_lang_String.h>
#include <java_net_DatagramPacket.h>

/* Need this to create strings... */
extern struct java_lang_String_ClassStruct java_lang_String_Class;
extern GC_REF(LayoutHead, java_lang_String_Layout);
/* and Exceptions */
extern struct java_io_IOException_ClassStruct java_io_IOException_Class;
extern GC_REF(LayoutHead,java_io_IOException_Layout);

/**
 * Create a socket.
 */
GC_VAR_FUNC_BEGIN(JInt, java_net_DatagramSocket_createSocket)
	GC_FUNC_ENTER;
{
	JInt result = (JInt) socket(PF_INET, SOCK_DGRAM, 0);
	GC_RETURN_VAR(result);
}
	GC_FUNC_LEAVE;
GC_VAR_FUNC_END(JInt, java_net_DatagramSocket_createSocket)

/**
 * Close the socket.
 */
GC_PROC_BEGIN(java_net_DatagramSocket_closeSocket_int, JInt usock)
	GC_FUNC_ENTER;
{
	close(usock);
}
	GC_FUNC_LEAVE;
GC_PROC_END(java_net_DatagramSocket_closeSocket)

GC_PROC_BEGIN(java_net_DatagramSocket_bind_int_java_lang_String_int,
		JInt usock,
		GC_PARAM(java_lang_String, addr),
		JInt port) EXCEPTION_THROWS
	GC_PARAM_REF(java_lang_String, addr);
	GC_PUSH_PARAM(addr);
	GC_FUNC_ENTER;
{
	struct addrinfo hints, *servinfo, *p;
	int res;
	char port_s[10];

	sprintf(port_s, "%d", (int)port);

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_DGRAM;
	hints.ai_flags = AI_PASSIVE;

	res = getaddrinfo(NULL, port_s, &hints, &servinfo);
	if (res == -1) {
		GC_REF(java_io_IOException, e);
		GC_REF(java_lang_String, s);
		GC_PUSH_ROOT(e);
		GC_PUSH_ROOT(s);
		GC_REF_FUNC_CALL(s,
				java_lang_String_String_CCharArray,
				&java_lang_String_Class,GC_PASS(java_lang_String_Layout),
				strerror(errno)/* NB: strerror is not reentrant! */
				);CHECK(0);
		GC_REF_FUNC_CALL(e,
				java_io_IOException_IOException_java_lang_String,
				&java_io_IOException_Class,
				GC_PASS(java_io_IOException_Layout), 
				GC_PASS(s));
		EXCEPTION_STORE_OBJ(e);
		EXCEPTION_THROW(0, &java_io_IOException_Class);
		GC_POP_ROOT(s);
		GC_POP_ROOT(e);

		/*printf("error %d while binding socket!\n", errno);*/
		GC_RETURN;
	}

	/* loop through all the results and bind to the first we can */
	p = servinfo;
	while (p != NULL) {
		if (bind(usock, p->ai_addr, p->ai_addrlen) != -1) {
			/* bind success! */
			break;
		}

		/* bind fail! Try next address... */
		p = p->ai_next;
	}

	if (p == NULL) {
		GC_REF(java_io_IOException, e);
		GC_REF(java_lang_String, s);
		GC_PUSH_ROOT(e);
		GC_PUSH_ROOT(s);
		GC_REF_FUNC_CALL(s,
				java_lang_String_String_CCharArray,
				&java_lang_String_Class,GC_PASS(java_lang_String_Layout),
				strerror(errno)/* NB: strerror is not reentrant! */
				);CHECK(0);
		GC_REF_FUNC_CALL(e,
				java_io_IOException_IOException_java_lang_String,
				&java_io_IOException_Class,
				GC_PASS(java_io_IOException_Layout), 
				GC_PASS(s));
		EXCEPTION_STORE_OBJ(e);
		EXCEPTION_THROW(0, &java_io_IOException_Class);
		GC_POP_ROOT(s);
		GC_POP_ROOT(e);

		/*printf("failed to bind socket!\n", errno);*/
	}
}
	GC_FUNC_LEAVE;
GC_PROC_END(java_net_DatagramSocket_bind_int_java_lang_String_int)

GC_PROC_BEGIN(java_net_DatagramSocket_connect_int_java_lang_String_int,
		JInt usock, GC_PARAM(java_lang_String, addr), JInt port)
	GC_PARAM_REF(java_lang_String, addr);
	GC_PUSH_PARAM(addr);
	GC_FUNC_ENTER;
{
	struct sockaddr_in ip4addr;
	ip4addr.sin_family = AF_INET;
	ip4addr.sin_port = htons(port);

	{
		GC_REF(JCharArray, data);
		GC_PUSH_ROOT(data);
		GC_GET_REF(data, addr, _java_lang_String_value);
		gc_lock();
		char* s = GC___PTR(data->ref)->data;
		/*printf("connecting socket to %s:%d\n", s, (int)port);*/
		inet_pton(AF_INET, s, &ip4addr.sin_addr);
		gc_unlock();
		GC_POP_ROOT(data);
	}


	int res = connect(usock, (struct sockaddr*)&ip4addr, sizeof ip4addr);
	if (res == -1) {
		GC_REF(java_io_IOException, e);
		GC_REF(java_lang_String, s);
		GC_PUSH_ROOT(e);
		GC_PUSH_ROOT(s);
		GC_REF_FUNC_CALL(s,
				java_lang_String_String_CCharArray,
				&java_lang_String_Class,GC_PASS(java_lang_String_Layout),
				strerror(errno)/* NB: strerror is not reentrant! */
				);CHECK(0);
		GC_REF_FUNC_CALL(e,
				java_io_IOException_IOException_java_lang_String,
				&java_io_IOException_Class,
				GC_PASS(java_io_IOException_Layout), 
				GC_PASS(s));
		EXCEPTION_STORE_OBJ(e);
		EXCEPTION_THROW(0, &java_io_IOException_Class);
		GC_POP_ROOT(s);
		GC_POP_ROOT(e);

		/*printf("error %d while connecting socket!\n", errno);*/
	}
}
	GC_FUNC_LEAVE;
GC_PROC_END(java_net_DatagramSocket_connect_int_java_lang_String_int)

GC_PROC_BEGIN(java_net_DatagramSocket_receive_int_java_net_DatagramPacket,
		JInt usock, GC_PARAM(java_net_DatagramPacket, packet)) EXCEPTION_THROWS
	GC_PARAM_REF(java_net_DatagramPacket, packet);
	GC_PUSH_PARAM(packet);
	GC_FUNC_ENTER;
{
	char str[INET_ADDRSTRLEN];
	char s[INET6_ADDRSTRLEN];
	struct sockaddr_storage from;
	int fromlen = sizeof from;

	JInt length;
	GC_GET(length, packet, _java_net_DatagramPacket_length);
	char* buf = malloc(length);

	int res = recvfrom(usock, buf, length, 0, (struct sockaddr*)&from, &fromlen);
	if (res == -1) {
		GC_REF(java_io_IOException, e);
		GC_REF(java_lang_String, s);
		GC_PUSH_ROOT(e);
		GC_PUSH_ROOT(s);
		GC_REF_FUNC_CALL(s,
				java_lang_String_String_CCharArray,
				&java_lang_String_Class,GC_PASS(java_lang_String_Layout),
				strerror(errno)/* NB: strerror is not reentrant! */
				);CHECK(0);
		GC_REF_FUNC_CALL(e,
				java_io_IOException_IOException_java_lang_String,
				&java_io_IOException_Class,
				GC_PASS(java_io_IOException_Layout), 
				GC_PASS(s));
		EXCEPTION_STORE_OBJ(e);
		EXCEPTION_THROW(0, &java_io_IOException_Class);
		GC_POP_ROOT(s);
		GC_POP_ROOT(e);

		/*printf("error %d while receiving from socket!\n", errno);*/
		GC_RETURN;
	}

	/* Create JByteArray to hold the data */
	length = res;
	GC_REF(JByteArray, data);
	GC_PUSH_ROOT(data);
	GC_REF_FUNC_CALL(data, newJByteArray, length);CHECK(0);
	int i;
	for (i = 0; i < length; ++i) {
		GC_SET_VAR_ARRAY(data, (JInt)i, buf[i]);
	}

	free(buf);

	/* Fetch the sender address and port */
	GC_REF(java_lang_String, addr);
	GC_PUSH_ROOT(addr);
	JInt port;
	if (from.ss_family == AF_INET) {
		struct sockaddr_in* ip4 = &from;
		inet_ntop(AF_INET, &(ip4->sin_addr), str, INET_ADDRSTRLEN);
		port = htons(ip4->sin_port);
	} else {
		struct sockaddr_in6* ip6 = &from;
		inet_ntop(AF_INET6, &(ip6->sin6_addr), str, INET_ADDRSTRLEN);
		port = ntohs(ip6->sin6_port);
	}
	GC_REF_FUNC_CALL(addr, java_lang_String_String_CCharArray, &java_lang_String_Class, GC_PASS(java_lang_String_Layout), str);

	/* Update the packet with new data */
	{
	struct java_net_DatagramPacket_ClassStruct* _tmp_method;
	GC_GET(_tmp_method, packet, class);
	GC_PROC_CALL(_tmp_method->methodTbl->setAddress_java_lang_String, GC_PASS(packet), GC_PASS(addr));
	GC_PROC_CALL(_tmp_method->methodTbl->setPort_int, GC_PASS(packet), (JInt)port);
	GC_PROC_CALL(_tmp_method->methodTbl->setData_byteA, GC_PASS(packet), GC_PASS(data));
	}CHECK(0);

	GC_POP_ROOT(addr);
	GC_POP_ROOT(data);
}
	GC_FUNC_LEAVE;
GC_PROC_END(java_net_DatagramSocket_receive_int_java_net_DatagramPacket)

GC_PROC_BEGIN(java_net_DatagramSocket_send_int_java_net_DatagramPacket_java_lang_String_int,
		JInt usock, GC_PARAM(java_net_DatagramPacket, packet), GC_PARAM(java_lang_String, addr), JInt port) EXCEPTION_THROWS
	GC_PARAM_REF(java_net_DatagramPacket, packet);
	GC_PUSH_PARAM(packet);
	GC_PARAM_REF(java_lang_String, addr);
	GC_PUSH_PARAM(addr);
	GC_FUNC_ENTER;
{
	struct sockaddr_in ip4addr;
	ip4addr.sin_family = AF_INET;
	ip4addr.sin_port = htons(port);

	{
		GC_REF(JCharArray, data);
		GC_PUSH_ROOT(data);
		GC_GET_REF(data, addr, _java_lang_String_value);
		gc_lock();
		char* s = GC___PTR(data->ref)->data;
		inet_pton(AF_INET, s, &ip4addr.sin_addr);
		/*printf("sending packet to %s:%d\n", s, port);*/
		gc_unlock();
		GC_POP_ROOT(data);
	}

	JInt length;
	JInt offset;
	GC_GET(length, packet, _java_net_DatagramPacket_length);
	GC_GET(offset, packet, _java_net_DatagramPacket_offset);
	char* buf = malloc(length);
	{
		GC_REF(JByteArray, data);
		GC_PUSH_ROOT(data);
		GC_GET_REF(data, packet, _java_net_DatagramPacket_buffer);
		gc_lock();
		char* s = GC___PTR(data->ref)->data;
		memcpy(buf, (void*)((int)s+(int)offset), length);
		gc_unlock();
		GC_POP_ROOT(data);
	}

	int total = 0;
	while (total < length) {
		int res = sendto(usock, buf+total, length, 0, (struct sockaddr*)&ip4addr, sizeof ip4addr); 
		if (res == -1) {
			GC_REF(java_io_IOException, e);
			GC_REF(java_lang_String, s);
			GC_PUSH_ROOT(e);
			GC_PUSH_ROOT(s);
			GC_REF_FUNC_CALL(s,
					java_lang_String_String_CCharArray,
					&java_lang_String_Class,GC_PASS(java_lang_String_Layout),
					strerror(errno)/* NB: strerror is not reentrant! */
					);CHECK(0);
			GC_REF_FUNC_CALL(e,
					java_io_IOException_IOException_java_lang_String,
					&java_io_IOException_Class,
					GC_PASS(java_io_IOException_Layout), 
					GC_PASS(s));
			EXCEPTION_STORE_OBJ(e);
			EXCEPTION_THROW(0, &java_io_IOException_Class);
			GC_POP_ROOT(s);
			GC_POP_ROOT(e);

			/*printf("error while sending to socket: %s\n", sterror(errno));*/
			break;
		} else {
			total += res;
		}
	}
	free(buf);
}
	GC_FUNC_LEAVE;
GC_PROC_END(java_net_DatagramSocket_send_int_java_net_DatagramPacket_java_lang_String_int)

GC_PROC_BEGIN(java_net_DatagramSocket_send_connected_int_java_net_DatagramPacket,
		JInt usock, GC_PARAM(java_net_DatagramPacket, packet)) EXCEPTION_THROWS
	GC_PARAM_REF(java_net_DatagramPacket, packet);
	GC_PUSH_PARAM(packet);
	GC_FUNC_ENTER;
{
	JInt length;
	JInt offset;
	GC_GET(length, packet, _java_net_DatagramPacket_length);
	GC_GET(offset, packet, _java_net_DatagramPacket_offset);
	char* buf = malloc(length);
	{
		GC_REF(JByteArray, data);
		GC_PUSH_ROOT(data);
		GC_GET_REF(data, packet, _java_net_DatagramPacket_buffer);
		gc_lock();
		char* s = GC___PTR(data->ref)->data;
		memcpy(buf, (void*)((int)s+(int)offset), length);
		gc_unlock();
		GC_POP_ROOT(data);
	}

	int total = 0;
	while (total < length) {
		int res = send(usock, buf+total, length, 0); 
		if (res == -1) {
			GC_REF(java_io_IOException, e);
			GC_REF(java_lang_String, s);
			GC_PUSH_ROOT(e);
			GC_PUSH_ROOT(s);
			GC_REF_FUNC_CALL(s,
					java_lang_String_String_CCharArray,
					&java_lang_String_Class,GC_PASS(java_lang_String_Layout),
					strerror(errno)/* NB: strerror is not reentrant! */
					);CHECK(0);
			GC_REF_FUNC_CALL(e,
					java_io_IOException_IOException_java_lang_String,
					&java_io_IOException_Class,
					GC_PASS(java_io_IOException_Layout), 
					GC_PASS(s));
			EXCEPTION_STORE_OBJ(e);
			EXCEPTION_THROW(0, &java_io_IOException_Class);
			GC_POP_ROOT(s);
			GC_POP_ROOT(e);
			break;
		} else {
			total += res;
		}
	}
	free(buf);
}
	GC_FUNC_LEAVE;
GC_PROC_END(java_net_DatagramSocket_send_connected_int_java_net_DatagramPacket)

#endif
