/**
 * This file is part of Ljrt.
 *
 */
/**
 * ServerSocket native implementation for the Ljrt class library.
 * The original author is probably Anders Nilsson.
 *
 * /Jesper
 */

#include <gc.h>
#include <gci_common.h>
#include <exception.h>
#include <javatypes.h>
#include <class_indices.h>
#include <java_lang_String.h>

#ifdef JAVA_IO_IOEXCEPTION_INDEX
#include <java_io_IOException.h>
#endif

#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <stdio.h>

#include <signal.h>

/* Need this to create strings... */
extern struct java_lang_String_ClassStruct java_lang_String_Class;
extern GC_REF(LayoutHead, java_lang_String_Layout);
/* and Exceptions */
extern struct java_io_IOException_ClassStruct java_io_IOException_Class;
extern GC_REF(LayoutHead,java_io_IOException_Layout);

/**
 * Returns NULL on success, or an error message on failure.
 *
 * FIXME: Returned string is not thread safe! strerror_r is not available on CRIS!
 */
const char* init_socket(int port, int* sockfd) 
{
	struct addrinfo		hints;
	struct addrinfo*	servinfo;
	struct addrinfo*	p;
	int			yes = 1;
	int			rv;
	char			port_s[64];

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE;

	snprintf(port_s, 64, "%d", port);

	rv = getaddrinfo(NULL, port_s, &hints, &servinfo);
	if (rv != 0)
		return gai_strerror(rv);

	for (p = servinfo; p != NULL; p = p->ai_next) {
		*sockfd = socket(p->ai_family, p->ai_socktype, p->ai_protocol);
		if (*sockfd == -1)
			continue;

		if (setsockopt(*sockfd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1) {
			/*perror("native connect (setsockopt)");*/
			close(*sockfd);/* should close? */
			return strerror(errno);/* should be fatal?? */
		}

		if (bind(*sockfd, p->ai_addr, p->ai_addrlen) == -1) {
			/*perror("native connect (bind)");*/
			close(*sockfd);
			continue;
		}
		/* done */
		break;
	}

	if (p == NULL)
		return "failed to bind socket";

	freeaddrinfo(servinfo);

	if (listen(*sockfd, 10) == -1) {
		/*perror("native connect (listen)");*/
		close(*sockfd);/* should close? */
		return strerror(errno);
	}

	/* success: return NULL */
	return NULL;
}

/**
 * Create and bind a socket to the given port.
 */
GC_VAR_FUNC_BEGIN(JInt,
		java_net_ServerSocket_createServerSocket_int,
		JInt port) EXCEPTION_THROWS
	GC_FUNC_ENTER;
{
	int		sockfd;
	const char*	error;

	error = init_socket((int)port, &sockfd);

	if (error != NULL) {
		GC_REF(java_io_IOException, exception);
		GC_REF(java_lang_String, errmsg);
		GC_PUSH_ROOT(exception);
		GC_PUSH_ROOT(errmsg);
		GC_REF_FUNC_CALL(errmsg,
				java_lang_String_String_CCharArray,
				&java_lang_String_Class,
				GC_PASS(java_lang_String_Layout),
				error);CHECK(0);

		GC_REF_FUNC_CALL(exception,
				java_io_IOException_IOException_java_lang_String,
				&java_io_IOException_Class,
				GC_PASS(java_io_IOException_Layout), 
				GC_PASS(errmsg));
		EXCEPTION_STORE_OBJ(exception);
		GC_POP_ROOT(errmsg);
		GC_POP_ROOT(exception);
		EXCEPTION_THROW(0, &java_io_IOException_Class);
	}
	GC_RETURN_VAR((JInt)sockfd);
}
	GC_FUNC_LEAVE;
GC_VAR_FUNC_END(JInt, java_net_ServerSocket_createServerSocket_int)

GC_VAR_FUNC_BEGIN(JInt,
		java_net_ServerSocket_accept_int,
		JInt sockfd) EXCEPTION_THROWS
	GC_FUNC_ENTER;
{
	struct sockaddr_storage	client_addr;
	socklen_t		sin_size;
	int			clientfd;

	sin_size = sizeof(client_addr);
	clientfd = accept((int)sockfd, (struct sockaddr*) &client_addr, &sin_size);
	if (clientfd == -1) {
		/* printf("accept failed: %d %s\n", errno, strerror(errno)); */
		GC_REF(java_io_IOException, exception);
		GC_PUSH_ROOT(exception);
		GC_REF_FUNC_CALL(exception,java_io_IOException_IOException,
				&java_io_IOException_Class,
				GC_PASS(java_io_IOException_Layout));
		EXCEPTION_STORE_OBJ(exception);
		GC_POP_ROOT(exception);
		EXCEPTION_THROW(0, &java_io_IOException_Class);
	}

	GC_RETURN_VAR((JInt)clientfd);
}
	GC_FUNC_LEAVE;
GC_VAR_FUNC_END(JInt, java_net_ServerSocket_accept)

/* Close the native socket. */
GC_PROC_BEGIN(java_net_ServerSocket_close_int,
		JInt sockfd) EXCEPTION_THROWS
	GC_FUNC_ENTER
{
	shutdown(sockfd, SHUT_RDWR);
	close(sockfd);
}
	GC_FUNC_LEAVE;
GC_PROC_END(java_net_ServerSocket_close_int)

/********************** HERE BE DRAGONS *****************************
 * implementation of native methods in corresponding stream classes *
 ********************* HERE BE DRAGONS ******************************/
GC_VAR_FUNC_BEGIN(JInt,
		java_io_SimpleTCPInputStream_readByte_int,
		JInt sockfd) EXCEPTION_THROWS
	GC_FUNC_ENTER;
{
	char	data;
	int	length;

	length = recv((int)sockfd, &data, 1, 0);
	if (length == 1) {
		GC_RETURN_VAR((JInt)data);
	} else {
		GC_RETURN_VAR((JInt)-1);
	}
}
	GC_FUNC_LEAVE;
GC_VAR_FUNC_END(JInt, java_io_SimpleTCPInputStream_readByte_int)

GC_VAR_FUNC_BEGIN(JInt,
		java_io_SimpleTCPInputStream_read_int_byteA_int_int,
		JInt sockfd,
		GC_PARAM(JByteArray, data),
		JInt off,
		JInt len) EXCEPTION_THROWS
	GC_PARAM_REF(JByteArray, data);
	GC_PUSH_PARAM(data);
	GC_FUNC_ENTER
{
	/* Test to see if there is a locking problem with read.
 	 * Introduces unnecessary copying, but other threads will never be
 	 * locked by read.
 	 */
	int	length = (int)len;
	char	tmpbuf[length+1];/* +1 reall needed here? */
	int	result;

	result = recv((int)sockfd, tmpbuf, length, 0);
	gc_lock();
	memmove(&GC___PTR(data->ref)->data[off], tmpbuf, length);
	gc_unlock();

	/*gc_lock();*/
	/*char*cbuf = &GC___PTR(data->ref)->data[off];*/
	/*int result = read(sockfd, cbuf, length);*/
	/*gc_unlock();*/

	GC_RETURN_VAR((JInt)result);
}
	GC_FUNC_LEAVE;
	GC_POP_PARAM(data);
GC_VAR_FUNC_END(JInt,java_io_SimpleTCPInputStream_readByte_int_JByteArray_int)

GC_VAR_FUNC_BEGIN(JInt,
		java_io_SimpleTCPOutputStream_writeByte_int_int,
		JInt sockfd,
		JInt b) EXCEPTION_THROWS
	GC_FUNC_ENTER
{
	int	length;
	char	data;
	
	data = (char)b;
	length = send((int)sockfd, &data, 1, 0);

	GC_RETURN_VAR((JInt)length);
}
	GC_FUNC_LEAVE;
GC_VAR_FUNC_END(JInt, java_io_SimpleTCPOutputStream_writeByte_int_int)

GC_VAR_FUNC_BEGIN(JInt,
		java_io_SimpleTCPOutputStream_writeByteA_int_byteA_int_int,
		JInt sockfd,
		GC_PARAM(JByteArray, data),
		JInt off,
		JInt len) EXCEPTION_THROWS
	GC_PARAM_REF(JByteArray, data);
	GC_PUSH_PARAM(data);
	GC_FUNC_ENTER
{
	int	length;
	char*	cbuf;

	gc_lock();
	/*printf("about to write on %d...\n", sockfd);*/
	/*printf("... data == %x\n",data);*/
	/*printf("... data == %x\n",data-ref);*/
	/*printf("cbuf (%d):  %s\n",  len, cbuf);*/
	/*printf("writeByteA fre %d %d\n",off,len); */
	length = (int)len;
	length = send((int)sockfd, &GC___PTR(data->ref)->data[off], length, 0);
	/*if(length != len)	 printf("writeByteA: (%d): %s\n", length, cbuf);*/
	/*	   printf("writeByteA efter\n"); */
	gc_unlock();

	GC_RETURN_VAR((JInt)length);
}
	GC_FUNC_LEAVE;
	GC_POP_PARAM(data);
GC_VAR_FUNC_END(JInt, java_io_SimpleTCPOutputStream_writeByteA_int_byteA_int_int)

GC_PROC_BEGIN(java_io_SimpleTCPOutputStream_close_int,
		JInt sockfd)
	GC_FUNC_ENTER;
{
	shutdown(sockfd, SHUT_RDWR);
	close(sockfd);
}
	GC_FUNC_LEAVE;
GC_PROC_END(java_io_SimpleTCPOutputStream_close_int)

static uint32_t wrap_get_host_by_name(const char* name) 
{
	struct hostent* he;
	
	he = gethostbyname(name);
#ifdef GC_PRINT_DEBUG
	printf("foo\n");
	printf("he->hname = %s\n \
			he->length = %d\n\
			he->haddr = %x %x %x %x\n",
			he->h_name,
			he->h_length,
			he->h_addr[0],
			he->h_addr[1],
			he->h_addr[2],
			he->h_addr[3]);
#endif
	return ((unsigned char)he->h_addr[0] << 24) | 
			((unsigned char)he->h_addr[1] << 16) | 
			((unsigned char)he->h_addr[2] << 8) | 
			(unsigned char)he->h_addr[3];
}

GC_VAR_FUNC_BEGIN(JInt,
		java_net_InetAddress_getLocalHostAddr)
	GC_FUNC_ENTER;
{
	uint32_t	result;

	result = wrap_get_host_by_name("localhost");
#ifdef GC_PRINT_DEBUG
	printf("bar:%ld\n",result);
#endif
	GC_RETURN_VAR(result);
}
	GC_FUNC_LEAVE;
GC_VAR_FUNC_END(JInt, java_net_InetAddress_getLocalHostAddr)

GC_VAR_FUNC_BEGIN(JInt,
		java_net_InetAddress_getHostByName_java_lang_String,
		GC_PARAM(java_lang_String, hostname))
	GC_PARAM_REF(java_lang_String, hostname);
	GC_PUSH_PARAM(hostname);
	GC_FUNC_ENTER
{
#ifdef GC_PRINT_DEBUG
	printf("getHostByName\n");
#endif
	uint32_t	result;
	char*		nodename;
	GC_REF(JCharArray, data);
	GC_PUSH_ROOT(data);
	GC_GET_REF(data, hostname, _java_lang_String_value);

	gc_lock();
	nodename = GC___PTR(data->ref)->data;

#ifdef GC_PRINT_DEBUG
	printf("getHostByName: name == %s\n", s);
#endif
	result = wrap_get_host_by_name(nodename);
	gc_unlock();
#ifdef GC_PRINT_DEBUG
	printf("bar:%ld\n",result);
#endif
	GC_RETURN_VAR((JInt)result);
}
	GC_FUNC_LEAVE;
	GC_POP_PARAM(hostname);
GC_VAR_FUNC_END(JInt, java_net_InetAddress_getHostByName_java_lang_String)

