#include <gc.h>
#include <gci_common.h>
#include <exception.h>
#include <javatypes.h>

#include <stdio.h>
#include <stdlib.h>

#include <strings.h>
#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <netdb.h>

#define PORT 80
#define HEADERSIZE 118
#define MAXDATASIZE 100

static int inited=0;

/**********************************************************
 * 
 * int se.lth.cs.camera.MotionDetector.nativeConnect()
 *
 * Connects to the motion detector.
 *
 * Returns: -1 if something went wrong otherwise the file descriptor.
 *
 **********************************************************/
GC_VAR_FUNC_BEGIN(JInt, se_lth_cs_camera_MotionDetector_nativeConnect)
GC_FUNC_ENTER
{
    int sockfd, dummy, numbytes; 
    char buf[HEADERSIZE+1];
    int ok = 1;
    struct hostent *he;
    struct sockaddr_in addr; // connector's address information 

    if ((he = gethostbyname("localhost")) == NULL) {  // get the host info 
	ok = 0;
	printf("gethostbyname\n");
    }

    if (ok && (sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
	ok = 0;
	printf("socket\n");
    }

    addr.sin_family = AF_INET;    // host byte order 
    addr.sin_port = htons(PORT);  // short, network byte order 
    addr.sin_addr = *((struct in_addr *)he->h_addr);
    memset(&(addr.sin_zero), '\0', 8);  // zero the rest of the struct 

    if (ok && connect(sockfd, (struct sockaddr *)&addr, sizeof(struct sockaddr)) == -1) {
	ok = 0;
	printf("connect\n");
    }

    char* header = "GET /axis-cgi/motion/motiondata.cgi HTTP/1.0\r\nAuthorization: Basic cnQ6c2lnZ2U\r\n\n";

    if (ok && send(sockfd, header, strlen(header), 0) == -1) {
	ok = 0;
	printf("send\n");
    }

    if (ok && (numbytes = recv(sockfd, buf, HEADERSIZE, MSG_WAITALL)) == -1) {
	perror("recv");
	return -1;
    }

    if (!ok) 
      sockfd = -1;

    GC_RETURN_VAR(sockfd);
}
GC_FUNC_LEAVE;
GC_VAR_FUNC_END(JInt, se_lth_cs_camera_MotionDetector_nativeConnect)


/**********************************************************
 * 
 * int se.lth.cs.camera.MotionDetector.nativeLevelUpdate(int fd)
 *
 * Connects to the motion detector and updates values.
 *
 * Returns: -1 if something went wrong otherwise the level value.
 *
 **********************************************************/
GC_VAR_FUNC_BEGIN(JInt, se_lth_cs_camera_MotionDetector_nativeLevelUpdate_int, JInt fd)
GC_FUNC_ENTER
{
    JInt level = -1;
    int numbytes = 0;
    char buf[MAXDATASIZE+1];
    int ok = 1;
    int marks[3];
    int nbr = 0;
    int readbytes = 0;

    while (ok && nbr < 3) {

	if (ok && (readbytes = recv(fd, buf + numbytes, 1, MSG_WAITALL)) == -1) {
	    ok = 0;
	}
	numbytes += readbytes;
	if (buf[numbytes-1] == '=') {
	    marks[nbr] = numbytes-1;
	    nbr++;
	}
    }

    if (ok && (readbytes = recv(fd, buf + numbytes, 3, MSG_WAITALL)) == -1) {
        ok = 0;
    }
    numbytes += readbytes;

    /* for DEBUG
    buf[numbytes] = '\0';
    printf("------ Data -------\n%s\n", buf);
    */

    if (ok && sscanf(buf + marks[1] + 1, "%d", &level) != 1) {
      ok = 0;
    }

    /* For DEBUG
    printf("Level: %d\n", level);
    */
    
    GC_RETURN_VAR(level);
}
GC_FUNC_LEAVE;
GC_VAR_FUNC_END(JInt, se_lth_cs_camera_MotionDetector_nativeLevelUpdate)


/**********************************************************
 * 
 * void se.lth.cs.camera.MotionDetector.nativeClose(int fd)
 *
 * Closes the connection to the motion detector.
 *
 **********************************************************/
GC_PROC_BEGIN(se_lth_cs_camera_MotionDetector_nativeClose_int, JInt fd)
GC_FUNC_ENTER
{
  if (inited)
    close(fd);
  else
    inited=1;
}
GC_FUNC_LEAVE;
GC_PROC_END(se_lth_cs_camera_MotionDetector_nativeClose_int)
