#include <class_indices.h>

#ifdef SE_LTH_CS_REALTIME_SEMAPHORE_MUTEXSEM_INDEX

#include <gc.h>
#include <pthread.h>
#include <javatypes.h>
#include <se_lth_cs_realtime_semaphore_MutexSem.h>

#ifdef GC_HAS_FINALIZERS
static void dealloc_mutex(se_lth_cs_realtime_semaphore_MutexSem *this){
#ifdef GC_PRINT_DEBUG
    printf("freed mutex semaphore: %x\n", this->_se_lth_cs_realtime_semaphore_MutexSem_target);
#endif
	free(this->_se_lth_cs_realtime_semaphore_MutexSem_target);		
}
#endif

typedef struct timespec timespec;

GC_VAR_FUNC_BEGIN(JLong,se_lth_cs_realtime_semaphore_MutexSem_createMutexSem_int,GC_PARAM(se_lth_cs_realtime_semaphore_MutexSem,this),int iValue)
     GC_PARAM_REF(se_lth_cs_realtime_semaphore_MutexSem,this);
     GC_PUSH_PARAM(this);
     GC_FUNC_ENTER;
{
	/* Create mutex attributes - recursive mutex */
        pthread_mutexattr_t* attr = (pthread_mutexattr_t*)malloc(sizeof(pthread_mutexattr_t));
	pthread_mutexattr_init(attr);
	pthread_mutexattr_settype(attr, PTHREAD_MUTEX_RECURSIVE_NP);
	
	/* Create mutex */
	pthread_mutex_t* mutex = (pthread_mutex_t*)malloc(sizeof(pthread_mutex_t));
	pthread_mutex_init(mutex, attr);

#ifdef GC_HAS_FINALIZERS
#ifdef GC_PRINT_DEBUG
    printf("Allocated mutex semaphore: %x\n", mutex);
#endif
  GC_SET_FINALIZER(this, dealloc_mutex);
#endif

  GC_RETURN_VAR((JLong)mutex);
}
GC_FUNC_LEAVE
GC_POP_PARAM(this);
GC_VAR_FUNC_END(JLong,se_lth_cs_realtime_semaphore_MutexSem_createMutexSem_int)

	
GC_PROC_BEGIN(se_lth_cs_realtime_semaphore_MutexSem_giveMutexSem_long,GC_PARAM(se_lth_cs_realtime_semaphore_MutexSem,this),JLong target)
     GC_PARAM_REF(se_lth_cs_realtime_semaphore_MutexSem,this);
     GC_PUSH_PARAM(this);
     GC_FUNC_ENTER;
{
	pthread_mutex_unlock((pthread_mutex_t*)target);
}
GC_FUNC_LEAVE
GC_POP_PARAM(this);
GC_PROC_END(_se_lth_cs_realtime_semaphore_MutexSem_giveMutexSem_long)

GC_PROC_BEGIN(se_lth_cs_realtime_semaphore_MutexSem_takeMutexSem_long,GC_PARAM(se_lth_cs_realtime_semaphore_MutexSem,this),JLong target)
     GC_PARAM_REF(se_lth_cs_realtime_semaphore_MutexSem,this);
     GC_PUSH_PARAM(this);
     GC_FUNC_ENTER;
{
	pthread_mutex_lock((pthread_mutex_t*)target);
}
GC_FUNC_LEAVE
GC_POP_PARAM(this);
GC_PROC_END(_se_lth_cs_realtime_semaphore_MutexSem_takeMutexSem_long)

GC_PROC_BEGIN(se_lth_cs_realtime_semaphore_MutexSem_takeMutexSem_long_long,GC_PARAM(se_lth_cs_realtime_semaphore_MutexSem,this),JLong target, JLong timeout)
	GC_PARAM_REF(se_lth_cs_realtime_semaphore_MutexSem,this);
	GC_PUSH_PARAM(this);
	GC_FUNC_ENTER;
{
	struct timespec	ts;
	
	ts.tv_sec  = timeout/1000;
	ts.tv_nsec = timeout*1000;
	
	pthread_mutex_timedlock((pthread_mutex_t*)target, &ts);
}
GC_FUNC_LEAVE
GC_POP_PARAM(this);
GC_PROC_END(_se_lth_cs_realtime_semaphore_MutexSem_takeMutexSem_long_long)
	
	
#endif
