#include <config.h>
#include "cpu_utilization_logger.h"

static struct proc_dir_entry *proc_logger;

//#define GC_THREAD 2

//COMMON
static int nbrOfIDs = 0;
static RT_TASK *IDbook[MAX_NBR_OF_ID];
static char **names[MAX_NBR_OF_ID];

#if DELAY_MEASURE > 0
// GC_LOCK_MEASURE
static RTIME startTimes[MAX_NBR_OF_ID];
static long maxDelay[MAX_NBR_OF_ID];
static long delayHistogram[MAX_NBR_OF_ID][DELAY_SLOTS + 1];
#endif

#if CPU_TIME > 0
static RTIME lastTime = 0;
static long totalTimes[MAX_NBR_OF_ID];
static long nanoPart[MAX_NBR_OF_ID];
static int runningThreadID = 0;
static int first = 1;
static int gc_running = 0;
static int gc_caller;
#endif

#ifdef CONTEXT_SWITCH_LOGGING
typedef struct csitem {
        unsigned long time;
        int  pid;
	int freemem;
} csitem;

#define CSLOGLENGTH 50000
csitem cslog[CSLOGLENGTH];
int cscnt = 0;

void log_context_switch(void)
{
	if(cscnt < CSLOGLENGTH) {
		RT_TASK* runner = rt_whoami();
		int k=0;
		while (k < nbrOfIDs && runner != IDbook[k]) {
			k++;
		}
		if (k < nbrOfIDs) {
			cslog[cscnt].time = sched_currentTimeMillis();
			cslog[cscnt].pid = k;
			cslog[cscnt].freemem = gc_freemem();
			cscnt++;
		}
	}
}

#endif

#if SAMPLE_RATE > 0
static int sampleCount, sampleCntTmp;
static RTIME sampleStart, sampleEnd;

void cpu_utilization_logger_log_sample(void) {
	sampleCntTmp++;
}

void cpu_utilization_logger_sample_start(void){
	sampleCntTmp = 0;
	sampleStart = rt_get_time_ns();
}

void cpu_utilization_logger_sample_end(void){
	sampleCount = sampleCntTmp;
	sampleEnd = rt_get_time_ns();
}

static int getSampleCount() {
	return sampleCount;
}

static long getSampleTime() {
	return (long) (sampleEnd - sampleStart);

}
static int getSampleRate() {
	return sampleCount / (long) (sampleEnd - sampleStart);
}
#endif

#if GC_CYCLE > 0
static int gc_cycle_count;
static int gc_panic_count;
void cpu_utilization_logger_gc_log_init(void){
	gc_cycle_count=0;
	gc_panic_count=0;
}
void cpu_utilization_logger_gc_log_conclude(void){
	gc_panic_count++;
}
void cpu_utilization_logger_gc_cycle_start(void){
	gc_cycle_count++;
}
static int getGCCycleCount(){
	return gc_cycle_count;
}
#endif
#if LOCK_COUNTING > 0
static int nbrOfLocks[MAX_NBR_OF_ID];
static int nbrOfLazyLocks[MAX_NBR_OF_ID];
#endif

int cpu_utilization_logger_read_proc(char *page, char **start, off_t off,
			   int count, int *eof, void *data) {
    PROC_PRINT_VARS;
    int i, j;

    PROC_PRINT("\t");
    for (i = 0; i < nbrOfIDs; i++) {
	if (names[i] == 0) {
	    PROC_PRINT("\t%d:%d",i, schedPr2javaPr(rt_get_prio(IDbook[i]), 0));
	} else {
	    PROC_PRINT("\t%s:%d", names[i], schedPr2javaPr(rt_get_prio(IDbook[i]), 0));	    
	}
    }
    PROC_PRINT("\n");
#if SAMPLE_RATE > 0
    PROC_PRINT("Sample count = %d", getSampleCount());   
    PROC_PRINT("\n");
#endif
#if GC_CYCLE > 0
    PROC_PRINT("GC cycle count = %d", getGCCycleCount());   
    PROC_PRINT("\n");
#endif
#if DELAY_MEASURE > 0
    // DELAY INFO //
    for (i = 0; i < DELAY_SLOTS; i++) {
	PROC_PRINT("%d - %d:", i * DELAY_SLOTSIZE, (i + 1) * DELAY_SLOTSIZE - 1);
	for (j = 0; j < nbrOfIDs; j++) {
	    PROC_PRINT("\t%ld", delayHistogram[j][i]);
	}
	PROC_PRINT("\n");
    }
    PROC_PRINT("ver %d:", DELAY_SLOTS * DELAY_SLOTSIZE);
    for (i = 0; i < nbrOfIDs; i++) {
	PROC_PRINT("\t%ld", delayHistogram[i][DELAY_SLOTS]);
    }
    PROC_PRINT("\nMax delay:");
    for (i = 0; i < nbrOfIDs; i++) {
	PROC_PRINT("\t%ld", maxDelay[i]);
    }
#elif LOCK_COUNTING > 0
    PROC_PRINT("Locks:\t");
    for (i = 0; i < nbrOfIDs; i++) {
	PROC_PRINT("\t%d", nbrOfLocks[i]);
    }
    PROC_PRINT("\n");    
    PROC_PRINT("Lazy Locks:\t");
    for (i = 0; i < nbrOfIDs; i++) {
	PROC_PRINT("\t%d", nbrOfLazyLocks[i]);
    }
#elif CPU_TIME > 0
    PROC_PRINT("Times:\t");
    for (i = 0; i < nbrOfIDs; i++) {
	PROC_PRINT("\t%d", totalTimes[i]);
    }
#endif

#ifdef CONTEXT_SWITCH_LOGGING
    PROC_PRINT("Context switches:\t");
    for (i = 0; i < cscnt; i++) {
	PROC_PRINT("\t%ld,%d,%d\n", cslog[i].time, cslog[i].pid, cslog[i].freemem);
    }
#endif


    PROC_PRINT("\n");    
    PROC_PRINT_DONE;
}

#if CPU_TIME > 0
void cpu_utilization_logger_signal_handler(void) {
    int i;
    RTIME switchTime = rt_get_time_ns();
    RT_TASK *runner = rt_whoami();
    for (i = 0; i < nbrOfIDs; i++) {
	if (IDbook[i] == runner) {
	    RTIME time_nano = switchTime - lastTime;
	    if (first) {
		first = 0;
	    } else {
		time_nano += nanoPart[runningThreadID];
		nanoPart[runningThreadID] = time_nano % 1048576;
		if (time_nano < 0) {
		    printk("overflow: lastThread%d, Thread%d\n", runningThreadID, i);
		}
		totalTimes[runningThreadID] += time_nano / 1048576;
	    }
	    if (gc_running && i == gc_caller) {
		runningThreadID = GC_THREAD;
	    } else {
		runningThreadID = i;
	    }
	    break;
	}
    }
    lastTime = rt_get_time_ns();
}

void cpu_utilization_logger_setGCRunning() {
    RTIME switchTime = rt_get_time_ns();
    RTIME time_nano = switchTime - lastTime;
    time_nano += nanoPart[runningThreadID];
    nanoPart[runningThreadID] = time_nano % 1048576;
    if (time_nano < 0) {
	printk("overflow: lastThread%d, Thread GCI\n", runningThreadID);
    }
    totalTimes[runningThreadID] += time_nano / 1048576;
    gc_caller = runningThreadID;
    runningThreadID = GC_THREAD;
    gc_running++;
    lastTime = rt_get_time_ns();
}

void cpu_utilization_logger_setGCnotRunning() {
    RTIME switchTime = rt_get_time_ns();
    RTIME time_nano = switchTime - lastTime;
    time_nano += nanoPart[GC_THREAD];
    nanoPart[GC_THREAD] = time_nano % 1048576;
    if (time_nano < 0) {
	printk("overflow: lastThread GCI, Thread %d\n", runningThreadID);
    }
    totalTimes[GC_THREAD] += time_nano / 1048576;
    gc_running--;
    if (gc_running == 0) {
	runningThreadID = gc_caller;
    }
    lastTime = rt_get_time_ns();    
}

void cpu_utilization_logger_printMyNumber() {
    int i;
    RT_TASK *runner = rt_whoami();
    for (i = 0; i < nbrOfIDs; i++) {
	if (IDbook[i] == runner) {
	    printk("Thread%d with priority %d and ID%x is running\n", i, schedPr2javaPr(rt_get_prio(IDbook[i])), runner);
	}
    }
}   
#endif

void cpu_utilization_logger_install() {
    proc_logger = create_proc_entry("gc_lock_stats", S_IFREG|S_IRUGO|S_IWUSR, 0);
    if (proc_logger) {
	proc_logger->read_proc = cpu_utilization_logger_read_proc;
    }
}

void cpu_utilization_logger_uninstall() {
    remove_proc_entry("gc_lock_stats", 0);
}

void cpu_utilization_logger_addNamedTask(RT_TASK *task, char *taskname) {
    if(nbrOfIDs < MAX_NBR_OF_ID) {
	int length;
	char *tmp;
	
	IDbook[nbrOfIDs] = task;
	// Start: Set taskname //
	length = 0;
	while(taskname[length++]);	
	names[nbrOfIDs] = (char*) kmalloc(length, 0);
	tmp = names[nbrOfIDs];
	while(*tmp++ = *taskname++);
	// End : Set taskname //
#if LOCK_CONTING > 0
	nbrOfLocks[nbrOfIDs] = 0;
	nbrOfLazyLocks[nbrOfIDs] = 0;
#elif CPU_TIME > 0
	totalTimes[nbrOfIDs] = 0;
	nanoPart[nbrOfIDs] = 0;
	rt_task_signal_handler(task, &cpu_utilization_logger_signal_handler);
#endif
	nbrOfIDs++;
//	printk("cpu_utilization_logger: Added %s\n", taskname);
    } else {
	printk("To many threads to measure, raise MAX_NBR_OF_ID\n");
    }
}

void cpu_utilization_logger_addTask(RT_TASK *task) {
    if(nbrOfIDs < MAX_NBR_OF_ID) {
	IDbook[nbrOfIDs] = task;
	names[nbrOfIDs] = 0;
#if LOCK_CONTING > 0
	nbrOfLocks[nbrOfIDs] = 0;
	nbrOfLazyLocks[nbrOfIDs] = 0;
#elif CPU_TIME > 0
	totalTimes[nbrOfIDs] = 0;
	nanoPart[nbrOfIDs] = 0;
	rt_task_signal_handler(task, &cpu_utilization_logger_signal_handler);
#endif
//	printk("Added nbr: %d\ with ID%xn", nbrOfIDs, task);
	nbrOfIDs++;
    } else {
	printk("To many threads to measure, raise MAX_NBR_OF_ID\n");
    }
}

#if DELAY_MEASURE > 0
void cpu_utilization_logger_logg(long delay) {
    int k = 0;
    RT_TASK* runner = rt_whoami();
    while (k < nbrOfIDs && runner != IDbook[k]) {
	k++;
    }
    if (k < nbrOfIDs) {
	if (delay > maxDelay[k]) {
	    maxDelay[k] = delay;
	}
	if (delay < 0) {
	} else if (delay / DELAY_SLOTSIZE > DELAY_SLOTS) {
	    delayHistogram[k][DELAY_SLOTS]++;
	} else {
	    delayHistogram[k][delay / DELAY_SLOTSIZE]++;
	}
    }
}

void cpu_utilization_logger_startLockTimer() {
    int k = 0;
    RT_TASK* runner = rt_whoami();
    while (k < nbrOfIDs && runner != IDbook[k]) {
	k++;
    }
    /*  if (k == GC_THREAD) {
	rt_preempt_always(0);
	}*/
    startTimes[k] = rt_get_time_ns();
}

void cpu_utilization_logger_stopLockTimer(RTIME stop) {
    int k = 0;
    RT_TASK* runner = rt_whoami();
/*    if (k == GC_THREAD) {
	rt_preempt_always(1);
	}*/
    while (k < nbrOfIDs && runner != IDbook[k]) {
	k++;
    }
    if (k < nbrOfIDs) {
	long delay = stop - startTimes[k];
	if (delay > maxDelay[k]) {
	    maxDelay[k] = delay;
	}
	if (delay < 0) {
	} else if (delay / DELAY_SLOTSIZE > DELAY_SLOTS) {
	    delayHistogram[k][DELAY_SLOTS]++;
	} else {
	    delayHistogram[k][delay / DELAY_SLOTSIZE]++;
	}
    }
}
#endif

#if LOCK_COUNTING > 0
void cpu_utilization_logger_loggLock() {
    int k = 0;
    RT_TASK* runner = rt_whoami();
    while (k < nbrOfIDs && runner != IDbook[k]) {
	k++;
    }
    if (k < nbrOfIDs) {
	nbrOfLocks[k]++;
    }
}
void cpu_utilization_logger_loggLazyLock() {
    int k = 0;
    RT_TASK* runner = rt_whoami();
    while (k < nbrOfIDs && runner != IDbook[k]) {
	k++;
    }
    if (k < nbrOfIDs) {
	nbrOfLazyLocks[k]++;
    }
}
#endif
