/***********************************************************************
 * File : java_langThread.h
 * Implementation of the main() method called by any program 
 * compiled with Java2C compiled. 
 * gc_lock needs to be defined in the compilation of the generated 
 * files to call the right
 * method, translation of the Java main.
 * 
 * Implementation of the native methods included in java.lang.Thread API
 * 
 * Functions corresponding to native Methods declared in java.lang.Thread
 *		createThread();
 *		start();
 *		<syncrhonized> interrupt();
 *		<syncrhonized> interrupted();
 * 	<static> currentThread();
 *		<static> sleep(long millis);
 *
 * This library needs :
 *		class_indices.h  	 <generated by Java2C>
 *		java_lang_Thread.h <generated by Java2C>
 *		java_types.h		 <included in Java2C releases>
 *		Garbage collector  directory <included in Java2C releases>
 *		misc. libraries :  pthread.h - time.h 
 ************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef CPU_UTILIZATION_LOGGING
#include "cpu_utilization_logger.h"
#endif

#include <gc.h>
#include <gci_thread.h>
#include <exception.h>
#include <javatypes.h>
#include <java_lang_Thread.h>
#include <java_lang_Runnable.h>
#include <iface.h>
#include <class_indices.h>

#if GC_THREAD != GC_THREAD_RTAIKERNEL
#include <errno.h>
#endif

#if GC_THREAD == GC_THREAD_RTAIKERNEL

#include <linux/module.h>
#include <linux/kernel.h>
#include <rtai.h>
#include <rtai_sched.h>

#ifdef SE_LTH_CS_UTIL_LINUX_KERNELSYMBOL_INDEX
#include <asm/uaccess.h>        // needed for get_fs()
#include <asm/processor.h>      // needed for mm_segment_t;

static mm_segment_t fs;
#endif

#elif GC_THREAD == GC_THREAD_PTHREAD

#include <pthread.h>
#include <sys/mman.h>

#ifdef __RTAI__
#include <rtai_usp_posix.h>
#include <rtai_lxrt_user.h>
#include <rtai_lxrt.h>
#endif  // __RTAI__

#endif  // GC_THREAD



#define NS_MS 	1000000
#define MS 	1000
#define NSEC 	1000000000

//message to send in different cases.
#define RTAIKERNEL_NOTIFY_MSG    10
#define RTAIKERNEL_INTERRUPT_MSG 20


MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;

extern GC_REF(LayoutHead,java_lang_Thread_Layout);
extern java_lang_Thread_Class;

/* Number of program's thread alive */
static int nbr_threads;

/* Pointer to the main rtai-thread */
static RT_TASK main_thread_task;

/* Struct linked list for keeping track of dead threads */
typedef struct death_row {
    struct death_row *next;
    RT_TASK *rt_task;
} death_row;

/* root node for dead threads list */
static death_row* dr_root = 0;


/* Mutex protecting access to nbr_threads local variable */
static SEM threads_interrupt_mutex;

/* Mutex protecting access to nbr_threads local variable */
static SEM threads_mutex; 

/* Mutex protecting access to nbr_threads local variable */
static SEM this_param_mutex; 

/* Condition to wake up C main thread when any other thread exits */
static SEM gc_threadhasdied_cond;

/* Semaphore to achieve mutual exclution on the "deat_row"-list */
static SEM death_row_sem;

/*test semaphore*/
//static SEM wait_notify_sem;

/* Variable that stores the range of priorities in posix */
static int pr_step_posix;

/* Variable that stores the range of priorities in JAVA */
static int pr_step_java;

/* Thread specific data key (TSD), needed in the implementation of currentThread */
//pthread_key_t tsd_key;

/* Attribute specifying that the priority is inherited from the creator thread */
//pthread_attr_t attr;

/* Data structure used by join(long millis) and join(long millis,int nanos) */
/*
struct timedjoin_data {
	pthread_t target;
	pthread_mutex_t lock;
	pthread_cond_t cond;
	struct timespec time;
};*/

/* Function increasing the number of C threads alive */
void inc_threads() {
  rt_sem_wait(&threads_mutex);
  nbr_threads++;
  rt_sem_signal(&threads_mutex);
  // printk("Nbr of threads++ -> %d\n",nbr_threads);
}

/* Function decreasing the number of C threads alive */
void dec_threads() {
  rt_sem_wait(&threads_mutex);
  nbr_threads--;
  rt_sem_signal(&gc_threadhasdied_cond);
  rt_sem_signal(&threads_mutex);
  // printk("Nbr of threads-- -> %d\n",nbr_threads);
}

/* Add death_awaiting thread to list */
void enter_death_row(RT_TASK* rt_task){
    death_row* temp;
#ifdef GC_PRINT_DEBUG
    printk("entering deathrow, address:%x\n",rt_whoami());
#endif
    temp = (death_row*) kmalloc(sizeof(death_row),0);
    rt_sem_wait(&death_row_sem);
    temp->next = dr_root;
    dr_root = temp;
    temp->rt_task = rt_task;
    dec_threads();
    rt_sem_signal(&death_row_sem);
}
/*
  dec_threads();
  printk("thread exiting: %x\n", rt_whoami());
  rt_task_suspend(rt_whoami());

*/
/* Frees allocated memory for all threads in death_row */
void free_killed_tasks(void){
    while(dr_root){
	int i;
    	death_row* temp;
	 rt_sem_wait(&death_row_sem);
	temp = dr_root;
	dr_root = dr_root->next;
	rt_sem_signal(&death_row_sem);
	//printk("suspending:%x\n",temp->rt_task);
	if(i=rt_task_suspend(temp->rt_task))
	    printk("suspending returned %d :%x\n",i, temp->rt_task);

#ifdef GC_PRINT_DEBUG
	printk("deleting:%x\n",temp->rt_task);
#endif
	if(i=rt_task_delete(temp->rt_task)){
#if GC_PRINT_WARNINGS
	    printk("deleting returned %d :%x\n",i, temp->rt_task);
#endif
   	}
	kfree(temp->rt_task);
	//printk("freeing\n");
	kfree(temp);
	//printk("freeing done\n");
    }
}


/* Function in order to simulate Java main thread behavior, it */
/* waits until every threazd has finished its work	       */

void wait_for_all_threads_to_die() {
  // printk("java prewait, address:%x \n",rt_whoami());
  rt_sem_wait(&threads_mutex);
  do {
	 rt_sem_signal(&threads_mutex);
	 rt_sem_wait(&gc_threadhasdied_cond);
	  free_killed_tasks();
	 rt_sem_wait(&threads_mutex);

    //printk("Nbr of live threads after: %d\n",nbr_threads);
  } while (nbr_threads);
  // printk("java after wait\n");
  rt_sem_signal(&threads_mutex);  
  // printk("java after wait 2\n");
}


GC_PROC_BEGIN(create_main_thread)
  GC_FUNC_ENTER
  GC_REF(java_lang_Thread,main_thread);
  GC_PUSH_ROOT(main_thread);
  // printk("java 2\n");
  GC_REF_FUNC_CALL(main_thread,java_lang_Thread_Thread,&java_lang_Thread_Class,GC_PASS(java_lang_Thread_Layout));  
  // printk("java 3\n");
  GC_POP_ROOT(main_thread);
  GC_FUNC_LEAVE;
GC_PROC_END(create_main_thread)

static void init_module_main_thread(int t) {
  GC_INIT(GC_HEAPSIZE);
  {
	nbr_threads = 0;
	rt_mount_rtai();
	rt_typed_sem_init(&threads_mutex,1,BIN_SEM);
	rt_typed_sem_init(&threads_interrupt_mutex,1,BIN_SEM);
	rt_typed_sem_init(&this_param_mutex,1,BIN_SEM);
	rt_typed_sem_init(&gc_threadhasdied_cond,1,BIN_SEM);
	rt_typed_sem_init(&death_row_sem,1,BIN_SEM);
	//rt_typed_sem_init(&wait_notify_sem,1,BIN_SEM);

	javaClasses_Init();
	// printk("java 1\n");
	GC_PROC_CALL(create_main_thread);
	// printk("java after create_main\n");
	GC__ARG_LOCK;
	JAVAMAIN();
	GC__ARG_UNLOCK;
#ifdef GC_PRINT_DEBUG
	printk("Waiting for threads to die\n");
#endif
	wait_for_all_threads_to_die(); 
	printk("All threads dead\n");
  }
  GC_EXIT;
  printk("exiting...\n");
}

int init_module() {
     printk("init_module 1\n");
    // the main thread can be statically allocated
    // main_thread_task = (RT_TASK*)kmalloc(sizeof(RT_TASK),0);

    rt_preempt_always(1);
    rt_set_oneshot_mode();
    start_rt_timer(0);

#ifdef SE_LTH_CS_UTIL_LINUX_KERNELSYMBOL_INDEX
    fs = get_fs();     /* save previous value */
    set_fs (get_ds()); /* use kernel limit */
#endif
    rt_task_init(&main_thread_task,init_module_main_thread,0,10000,1,0,0);
#ifdef CPU_UTILIZATION_LOGGER
    cpu_utilization_logger_install();
    cpu_utilization_logger_addNamedTask(rt_whoami(), "Start task");
    cpu_utilization_logger_addNamedTask(&main_thread_task, "Main task");
#endif
    rt_task_resume(&main_thread_task);
    
    return 0;	  
}

/* Method to call when a new thread is created, initialises the TSD and call run method */

void thread_rtaikernel_do_start(void *gc_param_ithis)
{
  /* GC_THREAD_INIT declares gc_thread_t gc_thread */
  GC_THREAD_INIT
#ifdef PREEMPTION_POINT_HACK
  gc_reschedule_handler();
#endif
    {
      const int __gc_is_ref_func = 0; //HERE BE DRAGONS!
      GC_PARAM_REF(java_lang_Thread,ithis);
      GC__GET_THREAD_ROOT;
      GC_PUSH_PARAM(ithis);
      GC_FUNC_ENTER   
      GC_REF(java_lang_Thread,this);
      GC_REF(java_lang_Runnable,trr);
      int pr;
#ifdef GC_PRINT_DEBUG
//      printk("rtaikernel_do_start: ithis = %x\n", ithis);
//      printk("                     ithis->ref = %x\n", ithis->ref);
#endif
      GC_PUSH_ROOT(this);
      GC_PUSH_ROOT(trr);
      GC_ASSIGN(this,ithis);
      // Now, "this" is in a real local reference and we can release the mutex
      rt_sem_signal(&this_param_mutex);
      GC_GET(pr,this,_java_lang_Thread_priority);
#ifdef GC_PRINT_DEBUG
      printk("Changing priority: JAVA: %d RTAI: %x - rtaikernel_do_start\n",pr,javaPr2schedPr(pr,0));
#endif
      rt_change_prio(rt_whoami(), javaPr2schedPr(pr,0));
//      printk("rt_get_prio(rt_whoami()) = %x, rt_whoami() = %x\n",rt_get_prio(rt_whoami()),rt_whoami());

      gc_thread.this = (GC___REF(ObjectHead)) this->ref;
      GC_ASSIGN(gc_thread.javaThis,this);
      gc_thread.semObj = 0;
      /*
      {
	int status;
	gc_thread_t *tsd;
	/* Data specific
	if ((status = pthread_setspecific(tsd_key,&gc_thread)) != 0) 	strerror(status);
	tsd = (gc_thread_t*) pthread_getspecific(tsd_key);
	GC_SET(this,_java_lang_Thread_TSD,tsd);
      }*/
      GC_GET_REF(trr,this,_java_lang_Thread_targetRun);
      if (!(GC_IS_NULL(trr))) {
	struct java_lang_Runnable_ClassStruct *_tmp_method;
	GC_GET(_tmp_method,trr,class);
	GC_PROC_CALL(_tmp_method->ifazTbl->java_lang_Runnable_methodTbl->run,GC_PASS(trr));
      } else {
	struct java_lang_Thread_ClassStruct *_tmp_method;
	GC_GET(_tmp_method,this,class);
	GC_PROC_CALL(_tmp_method->methodTbl->run,GC_PASS(this));
      }
      GC_POP_ROOT(trr);
      GC_POP_ROOT(this);
      GC_FUNC_LEAVE
      GC_POP_PARAM(ithis);
    }
  GC_THREAD_EXIT
#ifdef PREEMPTION_POINT_HACK
  hack_ensure_unlock();
#else
  gc_unlock();  //HACK...
#endif
  enter_death_row(rt_whoami());
//  dec_threads();
//  printk("thread exiting: %x\n", rt_whoami());
  rt_task_suspend(rt_whoami());
}

/* Aux Native functions */
/*result=((i * (abs(POSIX_SUP-POSIX_INF)))/(JAVA_SUP-JAVA_INF))-(abs(POSIX_INF)+((abs(POSIX_SUP-POSIX_INF))/(JAVA_SUP-JAVA_INF)));*/

GC_PROC_BEGIN(java_lang_Thread_inheritPriority,GC_PARAM(java_lang_Thread,this))
  GC_PARAM_REF(java_lang_Thread,this);
//  struct sched_param param;
  JInt result;
  JInt policy;
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER;
//  if (pthread_getschedparam_rt(pthread_self_rt(),&policy,&param) != 0) printf("Error pthread_getschedparam()\n");
  /* Posix priority to Java priority */
//  result = schedPr2javaPr(param.sched_priority,_RT_SCHEDULING_POSIX);
  result = 5;
  GC_SET(this,_java_lang_Thread_priority,result);
  GC_FUNC_LEAVE;
  GC_POP_PARAM(this);
GC_PROC_END(java_lang_Thread_inheritPriority)

GC_VAR_FUNC_BEGIN(JLong,java_lang_Thread_createThread,GC_PARAM(java_lang_Thread,this))
  GC_PARAM_REF(java_lang_Thread,this);
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER;
  // Dummy implementation for rtaikernel
  GC_RETURN_VAR(0);

  GC_FUNC_LEAVE;
  GC_POP_PARAM(this);
GC_VAR_FUNC_END(JLong,java_lang_Thread_createThread)

/* Native implementation of java.lang.start() */
GC_PROC_BEGIN(java_lang_Thread_start,GC_PARAM(java_lang_Thread,this))
     GC_PARAM_REF(java_lang_Thread,this);
     GC_PUSH_PARAM(this);
     GC_FUNC_ENTER 
	{
//	  int status;
//	  static int count;
	  int pr;
	  RT_TASK* task;
	  /* Alloc memory for RT_TASK. free is done in "free_killed_tasks" */
	  task = (RT_TASK*)kmalloc(sizeof(RT_TASK),0);
	  // printk("java_lang_thread_start\n");
	  inc_threads();
	  // printk("java_lang_thread_start 2\n");
	  GC_GET(pr,this,_java_lang_Thread_priority);
          // HERE BE DRAGONS? The lock removed irt ref-param optimization
          // and replaced with this_param_mutex...
	  //gc_lock();
	  // create rtai-thread
#ifdef GC_PRINT_DEBUG
	  printk("Initial Priority when creating thread: %d\n",pr);
#endif

	  // HERE BE DRAGONS!
	HACK_ENSURE_UNLOCK

//	  rt_task_init(task, thread_rtaikernel_do_start,(void*) GC_PASS(this),10000,javaPr2schedPr(pr,0),0,0);

//	  rt_task_init(task, thread_rtaikernel_do_start,(void*) GC_PASS(this),10000,0,0,0);

	  rt_sem_wait(&this_param_mutex);
	  rt_task_init(task, thread_rtaikernel_do_start,(void*) GC_PASS(this),10000,javaPr2schedPr(pr,0),0,SIGNAL_HANDLER);

#ifdef CPU_UTILIZATION_LOGGER
	  cpu_utilization_logger_addTask(task);
#endif

	  //printk("java_lang_thread_start 3\n Adress: %x\n",task);
	  rt_task_resume(task);
	  rt_sem_wait(&this_param_mutex);
	  rt_sem_signal(&this_param_mutex);
	  //printk("java_lang_thread_start - after resume");
          // HERE BE DRAGONS? The lock removed irt ref-param optimization
          // and replaced with this_param_mutex...
	  //gc_unlock();
	  GC_SET(this,_java_lang_Thread_target,(long) task);
	}

     GC_FUNC_LEAVE;
     GC_POP_PARAM(this);
GC_PROC_END(java_lang_Thread_start)

/* Auxiliar function called from the thread clean-up handler in order to call accordingly to the GCI 	 */
/* the interruption thrower itself	(interrupter_impl). */
/* to be DISABLE after a pthread_cancel() call															 */
#if 0

void interrupter(void *arg) {
   int status;
	/* Have you ever seen something so dirty?*/

	#ifdef __RTAI__
	do {{
			int __hs_hs_hs__;
			jmp_buf _buffer;
			{
			pthread_cleanup_pop_rt(0);
	#else
	#ifndef __SOLARIS__
	{
		int __hs_hs_hs__;
		jmp_buf _buffer;
 		pthread_cleanup_pop_rt(0);
	#endif
	#endif
	GC_PROC_CALL(interrupter_impl);
}



/* Function responsible for throwing the interruption InterruptedException, it needs to get the currentThread */
/* because is the responsible for setting interruptedFlag to false (To gain semantic equivalence with JAVA    */
/* It resets the cancel_state to ENABLE and DEFERRED 'cause in some gcc versions seems to be cleared	      */

GC_PROC_BEGIN(interrupter_impl) EXCEPTION_THROWS
  GC_REF(java_lang_Thread,tmp);
  GC_FUNC_ENTER;
  GC_PUSH_ROOT(tmp);
  GC_REF_FUNC_CALL(tmp,java_lang_Thread_currentThread);
  GC_SET(tmp,_java_lang_Thread_interruptedFlag,false);
  GC_POP_ROOT(tmp);
  if (pthread_setcancelstate_rt(PTHREAD_CANCEL_ENABLE,0) != 0) printf("Error!\n");
  if (pthread_setcanceltype_rt(PTHREAD_CANCEL_DEFERRED,0) != 0) printf("Error!\n");
  EXCEPTION_THROW(JAVA_LANG_INTERRUPTEDEXCEPTION_INDEX);
  GC_FUNC_LEAVE;
GC_PROC_END(interrupter_impl)

#endif

/* Native implementation of java.lang.Thread.sleep(long millis) */

GC_PROC_BEGIN(java_lang_Thread_sleep_long,JLong millis) EXCEPTION_THROWS;
  GC_FUNC_ENTER
  {
    unsigned int msg;
//      printk("sleep..%x:%d,%d\t",rt_whoami(), schedPr2javaPr(rt_get_prio(rt_whoami()), 0), schedPr2javaPr(rt_get_inher_prio(rt_whoami()), 0));
    HACK_ENSURE_UNLOCK
    rt_receive_timed(0,&msg,nano2count(millis*(JLong)1E6));
    if(msg == RTAIKERNEL_INTERRUPT_MSG){
	EXCEPTION_THROW(JAVA_LANG_INTERRUPTEDEXCEPTION_INDEX);
	//printk("woohoo - Interruptmessage works!\n");
	//Implementation of interruption
    }
//    rt_sleep(nano2count(millis*(JLong)1E6));
//      printk("awake..%x:%d,%d\t",rt_whoami(), schedPr2javaPr(rt_get_prio(rt_whoami()), 0), schedPr2javaPr(rt_get_inher_prio(rt_whoami()), 0));
  }
  GC_FUNC_LEAVE
GC_PROC_END(java_lang_Thread_sleep_long)


/* Native implementation of java.lang.Thread.sleep(long millis,int nanos) */

GC_PROC_BEGIN(java_lang_Thread_sleep_long_int,JLong millis,int nanos) EXCEPTION_THROWS;
  GC_FUNC_ENTER
  {
    unsigned int msg;
    HACK_ENSURE_UNLOCK
    rt_receive_timed(0,&msg,nano2count(millis*(JLong)1E6+nanos));
    if(msg == RTAIKERNEL_INTERRUPT_MSG){
	EXCEPTION_THROW(JAVA_LANG_INTERRUPTEDEXCEPTION_INDEX);
	//printk("woohoo - Interruptmessage works!\n");
	//Implementation of interruption
    }
  }
  GC_FUNC_LEAVE
GC_PROC_END(java_lang_Thread_sleep_long_int)


/* Native implementation of java.lang.Thread.interrupt() */
int debug;

GC_PROC_BEGIN(java_lang_Thread_interrupt,GC_PARAM(java_lang_Thread,this))
  GC_PARAM_REF(java_lang_Thread,this);
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER
  {
  	RT_TASK* t;

	rt_sem_wait(&threads_interrupt_mutex);
	GC_SET(this,_java_lang_Thread_interruptedFlag,true);
	rt_sem_signal(&threads_interrupt_mutex);
	GC_GET(t,this,_java_lang_Thread_target);
	rt_send_if(t, RTAIKERNEL_INTERRUPT_MSG);
  }
  GC_FUNC_LEAVE;
  GC_POP_PARAM(this);
GC_PROC_END(java_lang_Thread_interrupt)

/* Native implementation of java.lang.Thread.interrupted() */

GC_VAR_FUNC_BEGIN(JBoolean,java_lang_Thread_interrupted,GC_PARAM(java_lang_Thread,this))
  GC_PARAM_REF(java_lang_Thread,this);
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER
  {
  	JBoolean tmp;

  	rt_sem_wait(&threads_interrupt_mutex);
  	GC_GET(tmp,this,_java_lang_Thread_interruptedFlag);
  	GC_SET(this,_java_lang_Thread_interruptedFlag,false);
  	rt_sem_signal(&threads_interrupt_mutex);
  	GC_RETURN_VAR(tmp);
  }
  GC_FUNC_LEAVE;
  GC_POP_PARAM(this);
GC_VAR_FUNC_END(JBoolean,java_lang_Thread_interrupted)

// GC_VAR_FUNC_BEGIN(JBoolean,java_lang_Thread_isInterrupted,GC_PARAM(java_lang_Thread,this))
//   GC_PARAM_REF(java_lang_Thread,this);
//   GC_PUSH_PARAM(this);
//   GC_FUNC_ENTER
//   {
//   	JBoolean tmp;
// 
//   	rt_sem_wait(&threads_interrupt_mutex);
//   	GC_GET(tmp,this,_java_lang_Thread_interruptedFlag);
//   	rt_sem_signal(&threads_interrupt_mutex);
//   	GC_RETURN_VAR(tmp);
//   }
//   GC_FUNC_LEAVE;
//   GC_POP_PARAM(this);
// GC_VAR_FUNC_END(JBoolean,java_lang_Thread_isInterrupted)

/* Native implementation of java.lang.Thread.currentThread() */
GC_REF_FUNC_BEGIN(java_lang_Thread,java_lang_Thread_currentThread)
  GC_REF(java_lang_Thread,tmp);
  GC_FUNC_ENTER;
  GC_PUSH_ROOT(tmp);
  {
//  	gc_thread_t *obj;
//	obj = (gc_thread_t*) pthread_getspecific(tsd_key);
//	GC_ASSIGN(tmp,obj->javaThis);
//    GC_RETURN_REF(tmp);
  }
  GC_POP_ROOT(tmp);
  GC_FUNC_LEAVE;
GC_REF_FUNC_END(java_lang_Thread,java_lang_Thread_currentThread)

/* Native implementation of java.lang.Thread.setPriority() */

GC_PROC_BEGIN(java_lang_Thread_setPriority_int,GC_PARAM(java_lang_Thread,this),JInt priority)  EXCEPTION_THROWS
  GC_PARAM_REF(java_lang_Thread,this);
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER
  JLong tmp;
  if (priority > 10 || priority < 1) {
    EXCEPTION_THROW(JAVA_LANG_ILLEGALARGUMENTEXCEPTION_INDEX)
  } else {
    GC_SET(this,_java_lang_Thread_priority,priority);
	GC_GET(tmp,this,_java_lang_Thread_target);
	if(tmp){
#ifdef GC_PRINT_DEBUG
//      		printk("Changing priority: JAVA: %d RTAI: %x - setPriority(int) on %x\n",priority,javaPr2schedPr(priority,0),tmp);
#endif
		rt_change_prio((RT_TASK*)tmp,javaPr2schedPr(priority,0));
	}
  }
  GC_FUNC_LEAVE;
  GC_POP_PARAM(this);
GC_PROC_END(java_lang_Thread_setPriority_int)

/* Native implementation of java.lang.Thread.join() */

GC_PROC_BEGIN(java_lang_Thread_join,GC_PARAM(java_lang_Thread,this))
  GC_PARAM_REF(java_lang_Thread,this);
//  JLong tmp;
//  JInt status;
//#ifdef __RTAI__
//  gc_thread_t *tsd;
//#endif
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER;
//  GC_GET(tmp,this,_java_lang_Thread_target);
//  GC_POP_PARAM(this);
//#ifdef __RTAI__
//  tsd = (gc_thread_t*) pthread_getspecific(tsd_key);
//  tsd->semObj = -2;     
//#endif  
//  pthread_cleanup_push_rt(interrupter,0);
//  if ((status = pthread_join_rt(tmp,0)) != 0 && status != ESRCH && status != EINVAL) printf("Error pthread_join() %d [%d,%d]\n",status,ESRCH,EINVAL);
//  pthread_cleanup_pop_rt(0);
//#ifdef __RTAI__
//  tsd->semObj = 0;     
//#endif    
  GC_FUNC_LEAVE;
  GC_POP_PARAM(this);
GC_PROC_END(java_lang_Thread_join)


/* Auxiliar functions for join(long millis) and join(long millis,int nanos) */

//void interruptible_join(struct timedjoin_data* data) {
//	struct sched_param param;
//	int policy;
//   if (pthread_getschedparam_rt(pthread_self_rt(),&policy,&param) != 0) printf("Error pthread_getschedparam()\n");
//	RTAI_REGISTER_THREAD(param.sched_priority);
//	pthread_join_rt(data->target,0);
//	pthread_cond_signal_rt(&data->cond);
//	RTAI_UNREGISTER_THREAD
//}
//void timedjoin_end(pthread_t* th) {
	/* Cancellation of joinning thread*/
/*	#ifdef __RTAI__
	do {{
			int __hs_hs_hs__;
			jmp_buf _buffer;	    		
			{
			pthread_cleanup_pop_rt(0); 
	#else
	#ifndef __SOLARIS__
	{		
		int __hs_hs_hs__;
		jmp_buf _buffer;	
 		pthread_cleanup_pop_rt(0);
	#endif  
	#endif
	pthread_cancel_rt(*th);
	/* Free thread's resources 
	pthread_join_rt(*th,0);*/
//}

/* Native implementation of java.lang.Thread.join(long millis) */

GC_PROC_BEGIN(java_lang_Thread_join_long,GC_PARAM(java_lang_Thread,this),JLong millis)
  GC_PARAM_REF(java_lang_Thread,this);  
/*  struct timedjoin_data data;
  struct timeval tp;
  pthread_t th;	
  JBoolean has_been_interrupted;
  GC_REF(java_lang_Thread,tmp);
  gc_thread_t *tsd;*/
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER;
/*  pthread_mutex_init_rt(&data.lock,0);
  pthread_cond_init_rt(&data.cond,0);
  data.target = 0;
  GC_GET(data.target,this,_java_lang_Thread_target);
  GC_POP_PARAM(this);
  gettimeofday(&tp,0);
  data.time.tv_sec  = (millis / MS)+ tp.tv_sec;
  data.time.tv_nsec = (millis % MS) * NS_MS + tp.tv_usec * MS;
  if (pthread_create_rt(&th,0,(void*)interruptible_join,(void*)&data) != 0)   
  		printf("Error pthread_create()<join>\n");	
  pthread_cleanup_push_rt(interrupter,0);
  pthread_cleanup_push_rt(timedjoin_end,(void*)&th);
  tsd = (gc_thread_t*) pthread_getspecific(tsd_key);
  tsd->semObj = &data.cond;     
  pthread_cond_timedwait_rt(&data.cond,&data.lock,&data.time);*/
//  pthread_cleanup_pop_rt(0); /* timedjoin_end is executed always */
//  pthread_cleanup_pop_rt(0); /* interrupter only executed when this thread is cancelled */
/*  pthread_cancel_rt(th);
   tsd->semObj = 0;
	GC_PUSH_ROOT(tmp);
   GC_REF_FUNC_CALL(tmp,java_lang_Thread_currentThread);
   GC_GET(has_been_interrupted,tmp,_java_lang_Thread_interruptedFlag);
   GC_POP_ROOT(tmp); 
   if (has_been_interrupted) { 
		GC_PROC_CALL(interrupter_impl);		
	}*/
  GC_FUNC_LEAVE;
  GC_POP_PARAM(this);
GC_PROC_END(java_lang_Thread_join_long)

/* Native implementation of java.lang.Thread.join(long millis, int nanos) */

GC_PROC_BEGIN(java_lang_Thread_join_long_int,GC_PARAM(java_lang_Thread,this),JLong millis,JInt nanos)
  GC_PARAM_REF(java_lang_Thread,this);
/*  struct timedjoin_data data;
  struct timeval tp;
  pthread_t th;*/	  
  GC_PUSH_PARAM(this);
  GC_FUNC_ENTER;
/*  pthread_mutex_init_rt(&data.lock,0);
  pthread_cond_init_rt(&data.cond,0);
  data.target = 0;
  GC_GET(data.target,this,_java_lang_Thread_target);
  GC_POP_PARAM(this);
  gettimeofday(&tp,0);
  data.time.tv_sec  = (millis / MS) + tp.tv_sec;
  data.time.tv_nsec = (millis % MS) * NS_MS + nanos + tp.tv_usec * MS;
  if (pthread_create_rt(&th,0,(void*)interruptible_join,(void*)&data) != 0)   
  		printf("Error pthread_create()<join>\n");	
  pthread_cleanup_push_rt(interrupter,0);
  pthread_cleanup_push_rt(timedjoin_end,(void*)&th);
  pthread_cond_timedwait_rt(&data.cond,&data.lock,&data.time);*/
//  pthread_cleanup_pop_rt(1); /* timedjoin_end is executed always */
//  pthread_cleanup_pop_rt(0); /* interrupter only executed when this thread is cancelled */
  GC_FUNC_LEAVE;
  GC_POP_PARAM(this);
GC_PROC_END(java_lang_Thread_join_long_int)


GC_PROC_BEGIN(java_lang_Thread_yield)
  GC_FUNC_ENTER;
  rt_task_yield();
  GC_FUNC_LEAVE;
GC_PROC_END(java_lang_Thread_yield)

void cleanup_module(){
    rt_task_suspend(&main_thread_task);
    rt_task_delete(&main_thread_task);
#ifdef CPU_UTILIZATION_LOGGER
    cpu_utilization_logger_uninstall();
#endif
    //The main thread can be statically allocated 
    //kfree(main_thread_task);
    printk("Module cleaned\n");
#ifdef SE_LTH_CS_UTIL_LINUX_KERNELSYMBOL_INDEX
    set_fs(fs); /* restore before returning to user space */
#endif
    rt_umount_rtai();

}
