#include <avr/io.h>
#include <compat/ina90.h>
#include <avr/signal.h>

#include "AVR_LCD.h"

static volatile int cursor = 0;

void LCD_moveCursor(uint8_t pos) {

	if (pos != cursor) {
		int delta;
		uint8_t cmd;
		if (pos > cursor) {
			delta = pos - cursor;
			cmd = 0x14;
		} else {
			delta = cursor - pos;
			cmd = 0x10;
		}
		for (; delta > 0; --delta) {
			LCD_writeCtrlByte(cmd);
		}
		cursor = pos;
	}
}

void LCD_enableCursor() {

	LCD_writeCtrlByte(0x0E);
}

void LCD_disableCursor() {

	LCD_writeCtrlByte(0x0C);
}

void LCD_DelayMS(int i) {
	int j;
	for(j = 0; j < i*2; j++) {
		int k;
		for(k = 0; k < 685; k++) {
			_NOP();
		}
	}
}

void LCD_WriteDelay() {
	int i;
	for(i = 0; i < 70; i++) {
		_NOP();
	}
}

void LCD_writeDataByte(uint8_t c)
{
	output(PORTD, c);
	output(PORTB, input(PORTB) & ~RW);
	output(PORTB, input(PORTB) | RS);
	output(PORTB, input(PORTB) | ENABLE);
	_NOP();
	_NOP();
	output(PORTB, input(PORTB) & ~ENABLE);
	LCD_WriteDelay();
}

void LCD_writeCtrlByte(uint8_t c)
{
	output(PORTD, c);
	output(PORTB, input(PORTB) & ~RW);
	output(PORTB, input(PORTB) & ~RS);
	output(PORTB, input(PORTB) | ENABLE);
	_NOP();
	_NOP();
	output(PORTB, input(PORTB) & ~ENABLE);
	LCD_WriteDelay();
}

void LCD_setupPorts()
{
	/*output(DDRF, 0x00);*/
	/*output(PORTF, 0xFF);*/
	output(DDRD, 0xFF);
	output(DDRB, input(DDRB) | (0x01 | 0x04 | 0x08 |0x80));
	LCD_DelayMS(15);

	/* System Set */
	LCD_writeCtrlByte(0x38);
	/* Display off */
	LCD_writeCtrlByte(0x08);
	/* Clear Display */
	LCD_writeCtrlByte(0x01);
	LCD_DelayMS(2);
	/* Entry Mode Set */
	LCD_writeCtrlByte(0x06);
	LCD_DelayMS(2);
	/* Display on */
	LCD_writeCtrlByte(0x0c);
}

void LCD_clear()
{

	LCD_writeCtrlByte(0x01);
	LCD_DelayMS(2);
	cursor = 0;
}

void LCD_printChar(char ch)
{

	LCD_writeDataByte((uint8_t)ch);
	cursor++;
}

void LCD_printByte(uint8_t b)
{
	int i;
	int nibble;

	nibble = (b >> 4) & 0xF;
	if (nibble < 0xA) {
		LCD_printChar(nibble+'0');
	} else {
		LCD_printChar(nibble-0xA+'A');
	}
	nibble = b & 0xF;
	if (nibble < 0xA) {
		LCD_printChar(nibble+'0');
	} else {
		LCD_printChar(nibble-0xA+'A');
	}
}

void LCD_printInt(uint16_t a)
{
	int i;
	int nibble;

	for (i = 3; i >= 0; --i) {
		nibble = ((a >> (i<<2)) & 0xF);
		if (nibble < 0xA) {
			LCD_printChar(nibble+'0');
		} else {
			LCD_printChar(nibble-0xA+'A');
		}
	}
}

void LCD_printInt32(uint32_t a)
{
	int i;
	int nibble;

	for (i = 7; i >= 0; --i) {
		nibble = ((a >> (i<<2)) & 0xF);
		if (nibble < 0xA) {
			LCD_printChar(nibble+'0');
		} else {
			LCD_printChar(nibble-0xA+'A');
		}
	}
}

void LCD_printStr(const char* str)
{
	int i;
	for (i = 0; str[i] != '\0'; ++i) {
		LCD_printChar(str[i]);
	}
}

