#include <avr/interrupt.h>
#include <compat/ina90.h>
#include <config.h>
#include <stdint.h>

#include "AVR_Memory.h"
#include "AVR_Kernel.h"

volatile static intptr_t	heap_ptr = AVR_HEAP_END;
volatile static intptr_t	stack_ptr = IDLE_STACK_END;
volatile static int		heap_free = HEAP_SPACE_FREE;
volatile static int		stack_free = STACK_SPACE_FREE;

/* allocate memory from the external RAM */
void* ealloc(unsigned size)
{
	intptr_t p;

	cli();
	p = heap_ptr - size;
	heap_free -= size;

	if (heap_free < 0) { 
		/* print error message and halt */
		LCD_clear();
		LCD_printStr("out of memory");
		while (1) _NOP();
	}
	heap_ptr = p;
	sei();
	return p;
}

/* allocate memory from internal RAM */
void* ialloc(unsigned size)
{
	intptr_t p;

	cli();
	p = stack_ptr - size;
	stack_free -= size;

	if (stack_free < 0) {
		/* print error message and halt */
		LCD_clear();
		LCD_printStr("out of memory");
		while (1) _NOP();
	}
	stack_ptr = p;
	sei();
	return p;
}

