#include <stddef.h>

#include "AVR_Queue.h"
#include "AVR_Kernel.h"

void queue_init(queue_t* q)
{
	q->succ = NULL;
	q->last = q;
}

void pq_insert(queue_t* q, kthread_t* group)
{
	kthread_t*	iter;
	kthread_t*	next;
	int		prio;

	prio = -1;
	iter = q;
	while (iter->last->succ) {
		next = iter->last->succ;
		if (next->prio_dyn < group->prio_dyn)
			break;
		prio = next->prio_dyn;
		iter = next;
	}

	if (prio == group->prio_dyn) {
		group->last->succ = iter->last->succ;
		iter->last->succ = group;
		iter->last = group->last;
	} else {
		group->last->succ = iter->last->succ;
		iter->last->succ = group;
	}
}

kthread_t* pq_pop_one(queue_t* q)
{
	kthread_t*	thread;

	thread = q->succ;
	if (thread) {
		q->succ = thread->succ;
		if (q->succ) {
			if (thread->last != thread)
				q->succ->last = thread->last;
		}
		thread->last = thread;
		thread->succ = NULL;
	}
	return thread;
}

kthread_t* pq_pop_group(queue_t* q)
{
	kthread_t*	group;

	group = q->succ;
	if (group) {
		q->succ = group->last->succ;
		group->last->succ = NULL;
	}
	return group;
}

/* NB: can only insert one thread at a time */
void tq_insert(queue_t* q, kthread_t* thread)
{
	kthread_t*	iter;

	iter = q;
	while (iter->succ && iter->succ->time <= thread->time)
		iter = iter->succ;

	thread->succ = iter->succ;
	iter->succ = thread;
}

kthread_t* tq_pop_one(queue_t* q)
{
	kthread_t*	thread;

	thread = q->succ;
	if (thread && thread->time <= kernel_ticks) {
		q->succ = thread->succ;
		thread->last = thread;
		thread->succ = NULL;
		return thread;
	} else {
		return NULL;
	}
}

