#include <stddef.h>
#include <compat/ina90.h>
#include <avr/interrupt.h>

#include "AVR_Semaphore.h"
#include "AVR_Kernel.h"

semaphore_t* new_semaphore(int initial)
{
	semaphore_t*	sem;

	sem = ealloc(sizeof(semaphore_t));
	queue_init(&sem->queue);
	sem->counter = initial;
	return sem;
}

mutex_t* new_mutex() {
	mutex_t*	mutex;

	mutex = ealloc(sizeof(mutex_t));
	queue_init(&mutex->queue);
	mutex->owner = NULL;
	mutex->prio_ceil = 0;
	return mutex;
}

void mutex_register_thread(kthread_t* thread, mutex_t* mutex)
{
	cli();
	if (mutex->prio_ceil < thread->prio_fix)
		mutex->prio_ceil = thread->prio_fix;
	sei();
}

void semaphore_take(semaphore_t* sem)
{
	cli();
	while (sem->counter == 0) {
		/* wait for other thread to give semaphore */
		pq_insert(&sem->queue, current_thread);
		kernel_reschedule();
		cli();
	}
	sem->counter--;
	sei();
}

void semaphore_give(semaphore_t* sem)
{
	kthread_t*	group;

	cli();
	sem->counter++;

	/* wake up threads that are waiting for this semaphore */
	while (group = pq_pop_group(&sem->queue))
		pq_insert(&ready_queue, group);

	sei();
}  

void mutex_take(mutex_t* mutex)
{
	cli();
	while (mutex->owner) {
		pq_insert(&mutex->queue, current_thread);
		kernel_reschedule();
		cli();
	}
	if (current_thread->prio_dyn < mutex->prio_ceil)
		current_thread->prio_dyn = mutex->prio_ceil;
	mutex->owner = current_thread;
	sei();
}

void mutex_give(mutex_t* mutex)
{
	kthread_t*	group;

	cli();

	if (mutex->owner != current_thread) {
		/* Print error message and halt */
		LCD_clear();
		LCD_printStr("not mutex owner");
		while (1) _NOP();
	}

	mutex->owner = NULL;

	/* wake up threads that are waiting for this mutex */
	while (group = pq_pop_group(&mutex->queue))
		pq_insert(&ready_queue, group);

	/* NB: This does not follow the Immediate Inheritance Protocol
 	 * correctly! We should compute the new dynamic priority based
 	 * on other resources currently held by the process!
 	 */
	current_thread->prio_dyn = current_thread->prio_fix;

	pq_insert(&ready_queue, current_thread);
	kernel_reschedule();
}

