#include <stddef.h>
#include <config.h>

#include "AVR_Thread.h"
#include "AVR_Kernel.h"
#include "AVR_Memory.h"
#include "AVR_Queue.h"

static void thread_exit();

kthread_t* new_kthread(
		entrypt_t entry,
		int priority,
		unsigned stack_size)
{
	kthread_t*	thread;
	intptr_t	stack_end;
	uint8_t*	sp;

	thread = ealloc(sizeof(kthread_t));
	stack_end = (intptr_t) ialloc(stack_size);

	thread->succ = NULL;
	thread->last = thread;
	thread->time = 0;
	thread->prio_fix = priority;
	thread->prio_dyn = priority;
	thread->sp = stack_end + stack_size - 38;
#ifdef CHECK_STACK
	thread->stack_end = stack_end;
#endif

	/* initialize the thread stack */
	sp = (uint8_t*)thread->sp;
	sp[37] = ((intptr_t)&thread_exit) & 0xFF;	/* exitaddr lo */
	sp[36] = ((intptr_t)&thread_exit) >> 8;		/* exitaddr hi */
	sp[35] = (intptr_t)entry & 0xFF;		/* retaddr lo */
	sp[34] = (intptr_t)entry >> 8;			/* retaddr hi */
	sp[33] = 0x00;					/* r0 */
	sp[32] = 0x00;					/* r1 */
	sp[31] = 0x00;					/* r2 */
	sp[30] = 0x00;					/* r3 */
	sp[29] = 0x00;					/* r4 */
	sp[28] = 0x00;					/* r5 */
	sp[27] = 0x00;					/* r6 */
	sp[26] = 0x00;					/* r7 */
	sp[25] = 0x00;					/* r8 */
	sp[24] = 0x00;					/* r9 */
	sp[23] = 0x00;					/* r10 */
	sp[22] = 0x00;					/* r11 */
	sp[21] = 0x00;					/* r12 */
	sp[20] = 0x00;					/* r13 */
	sp[19] = 0x00;					/* r14 */
	sp[18] = 0x00;					/* r15 */
	sp[17] = 0x00;					/* r16 */
	sp[16] = 0x00;					/* r17 */
	sp[15] = 0x00;					/* r18 */
	sp[14] = 0x00;					/* r19 */
	sp[13] = 0x00;					/* r20 */
	sp[12] = 0x00;					/* r21 */
	sp[11] = 0x00;					/* r22 */
	sp[10] = 0x00;					/* r23 */
	sp[9] = 0x00;					/* r24 */
	sp[8] = 0x00;					/* r25 */
	sp[7] = 0x00;					/* r26 */
	sp[6] = 0x00;					/* r27 */
	sp[5] = 0x00;					/* r28 */
	sp[4] = 0x00;					/* r29 */
	sp[3] = 0x00;					/* r30 */
	sp[2] = 0x00;					/* r31 */
	sp[1] = 0x00;					/* SREG */

	return thread;
}

static void thread_exit()
{
	/* Rescheduling without putting the thread back in
 	 * the ready queue will kill this thread.
 	 * 
 	 * NB: The memory used for the thread is not returned
 	 * to the system!
 	 */
	kernel_reschedule();
}

