#include <gc.h>
#include <avr/io.h> 
#include <avr/interrupt.h> 
#include <javatypes.h>
#include <java_lang_Thread.h>
#include <java_lang_Object.h>
#include <gci_common.h>
#include <exception.h>
#include <config.h>

#include "AVR_Thread.h"
#include "AVR_Kernel.h"
#include "AVR_Semaphore.h"

#ifdef _AVR_UART
#include "AVR_UART.h"
#endif

static semaphore_t*	__pass_sem;
static void*		__this_param;

int main()
{
	/* stuff needed for GC */
	void *eFlag = 0;
	void **__eFlag__ = &eFlag;

	/* suspend interrupts */
	cli();

#ifdef _AVR_EXT_MEM
	/* init ext mem */
	MCUCR = 0x80;
#endif

	/* Setup LCDDisplay */
	LCD_setupPorts();
	LCD_printStr("init");

#ifdef _AVR_UART
	initRS232();
#endif

	/* Initialize GC */
	GC_INIT(GC_HEAPSIZE);  
	__pass_sem = new_semaphore(1); 
	javaClasses_Init(__eFlag__);

	/* start threading */
	kernel_start(); 

	/* Should never come here */
	GC_EXIT;   
	return 0;
}

void jthread_entry()
{
	void *eFlag = 0;
	void **__eFlag__ = &eFlag;
	void *gc_param_ithis = __this_param;

	cli();
	GC_THREAD_INIT;
	GC__GET_THREAD_ROOT;
	GC_PARAM_REF(java_lang_Thread, ithis);
	GC_PUSH_PARAM(ithis);
	sei();
	{
		GC_REF(java_lang_Thread, this);
		GC_PUSH_ROOT(this);
		GC_ASSIGN(this, ithis);

		semaphore_give(__pass_sem);
		gc_thread.this = (void *) GC_PASS(this);
		{
			struct java_lang_Thread_ClassStruct *_tmp_method;
			GC_GET(_tmp_method, this, class);
			GC_PROC_CALL(_tmp_method->methodTbl->run, GC_PASS(this));
		}
		GC_POP_ROOT(this);
	} 
	cli();
	GC_POP_PARAM(ithis); 
	GC_THREAD_EXIT;
} 

GC_VAR_FUNC_BEGIN(JLong,java_lang_Thread_createThread,GC_PARAM(java_lang_Thread,this))
	GC_PARAM_REF(java_lang_Thread,this);
	GC_PUSH_PARAM(this);
	GC_FUNC_ENTER;


	JLong _result_;
	_result_ = (JLong) new_kthread(&jthread_entry, PRIO_JAVA, JAVA_STACK_SIZE);
	GC_RETURN_VAR(_result_);

	GC_FUNC_LEAVE;
	GC_POP_PARAM(this);	
GC_VAR_FUNC_END(JLong,java_lang_Thread_createThread)

GC_PROC_BEGIN(java_lang_Thread_startThread_long,GC_PARAM(java_lang_Thread,this), JLong target)
	GC_PARAM_REF(java_lang_Thread,this);
	GC_PUSH_PARAM(this);
	GC_FUNC_ENTER
{
	semaphore_take(__pass_sem);
	cli();
	__this_param = GC_PASS(this);
	pq_insert(&ready_queue, target);
	sei();
	/* This is required to ensure that the activation record containing
	 * __this_param is not popped from the stack until it has been read
	 * by the child thread
	 */
	semaphore_take(__pass_sem);
	semaphore_give(__pass_sem);
}

GC_FUNC_LEAVE
GC_POP_PARAM(this);
GC_PROC_END(java_lang_Thread_start)

GC_PROC_BEGIN(java_lang_Thread_start,GC_PARAM(java_lang_Thread,this))
	GC_PARAM_REF(java_lang_Thread,this);
	GC_PUSH_PARAM(this);
	GC_FUNC_ENTER
{
	JLong target;
	GC_GET(target,this,_java_lang_Thread_target);
	GC_PROC_CALL(java_lang_Thread_startThread_long,GC_PASS(this),target);
}

GC_FUNC_LEAVE
GC_POP_PARAM(this);
GC_PROC_END(java_lang_Thread_start)

	/* Native implementation of java.lang.Thread.currentThread() */

GC_REF_FUNC_BEGIN(java_lang_Thread,java_lang_Thread_currentThread)
	GC_FUNC_ENTER
	GC_REF(java_lang_Thread,tmp);
	GC_PUSH_ROOT(tmp);
{	
	/*	ljrt_thread_t *obj; */
	/*	obj = (ljrt_thread_t*) pthread_getspecific(ljrt_tsd_key); */
	/*	GC_ASSIGN(tmp,obj->javathread); */
	GC_RETURN_REF(tmp);
}  
GC_POP_ROOT(tmp);
GC_FUNC_LEAVE;
GC_REF_FUNC_END(java_lang_Thread,java_lang_Thread_currentThread)

GC_PROC_BEGIN(java_lang_Thread_sleep_long,long millis)
	GC_FUNC_ENTER
{
	kernel_sleep(millis);	   
}
	GC_FUNC_LEAVE
GC_PROC_END(java_lang_Thread_sleep_long)

GC_PROC_BEGIN(java_lang_Thread_inheritPriority,GC_PARAM(java_lang_Thread,this))
	GC_PARAM_REF(java_lang_Thread,this);
	/*   struct sched_param param; */
	/*   JInt result; */
	/*   JInt policy; */
	GC_PUSH_PARAM(this);
	GC_FUNC_ENTER;
	/*   if (pthread_getschedparam_rt(pthread_self_rt(),&policy,&param) != 0) printf("Error pthread_getschedparam()\n"); */
	/*   Posix priority to Java priority */ 
	/*   result = schedPr2javaPr(param.sched_priority,_RT_SCHEDULING_POSIX); */
	/*   GC_SET(this,_java_lang_Thread_priority,result); */
	GC_FUNC_LEAVE;
	GC_POP_PARAM(this);
GC_PROC_END(java_lang_Thread_inheritPriority)

	/* GC_VAR_FUNC_BEGIN(JBoolean,java_lang_Thread_interrupted,GC_PARAM(java_lang_Thread,this)) */
	/*   GC_PARAM_REF(java_lang_Thread,this); */
	/*   JBoolean tmp; */
	/* /\*	 pthread_mutex_t* mutex; *\/ */
	/*   GC_PUSH_PARAM(this); */
	/*   GC_FUNC_ENTER; */
	/* /\*	 pthread_mutex_lock_rt(&threads_interrupt_mutex); *\/ */
	/*   GC_GET(tmp,this,_java_lang_Thread_interruptedFlag); */
	/*   GC_SET(this,_java_lang_Thread_interruptedFlag,false); */
	/* /\*	 pthread_mutex_unlock_rt(&threads_interrupt_mutex); *\/ */
	/*   GC_RETURN_VAR(tmp);     */
	/*   GC_FUNC_LEAVE; */
	/*   GC_POP_PARAM(this); */
	/* GC_VAR_FUNC_END(JBoolean,java_lang_Thread_interrupted) */

GC_PROC_BEGIN(java_lang_Thread_interrupt,GC_PARAM(java_lang_Thread,this))
	GC_PARAM_REF(java_lang_Thread,this);
	GC_PUSH_PARAM(this);
	GC_FUNC_ENTER;
{
	/*	pthread_t t;	 */
	/*	gc_thread_t *tsd_blocked_thread; */
	/*	pthread_mutex_lock_rt(&threads_interrupt_mutex);						 */
	GC_SET(this,_java_lang_Object_interruptedFlag,true);
	/*	pthread_mutex_unlock_rt(&threads_interrupt_mutex);// */
	/*	GC_GET(t,this,_java_lang_Thread_target); */
	/*    GC_GET(tsd_blocked_thread,this,_java_lang_Thread_TSD); */
	/*	if (tsd_blocked_thread == 0 || tsd_blocked_thread->semObj == 0) {		 */
	/* #ifndef __RTAI__		 */
	/*		if (t != pthread_self_rt()) {  */
	/*			if ((status = pthread_cancel_rt(t)) != 0 && status !=ESRCH) printf("Error pthread_cancel()\n");				 */
	/*		} */
	/* #endif */
	/*    // These conditionals only when RTAI is running */
	/* #ifdef __RTAI__ */
	/*	}	else if (tsd_blocked_thread->semObj == -2) { */
	/*		// Target thread is blocked in a join */
	/*		if (pthread_cancel_rt(t) != 0) printf("Error pthread_cancel()\n"); */
	/*	} else if (tsd_blocked_thread->semObj == -1) { */
	/*		// Do nothing - Flag already set	- Target thread is blocked in a sleep */
	/*		; */
	/* #endif */
	/*	} else { */
	/*		pthread_cond_signal_rt(tsd_blocked_thread->semObj); */
	/*	} */
	//pthread_mutex_unlock_rt(&threads_interrupt_mutex);
}   
GC_FUNC_LEAVE;	
GC_POP_PARAM(this);	
GC_PROC_END(java_lang_Thread_interrupt)

	GC_PROC_BEGIN(java_lang_Thread_setPriority_int,GC_PARAM(java_lang_Thread,this),JInt priority)  EXCEPTION_THROWS
	GC_PARAM_REF(java_lang_Thread,this);
	/*   struct sched_param param; */
	JInt policy;
	JLong tmp;
	JInt status;
	GC_PUSH_PARAM(this);
	GC_FUNC_ENTER;
	/*   if (priority > 10 || priority < 1) { */
	/*     EXCEPTION_THROW(JAVA_LANG_ILLEGALARGUMENTEXCEPTION_INDEX) */
	/*   } */
	/*   else { */
	/*     GC_SET(this,_java_lang_Thread_priority,priority); */
	/*	 GC_GET(tmp,this,_java_lang_Thread_target); */
	/*	 param.sched_priority = javaPr2schedPr(priority,_RT_SCHEDULING_POSIX); */
	/*	 if ((status = pthread_setschedparam_rt(tmp,_RT_SCHEDULING_POSIX,&param)) != 0 && status != ESRCH) { */
	/*		printf("pthread_setschedparam() %d[%d,%d]\n",param.sched_priority,sched_get_priority_min(_RT_SCHEDULING_POSIX),sched_get_priority_max(_RT_SCHEDULING_POSIX)); */
	/*	} */
	/*   } */
	GC_FUNC_LEAVE;
	GC_POP_PARAM(this);
GC_PROC_END(java_lang_Thread_setPriority_int)

GC_PROC_BEGIN(java_lang_Thread_join,GC_PARAM(java_lang_Thread,this))
	GC_PARAM_REF(java_lang_Thread,this);
	JLong tmp;
	JInt status;
#ifdef __RTAI__
	gc_thread_t *tsd;
#endif
	GC_PUSH_PARAM(this);
	GC_FUNC_ENTER;
	/*   GC_GET(tmp,this,_java_lang_Thread_target); */
	/*   GC_POP_PARAM(this); */
	/* #ifdef __RTAI__ */
	/*   tsd = (gc_thread_t*) pthread_getspecific(tsd_key); */
	/*   tsd->semObj = -2;	    */
	/* #endif   */
	/*   pthread_cleanup_push_rt(interrupter,0); */
	/*   if ((status = pthread_join_rt(tmp,0)) != 0 && status != ESRCH && status != EINVAL) printf("Error pthread_join() %d [%d,%d]\n",status,ESRCH,EINVAL); */
	/*   pthread_cleanup_pop_rt(0); */
	/* #ifdef __RTAI__ */
	/*   tsd->semObj = 0;	   */
	/* #endif     */
	GC_FUNC_LEAVE;	
	GC_POP_PARAM(this);
GC_PROC_END(java_lang_Thread_join)

GC_PROC_BEGIN(java_lang_Thread_join_long,GC_PARAM(java_lang_Thread,this),JLong millis)
	GC_PARAM_REF(java_lang_Thread,this);  
	/*   struct timedjoin_data data; */
	/*   struct timeval tp; */
	/*   pthread_t th;	 */
	JBoolean has_been_interrupted;
	GC_REF(java_lang_Thread,tmp);
	/*   gc_thread_t *tsd; */
	GC_PUSH_PARAM(this);
	GC_FUNC_ENTER;
	/*   pthread_mutex_init_rt(&data.lock,0); */
	/*   pthread_cond_init_rt(&data.cond,0); */
	/*   data.target = 0; */
	/*   GC_GET(data.target,this,_java_lang_Thread_target); */
	/*   GC_POP_PARAM(this); */
	/*   gettimeofday(&tp,0); */
	/*   data.time.tv_sec  = (millis / MS)+ tp.tv_sec; */
	/*   data.time.tv_nsec = (millis % MS) * NS_MS + tp.tv_usec * MS; */
	/*   if (pthread_create_rt(&th,0,(void*)interruptible_join,(void*)&data) != 0)	  */
	/*		printf("Error pthread_create()<join>\n");	 */
	/*   pthread_cleanup_push_rt(interrupter,0); */
	/*   pthread_cleanup_push_rt(timedjoin_end,(void*)&th); */
	/*   tsd = (gc_thread_t*) pthread_getspecific(tsd_key); */
	/*   tsd->semObj = &data.cond;	    */
	/*   pthread_cond_timedwait_rt(&data.cond,&data.lock,&data.time); */
	/*   pthread_cleanup_pop_rt(0); /* timedjoin_end is executed always */ 
	/*   pthread_cleanup_pop_rt(0); /* interrupter only executed when this thread is cancelled */ 
	/*   pthread_cancel_rt(th); */
	/*    tsd->semObj = 0; */
	/*	GC_PUSH_ROOT(tmp); */
	/*    GC_REF_FUNC_CALL(tmp,java_lang_Thread_currentThread); */
	/*    GC_GET(has_been_interrupted,tmp,_java_lang_Thread_interruptedFlag); */
	/*    GC_POP_ROOT(tmp);  */
	/*    if (has_been_interrupted) {  */
	/*		GC_PROC_CALL(interrupter_impl);		 */
	/*	} */
	GC_FUNC_LEAVE;
	GC_POP_PARAM(this);
GC_PROC_END(java_lang_Thread_join_long)

GC_PROC_BEGIN(java_lang_Thread_join_long_int,GC_PARAM(java_lang_Thread,this),JLong millis,JInt nanos)
	GC_PARAM_REF(java_lang_Thread,this);
	/*   struct timedjoin_data data; */
	/*   struct timeval tp; */
	/*   pthread_t th;	   */
	GC_PUSH_PARAM(this);
	GC_FUNC_ENTER;
	/*   pthread_mutex_init_rt(&data.lock,0); */
	/*   pthread_cond_init_rt(&data.cond,0); */
	/*   data.target = 0; */
	/*   GC_GET(data.target,this,_java_lang_Thread_target); */
	/*   GC_POP_PARAM(this); */
	/*   gettimeofday(&tp,0); */
	/*   data.time.tv_sec  = (millis / MS) + tp.tv_sec; */
	/*   data.time.tv_nsec = (millis % MS) * NS_MS + nanos + tp.tv_usec * MS; */
	/*   if (pthread_create_rt(&th,0,(void*)interruptible_join,(void*)&data) != 0)	  */
	/*		printf("Error pthread_create()<join>\n");	 */
	/*   pthread_cleanup_push_rt(interrupter,0); */
	/*   pthread_cleanup_push_rt(timedjoin_end,(void*)&th); */
	/*   pthread_cond_timedwait_rt(&data.cond,&data.lock,&data.time); */
	/*   pthread_cleanup_pop_rt(1); /* timedjoin_end is executed always */ 
	/*   pthread_cleanup_pop_rt(0); /* interrupter only executed when this thread is cancelled */ 
	GC_FUNC_LEAVE;
	GC_POP_PARAM(this);
GC_PROC_END(java_lang_Thread_join_long_int)

