
/* include sfr definitions for IO registers */
#include <avr/io.h>
#include <compat/ina90.h>
#include <avr/interrupt.h>
/* #include <iom128.h>          */
/* #include <iomacros.h>  */
/* #include <avr/io.h> */
/* #include <stdlib.h> */
/* #include "interrupt.h" */
// #include "sig-avr.h" 

#include "AVR_Kernel.h"
#include <config.h>

uint8_t disableInts()
{
	uint8_t status = SREG;
	cli();
	return SREG;
}

void enableInts(uint8_t status)
{
	SREG = status;
}

/* #define EXT_CLOCK */

volatile unsigned long tickCounter;            	// current time
byte threadState;                     	// state for the current running thread
byte intMask;
KernelThread *currentThread;           	// current running thread
KernelThread *idleCode; // idle code
Queue *timeQueue;			// time queue
Queue *threadQueues[PRIO_LEVELS+1];	// one queue for each priority level
Queue *intQueue0;			// interrupt queue
static Semaphore* sem;

// This is pretty ugly, but we need it to run the clock servicing thread
void ClockDeviceThread();

extern byte *periodicStack[PRIO_LEVELS];
extern void led(int i);
extern Semaphore *__pass_sem;

#include <gc.h>

extern int JAVAMAIN(void);

/* void enableInts(int intMask) { */
/*   outp(0x10, EIMSK); */
/*   outp(0x0, TIMSK); */
/*   sbi(SREG,7); */
/* } */

/* int disableInts(void) { */
/*   int intMask = ((int) inp(TIMSK)) + (((int)inp(EIMSK)) << 8); */
/*   outp(0,TIMSK); */
/*   outp(0,EIMSK); */
/*   cbi(SREG,7); */
/*   return intMask; */
/* } */

// Check the current thread's stack pointer for overflow
void CheckStack()
{
  while (currentThread->sp < currentThread->stack_top) {
    // Print error message
    LCD_moveCursor(0);
	LCD_printChar('s');
	LCD_printChar('t');
	LCD_printChar('a');
	LCD_printChar('c');
	LCD_printChar('k');
	LCD_printChar(' ');
	LCD_printChar('o');
	LCD_printChar('v');
	LCD_printChar('e');
	LCD_printChar('r');
	LCD_printChar('f');
	LCD_printChar('l');
	LCD_printChar('o');
	LCD_printChar('w');
    LCD_moveCursor(40);
	LCD_printInt(currentThread->id);
  }
}

void initKernel(void) {
  int t;
  for(t = 0; t <= PRIO_LEVELS; t++) {
    threadQueues[t] = (Queue *)myStackMalloc(sizeof(Queue));
    initQueue(threadQueues[t]);
  }

  for(t = 0; t < PRIO_LEVELS; t++) {
    periodicStack[t] = 0;
  }

  timeQueue = (Queue *)myStackMalloc(sizeof(Queue));
  initQueue(timeQueue);
  //intQueue0 = (Queue *)myMyMalloc(sizeof(Queue));
  //initQueue(intQueue0);
 
  tickCounter = 0;
  threadState = 0;

/*   intMask = 0x10; // Default Interrupt mask (EIMSK int 4 enable) */
/*   intMask = 0x00; // Disable all interrupts */
// intMask could just as well be a macro. It is only assigned to here.
// It acts as a constant.
#ifdef _AVR_EXT_CLOCK
  intMask = 0x10; // Default Interrupt mask (EIMSK int 4 enable)
#else
  intMask = 0x80; // Enable OCIE2 in TIMSK
#endif
}

void startKernel(void) {
/*   outp(~0x07,PORTB);      */
  initKernel();
/*   outp(~0x08,PORTB);      */
  sem = allocSemaphore(0,1); 
/*   outp(~0x09,PORTB);      */

  // Main Thread
  readyQueueInsertLast(0, allocThread(0, &JAVAMAIN, PRIO_JAVAMAIN, JAVA_STACK_SIZE));
  // The clock servicing thread does not require as large stack as the Java threads.
  readyQueueInsertLast(0, allocThread(0, &ClockDeviceThread, PRIO_JAVA, 256)); // Clock Thread  

  // Instead of rescheduling here and losing the C stack we can utilize
  // it for idle code. The idle loop starts after we've enabled interrupts.
  idleCode = myStackMalloc(sizeof(KernelThread));
  idleCode->dynamicPrio = PRIO_IDLE;
  idleCode->fixedPrio = PRIO_IDLE;
  // Assuming we need no more than 128 bytes of idle code stack space
  // this makes stack_top = 1100-80 = 1080.
  idleCode->stack_top = 0x1080;
  currentThread = idleCode;
  // The idle code stack pointer will be set on the first context switch.

  LCD_clear();

/*   transmitInt(inp(TCCR2)); */
/*   transmitInt(inp(OCR2)); */
/*   transmitInt(inp(TIMSK)); */
/*   transmitInt(inp(SREG)); */
/*   transmit('\n'); */
  setInts(0);
#ifdef _AVR_EXT_CLOCK
/*    outp(inp(EICRB) | 0x03, EICRB); */
  //EICRB |= 0x03;
  output(EICRB, input(EICRB) | 0x03);
#else
/*    outp(0x0c,TCCR2);	// (affects tick length)  */
/*    outp(0x9d,OCR2); // Compare value (affects tick length)    */
  //TCCR2 = 0x0c;
  //OCR2 = 0x9d;
  output(TCCR2, 0x0C);
  output(OCR2, 0x9D);
#endif
  // Enable interrupts to allow other threads to run
#ifdef _AVR_EXT_CLOCK
  setInts((0x00 << 8) | 0x10);// TIMSK << 8 | EIMSK
#else
  setInts((0x80 << 8) | 0x00);
#endif

  // Idle code:
  while (1) {
    _NOP();
  }
}

/* INTERRUPT(SIG_OUTPUT_COMPARE2) {} */

void waitUntil(long time) {
  int status = disableInts();
  currentThread->time = time;
  prioTimeQueueInsert(timeQueue, currentThread);
  threadState |= THREAD_SUSPENDED;
  switchContext();
  enableInts(status);
}

void sleep(long delay) {
  int status = disableInts();
  currentThread->time = tickCounter + delay;
  prioTimeQueueInsert(timeQueue, currentThread);
  threadState |= THREAD_SUSPENDED;
  switchContext();
  enableInts(status);
} 

long getTickTime(int this) {
  int status = disableInts();
  long retVal = tickCounter;
  enableInts(status);
  return retVal;
}

int getTickFrac(int this) {
  int status = disableInts();
  long retVal = (long)OCR2;
  retVal *= ( 2<<16 / 0x9D );
  enableInts(status);
  return retVal;
}
