#include "AVR_Semaphore.h"
#include <gc.h>
extern byte threadState;

extern void print(char);

Semaphore *allocSemaphore(int this, int initialValue) {
  Semaphore *s;

  s = myMalloc(sizeof(Semaphore));
/*   print('i'); */
  s->queue = myMalloc(sizeof(Queue));
/*   print('j'); */
  initQueue(s->queue);
/*   print('k'); */

/*   print('l'); */
  s->counter = initialValue;
  // s->owner not used for semaphores, only mutexes /Jesper
  s->owner = 0;
  return s;
}

Mutex *allocMutex(int this) {
  Mutex *s;
  s = myMalloc(sizeof(Mutex));
  s->owner = 0;
  s->ceiling = PRIO_LEVELS;
  s->ceiling = 0; // Temporary
  return s;
}

void registerThread(int this, KernelThread *thread, Mutex *mutex) {
  int status = disableInts();
  if(mutex->ceiling > thread->fixedPrio) {
    mutex->ceiling = thread->fixedPrio;
  }
  enableInts(status);
}

void takeBinarySemaphore(int this, Semaphore *s) {
  int status = disableInts();
  while(s->counter == 0) {
    prioQueueInsert(s->queue, currentThread);
    threadState |= THREAD_SUSPENDED;
    switchContext();
  }
  s->owner = currentThread;
  s->counter = 0;
  enableInts(status);
}

void giveBinarySemaphore(int this, Semaphore *s) {
  int status = disableInts();
  s->counter = 1;
  s->owner = prioQueueRemoveSingle(s->queue);
  if(s->owner != 0) {
    readyQueueInsertLast(0, s->owner);
  }
  s->owner = 0;
  enableInts(status);
}

void takeCountingSemaphore(int this, Semaphore *s) {
  int status = disableInts();
  while(s->counter == 0) {
    prioQueueInsert(s->queue, currentThread);
    threadState |= THREAD_SUSPENDED;
    switchContext();
  }
  s->owner = currentThread;
  s->counter--;
  enableInts(status);
}

void giveCountingSemaphore(int this, Semaphore *s) {
  int status = disableInts();
  s->counter++;
  s->owner = prioQueueRemoveSingle(s->queue);
  if(s->owner != 0) {
    readyQueueInsertLast(0, s->owner);
  }
  s->owner = 0;
  enableInts(status);
}  

void takeMutex(int this, Mutex *s) {
  int status = disableInts();
  while(s->owner != 0) {
    //readyQueueInsertLast(this, currentThread);
    // A call to switchContext will put currentThread last on the
    // readyQueue if the THREAD_SUSPENDED flag is cleared
    threadState &= ~THREAD_SUSPENDED;
    switchContext();
   }
  currentThread->dynamicPrio = s->ceiling;
  s->owner = currentThread;
  enableInts(status);
}

void giveMutex(int this, Mutex *s) {
  int status = disableInts();
  if(s->owner == currentThread) {
    s->owner = 0;
    currentThread->dynamicPrio = currentThread->fixedPrio;
    readyQueueInsert(this, currentThread);
    threadState |= THREAD_SUSPENDED;
    switchContext();
  }
  else {
    // Error try to give mutex which we do not own
    while (1) {
      // Print error message
      LCD_writeDataByte('!');
      LCD_writeDataByte('1');
    }
  }
  enableInts(status);
}

