#include "AVR_Thread.h"
#include <config.h>

extern byte threadState;
extern byte intMask;
extern void led(int i);

byte *periodicStack[PRIO_LEVELS];	// stacks for the periodic threads

static int next_thread = 0;

KernelThread *allocThread(int this, void (*codeStart)(void), int priority,
    unsigned stack_size) {
  // Since the thread is not yet put in the ready queue,
  // we do not need to disable interrupts.
  //int status = disableInts();
  KernelThread *t;
  int i;
  byte *stack = (byte *)myStackMalloc(stack_size+16); // Hack to not underflow the stack
  // the above hack is likely not needed any more. /Jesper
  for(i = 0; i < (stack_size+16); i++) {
    ((char *)stack)[i] = 0;
  }
  //byte lo;
  //byte hi;

  t = myStackMalloc(sizeof(KernelThread));
  for(i = 0; i < sizeof(KernelThread); i++) {
    ((char *)t)[i] = 0;
  }

  t->sp = (long)stack + stack_size - 44;
  t->stack_top = (long)stack;/*For error checking purposes.*/
  t->id = next_thread++;

  // Not used? /Jesper
  //lo = (byte)((((int)stack) + 384) & 0xff);
  //hi = (byte)((((int)stack) + 384) >> 8);

/*   stack[stack_size-17] = 0; // R19 */
/*   stack[stack_size-16] = 0; // R18 */
  // "this" is not used by any thread entry point. Remains from old code? /Jesper
  stack[stack_size-15] = this >> 8; // First param (MSB) = HIGH(THIS)
  //stack[stack_size-14] = hi; // R29 = HIGH(c_stack)
  //stack[stack_size-13] = lo; // R28 = LOW(c_stack)
  
#ifdef _AVR_EXT_CLOCK
  stack[stack_size-12] = 0x00;					// TIMSK 0x80
  stack[stack_size-11] = 0x10;					// EIMSK
#else
  stack[stack_size-12] = 0x80;					// TIMSK 0x80
  stack[stack_size-11] = 0x0;//intMask;					// EIMSK
#endif
  stack[stack_size-10] = 0x80;					// SR 
  stack[stack_size-9] =  this;	    // Last register , First param (LSB) = LOW(this)
  stack[stack_size-7] = (byte)((int)codeStart & 0xff);  // retAdrLo
  stack[stack_size-8] = (byte)((int)codeStart >> 8);	// retAdrHi
  stack[stack_size-6] = (byte)(((int)&deallocCode) >> 8); // deallocCodeHi
  stack[stack_size-5] = (byte)(((int)&deallocCode) & 0xff); // deallocCodeLo

  t->type = THREAD_ONESHOT;
  t->dynamicPrio = priority;
  t->fixedPrio = priority;

  /* Actually the following lines are not necessary since
   	 next, pred and last are overwritten when the thread is
   	 put in the ready queue.
   */
  //t->next = 0;
  //t->pred = 0;
  //t->last = t;

/* outp(~0x3E,PORTB); */
  //enableInts(status);
  return t;
}

void startThread(int this, KernelThread *t) {
  int status = disableInts();
  readyQueueInsertLast(this, t);
  enableInts(status);
}

KernelThread *allocPeriodicThread(int this, void (*codeStart)(void), int priority, int interval) {	
  int status = disableInts();
  PeriodicThread *t;
  t = myStackMalloc(sizeof(PeriodicThread));
  if(periodicStack[priority] == 0) {
    periodicStack[priority] = (byte *)myStackMalloc(STACK_SIZE);
  }

  t->type = THREAD_PERIODIC;
  t->dynamicPrio = priority;
  t->fixedPrio = priority;
  t->codeLo = (byte)((int)codeStart & 0xff);
  t->codeHi = (byte)((int)codeStart >> 8);
  t->next = 0;
  t->pred = 0;
  t->last = (KernelThread *)t;
  t->interval = interval;
  t->initialized = 0xff;
  t->this = this;
  enableInts(status);
  return (KernelThread *)t;
}

void startPeriodicThread(int this, PeriodicThread *t) {
  int status = disableInts();
  t->time = tickCounter + t->interval;
  prioTimeQueueInsert(timeQueue, (KernelThread *)t);
  enableInts(status);
}

KernelThread *allocInterrupt(int this, void (*codeStart)(void), int priority, int intNr) {
  int status = disableInts();
  PeriodicThread *t;
  Queue *iq = intQueue0;
  t = myStackMalloc(sizeof(PeriodicThread));

  if(periodicStack[priority] == 0) {
    periodicStack[priority] = (byte *)myStackMalloc(STACK_SIZE);
  }
  
  t->type = THREAD_INTERRUPT;
  t->dynamicPrio = priority;
  t->fixedPrio = priority;
  t->codeLo = (byte)((int)codeStart & 0xff);
  t->codeHi = (byte)((int)codeStart >> 8);
  t->next = 0;
  t->pred = 0;
  t->last = (KernelThread *)t;
  t->initialized = 0xff;
  t->this = this;
  t->intQueueHi = (byte)((int)iq >> 8);
  t->intQueueLo = (byte)(((int)iq) & 0xff);
  enableInts(status);
  return (KernelThread *)t;
}

void startInterrupt(int this, PeriodicThread *t) {
  int status = disableInts();
  prioQueueInsert(intQueue0, (KernelThread *)t);
  enableInts(status);
}

void initThread(PeriodicThread *t) {
  threadState &= ~THREAD_SUSPENDED; // Clear suspension flag
  threadState &= ~PREEMTION_DISABLED; // Clear preemtion disabled flag
  
  if((t->type == THREAD_PERIODIC || t->type == THREAD_INTERRUPT) && t->initialized != 0x00) {
    int offset = t->fixedPrio;
    byte *stack = periodicStack[offset];
    t->sp = (int)stack + STACK_SIZE - 44;
    stack[STACK_SIZE-17] = 0; // R1
    stack[STACK_SIZE-16] = t->this >> 8;
    stack[STACK_SIZE-15] = t->this;
    stack[STACK_SIZE-14] = 0; // hi?
    stack[STACK_SIZE-13] = 0; // lo?
    stack[STACK_SIZE-12] = 0x0;				// TIMSK
    stack[STACK_SIZE-11] = intMask;				// EIMSK
    stack[STACK_SIZE-10] = 0x80;				// SR
    stack[STACK_SIZE-9] = 0;					// Last register
    stack[STACK_SIZE-8] = t->codeHi;				// retAdrHi
    stack[STACK_SIZE-7] = t->codeLo;				// retAdrLo
    stack[STACK_SIZE-6] = (byte)(((int)&deallocCode) >> 8);	// deallocCodeHi
    stack[STACK_SIZE-5] = (byte)(((int)&deallocCode) & 0xff);	// deallocCodeLo
    // 4 padbytes used as deallocCode puts bytes on the stack
    t->initialized = 0x00;
    threadState |= PREEMTION_DISABLED; // Temporary 
  }
}

void deallocCode(void) {
  int status = disableInts();
  PeriodicThread *p = (PeriodicThread *)currentThread;
  if(currentThread->type == THREAD_PERIODIC) {
    p->initialized = 0xff;
    p->time += p->interval;
    p->next = 0;
    p->pred = 0;
    p->last = currentThread;
    prioTimeQueueInsert(timeQueue, currentThread);
  }
  else if(currentThread->type == THREAD_INTERRUPT) {
    p->initialized = 0xff;
    p->next = 0;
    p->pred = 0;
    p->last = currentThread;

    prioQueueInsert((Queue *)(((int)p->intQueueHi << 8) +
                    (int)p->intQueueLo), currentThread);
  }
  enableInts(status);
  reSchedule();
}
