#ifndef AVR_TYPES_H
#define AVR_TYPES_H

typedef unsigned char byte;

// Stack size used for interrupts and periodic threads
/* #define STACK_SIZE 300 */
#define STACK_SIZE 712

// Java stack size is defined through autotools


// In the current implementation of thread queues
// the "pred" ptr serves no purpose. It could be removed
// to save space and clock cycles. "this" is not used
// for anything either and could also be removed. /Jesper

typedef struct KernelThread {
  volatile struct KernelThread *next;		// Pointer to next element in a queue
  volatile struct KernelThread *pred;		// Pointer to predecessor in queue
  volatile struct KernelThread *last;		// Pointer to last element within the same "group" in queue
  volatile unsigned long time;		// Time to wait for
  volatile byte dynamicPrio;		// Current Priority
  volatile byte fixedPrio;		// Original Priority
  volatile long sp; // don't need two separate bytes to store the stack pointer
  volatile byte type;
  volatile long stack_top; /* useful for error checks */
  volatile int id; /* useful for error checks */
} KernelThread;

typedef struct PeriodicThread {
  volatile struct KernelThread *next;		// Pointer to next element in a queue
  volatile struct KernelThread *pred;		// Pointer to predecessor in queue
  volatile struct KernelThread *last;		// Pointer to last element within the same "group" in queue
  volatile unsigned long time;		// Time to wait for
  volatile byte dynamicPrio;
  volatile byte fixedPrio;
  volatile long sp;			// Stack Pointer of thread
  volatile byte type;			// ThreadType (OneShot, Periodic, Interrupt
  volatile unsigned stack_size; // for testing stack overflow
  volatile byte intQueueHi;
  volatile byte intQueueLo;
  volatile int  interval;
  volatile byte codeHi;
  volatile byte codeLo;
  volatile byte initialized;
  volatile int  this;// Why? /Jesper
} PeriodicThread;

typedef struct Queue {
  volatile struct KernelThread *next;		// Pointer to next element in queue
  volatile struct KernelThread *pred;		// Pointer to predecessor in queue
  volatile struct KernelThread *last;		// Pointer to last element within the same "group" in queue
} Queue;

typedef struct Semaphore {
  volatile struct Queue *queue;
  volatile struct KernelThread *owner;
  volatile byte counter;
} Semaphore;

typedef struct Mutex {
  volatile struct KernelThread *owner;
  volatile byte ceiling;
} Mutex;


typedef struct _Object_ {
  volatile int size;
  volatile struct _Object_ *next; // used as a marking stack
  volatile struct _Object_ **ptr; // Pointer to entry in objectTable
  volatile int *ptrInfo;
} _Object_;

#endif
